/* Generated by Cython 0.29.23 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "name": "nes.cycore.mos6502",
        "sources": [
            "nes/cycore/mos6502.pyx"
        ]
    },
    "module_name": "nes.cycore.mos6502"
}
END: Cython Metadata */

#ifndef PY_SSIZE_T_CLEAN
#define PY_SSIZE_T_CLEAN
#endif /* PY_SSIZE_T_CLEAN */
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03030000)
    #error Cython requires Python 2.6+ or Python 3.3+.
#else
#define CYTHON_ABI "0_29_23"
#define CYTHON_HEX_VERSION 0x001D17F0
#define CYTHON_FUTURE_DIVISION 1
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#define __PYX_COMMA ,
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x02070000
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #if PY_VERSION_HEX < 0x03050000
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_PYTYPE_LOOKUP
  #define CYTHON_USE_PYTYPE_LOOKUP 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
  #undef CYTHON_PEP489_MULTI_PHASE_INIT
  #define CYTHON_PEP489_MULTI_PHASE_INIT 0
  #undef CYTHON_USE_TP_FINALIZE
  #define CYTHON_USE_TP_FINALIZE 0
  #undef CYTHON_USE_DICT_VERSIONS
  #define CYTHON_USE_DICT_VERSIONS 0
  #undef CYTHON_USE_EXC_INFO_STACK
  #define CYTHON_USE_EXC_INFO_STACK 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYTYPE_LOOKUP
    #define CYTHON_USE_PYTYPE_LOOKUP 0
  #elif !defined(CYTHON_USE_PYTYPE_LOOKUP)
    #define CYTHON_USE_PYTYPE_LOOKUP 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
  #ifndef CYTHON_PEP489_MULTI_PHASE_INIT
    #define CYTHON_PEP489_MULTI_PHASE_INIT (PY_VERSION_HEX >= 0x03050000)
  #endif
  #ifndef CYTHON_USE_TP_FINALIZE
    #define CYTHON_USE_TP_FINALIZE (PY_VERSION_HEX >= 0x030400a1)
  #endif
  #ifndef CYTHON_USE_DICT_VERSIONS
    #define CYTHON_USE_DICT_VERSIONS (PY_VERSION_HEX >= 0x030600B1)
  #endif
  #ifndef CYTHON_USE_EXC_INFO_STACK
    #define CYTHON_USE_EXC_INFO_STACK (PY_VERSION_HEX >= 0x030700A3)
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
  #ifdef SIZEOF_VOID_P
    enum { __pyx_check_sizeof_voidp = 1 / (int)(SIZEOF_VOID_P == sizeof(void*)) };
  #endif
#endif
#ifndef __has_attribute
  #define __has_attribute(x) 0
#endif
#ifndef __has_cpp_attribute
  #define __has_cpp_attribute(x) 0
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)
#ifdef _MSC_VER
    #ifndef _MSC_STDINT_H_
        #if _MSC_VER < 1300
           typedef unsigned char     uint8_t;
           typedef unsigned int      uint32_t;
        #else
           typedef unsigned __int8   uint8_t;
           typedef unsigned __int32  uint32_t;
        #endif
    #endif
#else
   #include <stdint.h>
#endif
#ifndef CYTHON_FALLTHROUGH
  #if defined(__cplusplus) && __cplusplus >= 201103L
    #if __has_cpp_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH [[fallthrough]]
    #elif __has_cpp_attribute(clang::fallthrough)
      #define CYTHON_FALLTHROUGH [[clang::fallthrough]]
    #elif __has_cpp_attribute(gnu::fallthrough)
      #define CYTHON_FALLTHROUGH [[gnu::fallthrough]]
    #endif
  #endif
  #ifndef CYTHON_FALLTHROUGH
    #if __has_attribute(fallthrough)
      #define CYTHON_FALLTHROUGH __attribute__((fallthrough))
    #else
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
  #if defined(__clang__ ) && defined(__apple_build_version__)
    #if __apple_build_version__ < 7000000
      #undef  CYTHON_FALLTHROUGH
      #define CYTHON_FALLTHROUGH
    #endif
  #endif
#endif

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
#if PY_VERSION_HEX >= 0x030800A4 && PY_VERSION_HEX < 0x030800B2
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, 0, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#else
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
#endif
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_STACKLESS
  #define METH_STACKLESS 0
#endif
#if PY_VERSION_HEX <= 0x030700A3 || !defined(METH_FASTCALL)
  #ifndef METH_FASTCALL
     #define METH_FASTCALL 0x80
  #endif
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject *const *args, Py_ssize_t nargs);
  typedef PyObject *(*__Pyx_PyCFunctionFastWithKeywords) (PyObject *self, PyObject *const *args,
                                                          Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
  #define __Pyx_PyCFunctionFastWithKeywords _PyCFunctionFastWithKeywords
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX < 0x030400A1
  #define PyMem_RawMalloc(n)           PyMem_Malloc(n)
  #define PyMem_RawRealloc(p, n)       PyMem_Realloc(p, n)
  #define PyMem_RawFree(p)             PyMem_Free(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#if !CYTHON_FAST_THREAD_STATE || PY_VERSION_HEX < 0x02070000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#elif PY_VERSION_HEX >= 0x03060000
  #define __Pyx_PyThreadState_Current _PyThreadState_UncheckedGet()
#elif PY_VERSION_HEX >= 0x03000000
  #define __Pyx_PyThreadState_Current PyThreadState_GET()
#else
  #define __Pyx_PyThreadState_Current _PyThreadState_Current
#endif
#if PY_VERSION_HEX < 0x030700A2 && !defined(PyThread_tss_create) && !defined(Py_tss_NEEDS_INIT)
#include "pythread.h"
#define Py_tss_NEEDS_INIT 0
typedef int Py_tss_t;
static CYTHON_INLINE int PyThread_tss_create(Py_tss_t *key) {
  *key = PyThread_create_key();
  return 0;
}
static CYTHON_INLINE Py_tss_t * PyThread_tss_alloc(void) {
  Py_tss_t *key = (Py_tss_t *)PyObject_Malloc(sizeof(Py_tss_t));
  *key = Py_tss_NEEDS_INIT;
  return key;
}
static CYTHON_INLINE void PyThread_tss_free(Py_tss_t *key) {
  PyObject_Free(key);
}
static CYTHON_INLINE int PyThread_tss_is_created(Py_tss_t *key) {
  return *key != Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE void PyThread_tss_delete(Py_tss_t *key) {
  PyThread_delete_key(*key);
  *key = Py_tss_NEEDS_INIT;
}
static CYTHON_INLINE int PyThread_tss_set(Py_tss_t *key, void *value) {
  return PyThread_set_key_value(*key, value);
}
static CYTHON_INLINE void * PyThread_tss_get(Py_tss_t *key) {
  return PyThread_get_key_value(*key);
}
#endif
#if CYTHON_COMPILING_IN_CPYTHON || defined(_PyDict_NewPresized)
#define __Pyx_PyDict_NewPresized(n)  ((n <= 8) ? PyDict_New() : _PyDict_NewPresized(n))
#else
#define __Pyx_PyDict_NewPresized(n)  PyDict_New()
#endif
#if PY_MAJOR_VERSION >= 3 || CYTHON_FUTURE_DIVISION
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1 && CYTHON_USE_UNICODE_INTERNALS
#define __Pyx_PyDict_GetItemStr(dict, name)  _PyDict_GetItem_KnownHash(dict, name, ((PyASCIIObject *) name)->hash)
#else
#define __Pyx_PyDict_GetItemStr(dict, name)  PyDict_GetItem(dict, name)
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #if defined(PyUnicode_IS_READY) && defined(PyUnicode_GET_SIZE)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
  #else
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_LENGTH(u))
  #endif
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None || (PyString_Check(b) && !PyString_CheckExact(b)))) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None || (PyUnicode_Check(b) && !PyUnicode_CheckExact(b)))) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#ifndef PyObject_Unicode
  #define PyObject_Unicode             PyObject_Str
#endif
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#if PY_VERSION_HEX >= 0x030900A4
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_SET_REFCNT(obj, refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SET_SIZE(obj, size)
#else
  #define __Pyx_SET_REFCNT(obj, refcnt) Py_REFCNT(obj) = (refcnt)
  #define __Pyx_SET_SIZE(obj, size) Py_SIZE(obj) = (size)
#endif
#if CYTHON_ASSUME_SAFE_MACROS
  #define __Pyx_PySequence_SIZE(seq)  Py_SIZE(seq)
#else
  #define __Pyx_PySequence_SIZE(seq)  PySequence_Size(seq)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? ((void)(klass), PyMethod_New(func, self)) : __Pyx_NewRef(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef __Pyx_PyAsyncMethodsStruct
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif

#define __PYX_MARK_ERR_POS(f_index, lineno) \
    { __pyx_filename = __pyx_f[f_index]; (void)__pyx_filename; __pyx_lineno = lineno; (void)__pyx_lineno; __pyx_clineno = __LINE__; (void)__pyx_clineno; }
#define __PYX_ERR(f_index, lineno, Ln_error) \
    { __PYX_MARK_ERR_POS(f_index, lineno) goto Ln_error; }

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__nes__cycore__mos6502
#define __PYX_HAVE_API__nes__cycore__mos6502
/* Early includes */
#include "pythread.h"
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include "pystate.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#if defined(PYREX_WITHOUT_ASSERTIONS) && !defined(CYTHON_WITHOUT_ASSERTIONS)
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_UTF8 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT (PY_MAJOR_VERSION >= 3 && __PYX_DEFAULT_STRING_ENCODING_IS_UTF8)
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
static CYTHON_INLINE int __Pyx_is_valid_index(Py_ssize_t i, Py_ssize_t limit) {
    return (size_t) i < (size_t) limit;
}
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER)
    #define __Pyx_sst_abs(value) ((Py_ssize_t)_abs64(value))
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyBytes_AsWritableString(s)     ((char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableSString(s)    ((signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsWritableUString(s)    ((unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsString(s)     ((const char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsSString(s)    ((const signed char*) PyBytes_AS_STRING(s))
#define __Pyx_PyBytes_AsUString(s)    ((const unsigned char*) PyBytes_AS_STRING(s))
#define __Pyx_PyObject_AsWritableString(s)    ((char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsWritableUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsSString(s)    ((const signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((const unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u) {
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b);
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
#define __Pyx_PySequence_Tuple(obj)\
    (likely(PyTuple_CheckExact(obj)) ? __Pyx_NewRef(obj) : PySequence_Tuple(obj))
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c) + 1);
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */
static CYTHON_INLINE void __Pyx_pretend_to_initialize(void* ptr) { (void)ptr; }

static PyObject *__pyx_m = NULL;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_cython_runtime = NULL;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "nes/cycore/mos6502.pyx",
  "stringsource",
  "nes/cycore/apu.pxd",
  "nes/cycore/carts.pxd",
  "nes/cycore/system.pxd",
  "nes/cycore/ppu.pxd",
  "nes/cycore/memory.pxd",
};
/* ForceInitThreads.proto */
#ifndef __PYX_FORCE_INIT_THREADS
  #define __PYX_FORCE_INIT_THREADS 0
#endif

/* NoFastGil.proto */
#define __Pyx_PyGILState_Ensure PyGILState_Ensure
#define __Pyx_PyGILState_Release PyGILState_Release
#define __Pyx_FastGIL_Remember()
#define __Pyx_FastGIL_Forget()
#define __Pyx_FastGilFuncInit()

/* BufferFormatStructs.proto */
#define IS_UNSIGNED(type) (((type) -1) > 0)
struct __Pyx_StructField_;
#define __PYX_BUF_FLAGS_PACKED_STRUCT (1 << 0)
typedef struct {
  const char* name;
  struct __Pyx_StructField_* fields;
  size_t size;
  size_t arraysize[8];
  int ndim;
  char typegroup;
  char is_unsigned;
  int flags;
} __Pyx_TypeInfo;
typedef struct __Pyx_StructField_ {
  __Pyx_TypeInfo* type;
  const char* name;
  size_t offset;
} __Pyx_StructField;
typedef struct {
  __Pyx_StructField* field;
  size_t parent_offset;
} __Pyx_BufFmt_StackElem;
typedef struct {
  __Pyx_StructField root;
  __Pyx_BufFmt_StackElem* head;
  size_t fmt_offset;
  size_t new_count, enc_count;
  size_t struct_alignment;
  int is_complex;
  char enc_type;
  char new_packmode;
  char enc_packmode;
  char is_valid_array;
} __Pyx_BufFmt_Context;

/* Atomics.proto */
#include <pythread.h>
#ifndef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 1
#endif
#define __pyx_atomic_int_type int
#if CYTHON_ATOMICS && __GNUC__ >= 4 && (__GNUC_MINOR__ > 1 ||\
                    (__GNUC_MINOR__ == 1 && __GNUC_PATCHLEVEL >= 2)) &&\
                    !defined(__i386__)
    #define __pyx_atomic_incr_aligned(value, lock) __sync_fetch_and_add(value, 1)
    #define __pyx_atomic_decr_aligned(value, lock) __sync_fetch_and_sub(value, 1)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using GNU atomics"
    #endif
#elif CYTHON_ATOMICS && defined(_MSC_VER) && 0
    #include <Windows.h>
    #undef __pyx_atomic_int_type
    #define __pyx_atomic_int_type LONG
    #define __pyx_atomic_incr_aligned(value, lock) InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #pragma message ("Using MSVC atomics")
    #endif
#elif CYTHON_ATOMICS && (defined(__ICC) || defined(__INTEL_COMPILER)) && 0
    #define __pyx_atomic_incr_aligned(value, lock) _InterlockedIncrement(value)
    #define __pyx_atomic_decr_aligned(value, lock) _InterlockedDecrement(value)
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Using Intel atomics"
    #endif
#else
    #undef CYTHON_ATOMICS
    #define CYTHON_ATOMICS 0
    #ifdef __PYX_DEBUG_ATOMICS
        #warning "Not using atomics"
    #endif
#endif
typedef volatile __pyx_atomic_int_type __pyx_atomic_int;
#if CYTHON_ATOMICS
    #define __pyx_add_acquisition_count(memview)\
             __pyx_atomic_incr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_atomic_decr_aligned(__pyx_get_slice_count_pointer(memview), memview->lock)
#else
    #define __pyx_add_acquisition_count(memview)\
            __pyx_add_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
    #define __pyx_sub_acquisition_count(memview)\
            __pyx_sub_acquisition_count_locked(__pyx_get_slice_count_pointer(memview), memview->lock)
#endif

/* MemviewSliceStruct.proto */
struct __pyx_memoryview_obj;
typedef struct {
  struct __pyx_memoryview_obj *memview;
  char *data;
  Py_ssize_t shape[8];
  Py_ssize_t strides[8];
  Py_ssize_t suboffsets[8];
} __Pyx_memviewslice;
#define __Pyx_MemoryView_Len(m)  (m.shape[0])


/*--- Type declarations ---*/
struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope;
struct __pyx_obj_3nes_6cycore_3apu_APUUnit;
struct __pyx_obj_3nes_6cycore_3apu_APUTriangle;
struct __pyx_obj_3nes_6cycore_3apu_APUPulse;
struct __pyx_obj_3nes_6cycore_3apu_APUNoise;
struct __pyx_obj_3nes_6cycore_3apu_APUDMC;
struct __pyx_obj_3nes_6cycore_3apu_NESAPU;
struct __pyx_obj_3nes_6cycore_5carts_NESCart;
struct __pyx_obj_3nes_6cycore_5carts_NESCart0;
struct __pyx_obj_3nes_6cycore_5carts_NESCart1;
struct __pyx_obj_3nes_6cycore_5carts_NESCart2;
struct __pyx_obj_3nes_6cycore_5carts_NESCart4;
struct __pyx_obj_3nes_6cycore_6system_InterruptListener;
struct __pyx_obj_3nes_6cycore_6system_NES;
struct __pyx_obj_3nes_6cycore_3ppu_NESPPU;
struct __pyx_obj_3nes_6cycore_6memory_MemoryBase;
struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM;
struct __pyx_obj_3nes_6cycore_6memory_NESVRAM;
struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502;
struct __pyx_array_obj;
struct __pyx_MemviewEnum_obj;
struct __pyx_memoryview_obj;
struct __pyx_memoryviewslice_obj;

/* "apu.pxd":4
 * from .memory cimport NESMappedRAM
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # sound synthesis constants
 *     SAMPLE_RATE = 48000     # 48kHz sample rate
 */
enum  {
  __pyx_e_3nes_6cycore_3apu_SAMPLE_RATE = 0xBB80,
  __pyx_e_3nes_6cycore_3apu_SAMPLE_SCALE = 0x10000,
  __pyx_e_3nes_6cycore_3apu_SAMPLE_OFFSET = 0x8000,
  __pyx_e_3nes_6cycore_3apu_CPU_FREQ_HZ = 0x1B4F4D,
  __pyx_e_3nes_6cycore_3apu_MAX_CPU_CYCLES_PER_LOOP = 24,
  __pyx_e_3nes_6cycore_3apu_STATUS = 0x4015,
  __pyx_e_3nes_6cycore_3apu_FRAME_COUNTER = 0x4017,
  __pyx_e_3nes_6cycore_3apu_FOUR_STEP = 0,
  __pyx_e_3nes_6cycore_3apu_FIVE_STEP = 1,
  __pyx_e_3nes_6cycore_3apu_BIT_DMC_INTERRUPT = 7,
  __pyx_e_3nes_6cycore_3apu_BIT_FRAME_INTERRUPT = 6,
  __pyx_e_3nes_6cycore_3apu_BIT_DMC_ACTIVE = 4,
  __pyx_e_3nes_6cycore_3apu_BIT_LENGTH_NOISE = 3,
  __pyx_e_3nes_6cycore_3apu_BIT_LENGTH_TRIANGLE = 2,
  __pyx_e_3nes_6cycore_3apu_BIT_LENGTH_PULSE2 = 1,
  __pyx_e_3nes_6cycore_3apu_BIT_LENGTH_PULSE1 = 0,
  __pyx_e_3nes_6cycore_3apu_BIT_ENABLE_DMC = 4,
  __pyx_e_3nes_6cycore_3apu_BIT_ENABLE_NOISE = 3,
  __pyx_e_3nes_6cycore_3apu_BIT_ENABLE_TRIANGLE = 2,
  __pyx_e_3nes_6cycore_3apu_BIT_ENABLE_PULSE2 = 1,
  __pyx_e_3nes_6cycore_3apu_BIT_ENABLE_PULSE1 = 0,
  __pyx_e_3nes_6cycore_3apu_BIT_MODE = 7,
  __pyx_e_3nes_6cycore_3apu_BIT_IRQ_INHIBIT = 6,
  __pyx_e_3nes_6cycore_3apu_NUM_ENVELOPES = 3,
  __pyx_e_3nes_6cycore_3apu_PULSE0 = 0,
  __pyx_e_3nes_6cycore_3apu_PULSE1 = 1,
  __pyx_e_3nes_6cycore_3apu_NOISE = 2,
  __pyx_e_3nes_6cycore_3apu_APU_BUFFER_LENGTH = 0x10000,
  __pyx_e_3nes_6cycore_3apu_CHUNK_SIZE = 0x2710
};

/* "carts.pxd":7
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     BYTES_PER_KB = 1024
 * 
 */
enum  {
  __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB = 0x400
};

/* "carts.pxd":10
 *     BYTES_PER_KB = 1024
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # cart memory map - note that some memory is connected to CPU bus, others to PPU bus
 *     RAM_START = 0x6000      # address in the CPU memory space
 */
enum  {
  __pyx_e_3nes_6cycore_5carts_RAM_START = 0x6000,
  __pyx_e_3nes_6cycore_5carts_PRG_ROM_START = 0x8000,
  __pyx_e_3nes_6cycore_5carts_CHR_ROM_START = 0x0000
};

/* "carts.pxd":35
 * # would be nicer to have these in a named enum, but that seems to generate Cython compile error "Array dimension not
 * # integer", so it's not clear what that is ending up doing.  For now, just name the constants with the mapper number.
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
 *     M0_MAX_CART_RAM_SIZE = 8 * BYTES_PER_KB   # up to 8kB of RAM
 */
enum  {

  /* "carts.pxd":39
 *     M0_MAX_CART_RAM_SIZE = 8 * BYTES_PER_KB   # up to 8kB of RAM
 *     M0_MAX_PRG_ROM_SIZE = 32 * BYTES_PER_KB   # up to 32kB of program ROM
 *     M0_CHR_MEM_SIZE = 8 * BYTES_PER_KB        # always (?) 8kB of chr memory (can be RAM or always ROM?)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_e_3nes_6cycore_5carts_M0_MAX_CART_RAM_SIZE = (8 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M0_MAX_PRG_ROM_SIZE = (32 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M0_CHR_MEM_SIZE = (8 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB)
};

/* "carts.pxd":61
 * # ref: https://wiki.nesdev.com/w/index.php/MMC1
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
 *     M1_PRG_RAM_BANK_SIZE = 8 * BYTES_PER_KB    # 8kB RAM banks (can be zero of these)
 */
enum  {

  /* "carts.pxd":67
 *     M1_PRG_ROM_BANK_SIZE = 16 * BYTES_PER_KB   # Each bank is 16kB of program ROM
 *     M1_MAX_PRG_BANKS = 16                      # up to 256kB of program ROM (16x16kb) - 512kb MMC1 ROMs not supported
 *     M1_CHR_ROM_BANK_SIZE = 4 * BYTES_PER_KB    # chr rom in 4kB banks             # <<<<<<<<<<<<<<
 *     M1_MAX_CHR_BANKS = 32                      # up to 128kb of CHR rom in (up to 32 x 4kb)
 * 
 */
  __pyx_e_3nes_6cycore_5carts_M1_PRG_RAM_BANK_SIZE = (8 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M1_MAX_PRG_RAM_BANKS = 4,
  __pyx_e_3nes_6cycore_5carts_M1_PRG_ROM_BANK_SIZE = (16 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M1_MAX_PRG_BANKS = 16,
  __pyx_e_3nes_6cycore_5carts_M1_CHR_ROM_BANK_SIZE = (4 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M1_MAX_CHR_BANKS = 32,
  __pyx_e_3nes_6cycore_5carts_M1_PRG_RAM_START = 0x6000,
  __pyx_e_3nes_6cycore_5carts_M1_PRG_ROM_BANK0_START = 0x8000,
  __pyx_e_3nes_6cycore_5carts_M1_PRG_ROM_BANK1_START = 0xC000,
  __pyx_e_3nes_6cycore_5carts_M1_CTRL_REG_START = 0x8000,
  __pyx_e_3nes_6cycore_5carts_M1_CHR_REG_0_START = 0xA000,
  __pyx_e_3nes_6cycore_5carts_M1_CHR_REG_1_START = 0xC000,
  __pyx_e_3nes_6cycore_5carts_M1_PRG_REG_START = 0xE000,
  __pyx_e_3nes_6cycore_5carts_M1_CHR_ROM_BANK0_START = 0x0000,
  __pyx_e_3nes_6cycore_5carts_M1_CHR_ROM_BANK1_START = 0x1000
};

/* "carts.pxd":124
 * # ref: https://wiki.nesdev.com/w/index.php/UxROM
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
 *     M2_MAX_PRG_BANKS = 16                      # Max number of banks; up to 8 in UNROM, up to 16 in UOROM
 */
enum  {

  /* "carts.pxd":127
 *     # maximum memory sizes; these are allocated but might not all be used (but are so small it doesn't matter)
 *     M2_MAX_PRG_BANKS = 16                      # Max number of banks; up to 8 in UNROM, up to 16 in UOROM
 *     M2_PRG_ROM_BANK_SIZE = 16 * BYTES_PER_KB   # Each bank is 16kB of program ROM             # <<<<<<<<<<<<<<
 * 
 *     # memory map
 */
  __pyx_e_3nes_6cycore_5carts_M2_MAX_PRG_BANKS = 16,
  __pyx_e_3nes_6cycore_5carts_M2_PRG_ROM_BANK_SIZE = (16 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M2_FIXED_PRG_ROM_START = 0xC000
};

/* "carts.pxd":157
 * # ref: https://wiki.nesdev.com/w/index.php/MMC3
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     M4_PRG_RAM_SIZE = 8 * BYTES_PER_KB
 *     M4_PRG_ROM_BANK_SIZE = 8 * BYTES_PER_KB
 */
enum  {

  /* "carts.pxd":160
 *     M4_PRG_RAM_SIZE = 8 * BYTES_PER_KB
 *     M4_PRG_ROM_BANK_SIZE = 8 * BYTES_PER_KB
 *     M4_CHR_ROM_BANK_SIZE = 1 * BYTES_PER_KB             # <<<<<<<<<<<<<<
 * 
 *     M4_MAX_PRG_BANKS = 64
 */
  __pyx_e_3nes_6cycore_5carts_M4_PRG_RAM_SIZE = (8 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M4_PRG_ROM_BANK_SIZE = (8 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M4_CHR_ROM_BANK_SIZE = (1 * __pyx_e_3nes_6cycore_5carts_BYTES_PER_KB),
  __pyx_e_3nes_6cycore_5carts_M4_MAX_PRG_BANKS = 64,
  __pyx_e_3nes_6cycore_5carts_M4_MAX_CHR_BANKS = 0x100,
  __pyx_e_3nes_6cycore_5carts_M4_PRG_RAM_START = 0x6000,
  __pyx_e_3nes_6cycore_5carts_M4_PRG_ROM_START = 0x8000,
  __pyx_e_3nes_6cycore_5carts_BANK_REG_START = 0x8000,
  __pyx_e_3nes_6cycore_5carts_MIRROR_PROTECT_REG_START = 0xA000,
  __pyx_e_3nes_6cycore_5carts_IRQ_LATCH_RELOAD_REG_START = 0xC000,
  __pyx_e_3nes_6cycore_5carts_IRQ_ACTIVATE_START = 0xE000
};
struct __pyx_opt_args_3nes_6cycore_6system_3NES_run_frame_headless;

/* "system.pxd":8
 * from .mos6502 cimport MOS6502
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     OAM_DMA = 1
 *     DMC_DMA = 2
 */
enum  {
  __pyx_e_3nes_6cycore_6system_OAM_DMA = 1,
  __pyx_e_3nes_6cycore_6system_DMC_DMA = 2,
  __pyx_e_3nes_6cycore_6system_DMC_DMA_DURING_OAM_DMA = 3
};

/* "system.pxd":29
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     PPU_CYCLES_PER_CPU_CYCLE = 3
 *     TARGET_AUDIO_BUFFER_SAMPLES = 2400   # increase this if you get frequent audio glitches, decrease if sound is laggy
 */
enum  {
  __pyx_e_3nes_6cycore_6system_PPU_CYCLES_PER_CPU_CYCLE = 3,
  __pyx_e_3nes_6cycore_6system_TARGET_AUDIO_BUFFER_SAMPLES = 0x960,
  __pyx_e_3nes_6cycore_6system_AUDIO_CHUNK_SAMPLES = 0x190,
  __pyx_e_3nes_6cycore_6system_MAX_RATE_DELTA = 0xBB8,
  __pyx_e_3nes_6cycore_6system_TARGET_FPS = 60,
  __pyx_e_3nes_6cycore_6system_SYNC_NONE = 0,
  __pyx_e_3nes_6cycore_6system_SYNC_AUDIO = 1,
  __pyx_e_3nes_6cycore_6system_SYNC_PYGAME = 2,
  __pyx_e_3nes_6cycore_6system_SYNC_VSYNC = 3
};

/* "system.pxd":61
 *     cdef int step(self, int log_cpu)
 *     cpdef void run(self)
 *     cpdef object run_frame_headless(self, int run_frames=?, object controller1_state=?, object controller2_state=?)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_3nes_6cycore_6system_3NES_run_frame_headless {
  int __pyx_n;
  int run_frames;
  PyObject *controller1_state;
  PyObject *controller2_state;
};
struct __pyx_opt_args_3nes_6cycore_3ppu_6NESPPU_copy_screen_buffer_to;

/* "ppu.pxd":8
 * # this odd mechanism allows (integer) constants to be shared between pyx files via pxd files
 * # if these constants also need to be used from python use cpdef in place of cdef here
 * cdef enum:             # <<<<<<<<<<<<<<
 *     NUM_PPU_REGISTERS = 8
 * 
 */
enum  {
  __pyx_e_3nes_6cycore_3ppu_NUM_PPU_REGISTERS = 8,
  __pyx_e_3nes_6cycore_3ppu_PPU_CTRL = 0,
  __pyx_e_3nes_6cycore_3ppu_PPU_MASK = 1,
  __pyx_e_3nes_6cycore_3ppu_PPU_STATUS = 2,
  __pyx_e_3nes_6cycore_3ppu_OAM_ADDR = 3,
  __pyx_e_3nes_6cycore_3ppu_OAM_DATA = 4,
  __pyx_e_3nes_6cycore_3ppu_PPU_SCROLL = 5,
  __pyx_e_3nes_6cycore_3ppu_PPU_ADDR = 6,
  __pyx_e_3nes_6cycore_3ppu_PPU_DATA = 7,
  __pyx_e_3nes_6cycore_3ppu_VBLANK_MASK = 128,
  __pyx_e_3nes_6cycore_3ppu_SPRITE0_HIT_MASK = 64,
  __pyx_e_3nes_6cycore_3ppu_SPRITE_OVERFLOW_MASK = 32,
  __pyx_e_3nes_6cycore_3ppu_SPRITE_SIZE_MASK = 32,
  __pyx_e_3nes_6cycore_3ppu_BKG_PATTERN_TABLE_MASK = 16,
  __pyx_e_3nes_6cycore_3ppu_SPRITE_PATTERN_TABLE_MASK = 8,
  __pyx_e_3nes_6cycore_3ppu_VRAM_INCREMENT_MASK = 4,
  __pyx_e_3nes_6cycore_3ppu_BIT_NAMETABLE_X = 0,
  __pyx_e_3nes_6cycore_3ppu_BIT_NAMETABLE_Y = 1,
  __pyx_e_3nes_6cycore_3ppu_RENDERING_ENABLED_MASK = 24,
  __pyx_e_3nes_6cycore_3ppu_RENDER_SPRITES_MASK = 16,
  __pyx_e_3nes_6cycore_3ppu_RENDER_BACKGROUND_MASK = 8,
  __pyx_e_3nes_6cycore_3ppu_V_BLANK_BIT = 7,
  __pyx_e_3nes_6cycore_3ppu_RENDER_LEFT8_BKG_BIT = 1,
  __pyx_e_3nes_6cycore_3ppu_RENDER_LEFT8_SPRITES_BIT = 2,
  __pyx_e_3nes_6cycore_3ppu_PPU_SCROLL_X = 0,
  __pyx_e_3nes_6cycore_3ppu_PPU_SCROLL_Y = 1,
  __pyx_e_3nes_6cycore_3ppu_OAM_SIZE_BYTES = 0x100,
  __pyx_e_3nes_6cycore_3ppu_PIXELS_PER_LINE = 0x155,
  __pyx_e_3nes_6cycore_3ppu_SCREEN_WIDTH_PX = 0x100,
  __pyx_e_3nes_6cycore_3ppu_SCREEN_HEIGHT_PX = 0xF0,
  __pyx_e_3nes_6cycore_3ppu_VERTICAL_OVERSCAN_PX = 8,
  __pyx_e_3nes_6cycore_3ppu_HORIZONTAL_OVERSCAN_PX = 8,
  __pyx_e_3nes_6cycore_3ppu_PRERENDER_LINE = 0x105,
  __pyx_e_3nes_6cycore_3ppu_SCREEN_TILE_ROWS = 30,
  __pyx_e_3nes_6cycore_3ppu_SCREEN_TILE_COLS = 32,
  __pyx_e_3nes_6cycore_3ppu_TILE_HEIGHT_PX = 8,
  __pyx_e_3nes_6cycore_3ppu_TILE_WIDTH_PX = 8,
  __pyx_e_3nes_6cycore_3ppu_PATTERN_SIZE_BYTES = 16
};

/* "ppu.pxd":134
 *     # screen buffer copy and clear
 *     cdef int get_background_color(self)
 *     cpdef void copy_screen_buffer_to(self, unsigned int[:, :] dest, bint v_overscan=?, bint h_overscan=?)             # <<<<<<<<<<<<<<
 *     cdef void _clear_to_bkg(self)
 * 
 */
struct __pyx_opt_args_3nes_6cycore_3ppu_6NESPPU_copy_screen_buffer_to {
  int __pyx_n;
  int v_overscan;
  int h_overscan;
};

/* "memory.pxd":20
 * ######## NES Main Memoroy #############################################
 * 
 * cpdef enum:             # <<<<<<<<<<<<<<
 *     RAM_SIZE = 2048             # 2kB of internal RAM
 * 
 */
enum  {
  __pyx_e_3nes_6cycore_6memory_RAM_SIZE = 0x800,
  __pyx_e_3nes_6cycore_6memory_RAM_END = 0x2000,
  __pyx_e_3nes_6cycore_6memory_PPU_END = 0x4000,
  __pyx_e_3nes_6cycore_6memory_APU_END = 0x4018,
  __pyx_e_3nes_6cycore_6memory_APU_UNUSED_END = 0x4020,
  __pyx_e_3nes_6cycore_6memory_OAM_DMA_REG = 0x4014,
  __pyx_e_3nes_6cycore_6memory_CONTROLLER1 = 0x4016,
  __pyx_e_3nes_6cycore_6memory_CONTROLLER2 = 0x4017,
  __pyx_e_3nes_6cycore_6memory_CART_START = 0x4020
};

/* "memory.pxd":52
 * ######## NES VRAM #####################################################
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 *     PATTERN_TABLE_SIZE_BYTES = 4096   # provided by the rom
 * 
 */
enum  {
  __pyx_e_3nes_6cycore_6memory_PATTERN_TABLE_SIZE_BYTES = 0x1000,
  __pyx_e_3nes_6cycore_6memory_XX_NAMETABLES_SIZE_BYTES = 0x1000,
  __pyx_e_3nes_6cycore_6memory_PALETTE_SIZE_BYTES = 32,
  __pyx_e_3nes_6cycore_6memory_NAMETABLE_LENGTH_BYTES = 0x400,
  __pyx_e_3nes_6cycore_6memory_PALETTE_START = 0x3F00,
  __pyx_e_3nes_6cycore_6memory_NAMETABLE_START = 0x2000,
  __pyx_e_3nes_6cycore_6memory_ATTRIBUTE_TABLE_OFFSET = 0x3C0
};

/* "nes/cycore/mos6502.pxd":6
 * 
 * 
 * cdef enum:             # <<<<<<<<<<<<<<
 * 
 *     # Masks for the bits in the status register
 */
enum  {
  __pyx_e_3nes_6cycore_7mos6502_SR_N_MASK = 128,
  __pyx_e_3nes_6cycore_7mos6502_SR_V_MASK = 64,
  __pyx_e_3nes_6cycore_7mos6502_SR_X_MASK = 32,
  __pyx_e_3nes_6cycore_7mos6502_SR_B_MASK = 16,
  __pyx_e_3nes_6cycore_7mos6502_SR_D_MASK = 8,
  __pyx_e_3nes_6cycore_7mos6502_SR_I_MASK = 4,
  __pyx_e_3nes_6cycore_7mos6502_SR_Z_MASK = 2,
  __pyx_e_3nes_6cycore_7mos6502_SR_C_MASK = 1,
  __pyx_e_3nes_6cycore_7mos6502_STACK_PAGE = 0x0100,
  __pyx_e_3nes_6cycore_7mos6502_IRQ_BRK_VECTOR_ADDR = 0xFFFE,
  __pyx_e_3nes_6cycore_7mos6502_RESET_VECTOR_ADDR = 0xFFFC,
  __pyx_e_3nes_6cycore_7mos6502_NMI_VECTOR_ADDR = 0xFFFA,
  __pyx_e_3nes_6cycore_7mos6502_LO_BYTE = 0,
  __pyx_e_3nes_6cycore_7mos6502_HI_BYTE = 1,
  __pyx_e_3nes_6cycore_7mos6502_INTERRUPT_REQUEST_CYCLES = 7,
  __pyx_e_3nes_6cycore_7mos6502_OAM_DMA_CPU_CYCLES = 0x201,
  __pyx_e_3nes_6cycore_7mos6502_ARG_NONE = -1L
};

/* "apu.pxd":51
 * 
 * 
 * cdef class APUEnvelope:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint start_flag, loop_flag
 */
struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_3apu_APUEnvelope *__pyx_vtab;
  int start_flag;
  int loop_flag;
  unsigned int decay_level;
  unsigned int divider;
  unsigned int volume;
};


/* "apu.pxd":60
 * 
 * 
 * cdef class APUUnit:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint enable, ctr_halt
 */
struct __pyx_obj_3nes_6cycore_3apu_APUUnit {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit *__pyx_vtab;
  int enable;
  int ctr_halt;
  int length_ctr;
  short output[__pyx_e_3nes_6cycore_3apu_SAMPLE_RATE];
};


/* "apu.pxd":71
 * 
 * 
 * cdef class APUTriangle(APUUnit):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint linear_reload_flag
 */
struct __pyx_obj_3nes_6cycore_3apu_APUTriangle {
  struct __pyx_obj_3nes_6cycore_3apu_APUUnit __pyx_base;
  int linear_reload_flag;
  int linear_reload_value;
  int period;
  int linear_ctr;
  double phase;
};


/* "apu.pxd":83
 * 
 * 
 * cdef class APUPulse(APUUnit):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint constant_volume, is_unit_1
 */
struct __pyx_obj_3nes_6cycore_3apu_APUPulse {
  struct __pyx_obj_3nes_6cycore_3apu_APUUnit __pyx_base;
  int constant_volume;
  int is_unit_1;
  int period;
  int adjusted_period;
  int duty;
  double phase;
  struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope *env;
  int sweep_enable;
  int sweep_negate;
  int sweep_reload;
  int sweep_period;
  int sweep_shift;
  int sweep_divider;
  int duty_waveform[4][8];
};


/* "apu.pxd":101
 * 
 * 
 * cdef class APUNoise(APUUnit):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint constant_volume, mode
 */
struct __pyx_obj_3nes_6cycore_3apu_APUNoise {
  struct __pyx_obj_3nes_6cycore_3apu_APUUnit __pyx_base;
  int constant_volume;
  int mode;
  int period;
  int feedback;
  int timer;
  struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope *env;
  unsigned int timer_table[16];
};


/* "apu.pxd":117
 * 
 * 
 * cdef class APUDMC(APUUnit):             # <<<<<<<<<<<<<<
 *     """
 *     The DMC unit is pretty different to the other APU units
 */
struct __pyx_obj_3nes_6cycore_3apu_APUDMC {
  struct __pyx_obj_3nes_6cycore_3apu_APUUnit __pyx_base;
  struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *memory;
  struct __pyx_obj_3nes_6cycore_6system_InterruptListener *interrupt_listener;
  int irq_enable;
  int loop_flag;
  int silence;
  int interrupt_flag;
  unsigned int sample_address;
  unsigned int sample_length;
  unsigned int address;
  unsigned int bytes_remaining;
  unsigned int rate;
  unsigned int timer;
  int output_level;
  int bits_remaining;
  unsigned char sample;
  unsigned int rate_table[16];
};


/* "apu.pxd":139
 * 
 * 
 * cdef class NESAPU:             # <<<<<<<<<<<<<<
 *     """
 *     References:
 */
struct __pyx_obj_3nes_6cycore_3apu_NESAPU {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_3apu_NESAPU *__pyx_vtab;
  double master_volume;
  int cycles;
  int rate;
  int frame_segment;
  int _reset_timer_in;
  double samples_per_cycle;
  double samples_required;
  unsigned PY_LONG_LONG _buffer_start;
  unsigned PY_LONG_LONG _buffer_end;
  struct __pyx_obj_3nes_6cycore_6system_InterruptListener *interrupt_listener;
  short output[__pyx_e_3nes_6cycore_3apu_APU_BUFFER_LENGTH];
  short buffer[__pyx_e_3nes_6cycore_3apu_CHUNK_SIZE];
  int mode;
  int irq_inhibit;
  int frame_interrupt_flag;
  struct __pyx_obj_3nes_6cycore_3apu_APUTriangle *triangle;
  struct __pyx_obj_3nes_6cycore_3apu_APUPulse *pulse1;
  struct __pyx_obj_3nes_6cycore_3apu_APUPulse *pulse2;
  struct __pyx_obj_3nes_6cycore_3apu_APUNoise *noise;
  struct __pyx_obj_3nes_6cycore_3apu_APUDMC *dmc;
  int frame_counter;
};


/* "carts.pxd":17
 * 
 * 
 * cdef class NESCart:             # <<<<<<<<<<<<<<
 *     """
 *     NES Cartridge interface
 */
struct __pyx_obj_3nes_6cycore_5carts_NESCart {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart *__pyx_vtab;
  int nametable_mirror_pattern[4];
};


/* "carts.pxd":42
 * 
 * 
 * cdef class NESCart0(NESCart):             # <<<<<<<<<<<<<<
 *     """
 *     Basic NES Cartridge (Type 0 / MMC0)
 */
struct __pyx_obj_3nes_6cycore_5carts_NESCart0 {
  struct __pyx_obj_3nes_6cycore_5carts_NESCart __pyx_base;
  unsigned char ram[__pyx_e_3nes_6cycore_5carts_M0_MAX_CART_RAM_SIZE];
  unsigned char prg_rom[__pyx_e_3nes_6cycore_5carts_M0_MAX_PRG_ROM_SIZE];
  unsigned char chr_mem[__pyx_e_3nes_6cycore_5carts_M0_CHR_MEM_SIZE];
  int prg_start_addr;
  int prg_rom_size;
  int ram_size;
  int chr_mem_writeable;
  int prg_rom_writeable;
};


/* "carts.pxd":98
 * 
 * 
 * cdef class NESCart1(NESCart):             # <<<<<<<<<<<<<<
 *     cdef int num_prg_banks, num_chr_banks, num_prg_ram_banks   # number of available memory banks in this cart
 *     cdef bint chr_mem_writeable                 # whether or not CHR memory is writeable (i.e. RAM)
 */
struct __pyx_obj_3nes_6cycore_5carts_NESCart1 {
  struct __pyx_obj_3nes_6cycore_5carts_NESCart __pyx_base;
  int num_prg_banks;
  int num_chr_banks;
  int num_prg_ram_banks;
  int chr_mem_writeable;
  unsigned char ctrl;
  unsigned char chr_bank[2];
  unsigned char prg_bank;
  unsigned char prg_ram_bank;
  unsigned char shift;
  int shift_ctr;
  unsigned char banked_prg_rom[__pyx_e_3nes_6cycore_5carts_M1_MAX_PRG_BANKS][__pyx_e_3nes_6cycore_5carts_M1_PRG_ROM_BANK_SIZE];
  unsigned char banked_chr_rom[__pyx_e_3nes_6cycore_5carts_M1_MAX_CHR_BANKS][__pyx_e_3nes_6cycore_5carts_M1_CHR_ROM_BANK_SIZE];
  unsigned char ram[__pyx_e_3nes_6cycore_5carts_M1_MAX_PRG_RAM_BANKS][__pyx_e_3nes_6cycore_5carts_M1_PRG_RAM_BANK_SIZE];
};


/* "carts.pxd":135
 * 
 * 
 * cdef class NESCart2(NESCart0):             # <<<<<<<<<<<<<<
 *     """
 *     NES Cartridge Type 2 (UNROM / UOROM)
 */
struct __pyx_obj_3nes_6cycore_5carts_NESCart2 {
  struct __pyx_obj_3nes_6cycore_5carts_NESCart0 __pyx_base;
  unsigned char banked_prg_rom[__pyx_e_3nes_6cycore_5carts_M2_MAX_PRG_BANKS][__pyx_e_3nes_6cycore_5carts_M2_PRG_ROM_BANK_SIZE];
  unsigned char prg_bank;
  unsigned char num_prg_banks;
  int emulate_bus_conflicts;
};


/* "carts.pxd":177
 * 
 * 
 * cdef class NESCart4(NESCart):             # <<<<<<<<<<<<<<
 *     cdef InterruptListener interrupt_listener
 *     cdef int num_prg_banks, num_chr_banks       # number of available memory banks in this cart
 */
struct __pyx_obj_3nes_6cycore_5carts_NESCart4 {
  struct __pyx_obj_3nes_6cycore_5carts_NESCart __pyx_base;
  struct __pyx_obj_3nes_6cycore_6system_InterruptListener *interrupt_listener;
  int num_prg_banks;
  int num_chr_banks;
  int chr_mem_writeable;
  unsigned char bank_register[8];
  int chr_a12_inversion;
  int prg_bank_mode;
  unsigned char bank_select;
  int prg_ram_enable;
  int prg_ram_protect;
  int mirror_pattern_fixed;
  unsigned char irq_reload_value;
  int irq_reload;
  int irq_enabled;
  int irq_counter;
  unsigned char banked_prg_rom[__pyx_e_3nes_6cycore_5carts_M4_MAX_PRG_BANKS][__pyx_e_3nes_6cycore_5carts_M4_PRG_ROM_BANK_SIZE];
  unsigned char banked_chr_rom[__pyx_e_3nes_6cycore_5carts_M4_MAX_CHR_BANKS][__pyx_e_3nes_6cycore_5carts_M4_CHR_ROM_BANK_SIZE];
  unsigned char ram[__pyx_e_3nes_6cycore_5carts_M4_PRG_RAM_SIZE];
};


/* "system.pxd":14
 * 
 * 
 * cdef class InterruptListener:             # <<<<<<<<<<<<<<
 *     cdef bint _nmi, _irq
 *     cdef int dma_pause, dma_pause_count
 */
struct __pyx_obj_3nes_6cycore_6system_InterruptListener {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_6system_InterruptListener *__pyx_vtab;
  int _nmi;
  int _irq;
  int dma_pause;
  int dma_pause_count;
};


/* "system.pxd":43
 * 
 * 
 * cdef class NES:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_obj_3nes_6cycore_6system_NES {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_6system_NES *__pyx_vtab;
  struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *ppu;
  struct __pyx_obj_3nes_6cycore_3apu_NESAPU *apu;
  struct __pyx_obj_3nes_6cycore_5carts_NESCart *cart;
  struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *cpu;
  struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *memory;
  struct __pyx_obj_3nes_6cycore_6system_InterruptListener *interrupt_listener;
  PyObject *controller1;
  PyObject *controller2;
  PyObject *screen;
  int screen_scale;
  int sync_mode;
  int v_overscan;
  int h_overscan;
};


/* "ppu.pxd":77
 * 
 * ### NES PPU prototype ##################################################################################################
 * cdef class NESPPU:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # ppu registers
 */
struct __pyx_obj_3nes_6cycore_3ppu_NESPPU {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_3ppu_NESPPU *__pyx_vtab;
  unsigned char ppu_ctrl;
  unsigned char ppu_mask;
  unsigned char oam_addr;
  unsigned char oam_data;
  unsigned char _ppu_data_buffer;
  unsigned char _io_latch;
  unsigned char ppu_scroll[2];
  unsigned int ppu_addr;
  unsigned int _ppu_byte_latch;
  int line;
  int pixel;
  int in_vblank;
  int sprite_zero_hit;
  int sprite_overflow;
  int ignore_ppu_ctrl;
  int frames_since_reset;
  int time_at_new_frame;
  PY_LONG_LONG cycles_since_reset;
  unsigned char oam[__pyx_e_3nes_6cycore_3ppu_OAM_SIZE_BYTES];
  unsigned char _oam[32];
  int _active_sprite_addrs[8];
  unsigned char _sprite_bkg_priority[8];
  int _sprite_line[8];
  char _sprite_pattern[8][8];
  int _num_active_sprites;
  int irq_tick_triggers[68];
  unsigned int _pattern_lo;
  unsigned int _pattern_hi;
  int _effective_x;
  int _effective_y;
  int _palette[2][4];
  struct __pyx_obj_3nes_6cycore_6memory_NESVRAM *vram;
  struct __pyx_obj_3nes_6cycore_6system_InterruptListener *interrupt_listener;
  unsigned int screen_buffer[0x100][0xF0];
  unsigned int hex_palette[64];
  int transparent_color;
  int bkg_color;
  int _palette_cache[8][4];
  int _palette_cache_valid[8];
};


/* "memory.pxd":13
 * ######## memory base ##################################################
 * 
 * cdef class MemoryBase:             # <<<<<<<<<<<<<<
 *     cdef unsigned char read(self, int address)
 *     cdef void write(self, int address, unsigned char value)
 */
struct __pyx_obj_3nes_6cycore_6memory_MemoryBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_6memory_MemoryBase *__pyx_vtab;
};


/* "memory.pxd":34
 * 
 * 
 * cdef class NESMappedRAM(MemoryBase):             # <<<<<<<<<<<<<<
 *     cdef unsigned char ram[RAM_SIZE]
 *     cdef unsigned char _last_bus
 */
struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM {
  struct __pyx_obj_3nes_6cycore_6memory_MemoryBase __pyx_base;
  unsigned char ram[__pyx_e_3nes_6cycore_6memory_RAM_SIZE];
  unsigned char _last_bus;
  struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *ppu;
  struct __pyx_obj_3nes_6cycore_3apu_NESAPU *apu;
  struct __pyx_obj_3nes_6cycore_5carts_NESCart *cart;
  struct __pyx_obj_3nes_6cycore_6system_InterruptListener *interrupt_listener;
  PyObject *controller1;
  PyObject *controller2;
};


/* "memory.pxd":68
 *     ATTRIBUTE_TABLE_OFFSET = 0x3C0  # offset of the attribute table from the start of the corresponding nametable
 * 
 * cdef class NESVRAM(MemoryBase):             # <<<<<<<<<<<<<<
 *     cdef unsigned char _nametables[XX_NAMETABLES_SIZE_BYTES]
 *     cdef unsigned char palette_ram[PALETTE_SIZE_BYTES]
 */
struct __pyx_obj_3nes_6cycore_6memory_NESVRAM {
  struct __pyx_obj_3nes_6cycore_6memory_MemoryBase __pyx_base;
  unsigned char _nametables[__pyx_e_3nes_6cycore_6memory_XX_NAMETABLES_SIZE_BYTES];
  unsigned char palette_ram[__pyx_e_3nes_6cycore_6memory_PALETTE_SIZE_BYTES];
  struct __pyx_obj_3nes_6cycore_5carts_NESCart *cart;
  int nametable_mirror_pattern[4];
};


/* "nes/cycore/mos6502.pxd":36
 * 
 * 
 * cdef class MOS6502:             # <<<<<<<<<<<<<<
 *     cdef NESMappedRAM memory        # the main memory; important that type is decalared so can use c-calling
 *     cdef object instructions        # the instructions in the original python format; still used for logging etc.
 */
struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 {
  PyObject_HEAD
  struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *__pyx_vtab;
  struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *memory;
  PyObject *instructions;
  unsigned char A;
  unsigned char X;
  unsigned char Y;
  int PC;
  int SP;
  int N;
  int V;
  int D;
  int I;
  int Z;
  int C;
  PY_LONG_LONG cycles_since_reset;
  int aax_sets_flags;
  int undocumented_support_level;
  int stack_underflow_causes_exception;
  int instr_size_bytes[0x100];
};


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */
struct __pyx_array_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_array *__pyx_vtab;
  char *data;
  Py_ssize_t len;
  char *format;
  int ndim;
  Py_ssize_t *_shape;
  Py_ssize_t *_strides;
  Py_ssize_t itemsize;
  PyObject *mode;
  PyObject *_format;
  void (*callback_free_data)(void *);
  int free_data;
  int dtype_is_object;
};


/* "View.MemoryView":279
 * 
 * @cname('__pyx_MemviewEnum')
 * cdef class Enum(object):             # <<<<<<<<<<<<<<
 *     cdef object name
 *     def __init__(self, name):
 */
struct __pyx_MemviewEnum_obj {
  PyObject_HEAD
  PyObject *name;
};


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */
struct __pyx_memoryview_obj {
  PyObject_HEAD
  struct __pyx_vtabstruct_memoryview *__pyx_vtab;
  PyObject *obj;
  PyObject *_size;
  PyObject *_array_interface;
  PyThread_type_lock lock;
  __pyx_atomic_int acquisition_count[2];
  __pyx_atomic_int *acquisition_count_aligned_p;
  Py_buffer view;
  int flags;
  int dtype_is_object;
  __Pyx_TypeInfo *typeinfo;
};


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */
struct __pyx_memoryviewslice_obj {
  struct __pyx_memoryview_obj __pyx_base;
  __Pyx_memviewslice from_slice;
  PyObject *from_object;
  PyObject *(*to_object_func)(char *);
  int (*to_dtype_func)(char *, PyObject *);
};



/* "apu.pxd":51
 * 
 * 
 * cdef class APUEnvelope:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint start_flag, loop_flag
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_APUEnvelope {
  void (*update)(struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope *);
  void (*restart)(struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope *);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_APUEnvelope *__pyx_vtabptr_3nes_6cycore_3apu_APUEnvelope;


/* "apu.pxd":60
 * 
 * 
 * cdef class APUUnit:             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint enable, ctr_halt
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit {
  void (*update_length_ctr)(struct __pyx_obj_3nes_6cycore_3apu_APUUnit *);
  void (*set_enable)(struct __pyx_obj_3nes_6cycore_3apu_APUUnit *, int);
  void (*set_length_ctr)(struct __pyx_obj_3nes_6cycore_3apu_APUUnit *, int);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit *__pyx_vtabptr_3nes_6cycore_3apu_APUUnit;


/* "apu.pxd":71
 * 
 * 
 * cdef class APUTriangle(APUUnit):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint linear_reload_flag
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_APUTriangle {
  struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit __pyx_base;
  void (*write_register)(struct __pyx_obj_3nes_6cycore_3apu_APUTriangle *, int, unsigned char);
  void (*quarter_frame)(struct __pyx_obj_3nes_6cycore_3apu_APUTriangle *);
  void (*half_frame)(struct __pyx_obj_3nes_6cycore_3apu_APUTriangle *);
  int (*generate_sample)(struct __pyx_obj_3nes_6cycore_3apu_APUTriangle *);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_APUTriangle *__pyx_vtabptr_3nes_6cycore_3apu_APUTriangle;


/* "apu.pxd":83
 * 
 * 
 * cdef class APUPulse(APUUnit):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint constant_volume, is_unit_1
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_APUPulse {
  struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit __pyx_base;
  void (*write_register)(struct __pyx_obj_3nes_6cycore_3apu_APUPulse *, int, unsigned char);
  void (*sweep_update)(struct __pyx_obj_3nes_6cycore_3apu_APUPulse *);
  void (*quarter_frame)(struct __pyx_obj_3nes_6cycore_3apu_APUPulse *);
  void (*half_frame)(struct __pyx_obj_3nes_6cycore_3apu_APUPulse *);
  int (*generate_sample)(struct __pyx_obj_3nes_6cycore_3apu_APUPulse *);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_APUPulse *__pyx_vtabptr_3nes_6cycore_3apu_APUPulse;


/* "apu.pxd":101
 * 
 * 
 * cdef class APUNoise(APUUnit):             # <<<<<<<<<<<<<<
 *     cdef:
 *         bint constant_volume, mode
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_APUNoise {
  struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit __pyx_base;
  void (*write_register)(struct __pyx_obj_3nes_6cycore_3apu_APUNoise *, int, unsigned char);
  void (*update_cycles)(struct __pyx_obj_3nes_6cycore_3apu_APUNoise *, int);
  void (*quarter_frame)(struct __pyx_obj_3nes_6cycore_3apu_APUNoise *);
  void (*half_frame)(struct __pyx_obj_3nes_6cycore_3apu_APUNoise *);
  int (*generate_sample)(struct __pyx_obj_3nes_6cycore_3apu_APUNoise *);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_APUNoise *__pyx_vtabptr_3nes_6cycore_3apu_APUNoise;


/* "apu.pxd":117
 * 
 * 
 * cdef class APUDMC(APUUnit):             # <<<<<<<<<<<<<<
 *     """
 *     The DMC unit is pretty different to the other APU units
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_APUDMC {
  struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit __pyx_base;
  void (*write_register)(struct __pyx_obj_3nes_6cycore_3apu_APUDMC *, int, unsigned char);
  void (*update_cycles)(struct __pyx_obj_3nes_6cycore_3apu_APUDMC *, int);
  void (*read_advance)(struct __pyx_obj_3nes_6cycore_3apu_APUDMC *);
  int (*generate_sample)(struct __pyx_obj_3nes_6cycore_3apu_APUDMC *);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_APUDMC *__pyx_vtabptr_3nes_6cycore_3apu_APUDMC;


/* "apu.pxd":139
 * 
 * 
 * cdef class NESAPU:             # <<<<<<<<<<<<<<
 *     """
 *     References:
 */

struct __pyx_vtabstruct_3nes_6cycore_3apu_NESAPU {
  unsigned char (*read_register)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int);
  void (*write_register)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int, unsigned char);
  void (*_set_status)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, unsigned char);
  int (*run_cycles)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int);
  void (*quarter_frame_tick)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *);
  void (*half_frame_tick)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *);
  int (*mix)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int, int, int, int, int);
  void (*generate_sample)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *);
  __Pyx_memviewslice (*get_sound)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int, int __pyx_skip_dispatch);
  void (*set_volume)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, float, int __pyx_skip_dispatch);
  int (*buffer_remaining)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int __pyx_skip_dispatch);
  void (*set_rate)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int, int __pyx_skip_dispatch);
  int (*get_rate)(struct __pyx_obj_3nes_6cycore_3apu_NESAPU *, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_3nes_6cycore_3apu_NESAPU *__pyx_vtabptr_3nes_6cycore_3apu_NESAPU;


/* "carts.pxd":17
 * 
 * 
 * cdef class NESCart:             # <<<<<<<<<<<<<<
 *     """
 *     NES Cartridge interface
 */

struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart {
  unsigned char (*read)(struct __pyx_obj_3nes_6cycore_5carts_NESCart *, int);
  void (*write)(struct __pyx_obj_3nes_6cycore_5carts_NESCart *, int, unsigned char);
  unsigned char (*read_ppu)(struct __pyx_obj_3nes_6cycore_5carts_NESCart *, int);
  void (*write_ppu)(struct __pyx_obj_3nes_6cycore_5carts_NESCart *, int, unsigned char);
  void (*irq_tick)(struct __pyx_obj_3nes_6cycore_5carts_NESCart *);
};
static struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart *__pyx_vtabptr_3nes_6cycore_5carts_NESCart;


/* "carts.pxd":42
 * 
 * 
 * cdef class NESCart0(NESCart):             # <<<<<<<<<<<<<<
 *     """
 *     Basic NES Cartridge (Type 0 / MMC0)
 */

struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart0 {
  struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart __pyx_base;
};
static struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart0 *__pyx_vtabptr_3nes_6cycore_5carts_NESCart0;


/* "carts.pxd":98
 * 
 * 
 * cdef class NESCart1(NESCart):             # <<<<<<<<<<<<<<
 *     cdef int num_prg_banks, num_chr_banks, num_prg_ram_banks   # number of available memory banks in this cart
 *     cdef bint chr_mem_writeable                 # whether or not CHR memory is writeable (i.e. RAM)
 */

struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart1 {
  struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart __pyx_base;
  void (*_write_shift)(struct __pyx_obj_3nes_6cycore_5carts_NESCart1 *, int, unsigned char);
  void (*_set_nametable_mirror_pattern)(struct __pyx_obj_3nes_6cycore_5carts_NESCart1 *);
  int (*_get_chr_bank)(struct __pyx_obj_3nes_6cycore_5carts_NESCart1 *, int);
};
static struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart1 *__pyx_vtabptr_3nes_6cycore_5carts_NESCart1;


/* "carts.pxd":135
 * 
 * 
 * cdef class NESCart2(NESCart0):             # <<<<<<<<<<<<<<
 *     """
 *     NES Cartridge Type 2 (UNROM / UOROM)
 */

struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart2 {
  struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart0 __pyx_base;
};
static struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart2 *__pyx_vtabptr_3nes_6cycore_5carts_NESCart2;


/* "carts.pxd":177
 * 
 * 
 * cdef class NESCart4(NESCart):             # <<<<<<<<<<<<<<
 *     cdef InterruptListener interrupt_listener
 *     cdef int num_prg_banks, num_chr_banks       # number of available memory banks in this cart
 */

struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart4 {
  struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart __pyx_base;
  unsigned int (*_get_ppu_bank)(struct __pyx_obj_3nes_6cycore_5carts_NESCart4 *, int);
};
static struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart4 *__pyx_vtabptr_3nes_6cycore_5carts_NESCart4;


/* "system.pxd":14
 * 
 * 
 * cdef class InterruptListener:             # <<<<<<<<<<<<<<
 *     cdef bint _nmi, _irq
 *     cdef int dma_pause, dma_pause_count
 */

struct __pyx_vtabstruct_3nes_6cycore_6system_InterruptListener {
  void (*raise_nmi)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  void (*reset_nmi)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  void (*raise_irq)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  void (*reset_irq)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  void (*reset_dma_pause)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  void (*raise_dma_pause)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *, int);
  int (*any_active)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  int (*nmi_active)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
  int (*irq_active)(struct __pyx_obj_3nes_6cycore_6system_InterruptListener *);
};
static struct __pyx_vtabstruct_3nes_6cycore_6system_InterruptListener *__pyx_vtabptr_3nes_6cycore_6system_InterruptListener;


/* "system.pxd":43
 * 
 * 
 * cdef class NES:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_3nes_6cycore_6system_NES {
  int (*step)(struct __pyx_obj_3nes_6cycore_6system_NES *, int);
  void (*run)(struct __pyx_obj_3nes_6cycore_6system_NES *, int __pyx_skip_dispatch);
  PyObject *(*run_frame_headless)(struct __pyx_obj_3nes_6cycore_6system_NES *, int __pyx_skip_dispatch, struct __pyx_opt_args_3nes_6cycore_6system_3NES_run_frame_headless *__pyx_optional_args);
};
static struct __pyx_vtabstruct_3nes_6cycore_6system_NES *__pyx_vtabptr_3nes_6cycore_6system_NES;


/* "ppu.pxd":77
 * 
 * ### NES PPU prototype ##################################################################################################
 * cdef class NESPPU:             # <<<<<<<<<<<<<<
 *     cdef:
 *         # ppu registers
 */

struct __pyx_vtabstruct_3nes_6cycore_3ppu_NESPPU {
  void (*write_oam)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, unsigned char *);
  int (*get_background_color)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*copy_screen_buffer_to)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, __Pyx_memviewslice, int __pyx_skip_dispatch, struct __pyx_opt_args_3nes_6cycore_3ppu_6NESPPU_copy_screen_buffer_to *__pyx_optional_args);
  void (*_clear_to_bkg)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  unsigned char (*read_register)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, int);
  void (*write_register)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, int, unsigned char);
  unsigned char (*ppu_status)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  int (*run_cycles)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, int);
  void (*prerender_scanline)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*render_visible_scanline)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*increment_pixel)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*_increment_vram_address)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*_trigger_nmi)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*_reset_effective_y)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*_prefetch_active_sprites)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*_fill_sprite_latches)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, int);
  int (*_overlay_sprites)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, int);
  void (*fill_bkg_latches)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  int (*_get_bkg_pixel)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*decode_palette)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, int *, int, int);
  void (*invalidate_palette_cache)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *);
  void (*debug_render_nametables)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, __Pyx_memviewslice, int __pyx_skip_dispatch);
  void (*debug_render_tile)(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU *, __Pyx_memviewslice, int, int, int, int, int, int, int, int __pyx_skip_dispatch);
};
static struct __pyx_vtabstruct_3nes_6cycore_3ppu_NESPPU *__pyx_vtabptr_3nes_6cycore_3ppu_NESPPU;


/* "memory.pxd":13
 * ######## memory base ##################################################
 * 
 * cdef class MemoryBase:             # <<<<<<<<<<<<<<
 *     cdef unsigned char read(self, int address)
 *     cdef void write(self, int address, unsigned char value)
 */

struct __pyx_vtabstruct_3nes_6cycore_6memory_MemoryBase {
  unsigned char (*read)(struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *, int);
  void (*write)(struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *, int, unsigned char);
};
static struct __pyx_vtabstruct_3nes_6cycore_6memory_MemoryBase *__pyx_vtabptr_3nes_6cycore_6memory_MemoryBase;


/* "memory.pxd":34
 * 
 * 
 * cdef class NESMappedRAM(MemoryBase):             # <<<<<<<<<<<<<<
 *     cdef unsigned char ram[RAM_SIZE]
 *     cdef unsigned char _last_bus
 */

struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM {
  struct __pyx_vtabstruct_3nes_6cycore_6memory_MemoryBase __pyx_base;
  void (*run_oam_dma)(struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *, int);
};
static struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *__pyx_vtabptr_3nes_6cycore_6memory_NESMappedRAM;


/* "memory.pxd":68
 *     ATTRIBUTE_TABLE_OFFSET = 0x3C0  # offset of the attribute table from the start of the corresponding nametable
 * 
 * cdef class NESVRAM(MemoryBase):             # <<<<<<<<<<<<<<
 *     cdef unsigned char _nametables[XX_NAMETABLES_SIZE_BYTES]
 *     cdef unsigned char palette_ram[PALETTE_SIZE_BYTES]
 */

struct __pyx_vtabstruct_3nes_6cycore_6memory_NESVRAM {
  struct __pyx_vtabstruct_3nes_6cycore_6memory_MemoryBase __pyx_base;
};
static struct __pyx_vtabstruct_3nes_6cycore_6memory_NESVRAM *__pyx_vtabptr_3nes_6cycore_6memory_NESVRAM;


/* "nes/cycore/mos6502.pyx":6
 * from nes.instructions import INSTRUCTION_SET, NamedInstruction, AddressModes
 * 
 * cdef class MOS6502:             # <<<<<<<<<<<<<<
 *     """
 *     Software emulator for MOS Technologies 6502 CPU.  The chip in the NES is a modified version of the 6502 that does
 */

struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 {
  void (*make_instruction_data_tables)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *);
  int (*run_instr)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char, unsigned char *);
  int (*dma_pause)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_from_le)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char *);
  int (*_read_word)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*trigger_nmi)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *);
  int (*trigger_irq)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *);
  int (*run_next_instr)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *);
  unsigned char (*_status_to_byte)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int);
  void (*_status_from_byte)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char);
  void (*push_stack)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char);
  unsigned char (*pop_stack)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *);
  int (*_neg)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int);
  int (*_from_2sc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char);
  void (*_set_zn)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int);
  int (*_adc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_and)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_asl)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_jump_relative)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bcc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bcs)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_beq)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bit)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bmi)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bne)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bpl)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  PyObject *(*_interrupt)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_brk)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bvc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_bvs)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_clc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_cld)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_cli)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_clv)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  void (*_compare)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_cmp)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_cpx)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_cpy)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_dec)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_dex)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_dey)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_eor)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_inc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_inx)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_iny)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_jmp)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_jsr)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_lda)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_ldx)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_ldy)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_lsr)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_nop)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_ora)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_pha)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_php)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_pla)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_plp)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_rol)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_ror)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_rti)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_rts)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_sbc)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_sec)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_sed)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_sei)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_sta)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_stx)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_sty)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_tax)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_tay)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_tsx)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_txa)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_txs)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
  int (*_tya)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int);
};
static struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *__pyx_vtabptr_3nes_6cycore_7mos6502_MOS6502;


/* "View.MemoryView":105
 * 
 * @cname("__pyx_array")
 * cdef class array:             # <<<<<<<<<<<<<<
 * 
 *     cdef:
 */

struct __pyx_vtabstruct_array {
  PyObject *(*get_memview)(struct __pyx_array_obj *);
};
static struct __pyx_vtabstruct_array *__pyx_vtabptr_array;


/* "View.MemoryView":330
 * 
 * @cname('__pyx_memoryview')
 * cdef class memoryview(object):             # <<<<<<<<<<<<<<
 * 
 *     cdef object obj
 */

struct __pyx_vtabstruct_memoryview {
  char *(*get_item_pointer)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*is_slice)(struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_slice_assignment)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*setitem_slice_assign_scalar)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *);
  PyObject *(*setitem_indexed)(struct __pyx_memoryview_obj *, PyObject *, PyObject *);
  PyObject *(*convert_item_to_object)(struct __pyx_memoryview_obj *, char *);
  PyObject *(*assign_item_from_object)(struct __pyx_memoryview_obj *, char *, PyObject *);
};
static struct __pyx_vtabstruct_memoryview *__pyx_vtabptr_memoryview;


/* "View.MemoryView":965
 * 
 * @cname('__pyx_memoryviewslice')
 * cdef class _memoryviewslice(memoryview):             # <<<<<<<<<<<<<<
 *     "Internal class for passing memoryview slices to Python"
 * 
 */

struct __pyx_vtabstruct__memoryviewslice {
  struct __pyx_vtabstruct_memoryview __pyx_base;
};
static struct __pyx_vtabstruct__memoryviewslice *__pyx_vtabptr__memoryviewslice;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#define __Pyx_BUILD_ASSERT_EXPR(cond)\
    (sizeof(char [1 - 2*!(cond)]) - 1)
#ifndef Py_MEMBER_SIZE
#define Py_MEMBER_SIZE(type, member) sizeof(((type *)0)->member)
#endif
  static size_t __pyx_pyframe_localsplus_offset = 0;
  #include "frameobject.h"
  #define __Pxy_PyFrame_Initialize_Offsets()\
    ((void)__Pyx_BUILD_ASSERT_EXPR(sizeof(PyFrameObject) == offsetof(PyFrameObject, f_localsplus) + Py_MEMBER_SIZE(PyFrameObject, f_localsplus)),\
     (void)(__pyx_pyframe_localsplus_offset = ((size_t)PyFrame_Type.tp_basicsize) - Py_MEMBER_SIZE(PyFrameObject, f_localsplus)))
  #define __Pyx_PyFrame_GetLocalsplus(frame)\
    (assert(__pyx_pyframe_localsplus_offset), (PyObject **)(((char *)(frame)) + __pyx_pyframe_localsplus_offset))
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyObjectGetMethod.proto */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
#define __Pyx_unpack_tuple2(tuple, value1, value2, is_tuple, has_known_size, decref_tuple)\
    (likely(is_tuple || PyTuple_Check(tuple)) ?\
        (likely(has_known_size || PyTuple_GET_SIZE(tuple) == 2) ?\
            __Pyx_unpack_tuple2_exact(tuple, value1, value2, decref_tuple) :\
            (__Pyx_UnpackTupleError(tuple, 2), -1)) :\
        __Pyx_unpack_tuple2_generic(tuple, value1, value2, has_known_size, decref_tuple))
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
    PyObject* tuple, PyObject** value1, PyObject** value2, int decref_tuple);
static int __Pyx_unpack_tuple2_generic(
    PyObject* tuple, PyObject** value1, PyObject** value2, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* PyDictVersioning.proto */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
#define __PYX_DICT_VERSION_INIT  ((PY_UINT64_T) -1)
#define __PYX_GET_DICT_VERSION(dict)  (((PyDictObject*)(dict))->ma_version_tag)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)\
    (version_var) = __PYX_GET_DICT_VERSION(dict);\
    (cache_var) = (value);
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP) {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    if (likely(__PYX_GET_DICT_VERSION(DICT) == __pyx_dict_version)) {\
        (VAR) = __pyx_dict_cached_value;\
    } else {\
        (VAR) = __pyx_dict_cached_value = (LOOKUP);\
        __pyx_dict_version = __PYX_GET_DICT_VERSION(DICT);\
    }\
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj);
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj);
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version);
#else
#define __PYX_GET_DICT_VERSION(dict)  (0)
#define __PYX_UPDATE_DICT_CACHE(dict, value, cache_var, version_var)
#define __PYX_PY_DICT_LOOKUP_IF_MODIFIED(VAR, DICT, LOOKUP)  (VAR) = (LOOKUP);
#endif

/* GetModuleGlobalName.proto */
#if CYTHON_USE_DICT_VERSIONS
#define __Pyx_GetModuleGlobalName(var, name)  {\
    static PY_UINT64_T __pyx_dict_version = 0;\
    static PyObject *__pyx_dict_cached_value = NULL;\
    (var) = (likely(__pyx_dict_version == __PYX_GET_DICT_VERSION(__pyx_d))) ?\
        (likely(__pyx_dict_cached_value) ? __Pyx_NewRef(__pyx_dict_cached_value) : __Pyx_GetBuiltinName(name)) :\
        __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
#define __Pyx_GetModuleGlobalNameUncached(var, name)  {\
    PY_UINT64_T __pyx_dict_version;\
    PyObject *__pyx_dict_cached_value;\
    (var) = __Pyx__GetModuleGlobalName(name, &__pyx_dict_version, &__pyx_dict_cached_value);\
}
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value);
#else
#define __Pyx_GetModuleGlobalName(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
#define __Pyx_GetModuleGlobalNameUncached(var, name)  (var) = __Pyx__GetModuleGlobalName(name)
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = __Pyx_PyThreadState_Current;
#define __Pyx_PyErr_Occurred()  __pyx_tstate->curexc_type
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#define __Pyx_PyErr_Occurred()  PyErr_Occurred()
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_Clear() __Pyx_ErrRestore(NULL, NULL, NULL)
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_PyErr_SetNone(exc) (Py_INCREF(exc), __Pyx_ErrRestore((exc), NULL, NULL))
#else
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#endif
#else
#define __Pyx_PyErr_Clear() PyErr_Clear()
#define __Pyx_PyErr_SetNone(exc) PyErr_SetNone(exc)
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestoreInState(tstate, type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchInState(tstate, type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* WriteUnraisableException.proto */
static void __Pyx_WriteUnraisable(const char *name, int clineno,
                                  int lineno, const char *filename,
                                  int full_traceback, int nogil);

/* PyObjectCall2Args.proto */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AndObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAnd(op1, op2) : PyNumber_And(op1, op2))
#endif

/* ObjectGetItem.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key);
#else
#define __Pyx_PyObject_GetItem(obj, key)  PyObject_GetItem(obj, key)
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_SubtractObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceSubtract(op1, op2) : PyNumber_Subtract(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RshiftObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_RshiftObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceRshift(op1, op2) : PyNumber_Rshift(op1, op2))
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_OrObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_OrObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceOr(op1, op2) : PyNumber_Or(op1, op2))
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, long intval, int inplace, int zerodivision_check);
#else
#define __Pyx_PyInt_AddObjC(op1, op2, intval, inplace, zerodivision_check)\
    (inplace ? PyNumber_InPlaceAdd(op1, op2) : PyNumber_Add(op1, op2))
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* GetAttr3.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *, PyObject *, PyObject *);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* IncludeStringH.proto */
#include <string.h>

/* HasAttr.proto */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *, PyObject *);

/* GetTopmostException.proto */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem * __Pyx_PyErr_GetTopmostException(PyThreadState *tstate);
#endif

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* ArgTypeTest.proto */
#define __Pyx_ArgTypeTest(obj, type, none_allowed, name, exact)\
    ((likely((Py_TYPE(obj) == type) | (none_allowed && (obj == Py_None)))) ? 1 :\
        __Pyx__ArgTypeTest(obj, type, name, exact))
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact);

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* UnaryNegOverflows.proto */
#define UNARY_NEG_WOULD_OVERFLOW(x)\
        (((x) < 0) & ((unsigned long)(x) == 0-(unsigned long)(x)))

static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *); /*proto*/
/* decode_c_string_utf16.proto */
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 0;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16LE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = -1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}
static CYTHON_INLINE PyObject *__Pyx_PyUnicode_DecodeUTF16BE(const char *s, Py_ssize_t size, const char *errors) {
    int byteorder = 1;
    return PyUnicode_DecodeUTF16(s, size, errors, &byteorder);
}

/* decode_c_string.proto */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors));

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* FastTypeChecks.proto */
#if CYTHON_COMPILING_IN_CPYTHON
#define __Pyx_TypeCheck(obj, type) __Pyx_IsSubtype(Py_TYPE(obj), (PyTypeObject *)type)
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject *type);
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *type1, PyObject *type2);
#else
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyErr_GivenExceptionMatches(err, type) PyErr_GivenExceptionMatches(err, type)
#define __Pyx_PyErr_GivenExceptionMatches2(err, type1, type2) (PyErr_GivenExceptionMatches(err, type1) || PyErr_GivenExceptionMatches(err, type2))
#endif
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)

static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* ListExtend.proto */
static CYTHON_INLINE int __Pyx_PyList_Extend(PyObject* L, PyObject* v) {
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject* none = _PyList_Extend((PyListObject*)L, v);
    if (unlikely(!none))
        return -1;
    Py_DECREF(none);
    return 0;
#else
    return PyList_SetSlice(L, PY_SSIZE_T_MAX, PY_SSIZE_T_MAX, v);
#endif
}

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        __Pyx_SET_SIZE(list, len + 1);
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname);

/* PyObject_GenericGetAttrNoDict.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttrNoDict PyObject_GenericGetAttr
#endif

/* PyObject_GenericGetAttr.proto */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name);
#else
#define __Pyx_PyObject_GenericGetAttr PyObject_GenericGetAttr
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* PyObjectGetAttrStrNoError.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name);

/* SetupReduce.proto */
static int __Pyx_setup_reduce(PyObject* type_obj);

/* TypeImport.proto */
#ifndef __PYX_HAVE_RT_ImportType_proto
#define __PYX_HAVE_RT_ImportType_proto
enum __Pyx_ImportType_CheckSize {
   __Pyx_ImportType_CheckSize_Error = 0,
   __Pyx_ImportType_CheckSize_Warn = 1,
   __Pyx_ImportType_CheckSize_Ignore = 2
};
static PyTypeObject *__Pyx_ImportType(PyObject* module, const char *module_name, const char *class_name, size_t size, enum __Pyx_ImportType_CheckSize check_size);
#endif

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* CLineInTraceback.proto */
#ifdef CYTHON_CLINE_IN_TRACEBACK
#define __Pyx_CLineForTraceback(tstate, c_line)  (((CYTHON_CLINE_IN_TRACEBACK)) ? c_line : 0)
#else
static int __Pyx_CLineForTraceback(PyThreadState *tstate, int c_line);
#endif

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* GCCDiagnostics.proto */
#if defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 6))
#define __Pyx_HAS_GCC_DIAGNOSTIC
#endif

#if PY_MAJOR_VERSION < 3
    static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags);
    static void __Pyx_ReleaseBuffer(Py_buffer *view);
#else
    #define __Pyx_GetBuffer PyObject_GetBuffer
    #define __Pyx_ReleaseBuffer PyBuffer_Release
#endif


/* BufferStructDeclare.proto */
typedef struct {
  Py_ssize_t shape, strides, suboffsets;
} __Pyx_Buf_DimInfo;
typedef struct {
  size_t refcount;
  Py_buffer pybuffer;
} __Pyx_Buffer;
typedef struct {
  __Pyx_Buffer *rcbuffer;
  char *data;
  __Pyx_Buf_DimInfo diminfo[8];
} __Pyx_LocalBuf_ND;

/* MemviewSliceIsContig.proto */
static int __pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim);

/* OverlappingSlices.proto */
static int __pyx_slices_overlap(__Pyx_memviewslice *slice1,
                                __Pyx_memviewslice *slice2,
                                int ndim, size_t itemsize);

/* Capsule.proto */
static CYTHON_INLINE PyObject *__pyx_capsule_create(void *p, const char *sig);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value);

/* CIntFromPy.proto */
static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value);

/* CIntFromPy.proto */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* MemviewSliceCopyTemplate.proto */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object);

/* MemviewSliceInit.proto */
#define __Pyx_BUF_MAX_NDIMS %(BUF_MAX_NDIMS)d
#define __Pyx_MEMVIEW_DIRECT   1
#define __Pyx_MEMVIEW_PTR      2
#define __Pyx_MEMVIEW_FULL     4
#define __Pyx_MEMVIEW_CONTIG   8
#define __Pyx_MEMVIEW_STRIDED  16
#define __Pyx_MEMVIEW_FOLLOW   32
#define __Pyx_IS_C_CONTIG 1
#define __Pyx_IS_F_CONTIG 2
static int __Pyx_init_memviewslice(
                struct __pyx_memoryview_obj *memview,
                int ndim,
                __Pyx_memviewslice *memviewslice,
                int memview_is_new_reference);
static CYTHON_INLINE int __pyx_add_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
static CYTHON_INLINE int __pyx_sub_acquisition_count_locked(
    __pyx_atomic_int *acquisition_count, PyThread_type_lock lock);
#define __pyx_get_slice_count_pointer(memview) (memview->acquisition_count_aligned_p)
#define __pyx_get_slice_count(memview) (*__pyx_get_slice_count_pointer(memview))
#define __PYX_INC_MEMVIEW(slice, have_gil) __Pyx_INC_MEMVIEW(slice, have_gil, __LINE__)
#define __PYX_XDEC_MEMVIEW(slice, have_gil) __Pyx_XDEC_MEMVIEW(slice, have_gil, __LINE__)
static CYTHON_INLINE void __Pyx_INC_MEMVIEW(__Pyx_memviewslice *, int, int);
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *, int, int);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_dma_pause(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_pause_type, int __pyx_v_count); /* proto*/
static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502_make_instruction_data_tables(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__from_le(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char *__pyx_v_data); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__read_word(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, int __pyx_v_wrap_at_page); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_trigger_nmi(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_trigger_irq(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_run_next_instr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto*/
static unsigned char __pyx_f_3nes_6cycore_7mos6502_7MOS6502__status_to_byte(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_b_flag); /* proto*/
static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502__status_from_byte(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_sr_byte); /* proto*/
static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502_push_stack(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_v); /* proto*/
static unsigned char __pyx_f_3nes_6cycore_7mos6502_7MOS6502_pop_stack(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__neg(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_v); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__from_2sc(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_v); /* proto*/
static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502__set_zn(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_v); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__adc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__and(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__asl(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__jump_relative(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_condition, int __pyx_v_offset_2sc); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bcc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bcs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__beq(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bit(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bmi(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bne(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bpl(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static PyObject *__pyx_f_3nes_6cycore_7mos6502_7MOS6502__interrupt(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_interrupt_vector, int __pyx_v_is_brk); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__brk(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bvc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bvs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__clc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cld(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cli(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__clv(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502__compare(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_v0, int __pyx_v_v); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cmp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cpx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cpy(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__dec(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__dex(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__dey(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__eor(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__inc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__inx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__iny(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__jmp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__jsr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__lda(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ldx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ldy(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__lsr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__nop(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ora(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__pha(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__php(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__pla(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__plp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__rol(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ror(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__rti(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__rts(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sbc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sec(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sed(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sei(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sta(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__stx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sty(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tax(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tay(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tsx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__txa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__txs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tya(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___); /* proto*/
static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_run_instr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_opcode, unsigned char *__pyx_v_data); /* proto*/
static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self); /* proto*/
static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto*/
static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src); /* proto*/
static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/
static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp); /* proto*/
static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value); /* proto*/

/* Module declarations from 'nes.cycore.apu' */
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_APUEnvelope = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_APUUnit = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_APUTriangle = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_APUPulse = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_APUNoise = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_APUDMC = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_3apu_NESAPU = 0;

/* Module declarations from 'nes.cycore.carts' */
static PyTypeObject *__pyx_ptype_3nes_6cycore_5carts_NESCart = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_5carts_NESCart0 = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_5carts_NESCart1 = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_5carts_NESCart2 = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_5carts_NESCart4 = 0;

/* Module declarations from 'nes.cycore.system' */
static PyTypeObject *__pyx_ptype_3nes_6cycore_6system_InterruptListener = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_6system_NES = 0;

/* Module declarations from 'nes.cycore.ppu' */
static PyTypeObject *__pyx_ptype_3nes_6cycore_3ppu_NESPPU = 0;

/* Module declarations from 'nes.cycore.memory' */
static PyTypeObject *__pyx_ptype_3nes_6cycore_6memory_MemoryBase = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_6memory_NESMappedRAM = 0;
static PyTypeObject *__pyx_ptype_3nes_6cycore_6memory_NESVRAM = 0;

/* Module declarations from 'nes.cycore.mos6502' */
static PyTypeObject *__pyx_ptype_3nes_6cycore_7mos6502_MOS6502 = 0;
static PyTypeObject *__pyx_array_type = 0;
static PyTypeObject *__pyx_MemviewEnum_type = 0;
static PyTypeObject *__pyx_memoryview_type = 0;
static PyTypeObject *__pyx_memoryviewslice_type = 0;
static PyObject *generic = 0;
static PyObject *strided = 0;
static PyObject *indirect = 0;
static PyObject *contiguous = 0;
static PyObject *indirect_contiguous = 0;
static int __pyx_memoryview_thread_locks_used;
static PyThread_type_lock __pyx_memoryview_thread_locks[8];
static PyObject *__pyx_f_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502__set_state(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, PyObject *); /*proto*/
static CYTHON_INLINE PyObject *__Pyx_carray_to_py_int(int *, Py_ssize_t); /*proto*/
static CYTHON_INLINE PyObject *__Pyx_carray_to_tuple_int(int *, Py_ssize_t); /*proto*/
static int __Pyx_carray_from_py_int(PyObject *, int *, Py_ssize_t); /*proto*/
static struct __pyx_array_obj *__pyx_array_new(PyObject *, Py_ssize_t, char *, char *, char *); /*proto*/
static void *__pyx_align_pointer(void *, size_t); /*proto*/
static PyObject *__pyx_memoryview_new(PyObject *, int, int, __Pyx_TypeInfo *); /*proto*/
static CYTHON_INLINE int __pyx_memoryview_check(PyObject *); /*proto*/
static PyObject *_unellipsify(PyObject *, int); /*proto*/
static PyObject *assert_direct_dimensions(Py_ssize_t *, int); /*proto*/
static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *, PyObject *); /*proto*/
static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int *, Py_ssize_t, Py_ssize_t, Py_ssize_t, int, int, int, int); /*proto*/
static char *__pyx_pybuffer_index(Py_buffer *, char *, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memslice_transpose(__Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice, int, PyObject *(*)(char *), int (*)(char *, PyObject *), int); /*proto*/
static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *); /*proto*/
static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *, __Pyx_memviewslice *); /*proto*/
static Py_ssize_t abs_py_ssize_t(Py_ssize_t); /*proto*/
static char __pyx_get_best_slice_order(__Pyx_memviewslice *, int); /*proto*/
static void _copy_strided_to_strided(char *, Py_ssize_t *, char *, Py_ssize_t *, Py_ssize_t *, Py_ssize_t *, int, size_t); /*proto*/
static void copy_strided_to_strided(__Pyx_memviewslice *, __Pyx_memviewslice *, int, size_t); /*proto*/
static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *, int); /*proto*/
static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *, Py_ssize_t *, Py_ssize_t, int, char); /*proto*/
static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *, __Pyx_memviewslice *, char, int); /*proto*/
static int __pyx_memoryview_err_extents(int, Py_ssize_t, Py_ssize_t); /*proto*/
static int __pyx_memoryview_err_dim(PyObject *, char *, int); /*proto*/
static int __pyx_memoryview_err(PyObject *, char *); /*proto*/
static int __pyx_memoryview_copy_contents(__Pyx_memviewslice, __Pyx_memviewslice, int, int, int); /*proto*/
static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *, int, int); /*proto*/
static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *, int, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_refcount_objects_in_slice(char *, Py_ssize_t *, Py_ssize_t *, int, int); /*proto*/
static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *, int, size_t, void *, int); /*proto*/
static void __pyx_memoryview__slice_assign_scalar(char *, Py_ssize_t *, Py_ssize_t *, int, size_t, void *); /*proto*/
static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *, PyObject *); /*proto*/
#define __Pyx_MODULE_NAME "nes.cycore.mos6502"
extern int __pyx_module_is_main_nes__cycore__mos6502;
int __pyx_module_is_main_nes__cycore__mos6502 = 0;

/* Implementation of 'nes.cycore.mos6502' */
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_OverflowError;
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_TypeError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_IndexError;
static PyObject *__pyx_builtin_MemoryError;
static PyObject *__pyx_builtin_Ellipsis;
static PyObject *__pyx_builtin_id;
static const char __pyx_k_[] = "";
static const char __pyx_k_A[] = "A";
static const char __pyx_k_O[] = "O";
static const char __pyx_k_c[] = "c";
static const char __pyx_k__2[] = "{} ";
static const char __pyx_k__3[] = "---";
static const char __pyx_k__5[] = "   ";
static const char __pyx_k__6[] = "    ";
static const char __pyx_k__7[] = " ";
static const char __pyx_k__9[] = "_";
static const char __pyx_k_id[] = "id";
static const char __pyx_k__10[] = "__";
static const char __pyx_k_aac[] = "_aac";
static const char __pyx_k_aax[] = "_aax";
static const char __pyx_k_arg[] = "arg";
static const char __pyx_k_arr[] = "_arr";
static const char __pyx_k_asr[] = "_asr";
static const char __pyx_k_atx[] = "_atx";
static const char __pyx_k_axa[] = "_axa";
static const char __pyx_k_axs[] = "_axs";
static const char __pyx_k_dcp[] = "_dcp";
static const char __pyx_k_dop[] = "_dop";
static const char __pyx_k_isc[] = "_isc";
static const char __pyx_k_kil[] = "_kil";
static const char __pyx_k_lar[] = "_lar";
static const char __pyx_k_lax[] = "_lax";
static const char __pyx_k_new[] = "__new__";
static const char __pyx_k_obj[] = "obj";
static const char __pyx_k_rla[] = "_rla";
static const char __pyx_k_rra[] = "_rra";
static const char __pyx_k_slo[] = "_slo";
static const char __pyx_k_sre[] = "_sre";
static const char __pyx_k_sxa[] = "_sxa";
static const char __pyx_k_sya[] = "_sya";
static const char __pyx_k_top[] = "_top";
static const char __pyx_k_xaa[] = "_xaa";
static const char __pyx_k_xas[] = "_xas";
static const char __pyx_k_addr[] = "addr";
static const char __pyx_k_base[] = "base";
static const char __pyx_k_caps[] = "caps";
static const char __pyx_k_data[] = "data";
static const char __pyx_k_dict[] = "__dict__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mode[] = "mode";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_ndim[] = "ndim";
static const char __pyx_k_pack[] = "pack";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_step[] = "step";
static const char __pyx_k_stop[] = "stop";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_0_02X[] = "#${0:02X}";
static const char __pyx_k_0_04X[] = "${0:04X}";
static const char __pyx_k_ASCII[] = "ASCII";
static const char __pyx_k_class[] = "__class__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_flags[] = "flags";
static const char __pyx_k_instr[] = "instr";
static const char __pyx_k_items[] = "items";
static const char __pyx_k_modes[] = "modes";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_reset[] = "reset";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_start[] = "start";
static const char __pyx_k_upper[] = "upper";
static const char __pyx_k_cycles[] = "cycles";
static const char __pyx_k_encode[] = "encode";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_memory[] = "memory";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_pickle[] = "pickle";
static const char __pyx_k_reduce[] = "__reduce__";
static const char __pyx_k_struct[] = "struct";
static const char __pyx_k_unpack[] = "unpack";
static const char __pyx_k_update[] = "update";
static const char __pyx_k_0_02X_2[] = "${0:02X}";
static const char __pyx_k_0_02X_3[] = "{0:02X} ";
static const char __pyx_k_0_02X_4[] = "{0:02X}  ";
static const char __pyx_k_0_02X_X[] = "${0:02X},X";
static const char __pyx_k_0_02X_Y[] = "${0:02X},Y";
static const char __pyx_k_0_04X_2[] = "(${0:04X})";
static const char __pyx_k_0_04X_3[] = "{0:04X}  ";
static const char __pyx_k_0_04X_X[] = "${0:04X},X";
static const char __pyx_k_0_04X_Y[] = "${0:04X},Y";
static const char __pyx_k_IMPLIED[] = "IMPLIED";
static const char __pyx_k_MOS6502[] = "MOS6502";
static const char __pyx_k_fortran[] = "fortran";
static const char __pyx_k_memview[] = "memview";
static const char __pyx_k_ABSOLUTE[] = "ABSOLUTE";
static const char __pyx_k_Ellipsis[] = "Ellipsis";
static const char __pyx_k_INDIRECT[] = "INDIRECT";
static const char __pyx_k_RELATIVE[] = "RELATIVE";
static const char __pyx_k_ZEROPAGE[] = "ZEROPAGE";
static const char __pyx_k_bytecode[] = "bytecode";
static const char __pyx_k_getstate[] = "__getstate__";
static const char __pyx_k_itemsize[] = "itemsize";
static const char __pyx_k_ppu_line[] = "ppu_line";
static const char __pyx_k_pyx_type[] = "__pyx_type";
static const char __pyx_k_setstate[] = "__setstate__";
static const char __pyx_k_IMMEDIATE[] = "IMMEDIATE";
static const char __pyx_k_TypeError[] = "TypeError";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_immediate[] = "immediate";
static const char __pyx_k_ppu_pixel[] = "ppu_pixel";
static const char __pyx_k_pyx_state[] = "__pyx_state";
static const char __pyx_k_reduce_ex[] = "__reduce_ex__";
static const char __pyx_k_ABSOLUTE_X[] = "ABSOLUTE_X";
static const char __pyx_k_ABSOLUTE_Y[] = "ABSOLUTE_Y";
static const char __pyx_k_INDIRECT_X[] = "INDIRECT_X";
static const char __pyx_k_INDIRECT_Y[] = "INDIRECT_Y";
static const char __pyx_k_IndexError[] = "IndexError";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_ZEROPAGE_X[] = "ZEROPAGE_X";
static const char __pyx_k_ZEROPAGE_Y[] = "ZEROPAGE_Y";
static const char __pyx_k_pyx_result[] = "__pyx_result";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_size_bytes[] = "size_bytes";
static const char __pyx_k_0_02X_1_04X[] = "${0:02X} (-> ${1:04X})";
static const char __pyx_k_ACCUMULATOR[] = "ACCUMULATOR";
static const char __pyx_k_MemoryError[] = "MemoryError";
static const char __pyx_k_PickleError[] = "PickleError";
static const char __pyx_k_AddressModes[] = "AddressModes";
static const char __pyx_k_pyx_checksum[] = "__pyx_checksum";
static const char __pyx_k_stringsource[] = "stringsource";
static const char __pyx_k_02X_X_02X_04X[] = "(${:02X},X) @ {:02X} = {:04X}";
static const char __pyx_k_02X_Y_04X_04X[] = "(${:02X}),Y = {:04X} @ {:04X}";
static const char __pyx_k_OverflowError[] = "OverflowError";
static const char __pyx_k_pyx_getbuffer[] = "__pyx_getbuffer";
static const char __pyx_k_reduce_cython[] = "__reduce_cython__";
static const char __pyx_k_aax_sets_flags[] = "aax_sets_flags";
static const char __pyx_k_INSTRUCTION_SET[] = "INSTRUCTION_SET";
static const char __pyx_k_Stack_underflow[] = "Stack underflow";
static const char __pyx_k_View_MemoryView[] = "View.MemoryView";
static const char __pyx_k_allocate_buffer[] = "allocate_buffer";
static const char __pyx_k_dtype_is_object[] = "dtype_is_object";
static const char __pyx_k_pyx_PickleError[] = "__pyx_PickleError";
static const char __pyx_k_setstate_cython[] = "__setstate_cython__";
static const char __pyx_k_NamedInstruction[] = "NamedInstruction";
static const char __pyx_k_nes_instructions[] = "nes.instructions";
static const char __pyx_k_pyx_unpickle_Enum[] = "__pyx_unpickle_Enum";
static const char __pyx_k_cline_in_traceback[] = "cline_in_traceback";
static const char __pyx_k_format_instruction[] = "format_instruction";
static const char __pyx_k_make_bytecode_dict[] = "_make_bytecode_dict";
static const char __pyx_k_nes_cycore_mos6502[] = "nes.cycore.mos6502";
static const char __pyx_k_strided_and_direct[] = "<strided and direct>";
static const char __pyx_k_pyx_unpickle_MOS6502[] = "__pyx_unpickle_MOS6502";
static const char __pyx_k_strided_and_indirect[] = "<strided and indirect>";
static const char __pyx_k_contiguous_and_direct[] = "<contiguous and direct>";
static const char __pyx_k_MemoryView_of_r_object[] = "<MemoryView of %r object>";
static const char __pyx_k_MemoryView_of_r_at_0x_x[] = "<MemoryView of %r at 0x%x>";
static const char __pyx_k_contiguous_and_indirect[] = "<contiguous and indirect>";
static const char __pyx_k_Cannot_index_with_type_s[] = "Cannot index with type '%s'";
static const char __pyx_k_Invalid_shape_in_axis_d_d[] = "Invalid shape in axis %d: %d.";
static const char __pyx_k_undocumented_support_level[] = "undocumented_support_level";
static const char __pyx_k_itemsize_0_for_cython_array[] = "itemsize <= 0 for cython.array";
static const char __pyx_k_unable_to_allocate_array_data[] = "unable to allocate array data.";
static const char __pyx_k_strided_and_direct_or_indirect[] = "<strided and direct or indirect>";
static const char __pyx_k_A_02X_X_02X_Y_02X_P_02X_SP_02X_P[] = "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}";
static const char __pyx_k_Buffer_view_does_not_expose_stri[] = "Buffer view does not expose strides";
static const char __pyx_k_Can_only_create_a_buffer_that_is[] = "Can only create a buffer that is contiguous in memory.";
static const char __pyx_k_Cannot_assign_to_read_only_memor[] = "Cannot assign to read-only memoryview";
static const char __pyx_k_Cannot_create_writable_memory_vi[] = "Cannot create writable memory view from read-only memoryview";
static const char __pyx_k_Empty_shape_tuple_for_cython_arr[] = "Empty shape tuple for cython.array";
static const char __pyx_k_Incompatible_checksums_s_vs_0x19[] = "Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))";
static const char __pyx_k_Incompatible_checksums_s_vs_0xb0[] = "Incompatible checksums (%s vs 0xb068931 = (name))";
static const char __pyx_k_Indirect_dimensions_not_supporte[] = "Indirect dimensions not supported";
static const char __pyx_k_Invalid_mode_expected_c_or_fortr[] = "Invalid mode, expected 'c' or 'fortran', got %s";
static const char __pyx_k_KIL_instruction_Processor_halted[] = "KIL instruction.  Processor halted.";
static const char __pyx_k_Out_of_bounds_on_buffer_access_a[] = "Out of bounds on buffer access (axis %d)";
static const char __pyx_k_Unable_to_convert_item_to_object[] = "Unable to convert item to object";
static const char __pyx_k_got_differing_extents_in_dimensi[] = "got differing extents in dimension %d (got %d and %d)";
static const char __pyx_k_no_default___reduce___due_to_non[] = "no default __reduce__ due to non-trivial __cinit__";
static const char __pyx_k_stack_underflow_causes_exception[] = "stack_underflow_causes_exception";
static const char __pyx_k_unable_to_allocate_shape_and_str[] = "unable to allocate shape and strides.";
static PyObject *__pyx_kp_u_;
static PyObject *__pyx_kp_u_02X_X_02X_04X;
static PyObject *__pyx_kp_u_02X_Y_04X_04X;
static PyObject *__pyx_kp_u_0_02X;
static PyObject *__pyx_kp_u_0_02X_1_04X;
static PyObject *__pyx_kp_u_0_02X_2;
static PyObject *__pyx_kp_u_0_02X_3;
static PyObject *__pyx_kp_u_0_02X_4;
static PyObject *__pyx_kp_u_0_02X_X;
static PyObject *__pyx_kp_u_0_02X_Y;
static PyObject *__pyx_kp_u_0_04X;
static PyObject *__pyx_kp_u_0_04X_2;
static PyObject *__pyx_kp_u_0_04X_3;
static PyObject *__pyx_kp_u_0_04X_X;
static PyObject *__pyx_kp_u_0_04X_Y;
static PyObject *__pyx_n_u_A;
static PyObject *__pyx_n_s_ABSOLUTE;
static PyObject *__pyx_n_s_ABSOLUTE_X;
static PyObject *__pyx_n_s_ABSOLUTE_Y;
static PyObject *__pyx_n_s_ACCUMULATOR;
static PyObject *__pyx_n_s_ASCII;
static PyObject *__pyx_kp_u_A_02X_X_02X_Y_02X_P_02X_SP_02X_P;
static PyObject *__pyx_n_s_AddressModes;
static PyObject *__pyx_kp_s_Buffer_view_does_not_expose_stri;
static PyObject *__pyx_kp_s_Can_only_create_a_buffer_that_is;
static PyObject *__pyx_kp_s_Cannot_assign_to_read_only_memor;
static PyObject *__pyx_kp_s_Cannot_create_writable_memory_vi;
static PyObject *__pyx_kp_s_Cannot_index_with_type_s;
static PyObject *__pyx_n_s_Ellipsis;
static PyObject *__pyx_kp_s_Empty_shape_tuple_for_cython_arr;
static PyObject *__pyx_n_s_IMMEDIATE;
static PyObject *__pyx_n_s_IMPLIED;
static PyObject *__pyx_n_s_INDIRECT;
static PyObject *__pyx_n_s_INDIRECT_X;
static PyObject *__pyx_n_s_INDIRECT_Y;
static PyObject *__pyx_n_s_INSTRUCTION_SET;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0x19;
static PyObject *__pyx_kp_s_Incompatible_checksums_s_vs_0xb0;
static PyObject *__pyx_n_s_IndexError;
static PyObject *__pyx_kp_s_Indirect_dimensions_not_supporte;
static PyObject *__pyx_kp_s_Invalid_mode_expected_c_or_fortr;
static PyObject *__pyx_kp_s_Invalid_shape_in_axis_d_d;
static PyObject *__pyx_kp_u_KIL_instruction_Processor_halted;
static PyObject *__pyx_n_s_MOS6502;
static PyObject *__pyx_n_s_MemoryError;
static PyObject *__pyx_kp_s_MemoryView_of_r_at_0x_x;
static PyObject *__pyx_kp_s_MemoryView_of_r_object;
static PyObject *__pyx_n_s_NamedInstruction;
static PyObject *__pyx_n_b_O;
static PyObject *__pyx_kp_s_Out_of_bounds_on_buffer_access_a;
static PyObject *__pyx_n_s_OverflowError;
static PyObject *__pyx_n_s_PickleError;
static PyObject *__pyx_n_s_RELATIVE;
static PyObject *__pyx_kp_u_Stack_underflow;
static PyObject *__pyx_n_s_TypeError;
static PyObject *__pyx_kp_s_Unable_to_convert_item_to_object;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_n_s_View_MemoryView;
static PyObject *__pyx_n_s_ZEROPAGE;
static PyObject *__pyx_n_s_ZEROPAGE_X;
static PyObject *__pyx_n_s_ZEROPAGE_Y;
static PyObject *__pyx_n_s__10;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_kp_u__3;
static PyObject *__pyx_kp_u__5;
static PyObject *__pyx_kp_u__6;
static PyObject *__pyx_kp_u__7;
static PyObject *__pyx_n_s__9;
static PyObject *__pyx_n_s_aac;
static PyObject *__pyx_n_s_aax;
static PyObject *__pyx_n_s_aax_sets_flags;
static PyObject *__pyx_n_s_addr;
static PyObject *__pyx_n_s_allocate_buffer;
static PyObject *__pyx_n_s_arg;
static PyObject *__pyx_n_s_arr;
static PyObject *__pyx_n_s_asr;
static PyObject *__pyx_n_s_atx;
static PyObject *__pyx_n_s_axa;
static PyObject *__pyx_n_s_axs;
static PyObject *__pyx_n_s_base;
static PyObject *__pyx_n_s_bytecode;
static PyObject *__pyx_n_s_c;
static PyObject *__pyx_n_u_c;
static PyObject *__pyx_n_s_caps;
static PyObject *__pyx_n_s_class;
static PyObject *__pyx_n_s_cline_in_traceback;
static PyObject *__pyx_kp_s_contiguous_and_direct;
static PyObject *__pyx_kp_s_contiguous_and_indirect;
static PyObject *__pyx_n_s_cycles;
static PyObject *__pyx_n_s_data;
static PyObject *__pyx_n_s_dcp;
static PyObject *__pyx_n_s_dict;
static PyObject *__pyx_n_s_dop;
static PyObject *__pyx_n_s_dtype_is_object;
static PyObject *__pyx_n_s_encode;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_flags;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_format_instruction;
static PyObject *__pyx_n_s_fortran;
static PyObject *__pyx_n_u_fortran;
static PyObject *__pyx_n_s_getstate;
static PyObject *__pyx_kp_s_got_differing_extents_in_dimensi;
static PyObject *__pyx_n_s_id;
static PyObject *__pyx_n_s_immediate;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_instr;
static PyObject *__pyx_n_s_isc;
static PyObject *__pyx_n_s_items;
static PyObject *__pyx_n_s_itemsize;
static PyObject *__pyx_kp_s_itemsize_0_for_cython_array;
static PyObject *__pyx_n_s_kil;
static PyObject *__pyx_n_s_lar;
static PyObject *__pyx_n_s_lax;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_make_bytecode_dict;
static PyObject *__pyx_n_s_memory;
static PyObject *__pyx_n_s_memview;
static PyObject *__pyx_n_s_mode;
static PyObject *__pyx_n_s_modes;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_n_s_ndim;
static PyObject *__pyx_n_s_nes_cycore_mos6502;
static PyObject *__pyx_n_s_nes_instructions;
static PyObject *__pyx_n_s_new;
static PyObject *__pyx_kp_s_no_default___reduce___due_to_non;
static PyObject *__pyx_n_s_obj;
static PyObject *__pyx_n_s_pack;
static PyObject *__pyx_n_s_pickle;
static PyObject *__pyx_n_s_ppu_line;
static PyObject *__pyx_n_s_ppu_pixel;
static PyObject *__pyx_n_s_pyx_PickleError;
static PyObject *__pyx_n_s_pyx_checksum;
static PyObject *__pyx_n_s_pyx_getbuffer;
static PyObject *__pyx_n_s_pyx_result;
static PyObject *__pyx_n_s_pyx_state;
static PyObject *__pyx_n_s_pyx_type;
static PyObject *__pyx_n_s_pyx_unpickle_Enum;
static PyObject *__pyx_n_s_pyx_unpickle_MOS6502;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_reduce;
static PyObject *__pyx_n_s_reduce_cython;
static PyObject *__pyx_n_s_reduce_ex;
static PyObject *__pyx_n_s_reset;
static PyObject *__pyx_n_s_rla;
static PyObject *__pyx_n_s_rra;
static PyObject *__pyx_n_s_setstate;
static PyObject *__pyx_n_s_setstate_cython;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_n_s_size_bytes;
static PyObject *__pyx_n_s_slo;
static PyObject *__pyx_n_s_sre;
static PyObject *__pyx_n_s_stack_underflow_causes_exception;
static PyObject *__pyx_n_s_start;
static PyObject *__pyx_n_s_step;
static PyObject *__pyx_n_s_stop;
static PyObject *__pyx_kp_s_strided_and_direct;
static PyObject *__pyx_kp_s_strided_and_direct_or_indirect;
static PyObject *__pyx_kp_s_strided_and_indirect;
static PyObject *__pyx_kp_s_stringsource;
static PyObject *__pyx_n_s_struct;
static PyObject *__pyx_n_s_sxa;
static PyObject *__pyx_n_s_sya;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_top;
static PyObject *__pyx_kp_s_unable_to_allocate_array_data;
static PyObject *__pyx_kp_s_unable_to_allocate_shape_and_str;
static PyObject *__pyx_n_s_undocumented_support_level;
static PyObject *__pyx_n_s_unpack;
static PyObject *__pyx_n_s_update;
static PyObject *__pyx_n_s_upper;
static PyObject *__pyx_n_s_xaa;
static PyObject *__pyx_n_s_xas;
static int __pyx_pf_3nes_6cycore_7mos6502_7MOS6502___init__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_memory, PyObject *__pyx_v_undocumented_support_level, PyObject *__pyx_v_aax_sets_flags, PyObject *__pyx_v_stack_underflow_causes_exception); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_2reset(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_18cycles_since_reset___get__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_4_make_bytecode_dict(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_6format_instruction(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_instr, PyObject *__pyx_v_data, PyObject *__pyx_v_caps); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_8log_line(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_ppu_line, PyObject *__pyx_v_ppu_pixel); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_10set_reset_vector(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_reset_vector); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_12_dop(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__, CYTHON_UNUSED PyObject *__pyx_v___); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_14_top(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__, CYTHON_UNUSED PyObject *__pyx_v___); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_16_kil(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__, CYTHON_UNUSED PyObject *__pyx_v___); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_18_lax(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_immediate); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_20_aax(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_22_dcp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_24_isc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_26_slo(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_28_rla(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_30_rra(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_32_sre(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_34_arr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_36_asr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_38_atx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_40_aac(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_42_axa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_44_axs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_immediate); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_46_lar(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_48_sxa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_50_sya(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_52_xaa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_54_xas(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_56__reduce_cython__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_58__setstate_cython__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr); /* proto */
static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item); /* proto */
static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /* proto */
static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name); /* proto */
static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object); /* proto */
static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /* proto */
static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self); /* proto */
static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state); /* proto */
static PyObject *__pyx_tp_new_3nes_6cycore_7mos6502_MOS6502(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_238;
static PyObject *__pyx_int_255;
static PyObject *__pyx_int_65280;
static PyObject *__pyx_int_26325635;
static PyObject *__pyx_int_184977713;
static PyObject *__pyx_int_neg_1;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_slice__26;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__28;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_codeobj__31;
static PyObject *__pyx_codeobj__38;
/* Late includes */

/* "nes/cycore/mos6502.pyx":35
 * 
 *     """
 *     def __init__(self, memory, undocumented_support_level=1, aax_sets_flags=False, stack_underflow_causes_exception=True):             # <<<<<<<<<<<<<<
 * 
 *         # memory is user-supplied object with read and write methods, allowing for memory mappers, bank switching, etc.
 */

/* Python wrapper */
static int __pyx_pw_3nes_6cycore_7mos6502_7MOS6502_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_pw_3nes_6cycore_7mos6502_7MOS6502_1__init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_memory = 0;
  PyObject *__pyx_v_undocumented_support_level = 0;
  PyObject *__pyx_v_aax_sets_flags = 0;
  PyObject *__pyx_v_stack_underflow_causes_exception = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_memory,&__pyx_n_s_undocumented_support_level,&__pyx_n_s_aax_sets_flags,&__pyx_n_s_stack_underflow_causes_exception,0};
    PyObject* values[4] = {0,0,0,0};
    values[1] = ((PyObject *)__pyx_int_1);
    values[2] = ((PyObject *)Py_False);
    values[3] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_memory)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_undocumented_support_level);
          if (value) { values[1] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_aax_sets_flags);
          if (value) { values[2] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_stack_underflow_causes_exception);
          if (value) { values[3] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(0, 35, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_memory = values[0];
    __pyx_v_undocumented_support_level = values[1];
    __pyx_v_aax_sets_flags = values[2];
    __pyx_v_stack_underflow_causes_exception = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 0, 1, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 35, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502___init__(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_memory, __pyx_v_undocumented_support_level, __pyx_v_aax_sets_flags, __pyx_v_stack_underflow_causes_exception);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_3nes_6cycore_7mos6502_7MOS6502___init__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_memory, PyObject *__pyx_v_undocumented_support_level, PyObject *__pyx_v_aax_sets_flags, PyObject *__pyx_v_stack_underflow_causes_exception) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "nes/cycore/mos6502.pyx":38
 * 
 *         # memory is user-supplied object with read and write methods, allowing for memory mappers, bank switching, etc.
 *         self.memory = memory             # <<<<<<<<<<<<<<
 * 
 *         # create a map from bytecodes to functions for the instruction set
 */
  if (!(likely(((__pyx_v_memory) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_memory, __pyx_ptype_3nes_6cycore_6memory_NESMappedRAM))))) __PYX_ERR(0, 38, __pyx_L1_error)
  __pyx_t_1 = __pyx_v_memory;
  __Pyx_INCREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->memory);
  __Pyx_DECREF(((PyObject *)__pyx_v_self->memory));
  __pyx_v_self->memory = ((struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":41
 * 
 *         # create a map from bytecodes to functions for the instruction set
 *         self.instructions = self._make_bytecode_dict()             # <<<<<<<<<<<<<<
 *         self.make_instruction_data_tables()
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_make_bytecode_dict); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 41, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v_self->instructions);
  __Pyx_DECREF(__pyx_v_self->instructions);
  __pyx_v_self->instructions = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":42
 *         # create a map from bytecodes to functions for the instruction set
 *         self.instructions = self._make_bytecode_dict()
 *         self.make_instruction_data_tables()             # <<<<<<<<<<<<<<
 * 
 *         # User-accessible registers
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->make_instruction_data_tables(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":45
 * 
 *         # User-accessible registers
 *         self.A = 0  # accumulator             # <<<<<<<<<<<<<<
 *         self.X = 0  # X register
 *         self.Y = 0  # Y register
 */
  __pyx_v_self->A = 0;

  /* "nes/cycore/mos6502.pyx":46
 *         # User-accessible registers
 *         self.A = 0  # accumulator
 *         self.X = 0  # X register             # <<<<<<<<<<<<<<
 *         self.Y = 0  # Y register
 * 
 */
  __pyx_v_self->X = 0;

  /* "nes/cycore/mos6502.pyx":47
 *         self.A = 0  # accumulator
 *         self.X = 0  # X register
 *         self.Y = 0  # Y register             # <<<<<<<<<<<<<<
 * 
 *         # Indirectly accessible registers
 */
  __pyx_v_self->Y = 0;

  /* "nes/cycore/mos6502.pyx":50
 * 
 *         # Indirectly accessible registers
 *         self.PC = 0     # program counter (not valid until a reset() has occurred)             # <<<<<<<<<<<<<<
 *         self.SP = 0     # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)
 * 
 */
  __pyx_v_self->PC = 0;

  /* "nes/cycore/mos6502.pyx":51
 *         # Indirectly accessible registers
 *         self.PC = 0     # program counter (not valid until a reset() has occurred)
 *         self.SP = 0     # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)             # <<<<<<<<<<<<<<
 * 
 *         # Flags
 */
  __pyx_v_self->SP = 0;

  /* "nes/cycore/mos6502.pyx":56
 *         # (stored as an 8-bit register in the CPU)
 *         # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
 *         self.N = 0             # <<<<<<<<<<<<<<
 *         self.V = 0
 *         self.D = 0
 */
  __pyx_v_self->N = 0;

  /* "nes/cycore/mos6502.pyx":57
 *         # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
 *         self.N = 0
 *         self.V = 0             # <<<<<<<<<<<<<<
 *         self.D = 0
 *         self.I = 0
 */
  __pyx_v_self->V = 0;

  /* "nes/cycore/mos6502.pyx":58
 *         self.N = 0
 *         self.V = 0
 *         self.D = 0             # <<<<<<<<<<<<<<
 *         self.I = 0
 *         self.Z = 0
 */
  __pyx_v_self->D = 0;

  /* "nes/cycore/mos6502.pyx":59
 *         self.V = 0
 *         self.D = 0
 *         self.I = 0             # <<<<<<<<<<<<<<
 *         self.Z = 0
 *         self.C = 0
 */
  __pyx_v_self->I = 0;

  /* "nes/cycore/mos6502.pyx":60
 *         self.D = 0
 *         self.I = 0
 *         self.Z = 0             # <<<<<<<<<<<<<<
 *         self.C = 0
 * 
 */
  __pyx_v_self->Z = 0;

  /* "nes/cycore/mos6502.pyx":61
 *         self.I = 0
 *         self.Z = 0
 *         self.C = 0             # <<<<<<<<<<<<<<
 * 
 *         # CPU cycles since the processor was started or reset
 */
  __pyx_v_self->C = 0;

  /* "nes/cycore/mos6502.pyx":64
 * 
 *         # CPU cycles since the processor was started or reset
 *         self.cycles_since_reset = 0             # <<<<<<<<<<<<<<
 * 
 *         # control behaviour of the cpu
 */
  __pyx_v_self->cycles_since_reset = 0;

  /* "nes/cycore/mos6502.pyx":67
 * 
 *         # control behaviour of the cpu
 *         self.aax_sets_flags = aax_sets_flags             # <<<<<<<<<<<<<<
 *         self.undocumented_support_level = undocumented_support_level
 *         self.stack_underflow_causes_exception = stack_underflow_causes_exception
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_aax_sets_flags); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 67, __pyx_L1_error)
  __pyx_v_self->aax_sets_flags = __pyx_t_4;

  /* "nes/cycore/mos6502.pyx":68
 *         # control behaviour of the cpu
 *         self.aax_sets_flags = aax_sets_flags
 *         self.undocumented_support_level = undocumented_support_level             # <<<<<<<<<<<<<<
 *         self.stack_underflow_causes_exception = stack_underflow_causes_exception
 * 
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_undocumented_support_level); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 68, __pyx_L1_error)
  __pyx_v_self->undocumented_support_level = __pyx_t_4;

  /* "nes/cycore/mos6502.pyx":69
 *         self.aax_sets_flags = aax_sets_flags
 *         self.undocumented_support_level = undocumented_support_level
 *         self.stack_underflow_causes_exception = stack_underflow_causes_exception             # <<<<<<<<<<<<<<
 * 
 *         # reset the cpu
 */
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_stack_underflow_causes_exception); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 69, __pyx_L1_error)
  __pyx_v_self->stack_underflow_causes_exception = __pyx_t_4;

  /* "nes/cycore/mos6502.pyx":72
 * 
 *         # reset the cpu
 *         self.reset()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_reset); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_3 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_3)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_3) ? __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3) : __Pyx_PyObject_CallNoArg(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 72, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":35
 * 
 *     """
 *     def __init__(self, memory, undocumented_support_level=1, aax_sets_flags=False, stack_underflow_causes_exception=True):             # <<<<<<<<<<<<<<
 * 
 *         # memory is user-supplied object with read and write methods, allowing for memory mappers, bank switching, etc.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":75
 * 
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """
 *         Resets the CPU
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_3reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_2reset[] = "\n        Resets the CPU\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_3reset(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset (wrapper)", 0);
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_2reset(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_2reset(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("reset", 0);

  /* "nes/cycore/mos6502.pyx":80
 *         """
 *         # read the program counter from the RESET_VECTOR_ADDR
 *         self.PC = self._read_word(RESET_VECTOR_ADDR, wrap_at_page=False)             # <<<<<<<<<<<<<<
 * 
 *         # clear the registers
 */
  __pyx_v_self->PC = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, __pyx_e_3nes_6cycore_7mos6502_RESET_VECTOR_ADDR, 0);

  /* "nes/cycore/mos6502.pyx":83
 * 
 *         # clear the registers
 *         self.A = 0  # accumulator             # <<<<<<<<<<<<<<
 *         self.X = 0  # X register
 *         self.Y = 0  # Y register
 */
  __pyx_v_self->A = 0;

  /* "nes/cycore/mos6502.pyx":84
 *         # clear the registers
 *         self.A = 0  # accumulator
 *         self.X = 0  # X register             # <<<<<<<<<<<<<<
 *         self.Y = 0  # Y register
 * 
 */
  __pyx_v_self->X = 0;

  /* "nes/cycore/mos6502.pyx":85
 *         self.A = 0  # accumulator
 *         self.X = 0  # X register
 *         self.Y = 0  # Y register             # <<<<<<<<<<<<<<
 * 
 *         # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)
 */
  __pyx_v_self->Y = 0;

  /* "nes/cycore/mos6502.pyx":89
 *         # stack pointer (8-bit, but the stack is in page 1 of mem, starting at 0x01FF and counting backwards to 0x0100)
 *         # reset does three stack pops, so this is set to 0xFD after reset [9]
 *         self.SP = 0xFD             # <<<<<<<<<<<<<<
 * 
 *         # Flags
 */
  __pyx_v_self->SP = 0xFD;

  /* "nes/cycore/mos6502.pyx":93
 *         # Flags
 *         # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
 *         self.N = False             # <<<<<<<<<<<<<<
 *         self.V = False
 *         self.D = False
 */
  __pyx_v_self->N = 0;

  /* "nes/cycore/mos6502.pyx":94
 *         # (use _status_[to/from]_byte to convert to 8-bit byte for storage on stack)
 *         self.N = False
 *         self.V = False             # <<<<<<<<<<<<<<
 *         self.D = False
 *         self.I = True   # CPU should "come up with interrupt disable bit set" [7]
 */
  __pyx_v_self->V = 0;

  /* "nes/cycore/mos6502.pyx":95
 *         self.N = False
 *         self.V = False
 *         self.D = False             # <<<<<<<<<<<<<<
 *         self.I = True   # CPU should "come up with interrupt disable bit set" [7]
 *         self.Z = False
 */
  __pyx_v_self->D = 0;

  /* "nes/cycore/mos6502.pyx":96
 *         self.V = False
 *         self.D = False
 *         self.I = True   # CPU should "come up with interrupt disable bit set" [7]             # <<<<<<<<<<<<<<
 *         self.Z = False
 *         self.C = False
 */
  __pyx_v_self->I = 1;

  /* "nes/cycore/mos6502.pyx":97
 *         self.D = False
 *         self.I = True   # CPU should "come up with interrupt disable bit set" [7]
 *         self.Z = False             # <<<<<<<<<<<<<<
 *         self.C = False
 * 
 */
  __pyx_v_self->Z = 0;

  /* "nes/cycore/mos6502.pyx":98
 *         self.I = True   # CPU should "come up with interrupt disable bit set" [7]
 *         self.Z = False
 *         self.C = False             # <<<<<<<<<<<<<<
 * 
 *         # CPU cycles since the processor was started or reset
 */
  __pyx_v_self->C = 0;

  /* "nes/cycore/mos6502.pyx":102
 *         # CPU cycles since the processor was started or reset
 *         # reset takes 7 cycles [9]
 *         self.cycles_since_reset = 7             # <<<<<<<<<<<<<<
 * 
 *     cdef int dma_pause(self, int pause_type, int count):
 */
  __pyx_v_self->cycles_since_reset = 7;

  /* "nes/cycore/mos6502.pyx":75
 * 
 * 
 *     def reset(self):             # <<<<<<<<<<<<<<
 *         """
 *         Resets the CPU
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":104
 *         self.cycles_since_reset = 7
 * 
 *     cdef int dma_pause(self, int pause_type, int count):             # <<<<<<<<<<<<<<
 *         cdef int cycles
 *         if pause_type == OAM_DMA:
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_dma_pause(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_pause_type, int __pyx_v_count) {
  int __pyx_v_cycles;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("dma_pause", 0);

  /* "nes/cycore/mos6502.pyx":106
 *     cdef int dma_pause(self, int pause_type, int count):
 *         cdef int cycles
 *         if pause_type == OAM_DMA:             # <<<<<<<<<<<<<<
 *             cycles = OAM_DMA_CPU_CYCLES + self.cycles_since_reset % 2
 *         elif pause_type == DMC_DMA:
 */
  switch (__pyx_v_pause_type) {
    case __pyx_e_3nes_6cycore_6system_OAM_DMA:

    /* "nes/cycore/mos6502.pyx":107
 *         cdef int cycles
 *         if pause_type == OAM_DMA:
 *             cycles = OAM_DMA_CPU_CYCLES + self.cycles_since_reset % 2             # <<<<<<<<<<<<<<
 *         elif pause_type == DMC_DMA:
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 */
    __pyx_v_cycles = (__pyx_e_3nes_6cycore_7mos6502_OAM_DMA_CPU_CYCLES + (__pyx_v_self->cycles_since_reset % 2));

    /* "nes/cycore/mos6502.pyx":106
 *     cdef int dma_pause(self, int pause_type, int count):
 *         cdef int cycles
 *         if pause_type == OAM_DMA:             # <<<<<<<<<<<<<<
 *             cycles = OAM_DMA_CPU_CYCLES + self.cycles_since_reset % 2
 *         elif pause_type == DMC_DMA:
 */
    break;
    case __pyx_e_3nes_6cycore_6system_DMC_DMA:

    /* "nes/cycore/mos6502.pyx":110
 *         elif pause_type == DMC_DMA:
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 *             cycles = 4             # <<<<<<<<<<<<<<
 *         elif pause_type == DMC_DMA_DURING_OAM_DMA:
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":108
 *         if pause_type == OAM_DMA:
 *             cycles = OAM_DMA_CPU_CYCLES + self.cycles_since_reset % 2
 *         elif pause_type == DMC_DMA:             # <<<<<<<<<<<<<<
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 *             cycles = 4
 */
    break;
    case __pyx_e_3nes_6cycore_6system_DMC_DMA_DURING_OAM_DMA:

    /* "nes/cycore/mos6502.pyx":113
 *         elif pause_type == DMC_DMA_DURING_OAM_DMA:
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 *             cycles = 2 * count             # <<<<<<<<<<<<<<
 *         self.cycles_since_reset += cycles
 *         return cycles
 */
    __pyx_v_cycles = (2 * __pyx_v_count);

    /* "nes/cycore/mos6502.pyx":111
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 *             cycles = 4
 *         elif pause_type == DMC_DMA_DURING_OAM_DMA:             # <<<<<<<<<<<<<<
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 *             cycles = 2 * count
 */
    break;
    default: break;
  }

  /* "nes/cycore/mos6502.pyx":114
 *             #todo: there are some special cases detailed here: https://wiki.nesdev.com/w/index.php/APU_DMC
 *             cycles = 2 * count
 *         self.cycles_since_reset += cycles             # <<<<<<<<<<<<<<
 *         return cycles
 * 
 */
  __pyx_v_self->cycles_since_reset = (__pyx_v_self->cycles_since_reset + __pyx_v_cycles);

  /* "nes/cycore/mos6502.pyx":115
 *             cycles = 2 * count
 *         self.cycles_since_reset += cycles
 *         return cycles             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __pyx_r = __pyx_v_cycles;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":104
 *         self.cycles_since_reset = 7
 * 
 *     cdef int dma_pause(self, int pause_type, int count):             # <<<<<<<<<<<<<<
 *         cdef int cycles
 *         if pause_type == OAM_DMA:
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":118
 * 
 *     @property
 *     def cycles_since_reset(self):             # <<<<<<<<<<<<<<
 *         return self.cycles_since_reset
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_18cycles_since_reset_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_18cycles_since_reset_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_18cycles_since_reset___get__(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_18cycles_since_reset___get__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "nes/cycore/mos6502.pyx":119
 *     @property
 *     def cycles_since_reset(self):
 *         return self.cycles_since_reset             # <<<<<<<<<<<<<<
 * 
 *     def _make_bytecode_dict(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->cycles_since_reset); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 119, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":118
 * 
 *     @property
 *     def cycles_since_reset(self):             # <<<<<<<<<<<<<<
 *         return self.cycles_since_reset
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.cycles_since_reset.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":121
 *         return self.cycles_since_reset
 * 
 *     def _make_bytecode_dict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Translates the instruction sets into a bytecode->instr dictionary and links the instruction opcodes
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_5_make_bytecode_dict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_4_make_bytecode_dict[] = "\n        Translates the instruction sets into a bytecode->instr dictionary and links the instruction opcodes\n        to the functions in this object that execute them (this is done by instruction name).\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_5_make_bytecode_dict(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_make_bytecode_dict (wrapper)", 0);
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_4_make_bytecode_dict(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_4_make_bytecode_dict(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  PyObject *__pyx_v_instructions = NULL;
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_instr_set = NULL;
  PyObject *__pyx_v_mode = NULL;
  PyObject *__pyx_v_instr = NULL;
  PyObject *__pyx_v_bytecode = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  int __pyx_t_12;
  Py_ssize_t __pyx_t_13;
  PyObject *(*__pyx_t_14)(PyObject *);
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_make_bytecode_dict", 0);

  /* "nes/cycore/mos6502.pyx":126
 *         to the functions in this object that execute them (this is done by instruction name).
 *         """
 *         instructions = [None] * 256             # <<<<<<<<<<<<<<
 *         for _, instr_set in INSTRUCTION_SET.items():
 *             for mode, instr in instr_set.modes.items():
 */
  __pyx_t_1 = PyList_New(1 * 256); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 126, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  { Py_ssize_t __pyx_temp;
    for (__pyx_temp=0; __pyx_temp < 0x100; __pyx_temp++) {
      __Pyx_INCREF(Py_None);
      __Pyx_GIVEREF(Py_None);
      PyList_SET_ITEM(__pyx_t_1, __pyx_temp, Py_None);
    }
  }
  __pyx_v_instructions = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":127
 *         """
 *         instructions = [None] * 256
 *         for _, instr_set in INSTRUCTION_SET.items():             # <<<<<<<<<<<<<<
 *             for mode, instr in instr_set.modes.items():
 *                 if type(instr.bytecode) == list:
 */
  __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_INSTRUCTION_SET); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 127, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_6, &__pyx_t_5, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_instr_set, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "nes/cycore/mos6502.pyx":128
 *         instructions = [None] * 256
 *         for _, instr_set in INSTRUCTION_SET.items():
 *             for mode, instr in instr_set.modes.items():             # <<<<<<<<<<<<<<
 *                 if type(instr.bytecode) == list:
 *                     # some of the undocumented opcodes have multiple aliases, so in that case a list of
 */
    __pyx_t_8 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr_set, __pyx_n_s_modes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 128, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_9), (&__pyx_t_7)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 128, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_10;
    __pyx_t_10 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_9, &__pyx_t_8, &__pyx_t_10, &__pyx_t_6, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 128, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_mode, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_instr, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "nes/cycore/mos6502.pyx":129
 *         for _, instr_set in INSTRUCTION_SET.items():
 *             for mode, instr in instr_set.modes.items():
 *                 if type(instr.bytecode) == list:             # <<<<<<<<<<<<<<
 *                     # some of the undocumented opcodes have multiple aliases, so in that case a list of
 *                     # opcodes is supplied
 */
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_t_6)), ((PyObject *)(&PyList_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_t_12 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely(__pyx_t_12 < 0)) __PYX_ERR(0, 129, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      if (__pyx_t_12) {

        /* "nes/cycore/mos6502.pyx":132
 *                     # some of the undocumented opcodes have multiple aliases, so in that case a list of
 *                     # opcodes is supplied
 *                     for bytecode in instr.bytecode:             # <<<<<<<<<<<<<<
 *                         instructions[bytecode] = NamedInstruction(name=instr_set.name,
 *                                                                   bytecode=bytecode,
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 132, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        if (likely(PyList_CheckExact(__pyx_t_10)) || PyTuple_CheckExact(__pyx_t_10)) {
          __pyx_t_6 = __pyx_t_10; __Pyx_INCREF(__pyx_t_6); __pyx_t_13 = 0;
          __pyx_t_14 = NULL;
        } else {
          __pyx_t_13 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 132, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_6);
          __pyx_t_14 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 132, __pyx_L1_error)
        }
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        for (;;) {
          if (likely(!__pyx_t_14)) {
            if (likely(PyList_CheckExact(__pyx_t_6))) {
              if (__pyx_t_13 >= PyList_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_10 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_13); __Pyx_INCREF(__pyx_t_10); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 132, __pyx_L1_error)
              #else
              __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 132, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);
              #endif
            } else {
              if (__pyx_t_13 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
              #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
              __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_13); __Pyx_INCREF(__pyx_t_10); __pyx_t_13++; if (unlikely(0 < 0)) __PYX_ERR(0, 132, __pyx_L1_error)
              #else
              __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_13); __pyx_t_13++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 132, __pyx_L1_error)
              __Pyx_GOTREF(__pyx_t_10);
              #endif
            }
          } else {
            __pyx_t_10 = __pyx_t_14(__pyx_t_6);
            if (unlikely(!__pyx_t_10)) {
              PyObject* exc_type = PyErr_Occurred();
              if (exc_type) {
                if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
                else __PYX_ERR(0, 132, __pyx_L1_error)
              }
              break;
            }
            __Pyx_GOTREF(__pyx_t_10);
          }
          __Pyx_XDECREF_SET(__pyx_v_bytecode, __pyx_t_10);
          __pyx_t_10 = 0;

          /* "nes/cycore/mos6502.pyx":133
 *                     # opcodes is supplied
 *                     for bytecode in instr.bytecode:
 *                         instructions[bytecode] = NamedInstruction(name=instr_set.name,             # <<<<<<<<<<<<<<
 *                                                                   bytecode=bytecode,
 *                                                                   mode=mode,
 */
          __Pyx_GetModuleGlobalName(__pyx_t_10, __pyx_n_s_NamedInstruction); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __pyx_t_15 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr_set, __pyx_n_s_name); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_name, __pyx_t_16) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "nes/cycore/mos6502.pyx":134
 *                     for bytecode in instr.bytecode:
 *                         instructions[bytecode] = NamedInstruction(name=instr_set.name,
 *                                                                   bytecode=bytecode,             # <<<<<<<<<<<<<<
 *                                                                   mode=mode,
 *                                                                   size_bytes=instr.size_bytes,
 */
          if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_bytecode, __pyx_v_bytecode) < 0) __PYX_ERR(0, 133, __pyx_L1_error)

          /* "nes/cycore/mos6502.pyx":135
 *                         instructions[bytecode] = NamedInstruction(name=instr_set.name,
 *                                                                   bytecode=bytecode,
 *                                                                   mode=mode,             # <<<<<<<<<<<<<<
 *                                                                   size_bytes=instr.size_bytes,
 *                                                                   cycles=instr.cycles,
 */
          if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_mode, __pyx_v_mode) < 0) __PYX_ERR(0, 133, __pyx_L1_error)

          /* "nes/cycore/mos6502.pyx":136
 *                                                                   bytecode=bytecode,
 *                                                                   mode=mode,
 *                                                                   size_bytes=instr.size_bytes,             # <<<<<<<<<<<<<<
 *                                                                   cycles=instr.cycles,
 *                                                                  )
 */
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_size_bytes); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 136, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_size_bytes, __pyx_t_16) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "nes/cycore/mos6502.pyx":137
 *                                                                   mode=mode,
 *                                                                   size_bytes=instr.size_bytes,
 *                                                                   cycles=instr.cycles,             # <<<<<<<<<<<<<<
 *                                                                  )
 *                 else:
 */
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_cycles); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 137, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          if (PyDict_SetItem(__pyx_t_15, __pyx_n_s_cycles, __pyx_t_16) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "nes/cycore/mos6502.pyx":133
 *                     # opcodes is supplied
 *                     for bytecode in instr.bytecode:
 *                         instructions[bytecode] = NamedInstruction(name=instr_set.name,             # <<<<<<<<<<<<<<
 *                                                                   bytecode=bytecode,
 *                                                                   mode=mode,
 */
          __pyx_t_16 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_empty_tuple, __pyx_t_15); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          if (unlikely(PyObject_SetItem(__pyx_v_instructions, __pyx_v_bytecode, __pyx_t_16) < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;

          /* "nes/cycore/mos6502.pyx":132
 *                     # some of the undocumented opcodes have multiple aliases, so in that case a list of
 *                     # opcodes is supplied
 *                     for bytecode in instr.bytecode:             # <<<<<<<<<<<<<<
 *                         instructions[bytecode] = NamedInstruction(name=instr_set.name,
 *                                                                   bytecode=bytecode,
 */
        }
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

        /* "nes/cycore/mos6502.pyx":129
 *         for _, instr_set in INSTRUCTION_SET.items():
 *             for mode, instr in instr_set.modes.items():
 *                 if type(instr.bytecode) == list:             # <<<<<<<<<<<<<<
 *                     # some of the undocumented opcodes have multiple aliases, so in that case a list of
 *                     # opcodes is supplied
 */
        goto __pyx_L7;
      }

      /* "nes/cycore/mos6502.pyx":140
 *                                                                  )
 *                 else:
 *                     instructions[instr.bytecode] = NamedInstruction(name=instr_set.name,             # <<<<<<<<<<<<<<
 *                                                                     bytecode=instr.bytecode,
 *                                                                     mode=mode,
 */
      /*else*/ {
        __Pyx_GetModuleGlobalName(__pyx_t_6, __pyx_n_s_NamedInstruction); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_16 = __Pyx_PyDict_NewPresized(5); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr_set, __pyx_n_s_name); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_name, __pyx_t_15) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

        /* "nes/cycore/mos6502.pyx":141
 *                 else:
 *                     instructions[instr.bytecode] = NamedInstruction(name=instr_set.name,
 *                                                                     bytecode=instr.bytecode,             # <<<<<<<<<<<<<<
 *                                                                     mode=mode,
 *                                                                     size_bytes=instr.size_bytes,
 */
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 141, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_bytecode, __pyx_t_15) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

        /* "nes/cycore/mos6502.pyx":142
 *                     instructions[instr.bytecode] = NamedInstruction(name=instr_set.name,
 *                                                                     bytecode=instr.bytecode,
 *                                                                     mode=mode,             # <<<<<<<<<<<<<<
 *                                                                     size_bytes=instr.size_bytes,
 *                                                                     cycles=instr.cycles,
 */
        if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_mode, __pyx_v_mode) < 0) __PYX_ERR(0, 140, __pyx_L1_error)

        /* "nes/cycore/mos6502.pyx":143
 *                                                                     bytecode=instr.bytecode,
 *                                                                     mode=mode,
 *                                                                     size_bytes=instr.size_bytes,             # <<<<<<<<<<<<<<
 *                                                                     cycles=instr.cycles,
 *                                                                    )
 */
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_size_bytes); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 143, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_size_bytes, __pyx_t_15) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

        /* "nes/cycore/mos6502.pyx":144
 *                                                                     mode=mode,
 *                                                                     size_bytes=instr.size_bytes,
 *                                                                     cycles=instr.cycles,             # <<<<<<<<<<<<<<
 *                                                                    )
 *         return instructions
 */
        __pyx_t_15 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_cycles); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 144, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        if (PyDict_SetItem(__pyx_t_16, __pyx_n_s_cycles, __pyx_t_15) < 0) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

        /* "nes/cycore/mos6502.pyx":140
 *                                                                  )
 *                 else:
 *                     instructions[instr.bytecode] = NamedInstruction(name=instr_set.name,             # <<<<<<<<<<<<<<
 *                                                                     bytecode=instr.bytecode,
 *                                                                     mode=mode,
 */
        __pyx_t_15 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_empty_tuple, __pyx_t_16); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_15);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_16);
        if (unlikely(PyObject_SetItem(__pyx_v_instructions, __pyx_t_16, __pyx_t_15) < 0)) __PYX_ERR(0, 140, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
        __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
      }
      __pyx_L7:;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":146
 *                                                                     cycles=instr.cycles,
 *                                                                    )
 *         return instructions             # <<<<<<<<<<<<<<
 * 
 *     cdef void make_instruction_data_tables(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_instructions);
  __pyx_r = __pyx_v_instructions;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":121
 *         return self.cycles_since_reset
 * 
 *     def _make_bytecode_dict(self):             # <<<<<<<<<<<<<<
 *         """
 *         Translates the instruction sets into a bytecode->instr dictionary and links the instruction opcodes
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._make_bytecode_dict", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_instructions);
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_instr_set);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_XDECREF(__pyx_v_instr);
  __Pyx_XDECREF(__pyx_v_bytecode);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":148
 *         return instructions
 * 
 *     cdef void make_instruction_data_tables(self):             # <<<<<<<<<<<<<<
 *         """
 *         Fills in some data tables about the instructions in cdef variables; replaces the more pythonic version
 */

static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502_make_instruction_data_tables(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  CYTHON_UNUSED PyObject *__pyx_v__ = NULL;
  PyObject *__pyx_v_instr_set = NULL;
  CYTHON_UNUSED PyObject *__pyx_v_mode = NULL;
  PyObject *__pyx_v_instr = NULL;
  PyObject *__pyx_v_bytecodes = NULL;
  PyObject *__pyx_v_bytecode = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_t_11;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_t_13;
  Py_ssize_t __pyx_t_14;
  PyObject *(*__pyx_t_15)(PyObject *);
  Py_ssize_t __pyx_t_16;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("make_instruction_data_tables", 0);

  /* "nes/cycore/mos6502.pyx":153
 *         make_bytecode_dict.
 *         """
 *         for _, instr_set in INSTRUCTION_SET.items():             # <<<<<<<<<<<<<<
 *             for mode, instr in instr_set.modes.items():
 *                 bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]
 */
  __pyx_t_2 = 0;
  __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_INSTRUCTION_SET); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
    __PYX_ERR(0, 153, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_items, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 153, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_6, &__pyx_t_5, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 153, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XDECREF_SET(__pyx_v__, __pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XDECREF_SET(__pyx_v_instr_set, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "nes/cycore/mos6502.pyx":154
 *         """
 *         for _, instr_set in INSTRUCTION_SET.items():
 *             for mode, instr in instr_set.modes.items():             # <<<<<<<<<<<<<<
 *                 bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]
 *                 for bytecode in bytecodes:
 */
    __pyx_t_8 = 0;
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr_set, __pyx_n_s_modes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__pyx_t_6 == Py_None)) {
      PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%.30s'", "items");
      __PYX_ERR(0, 154, __pyx_L1_error)
    }
    __pyx_t_10 = __Pyx_dict_iterator(__pyx_t_6, 0, __pyx_n_s_items, (&__pyx_t_9), (&__pyx_t_7)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 154, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_5);
    __pyx_t_5 = __pyx_t_10;
    __pyx_t_10 = 0;
    while (1) {
      __pyx_t_11 = __Pyx_dict_iter_next(__pyx_t_5, __pyx_t_9, &__pyx_t_8, &__pyx_t_10, &__pyx_t_6, NULL, __pyx_t_7);
      if (unlikely(__pyx_t_11 == 0)) break;
      if (unlikely(__pyx_t_11 == -1)) __PYX_ERR(0, 154, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_mode, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_instr, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "nes/cycore/mos6502.pyx":155
 *         for _, instr_set in INSTRUCTION_SET.items():
 *             for mode, instr in instr_set.modes.items():
 *                 bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]             # <<<<<<<<<<<<<<
 *                 for bytecode in bytecodes:
 *                     self.instr_size_bytes[bytecode] = instr.size_bytes
 */
      __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 155, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __pyx_t_12 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_t_10)), ((PyObject *)(&PyList_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_12); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 155, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_13 = __Pyx_PyObject_IsTrue(__pyx_t_12); if (unlikely(__pyx_t_13 < 0)) __PYX_ERR(0, 155, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      if (__pyx_t_13) {
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_6 = __pyx_t_12;
        __pyx_t_12 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_bytecode); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_12);
        __pyx_t_10 = PyList_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 155, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __Pyx_GIVEREF(__pyx_t_12);
        PyList_SET_ITEM(__pyx_t_10, 0, __pyx_t_12);
        __pyx_t_12 = 0;
        __pyx_t_6 = __pyx_t_10;
        __pyx_t_10 = 0;
      }
      __Pyx_XDECREF_SET(__pyx_v_bytecodes, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "nes/cycore/mos6502.pyx":156
 *             for mode, instr in instr_set.modes.items():
 *                 bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]
 *                 for bytecode in bytecodes:             # <<<<<<<<<<<<<<
 *                     self.instr_size_bytes[bytecode] = instr.size_bytes
 * 
 */
      if (likely(PyList_CheckExact(__pyx_v_bytecodes)) || PyTuple_CheckExact(__pyx_v_bytecodes)) {
        __pyx_t_6 = __pyx_v_bytecodes; __Pyx_INCREF(__pyx_t_6); __pyx_t_14 = 0;
        __pyx_t_15 = NULL;
      } else {
        __pyx_t_14 = -1; __pyx_t_6 = PyObject_GetIter(__pyx_v_bytecodes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 156, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_15 = Py_TYPE(__pyx_t_6)->tp_iternext; if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 156, __pyx_L1_error)
      }
      for (;;) {
        if (likely(!__pyx_t_15)) {
          if (likely(PyList_CheckExact(__pyx_t_6))) {
            if (__pyx_t_14 >= PyList_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyList_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_10); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 156, __pyx_L1_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 156, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          } else {
            if (__pyx_t_14 >= PyTuple_GET_SIZE(__pyx_t_6)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_6, __pyx_t_14); __Pyx_INCREF(__pyx_t_10); __pyx_t_14++; if (unlikely(0 < 0)) __PYX_ERR(0, 156, __pyx_L1_error)
            #else
            __pyx_t_10 = PySequence_ITEM(__pyx_t_6, __pyx_t_14); __pyx_t_14++; if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 156, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_10);
            #endif
          }
        } else {
          __pyx_t_10 = __pyx_t_15(__pyx_t_6);
          if (unlikely(!__pyx_t_10)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 156, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_10);
        }
        __Pyx_XDECREF_SET(__pyx_v_bytecode, __pyx_t_10);
        __pyx_t_10 = 0;

        /* "nes/cycore/mos6502.pyx":157
 *                 bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]
 *                 for bytecode in bytecodes:
 *                     self.instr_size_bytes[bytecode] = instr.size_bytes             # <<<<<<<<<<<<<<
 * 
 *     def format_instruction(self, instr, data, caps=True):
 */
        __pyx_t_10 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_size_bytes); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 157, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_10);
        __pyx_t_11 = __Pyx_PyInt_As_int(__pyx_t_10); if (unlikely((__pyx_t_11 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        __pyx_t_16 = __Pyx_PyIndex_AsSsize_t(__pyx_v_bytecode); if (unlikely((__pyx_t_16 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(0, 157, __pyx_L1_error)
        (__pyx_v_self->instr_size_bytes[__pyx_t_16]) = __pyx_t_11;

        /* "nes/cycore/mos6502.pyx":156
 *             for mode, instr in instr_set.modes.items():
 *                 bytecodes = instr.bytecode if type(instr.bytecode) == list else [instr.bytecode]
 *                 for bytecode in bytecodes:             # <<<<<<<<<<<<<<
 *                     self.instr_size_bytes[bytecode] = instr.size_bytes
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":148
 *         return instructions
 * 
 *     cdef void make_instruction_data_tables(self):             # <<<<<<<<<<<<<<
 *         """
 *         Fills in some data tables about the instructions in cdef variables; replaces the more pythonic version
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_WriteUnraisable("nes.cycore.mos6502.MOS6502.make_instruction_data_tables", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v__);
  __Pyx_XDECREF(__pyx_v_instr_set);
  __Pyx_XDECREF(__pyx_v_mode);
  __Pyx_XDECREF(__pyx_v_instr);
  __Pyx_XDECREF(__pyx_v_bytecodes);
  __Pyx_XDECREF(__pyx_v_bytecode);
  __Pyx_RefNannyFinishContext();
}

/* "nes/cycore/mos6502.pyx":159
 *                     self.instr_size_bytes[bytecode] = instr.size_bytes
 * 
 *     def format_instruction(self, instr, data, caps=True):             # <<<<<<<<<<<<<<
 *         """
 *         Formats an instruction for logline in the form of nestest logs
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_7format_instruction(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_6format_instruction[] = "\n        Formats an instruction for logline in the form of nestest logs\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_7format_instruction(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_instr = 0;
  PyObject *__pyx_v_data = 0;
  PyObject *__pyx_v_caps = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("format_instruction (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_instr,&__pyx_n_s_data,&__pyx_n_s_caps,0};
    PyObject* values[3] = {0,0,0};
    values[2] = ((PyObject *)Py_True);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_instr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_data)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("format_instruction", 0, 2, 3, 1); __PYX_ERR(0, 159, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_caps);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "format_instruction") < 0)) __PYX_ERR(0, 159, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_instr = values[0];
    __pyx_v_data = values[1];
    __pyx_v_caps = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("format_instruction", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 159, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.format_instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_6format_instruction(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_instr, __pyx_v_data, __pyx_v_caps);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_6format_instruction(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_instr, PyObject *__pyx_v_data, PyObject *__pyx_v_caps) {
  PyObject *__pyx_v_line = NULL;
  PyObject *__pyx_v_name = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  unsigned char *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  int __pyx_t_10;
  unsigned char __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("format_instruction", 0);

  /* "nes/cycore/mos6502.pyx":163
 *         Formats an instruction for logline in the form of nestest logs
 *         """
 *         line = ""             # <<<<<<<<<<<<<<
 *         name = instr.name if not caps else instr.name.upper()
 *         line += '{} '.format(name)
 */
  __Pyx_INCREF(__pyx_kp_u_);
  __pyx_v_line = __pyx_kp_u_;

  /* "nes/cycore/mos6502.pyx":164
 *         """
 *         line = ""
 *         name = instr.name if not caps else instr.name.upper()             # <<<<<<<<<<<<<<
 *         line += '{} '.format(name)
 *         if instr.mode == AddressModes.IMMEDIATE:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_caps); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 164, __pyx_L1_error)
  if (((!__pyx_t_2) != 0)) {
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_name); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_name); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_4, __pyx_n_s_upper); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallNoArg(__pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_1 = __pyx_t_3;
    __pyx_t_3 = 0;
  }
  __pyx_v_name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":165
 *         line = ""
 *         name = instr.name if not caps else instr.name.upper()
 *         line += '{} '.format(name)             # <<<<<<<<<<<<<<
 *         if instr.mode == AddressModes.IMMEDIATE:
 *             line += '#${0:02X}'.format(data[0])
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u__2, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_5)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_5, __pyx_v_name) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_name);
  __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "nes/cycore/mos6502.pyx":166
 *         name = instr.name if not caps else instr.name.upper()
 *         line += '{} '.format(name)
 *         if instr.mode == AddressModes.IMMEDIATE:             # <<<<<<<<<<<<<<
 *             line += '#${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IMMEDIATE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 166, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":167
 *         line += '{} '.format(name)
 *         if instr.mode == AddressModes.IMMEDIATE:
 *             line += '#${0:02X}'.format(data[0])             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ZEROPAGE:
 *             line += '${0:02X}'.format(data[0])
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 167, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "nes/cycore/mos6502.pyx":166
 *         name = instr.name if not caps else instr.name.upper()
 *         line += '{} '.format(name)
 *         if instr.mode == AddressModes.IMMEDIATE:             # <<<<<<<<<<<<<<
 *             line += '#${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":168
 *         if instr.mode == AddressModes.IMMEDIATE:
 *             line += '#${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE:             # <<<<<<<<<<<<<<
 *             line += '${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_X:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ZEROPAGE); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 168, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":169
 *             line += '#${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE:
 *             line += '${0:02X}'.format(data[0])             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ZEROPAGE_X:
 *             line += '${0:02X},X'.format(data[0])
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_2, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":168
 *         if instr.mode == AddressModes.IMMEDIATE:
 *             line += '#${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE:             # <<<<<<<<<<<<<<
 *             line += '${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_X:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":170
 *         elif instr.mode == AddressModes.ZEROPAGE:
 *             line += '${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_X:             # <<<<<<<<<<<<<<
 *             line += '${0:02X},X'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ZEROPAGE_X); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":171
 *             line += '${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_X:
 *             line += '${0:02X},X'.format(data[0])             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:
 *             line += '${0:02X},Y'.format(data[0])
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_X, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 171, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "nes/cycore/mos6502.pyx":170
 *         elif instr.mode == AddressModes.ZEROPAGE:
 *             line += '${0:02X}'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_X:             # <<<<<<<<<<<<<<
 *             line += '${0:02X},X'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":172
 *         elif instr.mode == AddressModes.ZEROPAGE_X:
 *             line += '${0:02X},X'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:             # <<<<<<<<<<<<<<
 *             line += '${0:02X},Y'.format(data[0])
 *         elif instr.mode == AddressModes.ABSOLUTE:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ZEROPAGE_Y); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 172, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":173
 *             line += '${0:02X},X'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:
 *             line += '${0:02X},Y'.format(data[0])             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ABSOLUTE:
 *             line += '${0:04X}'.format(self._from_le(data))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_Y, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":172
 *         elif instr.mode == AddressModes.ZEROPAGE_X:
 *             line += '${0:02X},X'.format(data[0])
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:             # <<<<<<<<<<<<<<
 *             line += '${0:02X},Y'.format(data[0])
 *         elif instr.mode == AddressModes.ABSOLUTE:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":174
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:
 *             line += '${0:02X},Y'.format(data[0])
 *         elif instr.mode == AddressModes.ABSOLUTE:             # <<<<<<<<<<<<<<
 *             line += '${0:04X}'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_X:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ABSOLUTE); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":175
 *             line += '${0:02X},Y'.format(data[0])
 *         elif instr.mode == AddressModes.ABSOLUTE:
 *             line += '${0:04X}'.format(self._from_le(data))             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ABSOLUTE_X:
 *             line += '${0:04X},X'.format(self._from_le(data))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_04X, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 175, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 175, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "nes/cycore/mos6502.pyx":174
 *         elif instr.mode == AddressModes.ZEROPAGE_Y:
 *             line += '${0:02X},Y'.format(data[0])
 *         elif instr.mode == AddressModes.ABSOLUTE:             # <<<<<<<<<<<<<<
 *             line += '${0:04X}'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_X:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":176
 *         elif instr.mode == AddressModes.ABSOLUTE:
 *             line += '${0:04X}'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_X:             # <<<<<<<<<<<<<<
 *             line += '${0:04X},X'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ABSOLUTE_X); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 176, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":177
 *             line += '${0:04X}'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_X:
 *             line += '${0:04X},X'.format(self._from_le(data))             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:
 *             line += '${0:04X},Y'.format(self._from_le(data))
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_04X_X, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyObject_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 177, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_4, __pyx_t_5) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_5);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":176
 *         elif instr.mode == AddressModes.ABSOLUTE:
 *             line += '${0:04X}'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_X:             # <<<<<<<<<<<<<<
 *             line += '${0:04X},X'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":178
 *         elif instr.mode == AddressModes.ABSOLUTE_X:
 *             line += '${0:04X},X'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:             # <<<<<<<<<<<<<<
 *             line += '${0:04X},Y'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.INDIRECT_X:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_ABSOLUTE_Y); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_5, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":179
 *             line += '${0:04X},X'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:
 *             line += '${0:04X},Y'.format(self._from_le(data))             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.INDIRECT_X:
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_04X_Y, __pyx_n_s_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 179, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 179, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_5);
    __pyx_t_5 = 0;

    /* "nes/cycore/mos6502.pyx":178
 *         elif instr.mode == AddressModes.ABSOLUTE_X:
 *             line += '${0:04X},X'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:             # <<<<<<<<<<<<<<
 *             line += '${0:04X},Y'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.INDIRECT_X:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":180
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:
 *             line += '${0:04X},Y'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.INDIRECT_X:             # <<<<<<<<<<<<<<
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
 *         elif instr.mode == AddressModes.INDIRECT_Y:
 */
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INDIRECT_X); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_5, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 180, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":181
 *             line += '${0:04X},Y'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.INDIRECT_X:
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.INDIRECT_Y:
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
 */
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_02X_X_02X_04X, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->X); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyNumber_Add(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_AndObjC(__pyx_t_9, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, __pyx_t_10, 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_5, __pyx_t_4, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[4] = {__pyx_t_9, __pyx_t_5, __pyx_t_4, __pyx_t_8};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      if (__pyx_t_9) {
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_9); __pyx_t_9 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_7, 0+__pyx_t_10, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_7, 1+__pyx_t_10, __pyx_t_4);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_7, 2+__pyx_t_10, __pyx_t_8);
      __pyx_t_5 = 0;
      __pyx_t_4 = 0;
      __pyx_t_8 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":180
 *         elif instr.mode == AddressModes.ABSOLUTE_Y:
 *             line += '${0:04X},Y'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.INDIRECT_X:             # <<<<<<<<<<<<<<
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
 *         elif instr.mode == AddressModes.INDIRECT_Y:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":182
 *         elif instr.mode == AddressModes.INDIRECT_X:
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
 *         elif instr.mode == AddressModes.INDIRECT_Y:             # <<<<<<<<<<<<<<
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
 *         elif instr.mode == AddressModes.INDIRECT:
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INDIRECT_Y); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_3, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 182, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":183
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
 *         elif instr.mode == AddressModes.INDIRECT_Y:
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.INDIRECT:
 *             line += '(${0:04X})'.format(self._from_le(data))
 */
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_02X_Y_04X_04X, __pyx_n_s_format); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_3 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, __pyx_t_10, 1)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_t_4); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_From_long(((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, __pyx_t_10, 1) + __pyx_v_self->Y) & 0xFFFF)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_3, __pyx_t_8, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
      PyObject *__pyx_temp[4] = {__pyx_t_5, __pyx_t_3, __pyx_t_8, __pyx_t_4};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-__pyx_t_10, 3+__pyx_t_10); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_9 = PyTuple_New(3+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_10, __pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_10, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_9, 2+__pyx_t_10, __pyx_t_4);
      __pyx_t_3 = 0;
      __pyx_t_8 = 0;
      __pyx_t_4 = 0;
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 183, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    }
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "nes/cycore/mos6502.pyx":182
 *         elif instr.mode == AddressModes.INDIRECT_X:
 *             line += '(${:02X},X) @ {:02X} = {:04X}'.format(data[0], data[0], self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True))
 *         elif instr.mode == AddressModes.INDIRECT_Y:             # <<<<<<<<<<<<<<
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
 *         elif instr.mode == AddressModes.INDIRECT:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":184
 *         elif instr.mode == AddressModes.INDIRECT_Y:
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
 *         elif instr.mode == AddressModes.INDIRECT:             # <<<<<<<<<<<<<<
 *             line += '(${0:04X})'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.IMPLIED:
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_INDIRECT); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_7, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 184, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":185
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
 *         elif instr.mode == AddressModes.INDIRECT:
 *             line += '(${0:04X})'.format(self._from_le(data))             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.IMPLIED:
 *             line += ''
 */
    __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_04X_2, __pyx_n_s_format); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_AsWritableUString(__pyx_v_data); if (unlikely((!__pyx_t_6) && PyErr_Occurred())) __PYX_ERR(0, 185, __pyx_L1_error)
    __pyx_t_7 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_t_6)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_9))) {
      __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_9);
      if (likely(__pyx_t_4)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_9);
        __Pyx_INCREF(__pyx_t_4);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_9, function);
      }
    }
    __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_9, __pyx_t_4, __pyx_t_7) : __Pyx_PyObject_CallOneArg(__pyx_t_9, __pyx_t_7);
    __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_9);
    __pyx_t_9 = 0;

    /* "nes/cycore/mos6502.pyx":184
 *         elif instr.mode == AddressModes.INDIRECT_Y:
 *             line += '(${:02X}),Y = {:04X} @ {:04X}'.format(data[0], self._read_word(data[0], wrap_at_page=True), (self._read_word(data[0], wrap_at_page=True) + self.Y) & 0xFFFF)
 *         elif instr.mode == AddressModes.INDIRECT:             # <<<<<<<<<<<<<<
 *             line += '(${0:04X})'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.IMPLIED:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":186
 *         elif instr.mode == AddressModes.INDIRECT:
 *             line += '(${0:04X})'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.IMPLIED:             # <<<<<<<<<<<<<<
 *             line += ''
 *         elif instr.mode == AddressModes.ACCUMULATOR:
 */
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_GetModuleGlobalName(__pyx_t_1, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_IMPLIED); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_9, __pyx_t_7, Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 186, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":187
 *             line += '(${0:04X})'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.IMPLIED:
 *             line += ''             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.ACCUMULATOR:
 *             line += 'A'
 */
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_kp_u_); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "nes/cycore/mos6502.pyx":186
 *         elif instr.mode == AddressModes.INDIRECT:
 *             line += '(${0:04X})'.format(self._from_le(data))
 *         elif instr.mode == AddressModes.IMPLIED:             # <<<<<<<<<<<<<<
 *             line += ''
 *         elif instr.mode == AddressModes.ACCUMULATOR:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":188
 *         elif instr.mode == AddressModes.IMPLIED:
 *             line += ''
 *         elif instr.mode == AddressModes.ACCUMULATOR:             # <<<<<<<<<<<<<<
 *             line += 'A'
 *         elif instr.mode == AddressModes.RELATIVE:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GetModuleGlobalName(__pyx_t_7, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ACCUMULATOR); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = PyObject_RichCompare(__pyx_t_1, __pyx_t_9, Py_EQ); __Pyx_XGOTREF(__pyx_t_7); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_7); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":189
 *             line += ''
 *         elif instr.mode == AddressModes.ACCUMULATOR:
 *             line += 'A'             # <<<<<<<<<<<<<<
 *         elif instr.mode == AddressModes.RELATIVE:
 *             line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))
 */
    __pyx_t_7 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_n_u_A); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_7);
    __pyx_t_7 = 0;

    /* "nes/cycore/mos6502.pyx":188
 *         elif instr.mode == AddressModes.IMPLIED:
 *             line += ''
 *         elif instr.mode == AddressModes.ACCUMULATOR:             # <<<<<<<<<<<<<<
 *             line += 'A'
 *         elif instr.mode == AddressModes.RELATIVE:
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":190
 *         elif instr.mode == AddressModes.ACCUMULATOR:
 *             line += 'A'
 *         elif instr.mode == AddressModes.RELATIVE:             # <<<<<<<<<<<<<<
 *             line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))
 *         return line
 */
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_mode); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_GetModuleGlobalName(__pyx_t_9, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_RELATIVE); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = PyObject_RichCompare(__pyx_t_7, __pyx_t_1, Py_EQ); __Pyx_XGOTREF(__pyx_t_9); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":191
 *             line += 'A'
 *         elif instr.mode == AddressModes.RELATIVE:
 *             line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))             # <<<<<<<<<<<<<<
 *         return line
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_1_04X, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_7 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_11 = __Pyx_PyInt_As_unsigned_char(__pyx_t_7); if (unlikely((__pyx_t_11 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __pyx_t_7 = __Pyx_PyInt_From_int(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_2sc(__pyx_v_self, __pyx_t_11)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_4 = __Pyx_GetItemInt(__pyx_v_data, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_11 = __Pyx_PyInt_As_unsigned_char(__pyx_t_4); if (unlikely((__pyx_t_11 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_From_long(((__pyx_v_self->PC + 2) + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_2sc(__pyx_v_self, __pyx_t_11))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    __pyx_t_10 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
        __pyx_t_10 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_7, __pyx_t_4};
      __pyx_t_9 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_7, __pyx_t_4};
      __pyx_t_9 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-__pyx_t_10, 2+__pyx_t_10); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else
    #endif
    {
      __pyx_t_3 = PyTuple_New(2+__pyx_t_10); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_3, 0+__pyx_t_10, __pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_4);
      PyTuple_SET_ITEM(__pyx_t_3, 1+__pyx_t_10, __pyx_t_4);
      __pyx_t_7 = 0;
      __pyx_t_4 = 0;
      __pyx_t_9 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_3, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 191, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = PyNumber_InPlaceAdd(__pyx_v_line, __pyx_t_9); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __Pyx_DECREF_SET(__pyx_v_line, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "nes/cycore/mos6502.pyx":190
 *         elif instr.mode == AddressModes.ACCUMULATOR:
 *             line += 'A'
 *         elif instr.mode == AddressModes.RELATIVE:             # <<<<<<<<<<<<<<
 *             line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))
 *         return line
 */
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":192
 *         elif instr.mode == AddressModes.RELATIVE:
 *             line += '${0:02X} (-> ${1:04X})'.format(self._from_2sc(data[0]), self.PC + 2 + self._from_2sc(data[0]))
 *         return line             # <<<<<<<<<<<<<<
 * 
 *     def log_line(self, ppu_line="---", ppu_pixel="---"):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_line);
  __pyx_r = __pyx_v_line;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":159
 *                     self.instr_size_bytes[bytecode] = instr.size_bytes
 * 
 *     def format_instruction(self, instr, data, caps=True):             # <<<<<<<<<<<<<<
 *         """
 *         Formats an instruction for logline in the form of nestest logs
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.format_instruction", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_XDECREF(__pyx_v_name);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":194
 *         return line
 * 
 *     def log_line(self, ppu_line="---", ppu_pixel="---"):             # <<<<<<<<<<<<<<
 *         """
 *         Generates a log line in the format of nestest
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_9log_line(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_8log_line[] = "\n        Generates a log line in the format of nestest\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_9log_line(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_ppu_line = 0;
  PyObject *__pyx_v_ppu_pixel = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("log_line (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_ppu_line,&__pyx_n_s_ppu_pixel,0};
    PyObject* values[2] = {0,0};
    values[0] = ((PyObject *)__pyx_kp_u__3);
    values[1] = ((PyObject *)__pyx_kp_u__3);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ppu_line);
          if (value) { values[0] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  1:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_ppu_pixel);
          if (value) { values[1] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "log_line") < 0)) __PYX_ERR(0, 194, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_ppu_line = values[0];
    __pyx_v_ppu_pixel = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("log_line", 0, 0, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 194, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.log_line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_8log_line(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_ppu_line, __pyx_v_ppu_pixel);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_8log_line(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_ppu_line, PyObject *__pyx_v_ppu_pixel) {
  PyObject *__pyx_v_str = NULL;
  PyObject *__pyx_v_bytecode = NULL;
  PyObject *__pyx_v_instr = NULL;
  PyObject *__pyx_v_data = NULL;
  PyObject *__pyx_v_i = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("log_line", 0);

  /* "nes/cycore/mos6502.pyx":198
 *         Generates a log line in the format of nestest
 *         """
 *         str = "{0:04X}  ".format(self.PC)             # <<<<<<<<<<<<<<
 *         bytecode = self.memory.read(self.PC)
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_04X_3, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->PC); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_1 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v_str = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":199
 *         """
 *         str = "{0:04X}  ".format(self.PC)
 *         bytecode = self.memory.read(self.PC)             # <<<<<<<<<<<<<<
 * 
 *         instr = self.instructions[bytecode]
 */
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_char(((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_self->PC)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 199, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytecode = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":201
 *         bytecode = self.memory.read(self.PC)
 * 
 *         instr = self.instructions[bytecode]             # <<<<<<<<<<<<<<
 *         data = bytearray(2)
 *         for i in range(instr.size_bytes - 1):
 */
  __pyx_t_1 = __Pyx_PyObject_GetItem(__pyx_v_self->instructions, __pyx_v_bytecode); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_instr = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":202
 * 
 *         instr = self.instructions[bytecode]
 *         data = bytearray(2)             # <<<<<<<<<<<<<<
 *         for i in range(instr.size_bytes - 1):
 *             data[i] = self.memory.read(self.PC + 1 + i)
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)(&PyByteArray_Type)), __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_data = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":203
 *         instr = self.instructions[bytecode]
 *         data = bytearray(2)
 *         for i in range(instr.size_bytes - 1):             # <<<<<<<<<<<<<<
 *             data[i] = self.memory.read(self.PC + 1 + i)
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_instr, __pyx_n_s_size_bytes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_SubtractObjC(__pyx_t_1, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_builtin_range, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (likely(PyList_CheckExact(__pyx_t_1)) || PyTuple_CheckExact(__pyx_t_1)) {
    __pyx_t_2 = __pyx_t_1; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 203, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 203, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_i, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "nes/cycore/mos6502.pyx":204
 *         data = bytearray(2)
 *         for i in range(instr.size_bytes - 1):
 *             data[i] = self.memory.read(self.PC + 1 + i)             # <<<<<<<<<<<<<<
 * 
 *         str += "{0:02X} ".format(bytecode)
 */
    __pyx_t_1 = __Pyx_PyInt_From_long((__pyx_v_self->PC + 1)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = PyNumber_Add(__pyx_t_1, __pyx_v_i); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_7 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_7 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_From_unsigned_char(((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_7)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (unlikely(PyObject_SetItem(__pyx_v_data, __pyx_v_i, __pyx_t_3) < 0)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":203
 *         instr = self.instructions[bytecode]
 *         data = bytearray(2)
 *         for i in range(instr.size_bytes - 1):             # <<<<<<<<<<<<<<
 *             data[i] = self.memory.read(self.PC + 1 + i)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nes/cycore/mos6502.pyx":206
 *             data[i] = self.memory.read(self.PC + 1 + i)
 * 
 *         str += "{0:02X} ".format(bytecode)             # <<<<<<<<<<<<<<
 *         str += "{0:02X} ".format(data[0]) if len(data) > 0 else "   "
 *         str += "{0:02X}  ".format(data[1]) if len(data) > 1 else "    "
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_3, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_2 = (__pyx_t_1) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_1, __pyx_v_bytecode) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_v_bytecode);
  __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_str, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 206, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_str, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "nes/cycore/mos6502.pyx":207
 * 
 *         str += "{0:02X} ".format(bytecode)
 *         str += "{0:02X} ".format(data[0]) if len(data) > 0 else "   "             # <<<<<<<<<<<<<<
 *         str += "{0:02X}  ".format(data[1]) if len(data) > 1 else "    "
 *         str += self.format_instruction(instr, data)
 */
  __pyx_t_5 = PyByteArray_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 207, __pyx_L1_error)
  if (((__pyx_t_5 > 0) != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_3, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_From_int(((unsigned char)(PyByteArray_AS_STRING(__pyx_v_data)[0]))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_2 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = __pyx_t_2;
    __pyx_t_2 = 0;
  } else {
    __Pyx_INCREF(__pyx_kp_u__5);
    __pyx_t_3 = __pyx_kp_u__5;
  }
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_str, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 207, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_str, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "nes/cycore/mos6502.pyx":208
 *         str += "{0:02X} ".format(bytecode)
 *         str += "{0:02X} ".format(data[0]) if len(data) > 0 else "   "
 *         str += "{0:02X}  ".format(data[1]) if len(data) > 1 else "    "             # <<<<<<<<<<<<<<
 *         str += self.format_instruction(instr, data)
 * 
 */
  __pyx_t_5 = PyByteArray_GET_SIZE(__pyx_v_data); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 208, __pyx_L1_error)
  if (((__pyx_t_5 > 1) != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_0_02X_4, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyInt_From_int(((unsigned char)(PyByteArray_AS_STRING(__pyx_v_data)[1]))); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    __pyx_t_3 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_1, __pyx_t_8, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_2 = __pyx_t_3;
    __pyx_t_3 = 0;
  } else {
    __Pyx_INCREF(__pyx_kp_u__6);
    __pyx_t_2 = __pyx_kp_u__6;
  }
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_str, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 208, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_str, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "nes/cycore/mos6502.pyx":209
 *         str += "{0:02X} ".format(data[0]) if len(data) > 0 else "   "
 *         str += "{0:02X}  ".format(data[1]) if len(data) > 1 else "    "
 *         str += self.format_instruction(instr, data)             # <<<<<<<<<<<<<<
 * 
 *         while len(str) < 48:
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_format_instruction); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_instr, __pyx_v_data};
    __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
    PyObject *__pyx_temp[3] = {__pyx_t_1, __pyx_v_instr, __pyx_v_data};
    __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_GOTREF(__pyx_t_3);
  } else
  #endif
  {
    __pyx_t_4 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if (__pyx_t_1) {
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1); __pyx_t_1 = NULL;
    }
    __Pyx_INCREF(__pyx_v_instr);
    __Pyx_GIVEREF(__pyx_v_instr);
    PyTuple_SET_ITEM(__pyx_t_4, 0+__pyx_t_7, __pyx_v_instr);
    __Pyx_INCREF(__pyx_v_data);
    __Pyx_GIVEREF(__pyx_v_data);
    PyTuple_SET_ITEM(__pyx_t_4, 1+__pyx_t_7, __pyx_v_data);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 209, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_str, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_str, __pyx_t_2);
  __pyx_t_2 = 0;

  /* "nes/cycore/mos6502.pyx":211
 *         str += self.format_instruction(instr, data)
 * 
 *         while len(str) < 48:             # <<<<<<<<<<<<<<
 *             str += " "
 * 
 */
  while (1) {
    __pyx_t_5 = PyObject_Length(__pyx_v_str); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(0, 211, __pyx_L1_error)
    __pyx_t_9 = ((__pyx_t_5 < 48) != 0);
    if (!__pyx_t_9) break;

    /* "nes/cycore/mos6502.pyx":212
 * 
 *         while len(str) < 48:
 *             str += " "             # <<<<<<<<<<<<<<
 * 
 *         str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}".format(self.A,
 */
    __pyx_t_2 = PyNumber_InPlaceAdd(__pyx_v_str, __pyx_kp_u__7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF_SET(__pyx_v_str, __pyx_t_2);
    __pyx_t_2 = 0;
  }

  /* "nes/cycore/mos6502.pyx":214
 *             str += " "
 * 
 *         str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}".format(self.A,             # <<<<<<<<<<<<<<
 *                                                                                          self.X,
 *                                                                                          self.Y,
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_kp_u_A_02X_X_02X_Y_02X_P_02X_SP_02X_P, __pyx_n_s_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->A); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "nes/cycore/mos6502.pyx":215
 * 
 *         str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}".format(self.A,
 *                                                                                          self.X,             # <<<<<<<<<<<<<<
 *                                                                                          self.Y,
 *                                                                                          self._status_to_byte(b_flag=0),
 */
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->X); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 215, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);

  /* "nes/cycore/mos6502.pyx":216
 *         str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}".format(self.A,
 *                                                                                          self.X,
 *                                                                                          self.Y,             # <<<<<<<<<<<<<<
 *                                                                                          self._status_to_byte(b_flag=0),
 *                                                                                          self.SP,
 */
  __pyx_t_8 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->Y); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 216, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);

  /* "nes/cycore/mos6502.pyx":217
 *                                                                                          self.X,
 *                                                                                          self.Y,
 *                                                                                          self._status_to_byte(b_flag=0),             # <<<<<<<<<<<<<<
 *                                                                                          self.SP,
 *                                                                                          ppu_line,
 */
  __pyx_t_10 = __Pyx_PyInt_From_unsigned_char(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_status_to_byte(__pyx_v_self, 0)); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 217, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);

  /* "nes/cycore/mos6502.pyx":218
 *                                                                                          self.Y,
 *                                                                                          self._status_to_byte(b_flag=0),
 *                                                                                          self.SP,             # <<<<<<<<<<<<<<
 *                                                                                          ppu_line,
 *                                                                                          ppu_pixel,
 */
  __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->SP); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);

  /* "nes/cycore/mos6502.pyx":221
 *                                                                                          ppu_line,
 *                                                                                          ppu_pixel,
 *                                                                                          self.cycles_since_reset             # <<<<<<<<<<<<<<
 *                                                                                         )
 *         return str
 */
  __pyx_t_12 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->cycles_since_reset); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 221, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = NULL;
  __pyx_t_7 = 0;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_13 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_13)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_13);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
      __pyx_t_7 = 1;
    }
  }
  #if CYTHON_FAST_PYCALL
  if (PyFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[9] = {__pyx_t_13, __pyx_t_4, __pyx_t_1, __pyx_t_8, __pyx_t_10, __pyx_t_11, __pyx_v_ppu_line, __pyx_v_ppu_pixel, __pyx_t_12};
    __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 8+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  } else
  #endif
  #if CYTHON_FAST_PYCCALL
  if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
    PyObject *__pyx_temp[9] = {__pyx_t_13, __pyx_t_4, __pyx_t_1, __pyx_t_8, __pyx_t_10, __pyx_t_11, __pyx_v_ppu_line, __pyx_v_ppu_pixel, __pyx_t_12};
    __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_7, 8+__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_XDECREF(__pyx_t_13); __pyx_t_13 = 0;
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
  } else
  #endif
  {
    __pyx_t_14 = PyTuple_New(8+__pyx_t_7); if (unlikely(!__pyx_t_14)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_14);
    if (__pyx_t_13) {
      __Pyx_GIVEREF(__pyx_t_13); PyTuple_SET_ITEM(__pyx_t_14, 0, __pyx_t_13); __pyx_t_13 = NULL;
    }
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_14, 0+__pyx_t_7, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_14, 1+__pyx_t_7, __pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_14, 2+__pyx_t_7, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_14, 3+__pyx_t_7, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_11);
    PyTuple_SET_ITEM(__pyx_t_14, 4+__pyx_t_7, __pyx_t_11);
    __Pyx_INCREF(__pyx_v_ppu_line);
    __Pyx_GIVEREF(__pyx_v_ppu_line);
    PyTuple_SET_ITEM(__pyx_t_14, 5+__pyx_t_7, __pyx_v_ppu_line);
    __Pyx_INCREF(__pyx_v_ppu_pixel);
    __Pyx_GIVEREF(__pyx_v_ppu_pixel);
    PyTuple_SET_ITEM(__pyx_t_14, 6+__pyx_t_7, __pyx_v_ppu_pixel);
    __Pyx_GIVEREF(__pyx_t_12);
    PyTuple_SET_ITEM(__pyx_t_14, 7+__pyx_t_7, __pyx_t_12);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_t_8 = 0;
    __pyx_t_10 = 0;
    __pyx_t_11 = 0;
    __pyx_t_12 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_14, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_14); __pyx_t_14 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "nes/cycore/mos6502.pyx":214
 *             str += " "
 * 
 *         str += "A:{:02X} X:{:02X} Y:{:02X} P:{:02X} SP:{:02X} PPU:{},{} CYC:{:d}".format(self.A,             # <<<<<<<<<<<<<<
 *                                                                                          self.X,
 *                                                                                          self.Y,
 */
  __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_v_str, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF_SET(__pyx_v_str, __pyx_t_3);
  __pyx_t_3 = 0;

  /* "nes/cycore/mos6502.pyx":223
 *                                                                                          self.cycles_since_reset
 *                                                                                         )
 *         return str             # <<<<<<<<<<<<<<
 * 
 *     def set_reset_vector(self, reset_vector):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_str);
  __pyx_r = __pyx_v_str;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":194
 *         return line
 * 
 *     def log_line(self, ppu_line="---", ppu_pixel="---"):             # <<<<<<<<<<<<<<
 *         """
 *         Generates a log line in the format of nestest
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.log_line", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_str);
  __Pyx_XDECREF(__pyx_v_bytecode);
  __Pyx_XDECREF(__pyx_v_instr);
  __Pyx_XDECREF(__pyx_v_data);
  __Pyx_XDECREF(__pyx_v_i);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":225
 *         return str
 * 
 *     def set_reset_vector(self, reset_vector):             # <<<<<<<<<<<<<<
 *         """
 *         Sets the reset vector (at fixed mem address), which tells the program counter where to start on a reset
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_11set_reset_vector(PyObject *__pyx_v_self, PyObject *__pyx_v_reset_vector); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_10set_reset_vector[] = "\n        Sets the reset vector (at fixed mem address), which tells the program counter where to start on a reset\n        :param reset_vector: 16-bit address to which PC will be set after a reset\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_11set_reset_vector(PyObject *__pyx_v_self, PyObject *__pyx_v_reset_vector) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("set_reset_vector (wrapper)", 0);
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_10set_reset_vector(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), ((PyObject *)__pyx_v_reset_vector));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_10set_reset_vector(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_reset_vector) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  unsigned char __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("set_reset_vector", 0);

  /* "nes/cycore/mos6502.pyx":230
 *         :param reset_vector: 16-bit address to which PC will be set after a reset
 *         """
 *         self.memory.write(RESET_VECTOR_ADDR, reset_vector & 0x00FF)  # low byte             # <<<<<<<<<<<<<<
 *         self.memory.write(RESET_VECTOR_ADDR + 1, (reset_vector & 0xFF00) >> 8)  # high byte
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_v_reset_vector, __pyx_int_255, 0x00FF, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_As_unsigned_char(__pyx_t_1); if (unlikely((__pyx_t_2 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 230, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_e_3nes_6cycore_7mos6502_RESET_VECTOR_ADDR, __pyx_t_2);

  /* "nes/cycore/mos6502.pyx":231
 *         """
 *         self.memory.write(RESET_VECTOR_ADDR, reset_vector & 0x00FF)  # low byte
 *         self.memory.write(RESET_VECTOR_ADDR + 1, (reset_vector & 0xFF00) >> 8)  # high byte             # <<<<<<<<<<<<<<
 * 
 *     cdef int _from_le(self, unsigned char* data):
 */
  __pyx_t_1 = __Pyx_PyInt_AndObjC(__pyx_v_reset_vector, __pyx_int_65280, 0xFF00, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = __Pyx_PyInt_RshiftObjC(__pyx_t_1, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_PyInt_As_unsigned_char(__pyx_t_3); if (unlikely((__pyx_t_2 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 231, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_e_3nes_6cycore_7mos6502_RESET_VECTOR_ADDR + 1), __pyx_t_2);

  /* "nes/cycore/mos6502.pyx":225
 *         return str
 * 
 *     def set_reset_vector(self, reset_vector):             # <<<<<<<<<<<<<<
 *         """
 *         Sets the reset vector (at fixed mem address), which tells the program counter where to start on a reset
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.set_reset_vector", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":233
 *         self.memory.write(RESET_VECTOR_ADDR + 1, (reset_vector & 0xFF00) >> 8)  # high byte
 * 
 *     cdef int _from_le(self, unsigned char* data):             # <<<<<<<<<<<<<<
 *         """
 *         Create an integer from a little-endian two byte array
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__from_le(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char *__pyx_v_data) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_from_le", 0);

  /* "nes/cycore/mos6502.pyx":239
 *         :return: an integer in the range 0-65535
 *         """
 *         return (data[HI_BYTE] << 8) + data[LO_BYTE]             # <<<<<<<<<<<<<<
 * 
 *     cdef int _read_word(self, int addr, int wrap_at_page):
 */
  __pyx_r = (((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_HI_BYTE]) << 8) + (__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]));
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":233
 *         self.memory.write(RESET_VECTOR_ADDR + 1, (reset_vector & 0xFF00) >> 8)  # high byte
 * 
 *     cdef int _from_le(self, unsigned char* data):             # <<<<<<<<<<<<<<
 *         """
 *         Create an integer from a little-endian two byte array
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":241
 *         return (data[HI_BYTE] << 8) + data[LO_BYTE]
 * 
 *     cdef int _read_word(self, int addr, int wrap_at_page):             # <<<<<<<<<<<<<<
 *         """
 *         Read a word at an address and return it as an integer
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__read_word(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, int __pyx_v_wrap_at_page) {
  unsigned char __pyx_v_byte_lo;
  unsigned char __pyx_v_byte_hi;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_read_word", 0);

  /* "nes/cycore/mos6502.pyx":248
 *         cdef unsigned char byte_lo, byte_hi
 * 
 *         if wrap_at_page and (addr & 0xFF) == 0xFF:             # <<<<<<<<<<<<<<
 *             # will wrap at page boundary
 *             byte_lo = self.memory.read(addr)
 */
  __pyx_t_2 = (__pyx_v_wrap_at_page != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (((__pyx_v_addr & 0xFF) == 0xFF) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":250
 *         if wrap_at_page and (addr & 0xFF) == 0xFF:
 *             # will wrap at page boundary
 *             byte_lo = self.memory.read(addr)             # <<<<<<<<<<<<<<
 *             byte_hi = self.memory.read(addr & 0xFF00)  # read the second (hi) byte from the start of the page
 *         else:
 */
    __pyx_v_byte_lo = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);

    /* "nes/cycore/mos6502.pyx":251
 *             # will wrap at page boundary
 *             byte_lo = self.memory.read(addr)
 *             byte_hi = self.memory.read(addr & 0xFF00)  # read the second (hi) byte from the start of the page             # <<<<<<<<<<<<<<
 *         else:
 *             #data = self.memory.read_block(addr, bytes=2)
 */
    __pyx_v_byte_hi = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_v_addr & 0xFF00));

    /* "nes/cycore/mos6502.pyx":248
 *         cdef unsigned char byte_lo, byte_hi
 * 
 *         if wrap_at_page and (addr & 0xFF) == 0xFF:             # <<<<<<<<<<<<<<
 *             # will wrap at page boundary
 *             byte_lo = self.memory.read(addr)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":254
 *         else:
 *             #data = self.memory.read_block(addr, bytes=2)
 *             byte_lo = self.memory.read(addr)             # <<<<<<<<<<<<<<
 *             byte_hi = self.memory.read(addr + 1)
 *         return byte_lo + (byte_hi << 8)
 */
  /*else*/ {
    __pyx_v_byte_lo = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);

    /* "nes/cycore/mos6502.pyx":255
 *             #data = self.memory.read_block(addr, bytes=2)
 *             byte_lo = self.memory.read(addr)
 *             byte_hi = self.memory.read(addr + 1)             # <<<<<<<<<<<<<<
 *         return byte_lo + (byte_hi << 8)
 * 
 */
    __pyx_v_byte_hi = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_v_addr + 1));
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":256
 *             byte_lo = self.memory.read(addr)
 *             byte_hi = self.memory.read(addr + 1)
 *         return byte_lo + (byte_hi << 8)             # <<<<<<<<<<<<<<
 * 
 *     cdef int trigger_nmi(self):
 */
  __pyx_r = (__pyx_v_byte_lo + (__pyx_v_byte_hi << 8));
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":241
 *         return (data[HI_BYTE] << 8) + data[LO_BYTE]
 * 
 *     cdef int _read_word(self, int addr, int wrap_at_page):             # <<<<<<<<<<<<<<
 *         """
 *         Read a word at an address and return it as an integer
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":258
 *         return byte_lo + (byte_hi << 8)
 * 
 *     cdef int trigger_nmi(self):             # <<<<<<<<<<<<<<
 *         """
 *         Trigger a non maskable interrupt (NMI)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_trigger_nmi(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trigger_nmi", 0);

  /* "nes/cycore/mos6502.pyx":262
 *         Trigger a non maskable interrupt (NMI)
 *         """
 *         self._interrupt(NMI_VECTOR_ADDR, is_brk=False)             # <<<<<<<<<<<<<<
 *         self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
 *         return INTERRUPT_REQUEST_CYCLES
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_interrupt(__pyx_v_self, __pyx_e_3nes_6cycore_7mos6502_NMI_VECTOR_ADDR, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 262, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":263
 *         """
 *         self._interrupt(NMI_VECTOR_ADDR, is_brk=False)
 *         self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES             # <<<<<<<<<<<<<<
 *         return INTERRUPT_REQUEST_CYCLES
 * 
 */
  __pyx_v_self->cycles_since_reset = (__pyx_v_self->cycles_since_reset + __pyx_e_3nes_6cycore_7mos6502_INTERRUPT_REQUEST_CYCLES);

  /* "nes/cycore/mos6502.pyx":264
 *         self._interrupt(NMI_VECTOR_ADDR, is_brk=False)
 *         self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
 *         return INTERRUPT_REQUEST_CYCLES             # <<<<<<<<<<<<<<
 * 
 *     cdef int trigger_irq(self):
 */
  __pyx_r = __pyx_e_3nes_6cycore_7mos6502_INTERRUPT_REQUEST_CYCLES;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":258
 *         return byte_lo + (byte_hi << 8)
 * 
 *     cdef int trigger_nmi(self):             # <<<<<<<<<<<<<<
 *         """
 *         Trigger a non maskable interrupt (NMI)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("nes.cycore.mos6502.MOS6502.trigger_nmi", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":266
 *         return INTERRUPT_REQUEST_CYCLES
 * 
 *     cdef int trigger_irq(self):             # <<<<<<<<<<<<<<
 *         """
 *         Trigger a maskable hardware interrupt (IRQ); if interrupt disable bit (self.I) is set, ignore
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_trigger_irq(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("trigger_irq", 0);

  /* "nes/cycore/mos6502.pyx":270
 *         Trigger a maskable hardware interrupt (IRQ); if interrupt disable bit (self.I) is set, ignore
 *         """
 *         if not self.I:             # <<<<<<<<<<<<<<
 *             self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=False)
 *             self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
 */
  __pyx_t_1 = ((!(__pyx_v_self->I != 0)) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":271
 *         """
 *         if not self.I:
 *             self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=False)             # <<<<<<<<<<<<<<
 *             self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
 *             return INTERRUPT_REQUEST_CYCLES
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_interrupt(__pyx_v_self, __pyx_e_3nes_6cycore_7mos6502_IRQ_BRK_VECTOR_ADDR, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 271, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "nes/cycore/mos6502.pyx":272
 *         if not self.I:
 *             self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=False)
 *             self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES             # <<<<<<<<<<<<<<
 *             return INTERRUPT_REQUEST_CYCLES
 *         else:
 */
    __pyx_v_self->cycles_since_reset = (__pyx_v_self->cycles_since_reset + __pyx_e_3nes_6cycore_7mos6502_INTERRUPT_REQUEST_CYCLES);

    /* "nes/cycore/mos6502.pyx":273
 *             self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=False)
 *             self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
 *             return INTERRUPT_REQUEST_CYCLES             # <<<<<<<<<<<<<<
 *         else:
 *             return 0  # ignored!
 */
    __pyx_r = __pyx_e_3nes_6cycore_7mos6502_INTERRUPT_REQUEST_CYCLES;
    goto __pyx_L0;

    /* "nes/cycore/mos6502.pyx":270
 *         Trigger a maskable hardware interrupt (IRQ); if interrupt disable bit (self.I) is set, ignore
 *         """
 *         if not self.I:             # <<<<<<<<<<<<<<
 *             self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=False)
 *             self.cycles_since_reset += INTERRUPT_REQUEST_CYCLES
 */
  }

  /* "nes/cycore/mos6502.pyx":275
 *             return INTERRUPT_REQUEST_CYCLES
 *         else:
 *             return 0  # ignored!             # <<<<<<<<<<<<<<
 * 
 *     cdef int run_next_instr(self):
 */
  /*else*/ {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "nes/cycore/mos6502.pyx":266
 *         return INTERRUPT_REQUEST_CYCLES
 * 
 *     cdef int trigger_irq(self):             # <<<<<<<<<<<<<<
 *         """
 *         Trigger a maskable hardware interrupt (IRQ); if interrupt disable bit (self.I) is set, ignore
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_WriteUnraisable("nes.cycore.mos6502.MOS6502.trigger_irq", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":277
 *             return 0  # ignored!
 * 
 *     cdef int run_next_instr(self):             # <<<<<<<<<<<<<<
 *         """
 *         Decode and run the next instruction at the program counter, updating the number of processor
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_run_next_instr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  int __pyx_v_cycles;
  unsigned char __pyx_v_bytecode;
  unsigned char __pyx_v_data[2];
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("run_next_instr", 0);

  /* "nes/cycore/mos6502.pyx":287
 *         cdef unsigned char data[2]
 * 
 *         bytecode = self.memory.read(self.PC)             # <<<<<<<<<<<<<<
 *         # surprisingly, loop unrolling here helped quite a bit
 *         if self.instr_size_bytes[bytecode] >= 2:
 */
  __pyx_v_bytecode = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_self->PC);

  /* "nes/cycore/mos6502.pyx":289
 *         bytecode = self.memory.read(self.PC)
 *         # surprisingly, loop unrolling here helped quite a bit
 *         if self.instr_size_bytes[bytecode] >= 2:             # <<<<<<<<<<<<<<
 *             data[0] = self.memory.read(self.PC + 1)
 *         if self.instr_size_bytes[bytecode] == 3:
 */
  __pyx_t_1 = (((__pyx_v_self->instr_size_bytes[__pyx_v_bytecode]) >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":290
 *         # surprisingly, loop unrolling here helped quite a bit
 *         if self.instr_size_bytes[bytecode] >= 2:
 *             data[0] = self.memory.read(self.PC + 1)             # <<<<<<<<<<<<<<
 *         if self.instr_size_bytes[bytecode] == 3:
 *             data[1] = self.memory.read(self.PC + 2)
 */
    (__pyx_v_data[0]) = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_v_self->PC + 1));

    /* "nes/cycore/mos6502.pyx":289
 *         bytecode = self.memory.read(self.PC)
 *         # surprisingly, loop unrolling here helped quite a bit
 *         if self.instr_size_bytes[bytecode] >= 2:             # <<<<<<<<<<<<<<
 *             data[0] = self.memory.read(self.PC + 1)
 *         if self.instr_size_bytes[bytecode] == 3:
 */
  }

  /* "nes/cycore/mos6502.pyx":291
 *         if self.instr_size_bytes[bytecode] >= 2:
 *             data[0] = self.memory.read(self.PC + 1)
 *         if self.instr_size_bytes[bytecode] == 3:             # <<<<<<<<<<<<<<
 *             data[1] = self.memory.read(self.PC + 2)
 * 
 */
  __pyx_t_1 = (((__pyx_v_self->instr_size_bytes[__pyx_v_bytecode]) == 3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":292
 *             data[0] = self.memory.read(self.PC + 1)
 *         if self.instr_size_bytes[bytecode] == 3:
 *             data[1] = self.memory.read(self.PC + 2)             # <<<<<<<<<<<<<<
 * 
 *         # upon retrieving the opcode, the 6502 immediately increments PC by the opcode size:
 */
    (__pyx_v_data[1]) = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_v_self->PC + 2));

    /* "nes/cycore/mos6502.pyx":291
 *         if self.instr_size_bytes[bytecode] >= 2:
 *             data[0] = self.memory.read(self.PC + 1)
 *         if self.instr_size_bytes[bytecode] == 3:             # <<<<<<<<<<<<<<
 *             data[1] = self.memory.read(self.PC + 2)
 * 
 */
  }

  /* "nes/cycore/mos6502.pyx":295
 * 
 *         # upon retrieving the opcode, the 6502 immediately increments PC by the opcode size:
 *         self.PC += self.instr_size_bytes[bytecode]             # <<<<<<<<<<<<<<
 * 
 *         cycles = self.run_instr(bytecode, data)
 */
  __pyx_v_self->PC = (__pyx_v_self->PC + (__pyx_v_self->instr_size_bytes[__pyx_v_bytecode]));

  /* "nes/cycore/mos6502.pyx":297
 *         self.PC += self.instr_size_bytes[bytecode]
 * 
 *         cycles = self.run_instr(bytecode, data)             # <<<<<<<<<<<<<<
 *         self.cycles_since_reset += cycles
 *         return cycles
 */
  __pyx_v_cycles = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->run_instr(__pyx_v_self, __pyx_v_bytecode, __pyx_v_data);

  /* "nes/cycore/mos6502.pyx":298
 * 
 *         cycles = self.run_instr(bytecode, data)
 *         self.cycles_since_reset += cycles             # <<<<<<<<<<<<<<
 *         return cycles
 * 
 */
  __pyx_v_self->cycles_since_reset = (__pyx_v_self->cycles_since_reset + __pyx_v_cycles);

  /* "nes/cycore/mos6502.pyx":299
 *         cycles = self.run_instr(bytecode, data)
 *         self.cycles_since_reset += cycles
 *         return cycles             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned char _status_to_byte(self, bint b_flag):
 */
  __pyx_r = __pyx_v_cycles;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":277
 *             return 0  # ignored!
 * 
 *     cdef int run_next_instr(self):             # <<<<<<<<<<<<<<
 *         """
 *         Decode and run the next instruction at the program counter, updating the number of processor
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":301
 *         return cycles
 * 
 *     cdef unsigned char _status_to_byte(self, bint b_flag):             # <<<<<<<<<<<<<<
 *         """
 *         Puts the status register into an 8-bit value.  Bit 6 is set high always, bit 5 (the "B flag") is set according
 */

static unsigned char __pyx_f_3nes_6cycore_7mos6502_7MOS6502__status_to_byte(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_b_flag) {
  unsigned char __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_status_to_byte", 0);

  /* "nes/cycore/mos6502.pyx":314
 *                 + self.I * SR_I_MASK
 *                 + self.Z * SR_Z_MASK
 *                 + self.C * SR_C_MASK) & 0xFF   # this final and forces it to be treated as unsigned             # <<<<<<<<<<<<<<
 * 
 *     cdef void _status_from_byte(self, unsigned char sr_byte):
 */
  __pyx_r = (((((((((__pyx_v_self->N * __pyx_e_3nes_6cycore_7mos6502_SR_N_MASK) + (__pyx_v_self->V * __pyx_e_3nes_6cycore_7mos6502_SR_V_MASK)) + __pyx_e_3nes_6cycore_7mos6502_SR_X_MASK) + (__pyx_v_b_flag * __pyx_e_3nes_6cycore_7mos6502_SR_B_MASK)) + (__pyx_v_self->D * __pyx_e_3nes_6cycore_7mos6502_SR_D_MASK)) + (__pyx_v_self->I * __pyx_e_3nes_6cycore_7mos6502_SR_I_MASK)) + (__pyx_v_self->Z * __pyx_e_3nes_6cycore_7mos6502_SR_Z_MASK)) + (__pyx_v_self->C * __pyx_e_3nes_6cycore_7mos6502_SR_C_MASK)) & 0xFF);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":301
 *         return cycles
 * 
 *     cdef unsigned char _status_to_byte(self, bint b_flag):             # <<<<<<<<<<<<<<
 *         """
 *         Puts the status register into an 8-bit value.  Bit 6 is set high always, bit 5 (the "B flag") is set according
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":316
 *                 + self.C * SR_C_MASK) & 0xFF   # this final and forces it to be treated as unsigned
 * 
 *     cdef void _status_from_byte(self, unsigned char sr_byte):             # <<<<<<<<<<<<<<
 *         """
 *         Sets the processor status from an 8-bit value as found on the stack.
 */

static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502__status_from_byte(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_sr_byte) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_status_from_byte", 0);

  /* "nes/cycore/mos6502.pyx":321
 *         Bit 5 (B flag) is NEVER set in the status register (but is on the stack), so it is ignored here
 *         """
 *         self.N = (sr_byte & SR_N_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.V = (sr_byte & SR_V_MASK) > 0
 *         self.D = (sr_byte & SR_D_MASK) > 0
 */
  __pyx_v_self->N = ((__pyx_v_sr_byte & __pyx_e_3nes_6cycore_7mos6502_SR_N_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":322
 *         """
 *         self.N = (sr_byte & SR_N_MASK) > 0
 *         self.V = (sr_byte & SR_V_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.D = (sr_byte & SR_D_MASK) > 0
 *         self.I = (sr_byte & SR_I_MASK) > 0
 */
  __pyx_v_self->V = ((__pyx_v_sr_byte & __pyx_e_3nes_6cycore_7mos6502_SR_V_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":323
 *         self.N = (sr_byte & SR_N_MASK) > 0
 *         self.V = (sr_byte & SR_V_MASK) > 0
 *         self.D = (sr_byte & SR_D_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.I = (sr_byte & SR_I_MASK) > 0
 *         self.Z = (sr_byte & SR_Z_MASK) > 0
 */
  __pyx_v_self->D = ((__pyx_v_sr_byte & __pyx_e_3nes_6cycore_7mos6502_SR_D_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":324
 *         self.V = (sr_byte & SR_V_MASK) > 0
 *         self.D = (sr_byte & SR_D_MASK) > 0
 *         self.I = (sr_byte & SR_I_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.Z = (sr_byte & SR_Z_MASK) > 0
 *         self.C = (sr_byte & SR_C_MASK) > 0
 */
  __pyx_v_self->I = ((__pyx_v_sr_byte & __pyx_e_3nes_6cycore_7mos6502_SR_I_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":325
 *         self.D = (sr_byte & SR_D_MASK) > 0
 *         self.I = (sr_byte & SR_I_MASK) > 0
 *         self.Z = (sr_byte & SR_Z_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.C = (sr_byte & SR_C_MASK) > 0
 * 
 */
  __pyx_v_self->Z = ((__pyx_v_sr_byte & __pyx_e_3nes_6cycore_7mos6502_SR_Z_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":326
 *         self.I = (sr_byte & SR_I_MASK) > 0
 *         self.Z = (sr_byte & SR_Z_MASK) > 0
 *         self.C = (sr_byte & SR_C_MASK) > 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void push_stack(self, unsigned char v):
 */
  __pyx_v_self->C = ((__pyx_v_sr_byte & __pyx_e_3nes_6cycore_7mos6502_SR_C_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":316
 *                 + self.C * SR_C_MASK) & 0xFF   # this final and forces it to be treated as unsigned
 * 
 *     cdef void _status_from_byte(self, unsigned char sr_byte):             # <<<<<<<<<<<<<<
 *         """
 *         Sets the processor status from an 8-bit value as found on the stack.
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "nes/cycore/mos6502.pyx":328
 *         self.C = (sr_byte & SR_C_MASK) > 0
 * 
 *     cdef void push_stack(self, unsigned char v):             # <<<<<<<<<<<<<<
 *         """
 *         Push a byte value onto the stack
 */

static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502_push_stack(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_v) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("push_stack", 0);

  /* "nes/cycore/mos6502.pyx":332
 *         Push a byte value onto the stack
 *         """
 *         self.memory.write(STACK_PAGE + self.SP, v)             # <<<<<<<<<<<<<<
 *         self.SP = (self.SP - 1) & 0xFF
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_e_3nes_6cycore_7mos6502_STACK_PAGE + __pyx_v_self->SP), __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":333
 *         """
 *         self.memory.write(STACK_PAGE + self.SP, v)
 *         self.SP = (self.SP - 1) & 0xFF             # <<<<<<<<<<<<<<
 * 
 *     cdef unsigned char pop_stack(self):
 */
  __pyx_v_self->SP = ((__pyx_v_self->SP - 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":328
 *         self.C = (sr_byte & SR_C_MASK) > 0
 * 
 *     cdef void push_stack(self, unsigned char v):             # <<<<<<<<<<<<<<
 *         """
 *         Push a byte value onto the stack
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "nes/cycore/mos6502.pyx":335
 *         self.SP = (self.SP - 1) & 0xFF
 * 
 *     cdef unsigned char pop_stack(self):             # <<<<<<<<<<<<<<
 *         """
 *         Pop (aka 'pull' in 6502 parlance) a byte from the stack
 */

static unsigned char __pyx_f_3nes_6cycore_7mos6502_7MOS6502_pop_stack(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  unsigned char __pyx_v_v;
  unsigned char __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pop_stack", 0);

  /* "nes/cycore/mos6502.pyx":340
 *         """
 *         cdef unsigned char v
 *         if self.SP == 0xFF and self.stack_underflow_causes_exception:             # <<<<<<<<<<<<<<
 *             raise OverflowError("Stack underflow")
 *         self.SP = (self.SP + 1) & 0xFF
 */
  __pyx_t_2 = ((__pyx_v_self->SP == 0xFF) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->stack_underflow_causes_exception != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "nes/cycore/mos6502.pyx":341
 *         cdef unsigned char v
 *         if self.SP == 0xFF and self.stack_underflow_causes_exception:
 *             raise OverflowError("Stack underflow")             # <<<<<<<<<<<<<<
 *         self.SP = (self.SP + 1) & 0xFF
 *         v = self.memory.read(STACK_PAGE + self.SP)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_OverflowError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 341, __pyx_L1_error)

    /* "nes/cycore/mos6502.pyx":340
 *         """
 *         cdef unsigned char v
 *         if self.SP == 0xFF and self.stack_underflow_causes_exception:             # <<<<<<<<<<<<<<
 *             raise OverflowError("Stack underflow")
 *         self.SP = (self.SP + 1) & 0xFF
 */
  }

  /* "nes/cycore/mos6502.pyx":342
 *         if self.SP == 0xFF and self.stack_underflow_causes_exception:
 *             raise OverflowError("Stack underflow")
 *         self.SP = (self.SP + 1) & 0xFF             # <<<<<<<<<<<<<<
 *         v = self.memory.read(STACK_PAGE + self.SP)
 *         return v
 */
  __pyx_v_self->SP = ((__pyx_v_self->SP + 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":343
 *             raise OverflowError("Stack underflow")
 *         self.SP = (self.SP + 1) & 0xFF
 *         v = self.memory.read(STACK_PAGE + self.SP)             # <<<<<<<<<<<<<<
 *         return v
 * 
 */
  __pyx_v_v = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), (__pyx_e_3nes_6cycore_7mos6502_STACK_PAGE + __pyx_v_self->SP));

  /* "nes/cycore/mos6502.pyx":344
 *         self.SP = (self.SP + 1) & 0xFF
 *         v = self.memory.read(STACK_PAGE + self.SP)
 *         return v             # <<<<<<<<<<<<<<
 * 
 *     cdef int _neg(self, int v):
 */
  __pyx_r = __pyx_v_v;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":335
 *         self.SP = (self.SP - 1) & 0xFF
 * 
 *     cdef unsigned char pop_stack(self):             # <<<<<<<<<<<<<<
 *         """
 *         Pop (aka 'pull' in 6502 parlance) a byte from the stack
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_WriteUnraisable("nes.cycore.mos6502.MOS6502.pop_stack", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":346
 *         return v
 * 
 *     cdef int _neg(self, int v):             # <<<<<<<<<<<<<<
 *         """
 *         Is the value of v negative in 2s complement (i.e. is bit 7 high)?
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__neg(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_v) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_neg", 0);

  /* "nes/cycore/mos6502.pyx":351
 *         :return: true if v is negative (true if bit 7 of v is 1)
 *         """
 *         return (v & 0b10000000) > 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _from_2sc(self, unsigned char v):
 */
  __pyx_r = ((__pyx_v_v & 128) > 0);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":346
 *         return v
 * 
 *     cdef int _neg(self, int v):             # <<<<<<<<<<<<<<
 *         """
 *         Is the value of v negative in 2s complement (i.e. is bit 7 high)?
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":353
 *         return (v & 0b10000000) > 0
 * 
 *     cdef int _from_2sc(self, unsigned char v):             # <<<<<<<<<<<<<<
 *         """
 *         Convert a 2's complement number to a signed integer
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__from_2sc(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_v) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_from_2sc", 0);

  /* "nes/cycore/mos6502.pyx":359
 *         #neg = (v & 0b10000000) > 0
 *         #return (v & 0b01111111) if not neg else (v & 0b01111111) - 128
 *         return (v & 0b01111111) - (v & 0b10000000)             # <<<<<<<<<<<<<<
 * 
 *     cdef void _set_zn(self, int v):
 */
  __pyx_r = ((__pyx_v_v & 127) - (__pyx_v_v & 128));
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":353
 *         return (v & 0b10000000) > 0
 * 
 *     cdef int _from_2sc(self, unsigned char v):             # <<<<<<<<<<<<<<
 *         """
 *         Convert a 2's complement number to a signed integer
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":361
 *         return (v & 0b01111111) - (v & 0b10000000)
 * 
 *     cdef void _set_zn(self, int v):             # <<<<<<<<<<<<<<
 *         """
 *         Sets the Z and N flags from a result value v
 */

static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502__set_zn(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_v) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_set_zn", 0);

  /* "nes/cycore/mos6502.pyx":365
 *         Sets the Z and N flags from a result value v
 *         """
 *         self.Z = (v & 0xFF) == 0  # only care about the bottom 8 bits being zero             # <<<<<<<<<<<<<<
 *         self.N = (v & 0b10000000) > 0
 * 
 */
  __pyx_v_self->Z = ((__pyx_v_v & 0xFF) == 0);

  /* "nes/cycore/mos6502.pyx":366
 *         """
 *         self.Z = (v & 0xFF) == 0  # only care about the bottom 8 bits being zero
 *         self.N = (v & 0b10000000) > 0             # <<<<<<<<<<<<<<
 * 
 *     ##################  The instructions (and their helper functions)  #################################################
 */
  __pyx_v_self->N = ((__pyx_v_v & 128) > 0);

  /* "nes/cycore/mos6502.pyx":361
 *         return (v & 0b01111111) - (v & 0b10000000)
 * 
 *     cdef void _set_zn(self, int v):             # <<<<<<<<<<<<<<
 *         """
 *         Sets the Z and N flags from a result value v
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "nes/cycore/mos6502.pyx":370
 *     ##################  The instructions (and their helper functions)  #################################################
 * 
 *     cdef int _adc(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Add to accumulator with carry.  Result goes into accumulator.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__adc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_v_result;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_adc", 0);

  /* "nes/cycore/mos6502.pyx":378
 *         cdef int result
 * 
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 * 
 *         # binary mode (standard)
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":381
 * 
 *         # binary mode (standard)
 *         result = self.A + v + self.C             # <<<<<<<<<<<<<<
 *         self.C = result > 255
 *         self.V = (self._neg(self.A) == self._neg(v)) and (self._neg(self.A) != self._neg(result))
 */
  __pyx_v_result = ((__pyx_v_self->A + __pyx_v_v) + __pyx_v_self->C);

  /* "nes/cycore/mos6502.pyx":382
 *         # binary mode (standard)
 *         result = self.A + v + self.C
 *         self.C = result > 255             # <<<<<<<<<<<<<<
 *         self.V = (self._neg(self.A) == self._neg(v)) and (self._neg(self.A) != self._neg(result))
 * 
 */
  __pyx_v_self->C = (__pyx_v_result > 0xFF);

  /* "nes/cycore/mos6502.pyx":383
 *         result = self.A + v + self.C
 *         self.C = result > 255
 *         self.V = (self._neg(self.A) == self._neg(v)) and (self._neg(self.A) != self._neg(result))             # <<<<<<<<<<<<<<
 * 
 *         # status
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_self->A) == ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_v));
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_self->A) != ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_result));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_v_self->V = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":388
 *         # setting N and Z flags in the usual way is valid in BCD mode because "the N flag contains the high bit of
 *         # the result of the instruction" [5] (having said that, BCD mode is removed on the NES)
 *         self._set_zn(result)             # <<<<<<<<<<<<<<
 * 
 *         # result
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_result);

  /* "nes/cycore/mos6502.pyx":391
 * 
 *         # result
 *         self.A = result & 0xFF             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
  __pyx_v_self->A = (__pyx_v_result & 0xFF);

  /* "nes/cycore/mos6502.pyx":393
 *         self.A = result & 0xFF
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _and(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":370
 *     ##################  The instructions (and their helper functions)  #################################################
 * 
 *     cdef int _adc(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Add to accumulator with carry.  Result goes into accumulator.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":395
 *         return 0
 * 
 *     cdef int _and(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Bitwise AND of accumulator and arg.  Result goes into accumulator.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__and(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_and", 0);

  /* "nes/cycore/mos6502.pyx":400
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self.A = self.A & v
 *         self._set_zn(self.A)
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":401
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg
 *         self.A = self.A & v             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  __pyx_v_self->A = (__pyx_v_self->A & __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":402
 *         v = self.memory.read(arg) if not immediate else arg
 *         self.A = self.A & v
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":403
 *         self.A = self.A & v
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _asl(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":395
 *         return 0
 * 
 *     cdef int _and(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Bitwise AND of accumulator and arg.  Result goes into accumulator.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":405
 *         return 0
 * 
 *     cdef int _asl(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Shift left one bit.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__asl(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  unsigned char __pyx_v_res_8bit;
  int __pyx_v_result;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_asl", 0);

  /* "nes/cycore/mos6502.pyx":411
 *         cdef unsigned char v, res_8bit
 *         cdef int result
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)             # <<<<<<<<<<<<<<
 *         result = v << 1
 * 
 */
  if (((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0)) {
    __pyx_t_1 = __pyx_v_self->A;
  } else {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":412
 *         cdef int result
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)
 *         result = v << 1             # <<<<<<<<<<<<<<
 * 
 *         # status
 */
  __pyx_v_result = (__pyx_v_v << 1);

  /* "nes/cycore/mos6502.pyx":415
 * 
 *         # status
 *         self.C = result > 255    # top bit of original is put into carry flag             # <<<<<<<<<<<<<<
 *         self._set_zn(result)
 * 
 */
  __pyx_v_self->C = (__pyx_v_result > 0xFF);

  /* "nes/cycore/mos6502.pyx":416
 *         # status
 *         self.C = result > 255    # top bit of original is put into carry flag
 *         self._set_zn(result)             # <<<<<<<<<<<<<<
 * 
 *         res_8bit = result & 0xFF
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_result);

  /* "nes/cycore/mos6502.pyx":418
 *         self._set_zn(result)
 * 
 *         res_8bit = result & 0xFF             # <<<<<<<<<<<<<<
 * 
 *         if addr==ARG_NONE:
 */
  __pyx_v_res_8bit = (__pyx_v_result & 0xFF);

  /* "nes/cycore/mos6502.pyx":420
 *         res_8bit = result & 0xFF
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = res_8bit
 */
  __pyx_t_2 = ((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0);
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":422
 *         if addr==ARG_NONE:
 *             # target the accumulator
 *             self.A = res_8bit             # <<<<<<<<<<<<<<
 *         else:
 *             self.memory.write(addr, res_8bit)
 */
    __pyx_v_self->A = __pyx_v_res_8bit;

    /* "nes/cycore/mos6502.pyx":420
 *         res_8bit = result & 0xFF
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = res_8bit
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":424
 *             self.A = res_8bit
 *         else:
 *             self.memory.write(addr, res_8bit)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_res_8bit);
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":425
 *         else:
 *             self.memory.write(addr, res_8bit)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _jump_relative(self, int condition, int offset_2sc):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":405
 *         return 0
 * 
 *     cdef int _asl(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Shift left one bit.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":427
 *         return 0
 * 
 *     cdef int _jump_relative(self, int condition, int offset_2sc):             # <<<<<<<<<<<<<<
 *         """
 *         Jump by offset_2sc (in 2s complement) if the condition is true
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__jump_relative(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_condition, int __pyx_v_offset_2sc) {
  int __pyx_v_extra_cycles;
  int __pyx_v_prev_pc_page;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_jump_relative", 0);

  /* "nes/cycore/mos6502.pyx":434
 *         cdef int extra_cycles, prev_pc_page
 * 
 *         extra_cycles = 0             # <<<<<<<<<<<<<<
 *         if condition:
 *             #print("jump")
 */
  __pyx_v_extra_cycles = 0;

  /* "nes/cycore/mos6502.pyx":435
 * 
 *         extra_cycles = 0
 *         if condition:             # <<<<<<<<<<<<<<
 *             #print("jump")
 *             extra_cycles = 1  # taking a branch takes at least 1 extra cycle
 */
  __pyx_t_1 = (__pyx_v_condition != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":437
 *         if condition:
 *             #print("jump")
 *             extra_cycles = 1  # taking a branch takes at least 1 extra cycle             # <<<<<<<<<<<<<<
 *             prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?
 *             self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)
 */
    __pyx_v_extra_cycles = 1;

    /* "nes/cycore/mos6502.pyx":438
 *             #print("jump")
 *             extra_cycles = 1  # taking a branch takes at least 1 extra cycle
 *             prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?             # <<<<<<<<<<<<<<
 *             self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)
 *             if prev_pc_page != self.PC & 0xFF00:
 */
    __pyx_v_prev_pc_page = (__pyx_v_self->PC & 0xFF00);

    /* "nes/cycore/mos6502.pyx":439
 *             extra_cycles = 1  # taking a branch takes at least 1 extra cycle
 *             prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?
 *             self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)             # <<<<<<<<<<<<<<
 *             if prev_pc_page != self.PC & 0xFF00:
 *                 # but it takes two extra cycles if the memory page changes
 */
    __pyx_v_self->PC = (__pyx_v_self->PC + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_2sc(__pyx_v_self, __pyx_v_offset_2sc));

    /* "nes/cycore/mos6502.pyx":440
 *             prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?
 *             self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)
 *             if prev_pc_page != self.PC & 0xFF00:             # <<<<<<<<<<<<<<
 *                 # but it takes two extra cycles if the memory page changes
 *                 extra_cycles = 2
 */
    __pyx_t_1 = ((__pyx_v_prev_pc_page != (__pyx_v_self->PC & 0xFF00)) != 0);
    if (__pyx_t_1) {

      /* "nes/cycore/mos6502.pyx":442
 *             if prev_pc_page != self.PC & 0xFF00:
 *                 # but it takes two extra cycles if the memory page changes
 *                 extra_cycles = 2             # <<<<<<<<<<<<<<
 *         #else:
 *         #    print("no jump")
 */
      __pyx_v_extra_cycles = 2;

      /* "nes/cycore/mos6502.pyx":440
 *             prev_pc_page = self.PC & 0xFF00  # should this be the PC before the branch instruction or after it?
 *             self.PC += self._from_2sc(offset_2sc)  # jumps to the address of the branch + 2 + offset (which is correct)
 *             if prev_pc_page != self.PC & 0xFF00:             # <<<<<<<<<<<<<<
 *                 # but it takes two extra cycles if the memory page changes
 *                 extra_cycles = 2
 */
    }

    /* "nes/cycore/mos6502.pyx":435
 * 
 *         extra_cycles = 0
 *         if condition:             # <<<<<<<<<<<<<<
 *             #print("jump")
 *             extra_cycles = 1  # taking a branch takes at least 1 extra cycle
 */
  }

  /* "nes/cycore/mos6502.pyx":445
 *         #else:
 *         #    print("no jump")
 *         return extra_cycles             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bcc(self, int offset, int _):
 */
  __pyx_r = __pyx_v_extra_cycles;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":427
 *         return 0
 * 
 *     cdef int _jump_relative(self, int condition, int offset_2sc):             # <<<<<<<<<<<<<<
 *         """
 *         Jump by offset_2sc (in 2s complement) if the condition is true
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":447
 *         return extra_cycles
 * 
 *     cdef int _bcc(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on carry clear
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bcc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bcc", 0);

  /* "nes/cycore/mos6502.pyx":451
 *         Branch on carry clear
 *         """
 *         return self._jump_relative(not self.C, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bcs(self, int offset, int _):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, (!(__pyx_v_self->C != 0)), __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":447
 *         return extra_cycles
 * 
 *     cdef int _bcc(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on carry clear
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":453
 *         return self._jump_relative(not self.C, offset)
 * 
 *     cdef int _bcs(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on carry set
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bcs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bcs", 0);

  /* "nes/cycore/mos6502.pyx":457
 *         Branch on carry set
 *         """
 *         return self._jump_relative(self.C, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _beq(self, int offset, int _):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, __pyx_v_self->C, __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":453
 *         return self._jump_relative(not self.C, offset)
 * 
 *     cdef int _bcs(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on carry set
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":459
 *         return self._jump_relative(self.C, offset)
 * 
 *     cdef int _beq(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch if zero flag is set
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__beq(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_beq", 0);

  /* "nes/cycore/mos6502.pyx":463
 *         Branch if zero flag is set
 *         """
 *         return self._jump_relative(self.Z, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bit(self, int addr, int _):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, __pyx_v_self->Z, __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":459
 *         return self._jump_relative(self.C, offset)
 * 
 *     cdef int _beq(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch if zero flag is set
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":465
 *         return self._jump_relative(self.Z, offset)
 * 
 *     cdef int _bit(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Manipulates the status register by setting the N and V flags to those bits of the value in the address
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bit(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bit", 0);

  /* "nes/cycore/mos6502.pyx":471
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(addr)             # <<<<<<<<<<<<<<
 *         self.N = (v & SR_N_MASK) > 0
 *         self.V = (v & SR_V_MASK) > 0
 */
  __pyx_v_v = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);

  /* "nes/cycore/mos6502.pyx":472
 *         cdef unsigned char v
 *         v = self.memory.read(addr)
 *         self.N = (v & SR_N_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.V = (v & SR_V_MASK) > 0
 *         self.Z = (self.A & v) == 0
 */
  __pyx_v_self->N = ((__pyx_v_v & __pyx_e_3nes_6cycore_7mos6502_SR_N_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":473
 *         v = self.memory.read(addr)
 *         self.N = (v & SR_N_MASK) > 0
 *         self.V = (v & SR_V_MASK) > 0             # <<<<<<<<<<<<<<
 *         self.Z = (self.A & v) == 0
 *         return 0
 */
  __pyx_v_self->V = ((__pyx_v_v & __pyx_e_3nes_6cycore_7mos6502_SR_V_MASK) > 0);

  /* "nes/cycore/mos6502.pyx":474
 *         self.N = (v & SR_N_MASK) > 0
 *         self.V = (v & SR_V_MASK) > 0
 *         self.Z = (self.A & v) == 0             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->Z = ((__pyx_v_self->A & __pyx_v_v) == 0);

  /* "nes/cycore/mos6502.pyx":475
 *         self.V = (v & SR_V_MASK) > 0
 *         self.Z = (self.A & v) == 0
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bmi(self, int offset, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":465
 *         return self._jump_relative(self.Z, offset)
 * 
 *     cdef int _bit(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Manipulates the status register by setting the N and V flags to those bits of the value in the address
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":477
 *         return 0
 * 
 *     cdef int _bmi(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on result minus (i.e. negative flag N is set)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bmi(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bmi", 0);

  /* "nes/cycore/mos6502.pyx":481
 *         Branch on result minus (i.e. negative flag N is set)
 *         """
 *         return self._jump_relative(self.N, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bne(self, int offset, int _):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, __pyx_v_self->N, __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":477
 *         return 0
 * 
 *     cdef int _bmi(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on result minus (i.e. negative flag N is set)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":483
 *         return self._jump_relative(self.N, offset)
 * 
 *     cdef int _bne(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch if zero flag is not set
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bne(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bne", 0);

  /* "nes/cycore/mos6502.pyx":487
 *         Branch if zero flag is not set
 *         """
 *         return self._jump_relative(not self.Z, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bpl(self, int offset, int _):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, (!(__pyx_v_self->Z != 0)), __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":483
 *         return self._jump_relative(self.N, offset)
 * 
 *     cdef int _bne(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch if zero flag is not set
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":489
 *         return self._jump_relative(not self.Z, offset)
 * 
 *     cdef int _bpl(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on result positive (i.e. negative flag N is not set)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bpl(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bpl", 0);

  /* "nes/cycore/mos6502.pyx":494
 *         """
 *         #print(self.N, not self.N)
 *         return self._jump_relative(not self.N, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef _interrupt(self, int interrupt_vector, int is_brk):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, (!(__pyx_v_self->N != 0)), __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":489
 *         return self._jump_relative(not self.Z, offset)
 * 
 *     cdef int _bpl(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on result positive (i.e. negative flag N is not set)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":496
 *         return self._jump_relative(not self.N, offset)
 * 
 *     cdef _interrupt(self, int interrupt_vector, int is_brk):             # <<<<<<<<<<<<<<
 *         """
 *         Interrupt routine, followed (with variations) by NMI, IRQ and BRK
 */

static PyObject *__pyx_f_3nes_6cycore_7mos6502_7MOS6502__interrupt(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_interrupt_vector, int __pyx_v_is_brk) {
  int __pyx_v_v;
  int __pyx_v_addr;
  unsigned char __pyx_v_sr;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  long __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_interrupt", 0);

  /* "nes/cycore/mos6502.pyx":504
 *         cdef unsigned char sr
 *         # push PC (+1 if BRK) to the stack, high bit first
 *         v = self.PC + (1 if is_brk else 0)             # <<<<<<<<<<<<<<
 *         self.push_stack((v & 0xFF00) >> 8)  # high byte
 *         self.push_stack(v & 0xFF)  # low byte
 */
  if ((__pyx_v_is_brk != 0)) {
    __pyx_t_1 = 1;
  } else {
    __pyx_t_1 = 0;
  }
  __pyx_v_v = (__pyx_v_self->PC + __pyx_t_1);

  /* "nes/cycore/mos6502.pyx":505
 *         # push PC (+1 if BRK) to the stack, high bit first
 *         v = self.PC + (1 if is_brk else 0)
 *         self.push_stack((v & 0xFF00) >> 8)  # high byte             # <<<<<<<<<<<<<<
 *         self.push_stack(v & 0xFF)  # low byte
 *         # push the processor status to the stack
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, ((__pyx_v_v & 0xFF00) >> 8));

  /* "nes/cycore/mos6502.pyx":506
 *         v = self.PC + (1 if is_brk else 0)
 *         self.push_stack((v & 0xFF00) >> 8)  # high byte
 *         self.push_stack(v & 0xFF)  # low byte             # <<<<<<<<<<<<<<
 *         # push the processor status to the stack
 *         # BUT note that the B flag ON THE STACK COPY ONLY is now set
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, (__pyx_v_v & 0xFF));

  /* "nes/cycore/mos6502.pyx":509
 *         # push the processor status to the stack
 *         # BUT note that the B flag ON THE STACK COPY ONLY is now set
 *         sr = self._status_to_byte(b_flag=True if is_brk else False)             # <<<<<<<<<<<<<<
 *         self.push_stack(sr)
 *         addr = self._read_word(interrupt_vector, wrap_at_page=False)
 */
  if ((__pyx_v_is_brk != 0)) {
    __pyx_t_2 = 1;
  } else {
    __pyx_t_2 = 0;
  }
  __pyx_v_sr = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_status_to_byte(__pyx_v_self, __pyx_t_2);

  /* "nes/cycore/mos6502.pyx":510
 *         # BUT note that the B flag ON THE STACK COPY ONLY is now set
 *         sr = self._status_to_byte(b_flag=True if is_brk else False)
 *         self.push_stack(sr)             # <<<<<<<<<<<<<<
 *         addr = self._read_word(interrupt_vector, wrap_at_page=False)
 *         self.PC = addr
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, __pyx_v_sr);

  /* "nes/cycore/mos6502.pyx":511
 *         sr = self._status_to_byte(b_flag=True if is_brk else False)
 *         self.push_stack(sr)
 *         addr = self._read_word(interrupt_vector, wrap_at_page=False)             # <<<<<<<<<<<<<<
 *         self.PC = addr
 *         # Set the interrupt disable flag.  But there seems to be disagreement in the sources as to whether or not the
 */
  __pyx_v_addr = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, __pyx_v_interrupt_vector, 0);

  /* "nes/cycore/mos6502.pyx":512
 *         self.push_stack(sr)
 *         addr = self._read_word(interrupt_vector, wrap_at_page=False)
 *         self.PC = addr             # <<<<<<<<<<<<<<
 *         # Set the interrupt disable flag.  But there seems to be disagreement in the sources as to whether or not the
 *         # I flag is set if this is a BRK instruction.
 */
  __pyx_v_self->PC = __pyx_v_addr;

  /* "nes/cycore/mos6502.pyx":523
 *         #   https://www.csh.rit.edu/~moffitt/docs/6502.html#BRK
 *         # Since blargg's tests run on a hardware NES, we follow that behaviour
 *         self.I = True             # <<<<<<<<<<<<<<
 * 
 *     cdef int _brk(self, int _, int __):
 */
  __pyx_v_self->I = 1;

  /* "nes/cycore/mos6502.pyx":496
 *         return self._jump_relative(not self.N, offset)
 * 
 *     cdef _interrupt(self, int interrupt_vector, int is_brk):             # <<<<<<<<<<<<<<
 *         """
 *         Interrupt routine, followed (with variations) by NMI, IRQ and BRK
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":525
 *         self.I = True
 * 
 *     cdef int _brk(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Force break, which simulates an interrupt request (IRQ).
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__brk(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_brk", 0);

  /* "nes/cycore/mos6502.pyx":533
 *         the stack
 *         """
 *         self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=True)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_interrupt(__pyx_v_self, __pyx_e_3nes_6cycore_7mos6502_IRQ_BRK_VECTOR_ADDR, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 533, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "nes/cycore/mos6502.pyx":534
 *         """
 *         self._interrupt(IRQ_BRK_VECTOR_ADDR, is_brk=True)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bvc(self, int offset, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":525
 *         self.I = True
 * 
 *     cdef int _brk(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Force break, which simulates an interrupt request (IRQ).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_WriteUnraisable("nes.cycore.mos6502.MOS6502._brk", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":536
 *         return 0
 * 
 *     cdef int _bvc(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on on overflow clear (V == 0)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bvc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bvc", 0);

  /* "nes/cycore/mos6502.pyx":540
 *         Branch on on overflow clear (V == 0)
 *         """
 *         return self._jump_relative(not self.V, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _bvs(self, int offset, int _):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, (!(__pyx_v_self->V != 0)), __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":536
 *         return 0
 * 
 *     cdef int _bvc(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on on overflow clear (V == 0)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":542
 *         return self._jump_relative(not self.V, offset)
 * 
 *     cdef int _bvs(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on on overflow set (V == 1)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__bvs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_offset, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bvs", 0);

  /* "nes/cycore/mos6502.pyx":546
 *         Branch on on overflow set (V == 1)
 *         """
 *         return self._jump_relative(self.V, offset)             # <<<<<<<<<<<<<<
 * 
 *     cdef int _clc(self, int _, int __):
 */
  __pyx_r = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jump_relative(__pyx_v_self, __pyx_v_self->V, __pyx_v_offset);
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":542
 *         return self._jump_relative(not self.V, offset)
 * 
 *     cdef int _bvs(self, int offset, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Branch on on overflow set (V == 1)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":548
 *         return self._jump_relative(self.V, offset)
 * 
 *     cdef int _clc(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear carry flag (set C:=0)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__clc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clc", 0);

  /* "nes/cycore/mos6502.pyx":552
 *         Clear carry flag (set C:=0)
 *         """
 *         self.C = False             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->C = 0;

  /* "nes/cycore/mos6502.pyx":553
 *         """
 *         self.C = False
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _cld(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":548
 *         return self._jump_relative(self.V, offset)
 * 
 *     cdef int _clc(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear carry flag (set C:=0)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":555
 *         return 0
 * 
 *     cdef int _cld(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear decimal flag (set D:=0)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cld(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_cld", 0);

  /* "nes/cycore/mos6502.pyx":559
 *         Clear decimal flag (set D:=0)
 *         """
 *         self.D = False             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->D = 0;

  /* "nes/cycore/mos6502.pyx":560
 *         """
 *         self.D = False
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _cli(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":555
 *         return 0
 * 
 *     cdef int _cld(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear decimal flag (set D:=0)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":562
 *         return 0
 * 
 *     cdef int _cli(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear interrupt disable flag (set I:=0)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cli(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_cli", 0);

  /* "nes/cycore/mos6502.pyx":566
 *         Clear interrupt disable flag (set I:=0)
 *         """
 *         self.I = False             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->I = 0;

  /* "nes/cycore/mos6502.pyx":567
 *         """
 *         self.I = False
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _clv(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":562
 *         return 0
 * 
 *     cdef int _cli(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear interrupt disable flag (set I:=0)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":569
 *         return 0
 * 
 *     cdef int _clv(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear value flag (set V:=0)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__clv(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_clv", 0);

  /* "nes/cycore/mos6502.pyx":573
 *         Clear value flag (set V:=0)
 *         """
 *         self.V = False             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->V = 0;

  /* "nes/cycore/mos6502.pyx":574
 *         """
 *         self.V = False
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef void _compare(self, int v0, int v):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":569
 *         return 0
 * 
 *     cdef int _clv(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Clear value flag (set V:=0)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":576
 *         return 0
 * 
 *     cdef void _compare(self, int v0, int v):             # <<<<<<<<<<<<<<
 *         self._set_zn(v0 - v)
 *         self.C = v0 >= v
 */

static void __pyx_f_3nes_6cycore_7mos6502_7MOS6502__compare(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_v0, int __pyx_v_v) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_compare", 0);

  /* "nes/cycore/mos6502.pyx":577
 * 
 *     cdef void _compare(self, int v0, int v):
 *         self._set_zn(v0 - v)             # <<<<<<<<<<<<<<
 *         self.C = v0 >= v
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, (__pyx_v_v0 - __pyx_v_v));

  /* "nes/cycore/mos6502.pyx":578
 *     cdef void _compare(self, int v0, int v):
 *         self._set_zn(v0 - v)
 *         self.C = v0 >= v             # <<<<<<<<<<<<<<
 * 
 *     cdef int _cmp(self, int arg, int immediate):
 */
  __pyx_v_self->C = (__pyx_v_v0 >= __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":576
 *         return 0
 * 
 *     cdef void _compare(self, int v0, int v):             # <<<<<<<<<<<<<<
 *         self._set_zn(v0 - v)
 *         self.C = v0 >= v
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "nes/cycore/mos6502.pyx":580
 *         self.C = v0 >= v
 * 
 *     cdef int _cmp(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cmp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_cmp", 0);

  /* "nes/cycore/mos6502.pyx":586
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self._compare(self.A, v)
 *         return 0
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":587
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg
 *         self._compare(self.A, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_compare(__pyx_v_self, __pyx_v_self->A, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":588
 *         v = self.memory.read(arg) if not immediate else arg
 *         self._compare(self.A, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _cpx(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":580
 *         self.C = v0 >= v
 * 
 *     cdef int _cmp(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":590
 *         return 0
 * 
 *     cdef int _cpx(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cpx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_cpx", 0);

  /* "nes/cycore/mos6502.pyx":596
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self._compare(self.X, v)
 *         return 0
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":597
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg
 *         self._compare(self.X, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_compare(__pyx_v_self, __pyx_v_self->X, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":598
 *         v = self.memory.read(arg) if not immediate else arg
 *         self._compare(self.X, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _cpy(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":590
 *         return 0
 * 
 *     cdef int _cpx(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":600
 *         return 0
 * 
 *     cdef int _cpy(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__cpy(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_cpy", 0);

  /* "nes/cycore/mos6502.pyx":606
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self._compare(self.Y, v)
 *         return 0
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":607
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg
 *         self._compare(self.Y, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_compare(__pyx_v_self, __pyx_v_self->Y, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":608
 *         v = self.memory.read(arg) if not immediate else arg
 *         self._compare(self.Y, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _dec(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":600
 *         return 0
 * 
 *     cdef int _cpy(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Sets flags as if a subtraction A - v was performed.  The N flag is valid iff the numbers are signed.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":610
 *         return 0
 * 
 *     cdef int _dec(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Decrement memory in addr by 1
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__dec(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dec", 0);

  /* "nes/cycore/mos6502.pyx":618
 *         """
 *         cdef unsigned char v
 *         v = (self.memory.read(addr) - 1) & 0xFF             # <<<<<<<<<<<<<<
 *         self._set_zn(v)
 *         self.memory.write(addr, v)
 */
  __pyx_v_v = ((((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr) - 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":619
 *         cdef unsigned char v
 *         v = (self.memory.read(addr) - 1) & 0xFF
 *         self._set_zn(v)             # <<<<<<<<<<<<<<
 *         self.memory.write(addr, v)
 *         return 0
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":620
 *         v = (self.memory.read(addr) - 1) & 0xFF
 *         self._set_zn(v)
 *         self.memory.write(addr, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":621
 *         self._set_zn(v)
 *         self.memory.write(addr, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _dex(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":610
 *         return 0
 * 
 *     cdef int _dec(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Decrement memory in addr by 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":623
 *         return 0
 * 
 *     cdef int _dex(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Decrement X by 1
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__dex(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dex", 0);

  /* "nes/cycore/mos6502.pyx":630
 *         :return:
 *         """
 *         self.X = (self.X - 1) & 0xFF             # <<<<<<<<<<<<<<
 *         self._set_zn(self.X)
 *         return 0
 */
  __pyx_v_self->X = ((__pyx_v_self->X - 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":631
 *         """
 *         self.X = (self.X - 1) & 0xFF
 *         self._set_zn(self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->X);

  /* "nes/cycore/mos6502.pyx":632
 *         self.X = (self.X - 1) & 0xFF
 *         self._set_zn(self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _dey(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":623
 *         return 0
 * 
 *     cdef int _dex(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Decrement X by 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":634
 *         return 0
 * 
 *     cdef int _dey(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Decrement Y by 1
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__dey(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dey", 0);

  /* "nes/cycore/mos6502.pyx":641
 *         :return:
 *         """
 *         self.Y = (self.Y - 1) & 0xFF             # <<<<<<<<<<<<<<
 *         self._set_zn(self.Y)
 *         return 0
 */
  __pyx_v_self->Y = ((__pyx_v_self->Y - 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":642
 *         """
 *         self.Y = (self.Y - 1) & 0xFF
 *         self._set_zn(self.Y)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->Y);

  /* "nes/cycore/mos6502.pyx":643
 *         self.Y = (self.Y - 1) & 0xFF
 *         self._set_zn(self.Y)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _eor(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":634
 *         return 0
 * 
 *     cdef int _dey(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Decrement Y by 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":645
 *         return 0
 * 
 *     cdef int _eor(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         XOR A with value and put result back into A
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__eor(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_eor", 0);

  /* "nes/cycore/mos6502.pyx":653
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self.A = v ^ self.A
 *         self._set_zn(self.A)
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":654
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg
 *         self.A = v ^ self.A             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  __pyx_v_self->A = (__pyx_v_v ^ __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":655
 *         v = self.memory.read(arg) if not immediate else arg
 *         self.A = v ^ self.A
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":656
 *         self.A = v ^ self.A
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _inc(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":645
 *         return 0
 * 
 *     cdef int _eor(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         XOR A with value and put result back into A
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":658
 *         return 0
 * 
 *     cdef int _inc(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Increment memory in addr by 1
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__inc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_inc", 0);

  /* "nes/cycore/mos6502.pyx":666
 *         """
 *         cdef unsigned char v
 *         v = (self.memory.read(addr) + 1) & 0xFF             # <<<<<<<<<<<<<<
 *         self._set_zn(v)
 *         self.memory.write(addr, v)
 */
  __pyx_v_v = ((((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr) + 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":667
 *         cdef unsigned char v
 *         v = (self.memory.read(addr) + 1) & 0xFF
 *         self._set_zn(v)             # <<<<<<<<<<<<<<
 *         self.memory.write(addr, v)
 *         return 0
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":668
 *         v = (self.memory.read(addr) + 1) & 0xFF
 *         self._set_zn(v)
 *         self.memory.write(addr, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":669
 *         self._set_zn(v)
 *         self.memory.write(addr, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _inx(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":658
 *         return 0
 * 
 *     cdef int _inc(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Increment memory in addr by 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":671
 *         return 0
 * 
 *     cdef int _inx(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Increment X by 1
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__inx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_inx", 0);

  /* "nes/cycore/mos6502.pyx":678
 *         :return:
 *         """
 *         self.X = (self.X + 1) & 0xFF             # <<<<<<<<<<<<<<
 *         self._set_zn(self.X)
 *         return 0
 */
  __pyx_v_self->X = ((__pyx_v_self->X + 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":679
 *         """
 *         self.X = (self.X + 1) & 0xFF
 *         self._set_zn(self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->X);

  /* "nes/cycore/mos6502.pyx":680
 *         self.X = (self.X + 1) & 0xFF
 *         self._set_zn(self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _iny(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":671
 *         return 0
 * 
 *     cdef int _inx(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Increment X by 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":682
 *         return 0
 * 
 *     cdef int _iny(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Increment Y by 1
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__iny(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_iny", 0);

  /* "nes/cycore/mos6502.pyx":689
 *         :return:
 *         """
 *         self.Y = (self.Y + 1) & 0xFF             # <<<<<<<<<<<<<<
 *         self._set_zn(self.Y)
 *         return 0
 */
  __pyx_v_self->Y = ((__pyx_v_self->Y + 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":690
 *         """
 *         self.Y = (self.Y + 1) & 0xFF
 *         self._set_zn(self.Y)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->Y);

  /* "nes/cycore/mos6502.pyx":691
 *         self.Y = (self.Y + 1) & 0xFF
 *         self._set_zn(self.Y)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _jmp(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":682
 *         return 0
 * 
 *     cdef int _iny(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Increment Y by 1
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":693
 *         return 0
 * 
 *     cdef int _jmp(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Jump to the (16 bit) address addr.  In the case of jump instructions jmp and jsr,
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__jmp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_jmp", 0);

  /* "nes/cycore/mos6502.pyx":701
 *         an immediate address
 *         """
 *         self.PC = addr             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->PC = __pyx_v_addr;

  /* "nes/cycore/mos6502.pyx":702
 *         """
 *         self.PC = addr
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _jsr(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":693
 *         return 0
 * 
 *     cdef int _jmp(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Jump to the (16 bit) address addr.  In the case of jump instructions jmp and jsr,
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":704
 *         return 0
 * 
 *     cdef int _jsr(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Jump to a subroutine at the value in memory[addr].  Addr here is an absolute (16 bit) location.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__jsr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_jsr", 0);

  /* "nes/cycore/mos6502.pyx":715
 *         # save PC - 1 to the stack, high byte first
 *         cdef int v
 *         v = self.PC - 1             # <<<<<<<<<<<<<<
 *         self.push_stack((v & 0xFF00) >> 8)
 *         self.push_stack(v & 0x00FF)
 */
  __pyx_v_v = (__pyx_v_self->PC - 1);

  /* "nes/cycore/mos6502.pyx":716
 *         cdef int v
 *         v = self.PC - 1
 *         self.push_stack((v & 0xFF00) >> 8)             # <<<<<<<<<<<<<<
 *         self.push_stack(v & 0x00FF)
 *         self.PC = addr
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, ((__pyx_v_v & 0xFF00) >> 8));

  /* "nes/cycore/mos6502.pyx":717
 *         v = self.PC - 1
 *         self.push_stack((v & 0xFF00) >> 8)
 *         self.push_stack(v & 0x00FF)             # <<<<<<<<<<<<<<
 *         self.PC = addr
 *         return 0
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, (__pyx_v_v & 0x00FF));

  /* "nes/cycore/mos6502.pyx":718
 *         self.push_stack((v & 0xFF00) >> 8)
 *         self.push_stack(v & 0x00FF)
 *         self.PC = addr             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->PC = __pyx_v_addr;

  /* "nes/cycore/mos6502.pyx":719
 *         self.push_stack(v & 0x00FF)
 *         self.PC = addr
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _lda(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":704
 *         return 0
 * 
 *     cdef int _jsr(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Jump to a subroutine at the value in memory[addr].  Addr here is an absolute (16 bit) location.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":721
 *         return 0
 * 
 *     cdef int _lda(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Load A from value / memory
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__lda(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_lda", 0);

  /* "nes/cycore/mos6502.pyx":726
 *         :return:
 *         """
 *         self.A = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_self->A = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":727
 *         """
 *         self.A = self.memory.read(arg) if not immediate else arg
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":728
 *         self.A = self.memory.read(arg) if not immediate else arg
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _ldx(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":721
 *         return 0
 * 
 *     cdef int _lda(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Load A from value / memory
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":730
 *         return 0
 * 
 *     cdef int _ldx(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Load X from value / memory
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ldx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_ldx", 0);

  /* "nes/cycore/mos6502.pyx":735
 *         :return:
 *         """
 *         self.X = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self._set_zn(self.X)
 *         return 0
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_self->X = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":736
 *         """
 *         self.X = self.memory.read(arg) if not immediate else arg
 *         self._set_zn(self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->X);

  /* "nes/cycore/mos6502.pyx":737
 *         self.X = self.memory.read(arg) if not immediate else arg
 *         self._set_zn(self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _ldy(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":730
 *         return 0
 * 
 *     cdef int _ldx(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Load X from value / memory
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":739
 *         return 0
 * 
 *     cdef int _ldy(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Load Y from value / memory
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ldy(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_ldy", 0);

  /* "nes/cycore/mos6502.pyx":744
 *         :return:
 *         """
 *         self.Y = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self._set_zn(self.Y)
 *         return 0
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_self->Y = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":745
 *         """
 *         self.Y = self.memory.read(arg) if not immediate else arg
 *         self._set_zn(self.Y)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->Y);

  /* "nes/cycore/mos6502.pyx":746
 *         self.Y = self.memory.read(arg) if not immediate else arg
 *         self._set_zn(self.Y)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _lsr(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":739
 *         return 0
 * 
 *     cdef int _ldy(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Load Y from value / memory
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":748
 *         return 0
 * 
 *     cdef int _lsr(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Logical shift right
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__lsr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  unsigned char __pyx_v_result;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_lsr", 0);

  /* "nes/cycore/mos6502.pyx":756
 *         """
 *         cdef unsigned char v, result
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)             # <<<<<<<<<<<<<<
 *         result = (v >> 1) & 0xFF
 * 
 */
  if (((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0)) {
    __pyx_t_1 = __pyx_v_self->A;
  } else {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":757
 *         cdef unsigned char v, result
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)
 *         result = (v >> 1) & 0xFF             # <<<<<<<<<<<<<<
 * 
 *         # status
 */
  __pyx_v_result = ((__pyx_v_v >> 1) & 0xFF);

  /* "nes/cycore/mos6502.pyx":760
 * 
 *         # status
 *         self.C = v & 0x01             # <<<<<<<<<<<<<<
 *         self.N = False
 *         self.Z = result == 0
 */
  __pyx_v_self->C = (__pyx_v_v & 0x01);

  /* "nes/cycore/mos6502.pyx":761
 *         # status
 *         self.C = v & 0x01
 *         self.N = False             # <<<<<<<<<<<<<<
 *         self.Z = result == 0
 * 
 */
  __pyx_v_self->N = 0;

  /* "nes/cycore/mos6502.pyx":762
 *         self.C = v & 0x01
 *         self.N = False
 *         self.Z = result == 0             # <<<<<<<<<<<<<<
 * 
 *         if addr==ARG_NONE:
 */
  __pyx_v_self->Z = (__pyx_v_result == 0);

  /* "nes/cycore/mos6502.pyx":764
 *         self.Z = result == 0
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = result
 */
  __pyx_t_2 = ((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0);
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":766
 *         if addr==ARG_NONE:
 *             # target the accumulator
 *             self.A = result             # <<<<<<<<<<<<<<
 *         else:
 *             self.memory.write(addr, result)
 */
    __pyx_v_self->A = __pyx_v_result;

    /* "nes/cycore/mos6502.pyx":764
 *         self.Z = result == 0
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = result
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":768
 *             self.A = result
 *         else:
 *             self.memory.write(addr, result)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_result);
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":769
 *         else:
 *             self.memory.write(addr, result)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _nop(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":748
 *         return 0
 * 
 *     cdef int _lsr(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Logical shift right
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":771
 *         return 0
 * 
 *     cdef int _nop(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         No-op.  This one is easy.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__nop(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_nop", 0);

  /* "nes/cycore/mos6502.pyx":776
 *         :return:
 *         """
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _ora(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":771
 *         return 0
 * 
 *     cdef int _nop(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         No-op.  This one is easy.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":778
 *         return 0
 * 
 *     cdef int _ora(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Bitwise OR with accumulator; result put into accumulator.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ora(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_ora", 0);

  /* "nes/cycore/mos6502.pyx":784
 *         """
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         self.A = self.A | v
 *         self._set_zn(self.A)
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":785
 *         cdef unsigned char v
 *         v = self.memory.read(arg) if not immediate else arg
 *         self.A = self.A | v             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  __pyx_v_self->A = (__pyx_v_self->A | __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":786
 *         v = self.memory.read(arg) if not immediate else arg
 *         self.A = self.A | v
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":787
 *         self.A = self.A | v
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _pha(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":778
 *         return 0
 * 
 *     cdef int _ora(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Bitwise OR with accumulator; result put into accumulator.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":789
 *         return 0
 * 
 *     cdef int _pha(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Push A onto stack
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__pha(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pha", 0);

  /* "nes/cycore/mos6502.pyx":796
 *         :return:
 *         """
 *         self.push_stack(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":797
 *         """
 *         self.push_stack(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _php(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":789
 *         return 0
 * 
 *     cdef int _pha(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Push A onto stack
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":799
 *         return 0
 * 
 *     cdef int _php(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Push status register onto stack
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__php(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_php", 0);

  /* "nes/cycore/mos6502.pyx":808
 *         cdef unsigned char v
 * 
 *         v = self._status_to_byte(b_flag=True)  # b-flag should be set True in this case [10]             # <<<<<<<<<<<<<<
 *         self.push_stack(v)
 *         return 0
 */
  __pyx_v_v = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_status_to_byte(__pyx_v_self, 1);

  /* "nes/cycore/mos6502.pyx":809
 * 
 *         v = self._status_to_byte(b_flag=True)  # b-flag should be set True in this case [10]
 *         self.push_stack(v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->push_stack(__pyx_v_self, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":810
 *         v = self._status_to_byte(b_flag=True)  # b-flag should be set True in this case [10]
 *         self.push_stack(v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _pla(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":799
 *         return 0
 * 
 *     cdef int _php(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Push status register onto stack
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":812
 *         return 0
 * 
 *     cdef int _pla(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Pull A from stack
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__pla(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_pla", 0);

  /* "nes/cycore/mos6502.pyx":819
 *         :return:
 *         """
 *         self.A = self.pop_stack()             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  __pyx_v_self->A = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":820
 *         """
 *         self.A = self.pop_stack()
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":821
 *         self.A = self.pop_stack()
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _plp(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":812
 *         return 0
 * 
 *     cdef int _pla(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Pull A from stack
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":823
 *         return 0
 * 
 *     cdef int _plp(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Pull processor status from the stack
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__plp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  unsigned char __pyx_v_v;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_plp", 0);

  /* "nes/cycore/mos6502.pyx":829
 *         """
 *         cdef unsigned char v
 *         v = self.pop_stack()             # <<<<<<<<<<<<<<
 *         self._status_from_byte(v)
 *         return 0
 */
  __pyx_v_v = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":830
 *         cdef unsigned char v
 *         v = self.pop_stack()
 *         self._status_from_byte(v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_status_from_byte(__pyx_v_self, __pyx_v_v);

  /* "nes/cycore/mos6502.pyx":831
 *         v = self.pop_stack()
 *         self._status_from_byte(v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _rol(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":823
 *         return 0
 * 
 *     cdef int _plp(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Pull processor status from the stack
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":833
 *         return 0
 * 
 *     cdef int _rol(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Rotate one bit left (including carry bit in the rotation, i.e. C rotates into lsb and msb rotates into C)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__rol(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  int __pyx_v_result;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_rol", 0);

  /* "nes/cycore/mos6502.pyx":841
 *         cdef int result
 * 
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)             # <<<<<<<<<<<<<<
 *         result = (v << 1) + self.C
 * 
 */
  if (((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0)) {
    __pyx_t_1 = __pyx_v_self->A;
  } else {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":842
 * 
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)
 *         result = (v << 1) + self.C             # <<<<<<<<<<<<<<
 * 
 *         # status
 */
  __pyx_v_result = ((__pyx_v_v << 1) + __pyx_v_self->C);

  /* "nes/cycore/mos6502.pyx":845
 * 
 *         # status
 *         self.C = result > 255             # <<<<<<<<<<<<<<
 *         self._set_zn(result)
 * 
 */
  __pyx_v_self->C = (__pyx_v_result > 0xFF);

  /* "nes/cycore/mos6502.pyx":846
 *         # status
 *         self.C = result > 255
 *         self._set_zn(result)             # <<<<<<<<<<<<<<
 * 
 *         if addr==ARG_NONE:
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_result);

  /* "nes/cycore/mos6502.pyx":848
 *         self._set_zn(result)
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = result & 0xFF
 */
  __pyx_t_2 = ((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0);
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":850
 *         if addr==ARG_NONE:
 *             # target the accumulator
 *             self.A = result & 0xFF             # <<<<<<<<<<<<<<
 *         else:
 *             self.memory.write(addr, result & 0xFF)
 */
    __pyx_v_self->A = (__pyx_v_result & 0xFF);

    /* "nes/cycore/mos6502.pyx":848
 *         self._set_zn(result)
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = result & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":852
 *             self.A = result & 0xFF
 *         else:
 *             self.memory.write(addr, result & 0xFF)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, (__pyx_v_result & 0xFF));
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":853
 *         else:
 *             self.memory.write(addr, result & 0xFF)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _ror(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":833
 *         return 0
 * 
 *     cdef int _rol(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Rotate one bit left (including carry bit in the rotation, i.e. C rotates into lsb and msb rotates into C)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":855
 *         return 0
 * 
 *     cdef int _ror(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Rotate one bit right (including carry bit in the rotation, i.e. C rotates into msb and lsb rotates into C)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__ror(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  unsigned char __pyx_v_v;
  int __pyx_v_result;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_ror", 0);

  /* "nes/cycore/mos6502.pyx":862
 *         cdef unsigned char v
 *         cdef int result
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)             # <<<<<<<<<<<<<<
 *         result = (v >> 1) + (self.C << 7)
 * 
 */
  if (((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0)) {
    __pyx_t_1 = __pyx_v_self->A;
  } else {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr);
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":863
 *         cdef int result
 *         v = self.A if addr==ARG_NONE else self.memory.read(addr)
 *         result = (v >> 1) + (self.C << 7)             # <<<<<<<<<<<<<<
 * 
 *         # status
 */
  __pyx_v_result = ((__pyx_v_v >> 1) + (__pyx_v_self->C << 7));

  /* "nes/cycore/mos6502.pyx":866
 * 
 *         # status
 *         self.C = v & 0x01             # <<<<<<<<<<<<<<
 *         self._set_zn(result)
 * 
 */
  __pyx_v_self->C = (__pyx_v_v & 0x01);

  /* "nes/cycore/mos6502.pyx":867
 *         # status
 *         self.C = v & 0x01
 *         self._set_zn(result)             # <<<<<<<<<<<<<<
 * 
 *         if addr==ARG_NONE:
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_result);

  /* "nes/cycore/mos6502.pyx":869
 *         self._set_zn(result)
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = result & 0xFF
 */
  __pyx_t_2 = ((__pyx_v_addr == __pyx_e_3nes_6cycore_7mos6502_ARG_NONE) != 0);
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":871
 *         if addr==ARG_NONE:
 *             # target the accumulator
 *             self.A = result & 0xFF             # <<<<<<<<<<<<<<
 *         else:
 *             self.memory.write(addr, result & 0xFF)
 */
    __pyx_v_self->A = (__pyx_v_result & 0xFF);

    /* "nes/cycore/mos6502.pyx":869
 *         self._set_zn(result)
 * 
 *         if addr==ARG_NONE:             # <<<<<<<<<<<<<<
 *             # target the accumulator
 *             self.A = result & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":873
 *             self.A = result & 0xFF
 *         else:
 *             self.memory.write(addr, result & 0xFF)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  /*else*/ {
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, (__pyx_v_result & 0xFF));
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":874
 *         else:
 *             self.memory.write(addr, result & 0xFF)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _rti(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":855
 *         return 0
 * 
 *     cdef int _ror(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Rotate one bit right (including carry bit in the rotation, i.e. C rotates into msb and lsb rotates into C)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":876
 *         return 0
 * 
 *     cdef int _rti(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Return from an interrupt.  Gets the PC of the last execution point and the flag register from the
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__rti(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  unsigned char __pyx_v_flags;
  unsigned char __pyx_v_addr_lo;
  unsigned char __pyx_v_addr_hi;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rti", 0);

  /* "nes/cycore/mos6502.pyx":883
 *         """
 *         cdef unsigned char flags, addr_lo, addr_hi
 *         flags = self.pop_stack()             # <<<<<<<<<<<<<<
 *         addr_lo = self.pop_stack()
 *         addr_hi = self.pop_stack()
 */
  __pyx_v_flags = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":884
 *         cdef unsigned char flags, addr_lo, addr_hi
 *         flags = self.pop_stack()
 *         addr_lo = self.pop_stack()             # <<<<<<<<<<<<<<
 *         addr_hi = self.pop_stack()
 *         self._status_from_byte(flags)
 */
  __pyx_v_addr_lo = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":885
 *         flags = self.pop_stack()
 *         addr_lo = self.pop_stack()
 *         addr_hi = self.pop_stack()             # <<<<<<<<<<<<<<
 *         self._status_from_byte(flags)
 *         self.PC = (addr_hi << 8) + addr_lo
 */
  __pyx_v_addr_hi = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":886
 *         addr_lo = self.pop_stack()
 *         addr_hi = self.pop_stack()
 *         self._status_from_byte(flags)             # <<<<<<<<<<<<<<
 *         self.PC = (addr_hi << 8) + addr_lo
 *         return 0
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_status_from_byte(__pyx_v_self, __pyx_v_flags);

  /* "nes/cycore/mos6502.pyx":887
 *         addr_hi = self.pop_stack()
 *         self._status_from_byte(flags)
 *         self.PC = (addr_hi << 8) + addr_lo             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->PC = ((__pyx_v_addr_hi << 8) + __pyx_v_addr_lo);

  /* "nes/cycore/mos6502.pyx":888
 *         self._status_from_byte(flags)
 *         self.PC = (addr_hi << 8) + addr_lo
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _rts(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":876
 *         return 0
 * 
 *     cdef int _rti(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Return from an interrupt.  Gets the PC of the last execution point and the flag register from the
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":890
 *         return 0
 * 
 *     cdef int _rts(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Returns from a subroutine (see jsr for the jump subroutine)
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__rts(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  unsigned char __pyx_v_addr_lo;
  unsigned char __pyx_v_addr_hi;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rts", 0);

  /* "nes/cycore/mos6502.pyx":897
 *         cdef unsigned char addr_lo, addr_hi
 *         # fetch PC-1 from the stack, low byte first
 *         addr_lo = self.pop_stack()             # <<<<<<<<<<<<<<
 *         addr_hi = self.pop_stack()
 *         # restore PC to the stack
 */
  __pyx_v_addr_lo = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":898
 *         # fetch PC-1 from the stack, low byte first
 *         addr_lo = self.pop_stack()
 *         addr_hi = self.pop_stack()             # <<<<<<<<<<<<<<
 *         # restore PC to the stack
 *         self.PC = (addr_hi << 8) + addr_lo + 1
 */
  __pyx_v_addr_hi = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->pop_stack(__pyx_v_self);

  /* "nes/cycore/mos6502.pyx":900
 *         addr_hi = self.pop_stack()
 *         # restore PC to the stack
 *         self.PC = (addr_hi << 8) + addr_lo + 1             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->PC = (((__pyx_v_addr_hi << 8) + __pyx_v_addr_lo) + 1);

  /* "nes/cycore/mos6502.pyx":901
 *         # restore PC to the stack
 *         self.PC = (addr_hi << 8) + addr_lo + 1
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _sbc(self, int arg, int immediate):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":890
 *         return 0
 * 
 *     cdef int _rts(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Returns from a subroutine (see jsr for the jump subroutine)
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":903
 *         return 0
 * 
 *     cdef int _sbc(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Subtract memory plus carry flag from A and put the result in A (i.e. A := A - M - ~C).
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sbc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_arg, int __pyx_v_immediate) {
  unsigned char __pyx_v_v;
  unsigned char __pyx_v_borrow;
  int __pyx_v_result;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("_sbc", 0);

  /* "nes/cycore/mos6502.pyx":916
 *         cdef unsigned char v, borrow
 *         cdef int result
 *         v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *         borrow = not self.C   # incoming borrow is indicated by not Carry  (borrow is either 0 or 1)
 * 
 */
  if (((!(__pyx_v_immediate != 0)) != 0)) {
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_arg);
  } else {
    __pyx_t_1 = __pyx_v_arg;
  }
  __pyx_v_v = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":917
 *         cdef int result
 *         v = self.memory.read(arg) if not immediate else arg
 *         borrow = not self.C   # incoming borrow is indicated by not Carry  (borrow is either 0 or 1)             # <<<<<<<<<<<<<<
 * 
 *         result = self.A - v - borrow     # arithmetic is done unsigned
 */
  __pyx_v_borrow = (!(__pyx_v_self->C != 0));

  /* "nes/cycore/mos6502.pyx":919
 *         borrow = not self.C   # incoming borrow is indicated by not Carry  (borrow is either 0 or 1)
 * 
 *         result = self.A - v - borrow     # arithmetic is done unsigned             # <<<<<<<<<<<<<<
 *         self.C = result >= 0  # not Carry == borrow; borrow if res<0, so C==not(res<0) == res>=0
 *         # overflow - see notes in description above
 */
  __pyx_v_result = ((__pyx_v_self->A - __pyx_v_v) - __pyx_v_borrow);

  /* "nes/cycore/mos6502.pyx":920
 * 
 *         result = self.A - v - borrow     # arithmetic is done unsigned
 *         self.C = result >= 0  # not Carry == borrow; borrow if res<0, so C==not(res<0) == res>=0             # <<<<<<<<<<<<<<
 *         # overflow - see notes in description above
 *         # I think this is correct and that the borrow doesn't matter here
 */
  __pyx_v_self->C = (__pyx_v_result >= 0);

  /* "nes/cycore/mos6502.pyx":923
 *         # overflow - see notes in description above
 *         # I think this is correct and that the borrow doesn't matter here
 *         self.V = (self._neg(self.A) != self._neg(v)) and (self._neg(v) == self._neg(result))             # <<<<<<<<<<<<<<
 *         self._set_zn(result)
 *         self.A = result & 0xFF
 */
  __pyx_t_2 = (((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_self->A) != ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_v));
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = (((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_v) == ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_neg(__pyx_v_self, __pyx_v_result));
  __pyx_t_1 = __pyx_t_2;
  __pyx_L3_bool_binop_done:;
  __pyx_v_self->V = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":924
 *         # I think this is correct and that the borrow doesn't matter here
 *         self.V = (self._neg(self.A) != self._neg(v)) and (self._neg(v) == self._neg(result))
 *         self._set_zn(result)             # <<<<<<<<<<<<<<
 *         self.A = result & 0xFF
 *         return 0
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_result);

  /* "nes/cycore/mos6502.pyx":925
 *         self.V = (self._neg(self.A) != self._neg(v)) and (self._neg(v) == self._neg(result))
 *         self._set_zn(result)
 *         self.A = result & 0xFF             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->A = (__pyx_v_result & 0xFF);

  /* "nes/cycore/mos6502.pyx":926
 *         self._set_zn(result)
 *         self.A = result & 0xFF
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _sec(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":903
 *         return 0
 * 
 *     cdef int _sbc(self, int arg, int immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Subtract memory plus carry flag from A and put the result in A (i.e. A := A - M - ~C).
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":928
 *         return 0
 * 
 *     cdef int _sec(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Set carry flag
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sec(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sec", 0);

  /* "nes/cycore/mos6502.pyx":932
 *         Set carry flag
 *         """
 *         self.C = True             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->C = 1;

  /* "nes/cycore/mos6502.pyx":933
 *         """
 *         self.C = True
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _sed(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":928
 *         return 0
 * 
 *     cdef int _sec(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Set carry flag
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":935
 *         return 0
 * 
 *     cdef int _sed(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Set decimal flag
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sed(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sed", 0);

  /* "nes/cycore/mos6502.pyx":939
 *         Set decimal flag
 *         """
 *         self.D = True             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->D = 1;

  /* "nes/cycore/mos6502.pyx":940
 *         """
 *         self.D = True
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _sei(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":935
 *         return 0
 * 
 *     cdef int _sed(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Set decimal flag
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":942
 *         return 0
 * 
 *     cdef int _sei(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Set interrupt disable flag
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sei(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sei", 0);

  /* "nes/cycore/mos6502.pyx":946
 *         Set interrupt disable flag
 *         """
 *         self.I = True             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_v_self->I = 1;

  /* "nes/cycore/mos6502.pyx":947
 *         """
 *         self.I = True
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _sta(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":942
 *         return 0
 * 
 *     cdef int _sei(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Set interrupt disable flag
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":949
 *         return 0
 * 
 *     cdef int _sta(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Store A in memory.  No flags are set.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sta(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sta", 0);

  /* "nes/cycore/mos6502.pyx":954
 *         :return:
 *         """
 *         self.memory.write(addr, self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":955
 *         """
 *         self.memory.write(addr, self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _stx(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":949
 *         return 0
 * 
 *     cdef int _sta(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Store A in memory.  No flags are set.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":957
 *         return 0
 * 
 *     cdef int _stx(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Store X in memory.  No flags are set.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__stx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_stx", 0);

  /* "nes/cycore/mos6502.pyx":962
 *         :return:
 *         """
 *         self.memory.write(addr, self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_self->X);

  /* "nes/cycore/mos6502.pyx":963
 *         """
 *         self.memory.write(addr, self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _sty(self, int addr, int _):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":957
 *         return 0
 * 
 *     cdef int _stx(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Store X in memory.  No flags are set.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":965
 *         return 0
 * 
 *     cdef int _sty(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Store Y in memory.  No flags are set.
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__sty(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, int __pyx_v_addr, CYTHON_UNUSED int __pyx_v__) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sty", 0);

  /* "nes/cycore/mos6502.pyx":970
 *         :return:
 *         """
 *         self.memory.write(addr, self.Y)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_v_addr, __pyx_v_self->Y);

  /* "nes/cycore/mos6502.pyx":971
 *         """
 *         self.memory.write(addr, self.Y)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _tax(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":965
 *         return 0
 * 
 *     cdef int _sty(self, int addr, int _):             # <<<<<<<<<<<<<<
 *         """
 *         Store Y in memory.  No flags are set.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":973
 *         return 0
 * 
 *     cdef int _tax(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer A to X
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tax(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("_tax", 0);

  /* "nes/cycore/mos6502.pyx":978
 *         :return:
 *         """
 *         self.X = self.A             # <<<<<<<<<<<<<<
 *         self._set_zn(self.X)
 *         return 0
 */
  __pyx_t_1 = __pyx_v_self->A;
  __pyx_v_self->X = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":979
 *         """
 *         self.X = self.A
 *         self._set_zn(self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->X);

  /* "nes/cycore/mos6502.pyx":980
 *         self.X = self.A
 *         self._set_zn(self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _tay(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":973
 *         return 0
 * 
 *     cdef int _tax(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer A to X
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":982
 *         return 0
 * 
 *     cdef int _tay(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer A to Y
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tay(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("_tay", 0);

  /* "nes/cycore/mos6502.pyx":987
 *         :return:
 *         """
 *         self.Y = self.A             # <<<<<<<<<<<<<<
 *         self._set_zn(self.Y)
 *         return 0
 */
  __pyx_t_1 = __pyx_v_self->A;
  __pyx_v_self->Y = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":988
 *         """
 *         self.Y = self.A
 *         self._set_zn(self.Y)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->Y);

  /* "nes/cycore/mos6502.pyx":989
 *         self.Y = self.A
 *         self._set_zn(self.Y)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _tsx(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":982
 *         return 0
 * 
 *     cdef int _tay(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer A to Y
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":991
 *         return 0
 * 
 *     cdef int _tsx(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer SP to X
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tsx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_tsx", 0);

  /* "nes/cycore/mos6502.pyx":996
 *         :return:
 *         """
 *         self.X = self.SP             # <<<<<<<<<<<<<<
 *         self._set_zn(self.X)
 *         return 0
 */
  __pyx_t_1 = __pyx_v_self->SP;
  __pyx_v_self->X = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":997
 *         """
 *         self.X = self.SP
 *         self._set_zn(self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->X);

  /* "nes/cycore/mos6502.pyx":998
 *         self.X = self.SP
 *         self._set_zn(self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _txa(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":991
 *         return 0
 * 
 *     cdef int _tsx(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer SP to X
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1000
 *         return 0
 * 
 *     cdef int _txa(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer X to A
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__txa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("_txa", 0);

  /* "nes/cycore/mos6502.pyx":1005
 *         :return:
 *         """
 *         self.A = self.X             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  __pyx_t_1 = __pyx_v_self->X;
  __pyx_v_self->A = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":1006
 *         """
 *         self.A = self.X
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":1007
 *         self.A = self.X
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _txs(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1000
 *         return 0
 * 
 *     cdef int _txa(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer X to A
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1009
 *         return 0
 * 
 *     cdef int _txs(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer X to SP
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__txs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("_txs", 0);

  /* "nes/cycore/mos6502.pyx":1014
 *         :return:
 *         """
 *         self.SP = self.X             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_t_1 = __pyx_v_self->X;
  __pyx_v_self->SP = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":1015
 *         """
 *         self.SP = self.X
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int _tya(self, int _, int __):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1009
 *         return 0
 * 
 *     cdef int _txs(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer X to SP
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1017
 *         return 0
 * 
 *     cdef int _tya(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer Y to A
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502__tya(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED int __pyx_v__, CYTHON_UNUSED int __pyx_v___) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  __Pyx_RefNannySetupContext("_tya", 0);

  /* "nes/cycore/mos6502.pyx":1022
 *         :return:
 *         """
 *         self.A = self.Y             # <<<<<<<<<<<<<<
 *         self._set_zn(self.A)
 *         return 0
 */
  __pyx_t_1 = __pyx_v_self->Y;
  __pyx_v_self->A = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":1023
 *         """
 *         self.A = self.Y
 *         self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

  /* "nes/cycore/mos6502.pyx":1024
 *         self.A = self.Y
 *         self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     ################################## The undocumented instructions (see [13] and others) #############################
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1017
 *         return 0
 * 
 *     cdef int _tya(self, int _, int __):             # <<<<<<<<<<<<<<
 *         """
 *         Transfer Y to A
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1028
 *     ################################## The undocumented instructions (see [13] and others) #############################
 * 
 *     def _dop(self, _, __):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_13_dop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_12_dop[] = "\n        Undocumented.\n        Double NOP instruction.  Luckily it is easy to implement!\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_13_dop(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  CYTHON_UNUSED PyObject *__pyx_v___ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dop (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__9,&__pyx_n_s__10,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__10)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_dop", 1, 2, 2, 1); __PYX_ERR(0, 1028, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_dop") < 0)) __PYX_ERR(0, 1028, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v__ = values[0];
    __pyx_v___ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_dop", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1028, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._dop", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_12_dop(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v__, __pyx_v___);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_12_dop(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__, CYTHON_UNUSED PyObject *__pyx_v___) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dop", 0);

  /* "nes/cycore/mos6502.pyx":1033
 *         Double NOP instruction.  Luckily it is easy to implement!
 *         """
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _top(self, _, __):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1028
 *     ################################## The undocumented instructions (see [13] and others) #############################
 * 
 *     def _dop(self, _, __):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1035
 *         return 0
 * 
 *     def _top(self, _, __):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_15_top(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_14_top[] = "\n        Undocumented.\n        Triple NOP instruction.\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_15_top(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  CYTHON_UNUSED PyObject *__pyx_v___ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_top (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__9,&__pyx_n_s__10,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__10)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_top", 1, 2, 2, 1); __PYX_ERR(0, 1035, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_top") < 0)) __PYX_ERR(0, 1035, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v__ = values[0];
    __pyx_v___ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_top", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1035, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._top", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_14_top(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v__, __pyx_v___);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_14_top(CYTHON_UNUSED struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__, CYTHON_UNUSED PyObject *__pyx_v___) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_top", 0);

  /* "nes/cycore/mos6502.pyx":1040
 *         Triple NOP instruction.
 *         """
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _kil(self, _, __):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1035
 *         return 0
 * 
 *     def _top(self, _, __):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1042
 *         return 0
 * 
 *     def _kil(self, _, __):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_17_kil(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_16_kil[] = "\n        Undocumented.\n        Shuts down the processor.  No recovery is possible after this has executed, short of a reset.\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_17_kil(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  CYTHON_UNUSED PyObject *__pyx_v___ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_kil (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s__9,&__pyx_n_s__10,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__10)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_kil", 1, 2, 2, 1); __PYX_ERR(0, 1042, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_kil") < 0)) __PYX_ERR(0, 1042, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v__ = values[0];
    __pyx_v___ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_kil", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1042, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._kil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_16_kil(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v__, __pyx_v___);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_16_kil(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v__, CYTHON_UNUSED PyObject *__pyx_v___) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_kil", 0);

  /* "nes/cycore/mos6502.pyx":1047
 *         Shuts down the processor.  No recovery is possible after this has executed, short of a reset.
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("KIL instruction.  Processor halted.")
 *         return 0
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "nes/cycore/mos6502.pyx":1048
 *         """
 *         if self.undocumented_support_level >= 1:
 *             raise ValueError("KIL instruction.  Processor halted.")             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1048, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 1048, __pyx_L1_error)

    /* "nes/cycore/mos6502.pyx":1047
 *         Shuts down the processor.  No recovery is possible after this has executed, short of a reset.
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             raise ValueError("KIL instruction.  Processor halted.")
 *         return 0
 */
  }

  /* "nes/cycore/mos6502.pyx":1049
 *         if self.undocumented_support_level >= 1:
 *             raise ValueError("KIL instruction.  Processor halted.")
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _lax(self, arg, immediate):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1042
 *         return 0
 * 
 *     def _kil(self, _, __):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._kil", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1051
 *         return 0
 * 
 *     def _lax(self, arg, immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_19_lax(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_18_lax[] = "\n        Undocumented.\n        Load A and X from value / memory\n        :return:\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_19_lax(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  PyObject *__pyx_v_immediate = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_lax (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s_immediate,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_immediate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_lax", 1, 2, 2, 1); __PYX_ERR(0, 1051, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_lax") < 0)) __PYX_ERR(0, 1051, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v_immediate = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_lax", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1051, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._lax", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_18_lax(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v_immediate);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_18_lax(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_immediate) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  unsigned char __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_lax", 0);

  /* "nes/cycore/mos6502.pyx":1057
 *         :return:
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = self.memory.read(arg) if not immediate else arg
 *             self.A = v
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1058
 *         """
 *         if self.undocumented_support_level >= 1:
 *             v = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 *             self.A = v
 *             self.X = v
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_immediate); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 1058, __pyx_L1_error)
    if (((!__pyx_t_1) != 0)) {
      __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_arg); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1058, __pyx_L1_error)
      __pyx_t_4 = __Pyx_PyInt_From_unsigned_char(((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_3)); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1058, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_2 = __pyx_t_4;
      __pyx_t_4 = 0;
    } else {
      __Pyx_INCREF(__pyx_v_arg);
      __pyx_t_2 = __pyx_v_arg;
    }
    __pyx_v_v = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "nes/cycore/mos6502.pyx":1059
 *         if self.undocumented_support_level >= 1:
 *             v = self.memory.read(arg) if not immediate else arg
 *             self.A = v             # <<<<<<<<<<<<<<
 *             self.X = v
 *             self._set_zn(self.A)
 */
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_char(__pyx_v_v); if (unlikely((__pyx_t_5 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1059, __pyx_L1_error)
    __pyx_v_self->A = __pyx_t_5;

    /* "nes/cycore/mos6502.pyx":1060
 *             v = self.memory.read(arg) if not immediate else arg
 *             self.A = v
 *             self.X = v             # <<<<<<<<<<<<<<
 *             self._set_zn(self.A)
 *         return 0
 */
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_char(__pyx_v_v); if (unlikely((__pyx_t_5 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1060, __pyx_L1_error)
    __pyx_v_self->X = __pyx_t_5;

    /* "nes/cycore/mos6502.pyx":1061
 *             self.A = v
 *             self.X = v
 *             self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

    /* "nes/cycore/mos6502.pyx":1057
 *         :return:
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = self.memory.read(arg) if not immediate else arg
 *             self.A = v
 */
  }

  /* "nes/cycore/mos6502.pyx":1062
 *             self.X = v
 *             self._set_zn(self.A)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _aax(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1051
 *         return 0
 * 
 *     def _lax(self, arg, immediate):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._lax", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1064
 *         return 0
 * 
 *     def _aax(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_21_aax(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_20_aax[] = "\n        Undocumented.\n        AND X register with accumulator and store result in memory.\n        Status flags: N,Z  (or are they?  nestest does not think so...)\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_21_aax(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_aax (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_aax", 1, 2, 2, 1); __PYX_ERR(0, 1064, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_aax") < 0)) __PYX_ERR(0, 1064, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_aax", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1064, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._aax", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_20_aax(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_20_aax(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  int __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_aax", 0);

  /* "nes/cycore/mos6502.pyx":1070
 *         Status flags: N,Z  (or are they?  nestest does not think so...)
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = self.A & self.X
 *             self.memory.write(addr, v)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1071
 *         """
 *         if self.undocumented_support_level >= 1:
 *             v = self.A & self.X             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, v)
 *             if self.aax_sets_flags:
 */
    __pyx_v_v = (__pyx_v_self->A & __pyx_v_self->X);

    /* "nes/cycore/mos6502.pyx":1072
 *         if self.undocumented_support_level >= 1:
 *             v = self.A & self.X
 *             self.memory.write(addr, v)             # <<<<<<<<<<<<<<
 *             if self.aax_sets_flags:
 *                 # the behaviour of this seems unclear, nestest does not think it does this, [13] does
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1072, __pyx_L1_error)
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2, __pyx_v_v);

    /* "nes/cycore/mos6502.pyx":1073
 *             v = self.A & self.X
 *             self.memory.write(addr, v)
 *             if self.aax_sets_flags:             # <<<<<<<<<<<<<<
 *                 # the behaviour of this seems unclear, nestest does not think it does this, [13] does
 *                 self._set_zn(v)
 */
    __pyx_t_1 = (__pyx_v_self->aax_sets_flags != 0);
    if (__pyx_t_1) {

      /* "nes/cycore/mos6502.pyx":1075
 *             if self.aax_sets_flags:
 *                 # the behaviour of this seems unclear, nestest does not think it does this, [13] does
 *                 self._set_zn(v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_v);

      /* "nes/cycore/mos6502.pyx":1073
 *             v = self.A & self.X
 *             self.memory.write(addr, v)
 *             if self.aax_sets_flags:             # <<<<<<<<<<<<<<
 *                 # the behaviour of this seems unclear, nestest does not think it does this, [13] does
 *                 self._set_zn(v)
 */
    }

    /* "nes/cycore/mos6502.pyx":1070
 *         Status flags: N,Z  (or are they?  nestest does not think so...)
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = self.A & self.X
 *             self.memory.write(addr, v)
 */
  }

  /* "nes/cycore/mos6502.pyx":1076
 *                 # the behaviour of this seems unclear, nestest does not think it does this, [13] does
 *                 self._set_zn(v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _dcp(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1064
 *         return 0
 * 
 *     def _aax(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._aax", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1078
 *         return 0
 * 
 *     def _dcp(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_23_dcp(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_22_dcp[] = "\n        Undocumented.\n        This opcode DECs the contents of a memory location and then CMPs the result with the A register.\n        Equivalent to DEC oper, CMP oper [14]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_23_dcp(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_dcp (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_dcp", 1, 2, 2, 1); __PYX_ERR(0, 1078, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_dcp") < 0)) __PYX_ERR(0, 1078, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_dcp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1078, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._dcp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_22_dcp(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_22_dcp(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  long __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_dcp", 0);

  /* "nes/cycore/mos6502.pyx":1084
 *         Equivalent to DEC oper, CMP oper [14]
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = (self.memory.read(addr) - 1) & 0xFF
 *             self.memory.write(addr, v)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1085
 *         """
 *         if self.undocumented_support_level >= 1:
 *             v = (self.memory.read(addr) - 1) & 0xFF             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, v)
 *             self._compare(self.A, v)
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1085, __pyx_L1_error)
    __pyx_v_v = ((((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2) - 1) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1086
 *         if self.undocumented_support_level >= 1:
 *             v = (self.memory.read(addr) - 1) & 0xFF
 *             self.memory.write(addr, v)             # <<<<<<<<<<<<<<
 *             self._compare(self.A, v)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1086, __pyx_L1_error)
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2, __pyx_v_v);

    /* "nes/cycore/mos6502.pyx":1087
 *             v = (self.memory.read(addr) - 1) & 0xFF
 *             self.memory.write(addr, v)
 *             self._compare(self.A, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_compare(__pyx_v_self, __pyx_v_self->A, __pyx_v_v);

    /* "nes/cycore/mos6502.pyx":1084
 *         Equivalent to DEC oper, CMP oper [14]
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = (self.memory.read(addr) - 1) & 0xFF
 *             self.memory.write(addr, v)
 */
  }

  /* "nes/cycore/mos6502.pyx":1088
 *             self.memory.write(addr, v)
 *             self._compare(self.A, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _isc(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1078
 *         return 0
 * 
 *     def _dcp(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._dcp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1090
 *         return 0
 * 
 *     def _isc(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_25_isc(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_24_isc[] = "\n        Undocumented.\n        Increase memory by one, then subtract memory from accumulator (with borrow). Status flags: N,V,Z,C\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_25_isc(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_isc (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_isc", 1, 2, 2, 1); __PYX_ERR(0, 1090, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_isc") < 0)) __PYX_ERR(0, 1090, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_isc", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1090, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._isc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_24_isc(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_24_isc(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  long __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_isc", 0);

  /* "nes/cycore/mos6502.pyx":1095
 *         Increase memory by one, then subtract memory from accumulator (with borrow). Status flags: N,V,Z,C
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = (self.memory.read(addr) + 1) & 0xFF
 *             self.memory.write(addr, v)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1096
 *         """
 *         if self.undocumented_support_level >= 1:
 *             v = (self.memory.read(addr) + 1) & 0xFF             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, v)
 *             self._sbc(v, immediate=True)
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1096, __pyx_L1_error)
    __pyx_v_v = ((((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2) + 1) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1097
 *         if self.undocumented_support_level >= 1:
 *             v = (self.memory.read(addr) + 1) & 0xFF
 *             self.memory.write(addr, v)             # <<<<<<<<<<<<<<
 *             self._sbc(v, immediate=True)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1097, __pyx_L1_error)
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2, __pyx_v_v);

    /* "nes/cycore/mos6502.pyx":1098
 *             v = (self.memory.read(addr) + 1) & 0xFF
 *             self.memory.write(addr, v)
 *             self._sbc(v, immediate=True)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_v, 1));

    /* "nes/cycore/mos6502.pyx":1095
 *         Increase memory by one, then subtract memory from accumulator (with borrow). Status flags: N,V,Z,C
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             v = (self.memory.read(addr) + 1) & 0xFF
 *             self.memory.write(addr, v)
 */
  }

  /* "nes/cycore/mos6502.pyx":1099
 *             self.memory.write(addr, v)
 *             self._sbc(v, immediate=True)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _slo(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1090
 *         return 0
 * 
 *     def _isc(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._isc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1101
 *         return 0
 * 
 *     def _slo(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_27_slo(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_26_slo[] = "\n        Undocumented.\n        Shift left one bit in memory, then OR accumulator with memory.  Status flags: N,Z,C\n        Equivalent to ASL, ORA\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_27_slo(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_slo (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_slo", 1, 2, 2, 1); __PYX_ERR(0, 1101, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_slo") < 0)) __PYX_ERR(0, 1101, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_slo", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1101, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._slo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_26_slo(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_26_slo(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_slo", 0);

  /* "nes/cycore/mos6502.pyx":1107
 *         Equivalent to ASL, ORA
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._asl(addr, ARG_NONE)
 *             self._ora(self.memory.read(addr), immediate=True)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1108
 *         """
 *         if self.undocumented_support_level >= 1:
 *             self._asl(addr, ARG_NONE)             # <<<<<<<<<<<<<<
 *             self._ora(self.memory.read(addr), immediate=True)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1108, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_asl(__pyx_v_self, __pyx_t_2, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE));

    /* "nes/cycore/mos6502.pyx":1109
 *         if self.undocumented_support_level >= 1:
 *             self._asl(addr, ARG_NONE)
 *             self._ora(self.memory.read(addr), immediate=True)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1109, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2), 1));

    /* "nes/cycore/mos6502.pyx":1107
 *         Equivalent to ASL, ORA
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._asl(addr, ARG_NONE)
 *             self._ora(self.memory.read(addr), immediate=True)
 */
  }

  /* "nes/cycore/mos6502.pyx":1110
 *             self._asl(addr, ARG_NONE)
 *             self._ora(self.memory.read(addr), immediate=True)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _rla(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1101
 *         return 0
 * 
 *     def _slo(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._slo", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1112
 *         return 0
 * 
 *     def _rla(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_29_rla(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_28_rla[] = "\n        Undocumented.\n        RLA ROLs the contents of a memory location and then ANDs the result with the accumulator.\n        Equivalent to ROL, AND\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_29_rla(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rla (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rla", 1, 2, 2, 1); __PYX_ERR(0, 1112, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_rla") < 0)) __PYX_ERR(0, 1112, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_rla", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1112, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._rla", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_28_rla(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_28_rla(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rla", 0);

  /* "nes/cycore/mos6502.pyx":1118
 *         Equivalent to ROL, AND
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._rol(addr, ARG_NONE)
 *             self._and(self.memory.read(addr), immediate=True)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1119
 *         """
 *         if self.undocumented_support_level >= 1:
 *             self._rol(addr, ARG_NONE)             # <<<<<<<<<<<<<<
 *             self._and(self.memory.read(addr), immediate=True)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1119, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rol(__pyx_v_self, __pyx_t_2, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE));

    /* "nes/cycore/mos6502.pyx":1120
 *         if self.undocumented_support_level >= 1:
 *             self._rol(addr, ARG_NONE)
 *             self._and(self.memory.read(addr), immediate=True)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1120, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2), 1));

    /* "nes/cycore/mos6502.pyx":1118
 *         Equivalent to ROL, AND
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._rol(addr, ARG_NONE)
 *             self._and(self.memory.read(addr), immediate=True)
 */
  }

  /* "nes/cycore/mos6502.pyx":1121
 *             self._rol(addr, ARG_NONE)
 *             self._and(self.memory.read(addr), immediate=True)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _rra(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1112
 *         return 0
 * 
 *     def _rla(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._rla", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1123
 *         return 0
 * 
 *     def _rra(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_31_rra(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_30_rra[] = "\n        Undocumented.\n        RRA RORs the contents of a memory location and then ADCs the result with the accumulator.\n        Equivalent to ROR, ADC\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_31_rra(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rra (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rra", 1, 2, 2, 1); __PYX_ERR(0, 1123, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_rra") < 0)) __PYX_ERR(0, 1123, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_rra", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1123, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._rra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_30_rra(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_30_rra(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_rra", 0);

  /* "nes/cycore/mos6502.pyx":1129
 *         Equivalent to ROR, ADC
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._ror(addr, ARG_NONE)
 *             self._adc(self.memory.read(addr), immediate=True)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1130
 *         """
 *         if self.undocumented_support_level >= 1:
 *             self._ror(addr, ARG_NONE)             # <<<<<<<<<<<<<<
 *             self._adc(self.memory.read(addr), immediate=True)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1130, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_t_2, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE));

    /* "nes/cycore/mos6502.pyx":1131
 *         if self.undocumented_support_level >= 1:
 *             self._ror(addr, ARG_NONE)
 *             self._adc(self.memory.read(addr), immediate=True)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1131, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2), 1));

    /* "nes/cycore/mos6502.pyx":1129
 *         Equivalent to ROR, ADC
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._ror(addr, ARG_NONE)
 *             self._adc(self.memory.read(addr), immediate=True)
 */
  }

  /* "nes/cycore/mos6502.pyx":1132
 *             self._ror(addr, ARG_NONE)
 *             self._adc(self.memory.read(addr), immediate=True)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _sre(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1123
 *         return 0
 * 
 *     def _rra(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._rra", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1134
 *         return 0
 * 
 *     def _sre(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_33_sre(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_32_sre[] = "\n        Undocumented.\n        Shift right one bit in memory, then EOR accumulator with memory.\n        Equivalent to LSR, EOR\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_33_sre(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sre (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sre", 1, 2, 2, 1); __PYX_ERR(0, 1134, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_sre") < 0)) __PYX_ERR(0, 1134, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_sre", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1134, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._sre", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_32_sre(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_32_sre(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sre", 0);

  /* "nes/cycore/mos6502.pyx":1140
 *         Equivalent to LSR, EOR
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._lsr(addr, ARG_NONE)
 *             self._eor(self.memory.read(addr), immediate=True)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1141
 *         """
 *         if self.undocumented_support_level >= 1:
 *             self._lsr(addr, ARG_NONE)             # <<<<<<<<<<<<<<
 *             self._eor(self.memory.read(addr), immediate=True)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1141, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_t_2, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE));

    /* "nes/cycore/mos6502.pyx":1142
 *         if self.undocumented_support_level >= 1:
 *             self._lsr(addr, ARG_NONE)
 *             self._eor(self.memory.read(addr), immediate=True)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1142, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2), 1));

    /* "nes/cycore/mos6502.pyx":1140
 *         Equivalent to LSR, EOR
 *         """
 *         if self.undocumented_support_level >= 1:             # <<<<<<<<<<<<<<
 *             self._lsr(addr, ARG_NONE)
 *             self._eor(self.memory.read(addr), immediate=True)
 */
  }

  /* "nes/cycore/mos6502.pyx":1143
 *             self._lsr(addr, ARG_NONE)
 *             self._eor(self.memory.read(addr), immediate=True)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     ##################### The arcane undocumented instructions (see [13] and others) ###################################
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1134
 *         return 0
 * 
 *     def _sre(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._sre", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1147
 *     ##################### The arcane undocumented instructions (see [13] and others) ###################################
 * 
 *     def _arr(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.  Mostly used for piracy (jk).
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_35_arr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_34_arr[] = "\n        Undocumented.  Mostly used for piracy (jk).\n        ANDs the contents of the A register with an immediate value and then RORs the result, and check bit 5 and 6:\n            If both bits are 1: set C, clear V.\n            If both bits are 0: clear C and V.\n            If only bit 5 is 1: set V, clear C.\n            If only bit 6 is 1: set C and V.\n        Equivalent to AND oper, ROR A\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_35_arr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_arr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_arr", 1, 2, 2, 1); __PYX_ERR(0, 1147, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_arr") < 0)) __PYX_ERR(0, 1147, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_arr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1147, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._arr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_34_arr(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_34_arr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_arr", 0);

  /* "nes/cycore/mos6502.pyx":1157
 *         Equivalent to AND oper, ROR A
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             self._and(arg, immediate=True)
 *             self._ror(ARG_NONE, ARG_NONE)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1158
 *         """
 *         if self.undocumented_support_level >= 2:
 *             self._and(arg, immediate=True)             # <<<<<<<<<<<<<<
 *             self._ror(ARG_NONE, ARG_NONE)
 *             if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_arg); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1158, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_t_2, 1));

    /* "nes/cycore/mos6502.pyx":1159
 *         if self.undocumented_support_level >= 2:
 *             self._and(arg, immediate=True)
 *             self._ror(ARG_NONE, ARG_NONE)             # <<<<<<<<<<<<<<
 *             if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):
 *                 self.C = True
 */
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE));

    /* "nes/cycore/mos6502.pyx":1160
 *             self._and(arg, immediate=True)
 *             self._ror(ARG_NONE, ARG_NONE)
 *             if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):             # <<<<<<<<<<<<<<
 *                 self.C = True
 *                 self.V = False
 */
    __pyx_t_3 = (((__pyx_v_self->A >> 5) & 1) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_3 = (((__pyx_v_self->A >> 6) & 1) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;
    if (__pyx_t_1) {

      /* "nes/cycore/mos6502.pyx":1161
 *             self._ror(ARG_NONE, ARG_NONE)
 *             if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):
 *                 self.C = True             # <<<<<<<<<<<<<<
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:
 */
      __pyx_v_self->C = 1;

      /* "nes/cycore/mos6502.pyx":1162
 *             if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):
 *                 self.C = True
 *                 self.V = False             # <<<<<<<<<<<<<<
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:
 *                 self.C = False
 */
      __pyx_v_self->V = 0;

      /* "nes/cycore/mos6502.pyx":1160
 *             self._and(arg, immediate=True)
 *             self._ror(ARG_NONE, ARG_NONE)
 *             if ((self.A >> 5) & 1) and ((self.A >> 6) & 1):             # <<<<<<<<<<<<<<
 *                 self.C = True
 *                 self.V = False
 */
      goto __pyx_L4;
    }

    /* "nes/cycore/mos6502.pyx":1163
 *                 self.C = True
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:             # <<<<<<<<<<<<<<
 *                 self.C = False
 *                 self.V = False
 */
    __pyx_t_3 = ((((__pyx_v_self->A >> 5) & 1) == 0) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_3 = ((((__pyx_v_self->A >> 6) & 1) == 0) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L7_bool_binop_done:;
    if (__pyx_t_1) {

      /* "nes/cycore/mos6502.pyx":1164
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:
 *                 self.C = False             # <<<<<<<<<<<<<<
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:
 */
      __pyx_v_self->C = 0;

      /* "nes/cycore/mos6502.pyx":1165
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:
 *                 self.C = False
 *                 self.V = False             # <<<<<<<<<<<<<<
 *             elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:
 *                 self.C = False
 */
      __pyx_v_self->V = 0;

      /* "nes/cycore/mos6502.pyx":1163
 *                 self.C = True
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1) == 0:             # <<<<<<<<<<<<<<
 *                 self.C = False
 *                 self.V = False
 */
      goto __pyx_L4;
    }

    /* "nes/cycore/mos6502.pyx":1166
 *                 self.C = False
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:             # <<<<<<<<<<<<<<
 *                 self.C = False
 *                 self.V = True
 */
    __pyx_t_3 = (((__pyx_v_self->A >> 5) & 1) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_3 = ((((__pyx_v_self->A >> 6) & 1) == 0) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_1) {

      /* "nes/cycore/mos6502.pyx":1167
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:
 *                 self.C = False             # <<<<<<<<<<<<<<
 *                 self.V = True
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):
 */
      __pyx_v_self->C = 0;

      /* "nes/cycore/mos6502.pyx":1168
 *             elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:
 *                 self.C = False
 *                 self.V = True             # <<<<<<<<<<<<<<
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):
 *                 self.C = True
 */
      __pyx_v_self->V = 1;

      /* "nes/cycore/mos6502.pyx":1166
 *                 self.C = False
 *                 self.V = False
 *             elif ((self.A >> 5) & 1) and ((self.A >> 6) & 1) == 0:             # <<<<<<<<<<<<<<
 *                 self.C = False
 *                 self.V = True
 */
      goto __pyx_L4;
    }

    /* "nes/cycore/mos6502.pyx":1169
 *                 self.C = False
 *                 self.V = True
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):             # <<<<<<<<<<<<<<
 *                 self.C = True
 *                 self.V = True
 */
    __pyx_t_3 = ((((__pyx_v_self->A >> 5) & 1) == 0) != 0);
    if (__pyx_t_3) {
    } else {
      __pyx_t_1 = __pyx_t_3;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_3 = (((__pyx_v_self->A >> 6) & 1) != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L11_bool_binop_done:;
    if (__pyx_t_1) {

      /* "nes/cycore/mos6502.pyx":1170
 *                 self.V = True
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):
 *                 self.C = True             # <<<<<<<<<<<<<<
 *                 self.V = True
 *         return 0
 */
      __pyx_v_self->C = 1;

      /* "nes/cycore/mos6502.pyx":1171
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):
 *                 self.C = True
 *                 self.V = True             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      __pyx_v_self->V = 1;

      /* "nes/cycore/mos6502.pyx":1169
 *                 self.C = False
 *                 self.V = True
 *             elif ((self.A >> 5) & 1) == 0 and ((self.A >> 6) & 1):             # <<<<<<<<<<<<<<
 *                 self.C = True
 *                 self.V = True
 */
    }
    __pyx_L4:;

    /* "nes/cycore/mos6502.pyx":1157
 *         Equivalent to AND oper, ROR A
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             self._and(arg, immediate=True)
 *             self._ror(ARG_NONE, ARG_NONE)
 */
  }

  /* "nes/cycore/mos6502.pyx":1172
 *                 self.C = True
 *                 self.V = True
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _asr(self, arg, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1147
 *     ##################### The arcane undocumented instructions (see [13] and others) ###################################
 * 
 *     def _arr(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.  Mostly used for piracy (jk).
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._arr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1174
 *         return 0
 * 
 *     def _asr(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_37_asr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_36_asr[] = "\n        UNTESTED\n        Undocumented.\n        ANDs the contents of the A register with an immediate value and then LSRs the result.\n        Equivalent to AND oper, LSR A\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_37_asr(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_asr (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_asr", 1, 2, 2, 1); __PYX_ERR(0, 1174, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_asr") < 0)) __PYX_ERR(0, 1174, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_asr", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1174, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._asr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_36_asr(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_36_asr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_asr", 0);

  /* "nes/cycore/mos6502.pyx":1181
 *         Equivalent to AND oper, LSR A
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             self._and(arg, immediate=True)
 *             self._lsr(ARG_NONE, ARG_NONE)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1182
 *         """
 *         if self.undocumented_support_level >= 2:
 *             self._and(arg, immediate=True)             # <<<<<<<<<<<<<<
 *             self._lsr(ARG_NONE, ARG_NONE)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_arg); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1182, __pyx_L1_error)
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_t_2, 1));

    /* "nes/cycore/mos6502.pyx":1183
 *         if self.undocumented_support_level >= 2:
 *             self._and(arg, immediate=True)
 *             self._lsr(ARG_NONE, ARG_NONE)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE, __pyx_e_3nes_6cycore_7mos6502_ARG_NONE));

    /* "nes/cycore/mos6502.pyx":1181
 *         Equivalent to AND oper, LSR A
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             self._and(arg, immediate=True)
 *             self._lsr(ARG_NONE, ARG_NONE)
 */
  }

  /* "nes/cycore/mos6502.pyx":1184
 *             self._and(arg, immediate=True)
 *             self._lsr(ARG_NONE, ARG_NONE)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _atx(self, arg, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1174
 *         return 0
 * 
 *     def _asr(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._asr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1186
 *         return 0
 * 
 *     def _atx(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_39_atx(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_38_atx[] = "\n        Undocumented.\n        FAILS BLARGG'S INSTRUCTION TEST 03-immediate (both variants here)\n\n        Following http://nesdev.com/undocumented_opcodes.txt:\n            AND byte with accumulator, then transfer accumulator to X register.\n            Status flags: N,Z\n\n        Following http://www.ffd2.com/fridge/docs/6502-NMOS.extra.opcodes (where this is OAL):\n            ORs the A register with #$EE, ANDs the result with an immediate value, and then stores the result in\n            both A and X.\n            Equivalent to ORA #$EE, AND oper, TAX\n\n        The latter causes blargg's test 03-immediate to fail for opcode 0xAB\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_39_atx(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_atx (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_atx", 1, 2, 2, 1); __PYX_ERR(0, 1186, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_atx") < 0)) __PYX_ERR(0, 1186, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_atx", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1186, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._atx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_38_atx(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_38_atx(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  unsigned char __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_atx", 0);

  /* "nes/cycore/mos6502.pyx":1202
 *         The latter causes blargg's test 03-immediate to fail for opcode 0xAB
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             #self.A &= arg
 *             #self.X = self.A
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1206
 *             #self.X = self.A
 *             #self._set_zn(self.A)
 *             self._and(arg | 0xEE, immediate=True)             # <<<<<<<<<<<<<<
 *             self.X = self.A
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_OrObjC(__pyx_v_arg, __pyx_int_238, 0xEE, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1206, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    (void)(((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_t_3, 1));

    /* "nes/cycore/mos6502.pyx":1207
 *             #self._set_zn(self.A)
 *             self._and(arg | 0xEE, immediate=True)
 *             self.X = self.A             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_4 = __pyx_v_self->A;
    __pyx_v_self->X = __pyx_t_4;

    /* "nes/cycore/mos6502.pyx":1202
 *         The latter causes blargg's test 03-immediate to fail for opcode 0xAB
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             #self.A &= arg
 *             #self.X = self.A
 */
  }

  /* "nes/cycore/mos6502.pyx":1208
 *             self._and(arg | 0xEE, immediate=True)
 *             self.X = self.A
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _aac(self, arg, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1186
 *         return 0
 * 
 *     def _atx(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         Undocumented.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._atx", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1210
 *         return 0
 * 
 *     def _aac(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_41_aac(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_40_aac[] = "\n        UNTESTED\n        Undocumented.\n        ANDs the contents of the A register with an immediate value and then\n        moves bit 7 of A into the Carry flag.  This opcode works basically\n        identically to AND #immed. except that the Carry flag is set to the same\n        state that the Negative flag is set to.  [14]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_41_aac(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_aac (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_aac", 1, 2, 2, 1); __PYX_ERR(0, 1210, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_aac") < 0)) __PYX_ERR(0, 1210, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_aac", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1210, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._aac", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_40_aac(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_40_aac(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned char __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_aac", 0);

  /* "nes/cycore/mos6502.pyx":1219
 *         state that the Negative flag is set to.  [14]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             self.A = self.A & arg
 *             self._set_zn(self.A)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1220
 *         """
 *         if self.undocumented_support_level >= 2:
 *             self.A = self.A & arg             # <<<<<<<<<<<<<<
 *             self._set_zn(self.A)
 *             self.C = (self.A & 0b10000000) > 0
 */
    __pyx_t_2 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->A); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_And(__pyx_t_2, __pyx_v_arg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_unsigned_char(__pyx_t_3); if (unlikely((__pyx_t_4 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1220, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->A = __pyx_t_4;

    /* "nes/cycore/mos6502.pyx":1221
 *         if self.undocumented_support_level >= 2:
 *             self.A = self.A & arg
 *             self._set_zn(self.A)             # <<<<<<<<<<<<<<
 *             self.C = (self.A & 0b10000000) > 0
 *         return 0
 */
    ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->A);

    /* "nes/cycore/mos6502.pyx":1222
 *             self.A = self.A & arg
 *             self._set_zn(self.A)
 *             self.C = (self.A & 0b10000000) > 0             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_v_self->C = ((__pyx_v_self->A & 128) > 0);

    /* "nes/cycore/mos6502.pyx":1219
 *         state that the Negative flag is set to.  [14]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             self.A = self.A & arg
 *             self._set_zn(self.A)
 */
  }

  /* "nes/cycore/mos6502.pyx":1223
 *             self._set_zn(self.A)
 *             self.C = (self.A & 0b10000000) > 0
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _axa(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1210
 *         return 0
 * 
 *     def _aac(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._aac", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1225
 *         return 0
 * 
 *     def _axa(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_43_axa(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_42_axa[] = "\n        UNTESTED\n        Undocumented.\n        Sores the result of A AND X AND the high byte of the target address of the operand +1 in memory [14, 15]\n        WARNING:  There seem to be multiple conflicting descriptions of this opcode. (see e.g. [13])\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_43_axa(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_axa (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_axa", 1, 2, 2, 1); __PYX_ERR(0, 1225, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_axa") < 0)) __PYX_ERR(0, 1225, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_axa", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1225, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._axa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_42_axa(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_42_axa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_v_hp1 = NULL;
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  unsigned char __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_axa", 0);

  /* "nes/cycore/mos6502.pyx":1232
 *         WARNING:  There seem to be multiple conflicting descriptions of this opcode. (see e.g. [13])
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             v = (self.X & self.A) & hp1
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1233
 *         """
 *         if self.undocumented_support_level >= 2:
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF             # <<<<<<<<<<<<<<
 *             v = (self.X & self.A) & hp1
 *             self.memory.write(addr, v)
 */
    __pyx_t_2 = __Pyx_PyInt_AndObjC(__pyx_v_addr, __pyx_int_65280, 0xFF00, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_RshiftObjC(__pyx_t_2, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_t_2, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1233, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_hp1 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":1234
 *         if self.undocumented_support_level >= 2:
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             v = (self.X & self.A) & hp1             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, v)
 *         return 0
 */
    __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_self->X & __pyx_v_self->A)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_And(__pyx_t_3, __pyx_v_hp1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1234, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_v = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "nes/cycore/mos6502.pyx":1235
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             v = (self.X & self.A) & hp1
 *             self.memory.write(addr, v)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1235, __pyx_L1_error)
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_char(__pyx_v_v); if (unlikely((__pyx_t_5 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1235, __pyx_L1_error)
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_4, __pyx_t_5);

    /* "nes/cycore/mos6502.pyx":1232
 *         WARNING:  There seem to be multiple conflicting descriptions of this opcode. (see e.g. [13])
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             v = (self.X & self.A) & hp1
 */
  }

  /* "nes/cycore/mos6502.pyx":1236
 *             v = (self.X & self.A) & hp1
 *             self.memory.write(addr, v)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _axs(self, arg, immediate):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1225
 *         return 0
 * 
 *     def _axa(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._axa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hp1);
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1238
 *         return 0
 * 
 *     def _axs(self, arg, immediate):             # <<<<<<<<<<<<<<
 *         """
 *         FAILS BLARGG'S INSTRUCTION TEST 03-immediate
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_45_axs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_44_axs[] = "\n        FAILS BLARGG'S INSTRUCTION TEST 03-immediate\n        Undocumented.\n        AND the contents of the A and X registers (without changing the contents of either register) and\n        stores the result in memory. [16]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_45_axs(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  PyObject *__pyx_v_immediate = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_axs (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s_immediate,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_immediate)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_axs", 1, 2, 2, 1); __PYX_ERR(0, 1238, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_axs") < 0)) __PYX_ERR(0, 1238, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v_immediate = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_axs", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1238, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._axs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_44_axs(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v_immediate);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_44_axs(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, PyObject *__pyx_v_immediate) {
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  unsigned char __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_axs", 0);

  /* "nes/cycore/mos6502.pyx":1245
 *         stores the result in memory. [16]
 *         """
 *         self.A = self.memory.read(arg) if not immediate else arg             # <<<<<<<<<<<<<<
 * 
 *         if self.undocumented_support_level >= 2:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_immediate); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1245, __pyx_L1_error)
  if (((!__pyx_t_2) != 0)) {
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_arg); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1245, __pyx_L1_error)
    __pyx_t_1 = ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_3);
  } else {
    __pyx_t_4 = __Pyx_PyInt_As_unsigned_char(__pyx_v_arg); if (unlikely((__pyx_t_4 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1245, __pyx_L1_error)
    __pyx_t_1 = __pyx_t_4;
  }
  __pyx_v_self->A = __pyx_t_1;

  /* "nes/cycore/mos6502.pyx":1247
 *         self.A = self.memory.read(arg) if not immediate else arg
 * 
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 * 
 *             if immediate:
 */
  __pyx_t_2 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_2) {

    /* "nes/cycore/mos6502.pyx":1249
 *         if self.undocumented_support_level >= 2:
 * 
 *             if immediate:             # <<<<<<<<<<<<<<
 *                 # AND X register with accumulator and store result in X regis-ter, then subtract byte from X register
 *                 # (without borrow).  Status flags: N,Z,C   (http://nesdev.com/undocumented_opcodes.txt)
 */
    __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_immediate); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(0, 1249, __pyx_L1_error)
    if (__pyx_t_2) {

      /* "nes/cycore/mos6502.pyx":1252
 *                 # AND X register with accumulator and store result in X regis-ter, then subtract byte from X register
 *                 # (without borrow).  Status flags: N,Z,C   (http://nesdev.com/undocumented_opcodes.txt)
 *                 self.X = self.A & self.X             # <<<<<<<<<<<<<<
 *                 t = self.X - arg
 *                 self.C = t >= 0
 */
      __pyx_v_self->X = (__pyx_v_self->A & __pyx_v_self->X);

      /* "nes/cycore/mos6502.pyx":1253
 *                 # (without borrow).  Status flags: N,Z,C   (http://nesdev.com/undocumented_opcodes.txt)
 *                 self.X = self.A & self.X
 *                 t = self.X - arg             # <<<<<<<<<<<<<<
 *                 self.C = t >= 0
 *                 self.X = t & 0xFF
 */
      __pyx_t_5 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->X); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyNumber_Subtract(__pyx_t_5, __pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1253, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_t = __pyx_t_6;
      __pyx_t_6 = 0;

      /* "nes/cycore/mos6502.pyx":1254
 *                 self.X = self.A & self.X
 *                 t = self.X - arg
 *                 self.C = t >= 0             # <<<<<<<<<<<<<<
 *                 self.X = t & 0xFF
 *                 self._set_zn(self.X)
 */
      __pyx_t_6 = PyObject_RichCompare(__pyx_v_t, __pyx_int_0, Py_GE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1254, __pyx_L1_error)
      __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1254, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_self->C = __pyx_t_3;

      /* "nes/cycore/mos6502.pyx":1255
 *                 t = self.X - arg
 *                 self.C = t >= 0
 *                 self.X = t & 0xFF             # <<<<<<<<<<<<<<
 *                 self._set_zn(self.X)
 *             else:
 */
      __pyx_t_6 = __Pyx_PyInt_AndObjC(__pyx_v_t, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1255, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_1 = __Pyx_PyInt_As_unsigned_char(__pyx_t_6); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1255, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __pyx_v_self->X = __pyx_t_1;

      /* "nes/cycore/mos6502.pyx":1256
 *                 self.C = t >= 0
 *                 self.X = t & 0xFF
 *                 self._set_zn(self.X)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.memory.write(arg, self.A & self.X)
 */
      ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, __pyx_v_self->X);

      /* "nes/cycore/mos6502.pyx":1249
 *         if self.undocumented_support_level >= 2:
 * 
 *             if immediate:             # <<<<<<<<<<<<<<
 *                 # AND X register with accumulator and store result in X regis-ter, then subtract byte from X register
 *                 # (without borrow).  Status flags: N,Z,C   (http://nesdev.com/undocumented_opcodes.txt)
 */
      goto __pyx_L4;
    }

    /* "nes/cycore/mos6502.pyx":1258
 *                 self._set_zn(self.X)
 *             else:
 *                 self.memory.write(arg, self.A & self.X)             # <<<<<<<<<<<<<<
 *                 self._set_zn(self.A & self.X)
 *         return 0
 */
    /*else*/ {
      __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_v_arg); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1258, __pyx_L1_error)
      ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_3, (__pyx_v_self->A & __pyx_v_self->X));

      /* "nes/cycore/mos6502.pyx":1259
 *             else:
 *                 self.memory.write(arg, self.A & self.X)
 *                 self._set_zn(self.A & self.X)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
      ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_set_zn(__pyx_v_self, (__pyx_v_self->A & __pyx_v_self->X));
    }
    __pyx_L4:;

    /* "nes/cycore/mos6502.pyx":1247
 *         self.A = self.memory.read(arg) if not immediate else arg
 * 
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 * 
 *             if immediate:
 */
  }

  /* "nes/cycore/mos6502.pyx":1260
 *                 self.memory.write(arg, self.A & self.X)
 *                 self._set_zn(self.A & self.X)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _lar(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1238
 *         return 0
 * 
 *     def _axs(self, arg, immediate):             # <<<<<<<<<<<<<<
 *         """
 *         FAILS BLARGG'S INSTRUCTION TEST 03-immediate
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._axs", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1262
 *         return 0
 * 
 *     def _lar(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_47_lar(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_46_lar[] = "\n        UNTESTED\n        Undocumented.\n        AND memory with stack pointer, transfer result to accumulator, X register and stack pointer. [16]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_47_lar(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_lar (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_lar", 1, 2, 2, 1); __PYX_ERR(0, 1262, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_lar") < 0)) __PYX_ERR(0, 1262, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_lar", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1262, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._lar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_46_lar(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_46_lar(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  int __pyx_v_v;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_lar", 0);

  /* "nes/cycore/mos6502.pyx":1268
 *         AND memory with stack pointer, transfer result to accumulator, X register and stack pointer. [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             v = self.memory.read(addr) & self.SP
 *             self.A = v
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1269
 *         """
 *         if self.undocumented_support_level >= 2:
 *             v = self.memory.read(addr) & self.SP             # <<<<<<<<<<<<<<
 *             self.A = v
 *             self.X = v
 */
    __pyx_t_2 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1269, __pyx_L1_error)
    __pyx_v_v = (((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.read(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_2) & __pyx_v_self->SP);

    /* "nes/cycore/mos6502.pyx":1270
 *         if self.undocumented_support_level >= 2:
 *             v = self.memory.read(addr) & self.SP
 *             self.A = v             # <<<<<<<<<<<<<<
 *             self.X = v
 *             self.SP = v
 */
    __pyx_v_self->A = __pyx_v_v;

    /* "nes/cycore/mos6502.pyx":1271
 *             v = self.memory.read(addr) & self.SP
 *             self.A = v
 *             self.X = v             # <<<<<<<<<<<<<<
 *             self.SP = v
 *         return 0
 */
    __pyx_v_self->X = __pyx_v_v;

    /* "nes/cycore/mos6502.pyx":1272
 *             self.A = v
 *             self.X = v
 *             self.SP = v             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_v_self->SP = __pyx_v_v;

    /* "nes/cycore/mos6502.pyx":1268
 *         AND memory with stack pointer, transfer result to accumulator, X register and stack pointer. [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             v = self.memory.read(addr) & self.SP
 *             self.A = v
 */
  }

  /* "nes/cycore/mos6502.pyx":1273
 *             self.X = v
 *             self.SP = v
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _sxa(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1262
 *         return 0
 * 
 *     def _lar(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._lar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1275
 *         return 0
 * 
 *     def _sxa(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         FAILS BLARGG'S TEST 07-abs_xy
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_49_sxa(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_48_sxa[] = "\n        FAILS BLARGG'S TEST 07-abs_xy\n        Undocumented.\n        AND X register with the high byte of the target address of the argument + 1. Store the result in memory. [16]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_49_sxa(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sxa (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sxa", 1, 2, 2, 1); __PYX_ERR(0, 1275, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_sxa") < 0)) __PYX_ERR(0, 1275, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_sxa", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1275, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._sxa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_48_sxa(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_48_sxa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_v_hp1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  unsigned char __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sxa", 0);

  /* "nes/cycore/mos6502.pyx":1281
 *         AND X register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.memory.write(addr, self.X & hp1)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1282
 *         """
 *         if self.undocumented_support_level >= 2:
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, self.X & hp1)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_AndObjC(__pyx_v_addr, __pyx_int_65280, 0xFF00, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_RshiftObjC(__pyx_t_2, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_t_2, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1282, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_hp1 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":1283
 *         if self.undocumented_support_level >= 2:
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.memory.write(addr, self.X & hp1)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1283, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->X); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_And(__pyx_t_3, __pyx_v_hp1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_char(__pyx_t_2); if (unlikely((__pyx_t_5 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1283, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_4, __pyx_t_5);

    /* "nes/cycore/mos6502.pyx":1281
 *         AND X register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.memory.write(addr, self.X & hp1)
 */
  }

  /* "nes/cycore/mos6502.pyx":1284
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.memory.write(addr, self.X & hp1)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _sya(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1275
 *         return 0
 * 
 *     def _sxa(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         FAILS BLARGG'S TEST 07-abs_xy
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._sxa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hp1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1286
 *         return 0
 * 
 *     def _sya(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         FAILS BLARGG'S TEST 07-abs_xy
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_51_sya(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_50_sya[] = "\n        FAILS BLARGG'S TEST 07-abs_xy\n        Undocumented.\n        AND Y register with the high byte of the target address of the argument + 1. Store the result in memory. [16]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_51_sya(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sya (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sya", 1, 2, 2, 1); __PYX_ERR(0, 1286, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_sya") < 0)) __PYX_ERR(0, 1286, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_sya", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1286, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._sya", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_50_sya(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_50_sya(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_v_v = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  unsigned char __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_sya", 0);

  /* "nes/cycore/mos6502.pyx":1292
 *         AND Y register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)
 *             self.memory.write(addr, v & 0xFF)
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1293
 *         """
 *         if self.undocumented_support_level >= 2:
 *             v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, v & 0xFF)
 *         return 0
 */
    __pyx_t_2 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->Y); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_v_addr, __pyx_int_65280, 0xFF00, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyInt_RshiftObjC(__pyx_t_3, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_AddObjC(__pyx_t_4, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyInt_AndObjC(__pyx_t_3, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_And(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1293, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_v_v = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":1294
 *         if self.undocumented_support_level >= 2:
 *             v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)
 *             self.memory.write(addr, v & 0xFF)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1294, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_v_v, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1294, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_As_unsigned_char(__pyx_t_3); if (unlikely((__pyx_t_6 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1294, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_5, __pyx_t_6);

    /* "nes/cycore/mos6502.pyx":1292
 *         AND Y register with the high byte of the target address of the argument + 1. Store the result in memory. [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)
 *             self.memory.write(addr, v & 0xFF)
 */
  }

  /* "nes/cycore/mos6502.pyx":1295
 *             v = self.Y & ((((addr & 0xFF00) >> 8) + 1) & 0xFF)
 *             self.memory.write(addr, v & 0xFF)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _xaa(self, arg, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1286
 *         return 0
 * 
 *     def _sya(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         FAILS BLARGG'S TEST 07-abs_xy
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._sya", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_v);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1297
 *         return 0
 * 
 *     def _xaa(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_53_xaa(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_52_xaa[] = "\n        UNTESTED\n        Undocumented.\n        This opcode ORs the A register with CONST, ANDs the result with X. ANDs the result with an immediate value, and\n        then stores the result in A.\n        Known to be extremely unstable, including temperature dependent! [16]\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_53_xaa(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_arg = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_xaa (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_arg,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_arg)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_xaa", 1, 2, 2, 1); __PYX_ERR(0, 1297, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_xaa") < 0)) __PYX_ERR(0, 1297, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_arg = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_xaa", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1297, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._xaa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_52_xaa(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_arg, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_52_xaa(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_arg, CYTHON_UNUSED PyObject *__pyx_v__) {
  long __pyx_v_CONST;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  unsigned char __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_xaa", 0);

  /* "nes/cycore/mos6502.pyx":1305
 *         Known to be extremely unstable, including temperature dependent! [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             CONST = 0x00
 *             self.A = (self.A | CONST) & self.X & arg
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1306
 *         """
 *         if self.undocumented_support_level >= 2:
 *             CONST = 0x00             # <<<<<<<<<<<<<<
 *             self.A = (self.A | CONST) & self.X & arg
 *         return 0
 */
    __pyx_v_CONST = 0x00;

    /* "nes/cycore/mos6502.pyx":1307
 *         if self.undocumented_support_level >= 2:
 *             CONST = 0x00
 *             self.A = (self.A | CONST) & self.X & arg             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(((__pyx_v_self->A | __pyx_v_CONST) & __pyx_v_self->X)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_And(__pyx_t_2, __pyx_v_arg); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1307, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyInt_As_unsigned_char(__pyx_t_3); if (unlikely((__pyx_t_4 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1307, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_self->A = __pyx_t_4;

    /* "nes/cycore/mos6502.pyx":1305
 *         Known to be extremely unstable, including temperature dependent! [16]
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             CONST = 0x00
 *             self.A = (self.A | CONST) & self.X & arg
 */
  }

  /* "nes/cycore/mos6502.pyx":1308
 *             CONST = 0x00
 *             self.A = (self.A | CONST) & self.X & arg
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def _xas(self, addr, _):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1297
 *         return 0
 * 
 *     def _xaa(self, arg, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._xaa", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1310
 *         return 0
 * 
 *     def _xas(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_55_xas(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_54_xas[] = "\n        UNTESTED\n        Undocumented.\n        ANDs the contents of the A and X registers (without changing the contents\n        of either register) and transfers the result to the stack pointer. It then ANDs that result with the\n        contents of the high byte of the target address of the operand +1 and stores that final result in\n        memory\n        ";
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_55_xas(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_addr = 0;
  CYTHON_UNUSED PyObject *__pyx_v__ = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_xas (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_addr,&__pyx_n_s__9,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_addr)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s__9)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_xas", 1, 2, 2, 1); __PYX_ERR(0, 1310, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_xas") < 0)) __PYX_ERR(0, 1310, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_addr = values[0];
    __pyx_v__ = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_xas", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1310, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._xas", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_54_xas(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), __pyx_v_addr, __pyx_v__);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_54_xas(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v_addr, CYTHON_UNUSED PyObject *__pyx_v__) {
  PyObject *__pyx_v_hp1 = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  unsigned char __pyx_t_5;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_xas", 0);

  /* "nes/cycore/mos6502.pyx":1319
 *         memory
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.SP = self.A & self.X
 */
  __pyx_t_1 = ((__pyx_v_self->undocumented_support_level >= 2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1320
 *         """
 *         if self.undocumented_support_level >= 2:
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF             # <<<<<<<<<<<<<<
 *             self.SP = self.A & self.X
 *             self.memory.write(addr, self.SP & hp1)
 */
    __pyx_t_2 = __Pyx_PyInt_AndObjC(__pyx_v_addr, __pyx_int_65280, 0xFF00, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyInt_RshiftObjC(__pyx_t_2, __pyx_int_8, 8, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyInt_AndObjC(__pyx_t_2, __pyx_int_255, 0xFF, 0, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1320, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_hp1 = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "nes/cycore/mos6502.pyx":1321
 *         if self.undocumented_support_level >= 2:
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.SP = self.A & self.X             # <<<<<<<<<<<<<<
 *             self.memory.write(addr, self.SP & hp1)
 *         return 0
 */
    __pyx_v_self->SP = (__pyx_v_self->A & __pyx_v_self->X);

    /* "nes/cycore/mos6502.pyx":1322
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.SP = self.A & self.X
 *             self.memory.write(addr, self.SP & hp1)             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_v_addr); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L1_error)
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->SP); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyNumber_And(__pyx_t_3, __pyx_v_hp1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_unsigned_char(__pyx_t_2); if (unlikely((__pyx_t_5 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(0, 1322, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    ((struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM *)__pyx_v_self->memory->__pyx_base.__pyx_vtab)->__pyx_base.write(((struct __pyx_obj_3nes_6cycore_6memory_MemoryBase *)__pyx_v_self->memory), __pyx_t_4, __pyx_t_5);

    /* "nes/cycore/mos6502.pyx":1319
 *         memory
 *         """
 *         if self.undocumented_support_level >= 2:             # <<<<<<<<<<<<<<
 *             hp1 = (((addr & 0xFF00) >> 8) + 1) & 0xFF
 *             self.SP = self.A & self.X
 */
  }

  /* "nes/cycore/mos6502.pyx":1323
 *             self.SP = self.A & self.X
 *             self.memory.write(addr, self.SP & hp1)
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     cdef int run_instr(self, unsigned char opcode, unsigned char data[2]):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_int_0);
  __pyx_r = __pyx_int_0;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1310
 *         return 0
 * 
 *     def _xas(self, addr, _):             # <<<<<<<<<<<<<<
 *         """
 *         UNTESTED
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502._xas", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_hp1);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "nes/cycore/mos6502.pyx":1325
 *         return 0
 * 
 *     cdef int run_instr(self, unsigned char opcode, unsigned char data[2]):             # <<<<<<<<<<<<<<
 *         # *********** AUTOGENERATED BY meta.py - DO NOT EDIT DIRECTLY ***********
 *         cdef int cycles=0, arg, immediate=False, m
 */

static int __pyx_f_3nes_6cycore_7mos6502_7MOS6502_run_instr(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, unsigned char __pyx_v_opcode, unsigned char *__pyx_v_data) {
  int __pyx_v_cycles;
  int __pyx_v_arg;
  int __pyx_v_immediate;
  int __pyx_v_m;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  long __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("run_instr", 0);

  /* "nes/cycore/mos6502.pyx":1327
 *     cdef int run_instr(self, unsigned char opcode, unsigned char data[2]):
 *         # *********** AUTOGENERATED BY meta.py - DO NOT EDIT DIRECTLY ***********
 *         cdef int cycles=0, arg, immediate=False, m             # <<<<<<<<<<<<<<
 * 
 *         #print("{:02X} {:02X} {:02X}".format(opcode, data[0], data[1]))
 */
  __pyx_v_cycles = 0;
  __pyx_v_immediate = 0;

  /* "nes/cycore/mos6502.pyx":1331
 *         #print("{:02X} {:02X} {:02X}".format(opcode, data[0], data[1]))
 * 
 *         if opcode==0x69:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x69) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1332
 * 
 *         if opcode==0x69:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1333
 *         if opcode==0x69:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1334
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x65:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1335
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x65:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1331
 *         #print("{:02X} {:02X} {:02X}".format(opcode, data[0], data[1]))
 * 
 *         if opcode==0x69:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1336
 *             immediate = True
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x65:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x65) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1337
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x65:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1338
 *         elif opcode==0x65:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x75:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1339
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x75:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1336
 *             immediate = True
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x65:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1340
 *             arg = data[0]
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x75:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x75) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1341
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x75:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1342
 *         elif opcode==0x75:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x6D:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1343
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x6D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1340
 *             arg = data[0]
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x75:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1344
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x6D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x6D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1345
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x6D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1346
 *         elif opcode==0x6D:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x7D:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1347
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x7D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1344
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x6D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1348
 *             arg = self._from_le(data)
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x7D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x7D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1349
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x7D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1350
 *         elif opcode==0x7D:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1351
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x79:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1352
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x79:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1348
 *             arg = self._from_le(data)
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x7D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1353
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x79:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x79) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1354
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x79:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1355
 *         elif opcode==0x79:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1356
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x61:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1357
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x61:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1353
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x79:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1358
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x61:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x61) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1359
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x61:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1360
 *         elif opcode==0x61:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x71:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1361
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x71:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1358
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x61:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1362
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x71:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x71) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1363
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x71:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1364
 *         elif opcode==0x71:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1365
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1366
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x29:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1367
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x29:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_adc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1362
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x71:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1368
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x29:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x29) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1369
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x29:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1370
 *         elif opcode==0x29:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1371
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x25:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1372
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x25:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1368
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._adc(arg, immediate)
 *         elif opcode==0x29:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1373
 *             immediate = True
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x25:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x25) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1374
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x25:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1375
 *         elif opcode==0x25:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x35:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1376
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x35:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1373
 *             immediate = True
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x25:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1377
 *             arg = data[0]
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x35:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x35) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1378
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x35:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1379
 *         elif opcode==0x35:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x2D:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1380
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x2D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1377
 *             arg = data[0]
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x35:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1381
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x2D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x2D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1382
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x2D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1383
 *         elif opcode==0x2D:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x3D:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1384
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x3D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1381
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x2D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1385
 *             arg = self._from_le(data)
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x3D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x3D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1386
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x3D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1387
 *         elif opcode==0x3D:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1388
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x39:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1389
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x39:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1385
 *             arg = self._from_le(data)
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x3D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1390
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x39:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x39) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1391
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x39:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1392
 *         elif opcode==0x39:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1393
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x21:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1394
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x21:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1390
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x39:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1395
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x21:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x21) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1396
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x21:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1397
 *         elif opcode==0x21:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x31:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1398
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x31:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1395
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x21:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1399
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x31:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x31) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1400
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x31:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1401
 *         elif opcode==0x31:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1402
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1403
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x0A:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1404
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x0A:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_and(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1399
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x31:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1405
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x0A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x0A) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1406
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x0A:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._asl(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1407
 *         elif opcode==0x0A:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x06:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1408
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._asl(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x06:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_asl(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1405
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._and(arg, immediate)
 *         elif opcode==0x0A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1409
 *             arg = ARG_NONE
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x06:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x06) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1410
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x06:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._asl(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1411
 *         elif opcode==0x06:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x16:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1412
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._asl(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x16:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_asl(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1409
 *             arg = ARG_NONE
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x06:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1413
 *             arg = data[0]
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x16:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x16) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1414
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x16:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._asl(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1415
 *         elif opcode==0x16:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x0E:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1416
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._asl(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x0E:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_asl(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1413
 *             arg = data[0]
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x16:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1417
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x0E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x0E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1418
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x0E:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._asl(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1419
 *         elif opcode==0x0E:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x1E:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1420
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._asl(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x1E:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_asl(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1417
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x0E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1421
 *             arg = self._from_le(data)
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x1E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x1E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1422
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x1E:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._asl(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1423
 *         elif opcode==0x1E:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x90:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1424
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._asl(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x90:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_asl(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1421
 *             arg = self._from_le(data)
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x1E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1425
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x90:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x90) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1426
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x90:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1427
 *         elif opcode==0x90:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bcc(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1429
 *             arg = data[0]
 * 
 *             cycles += self._bcc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB0:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bcc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1425
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._asl(arg, immediate)
 *         elif opcode==0x90:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1430
 * 
 *             cycles += self._bcc(arg, immediate)
 *         elif opcode==0xB0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB0) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1431
 *             cycles += self._bcc(arg, immediate)
 *         elif opcode==0xB0:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1432
 *         elif opcode==0xB0:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bcs(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1434
 *             arg = data[0]
 * 
 *             cycles += self._bcs(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF0:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bcs(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1430
 * 
 *             cycles += self._bcc(arg, immediate)
 *         elif opcode==0xB0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1435
 * 
 *             cycles += self._bcs(arg, immediate)
 *         elif opcode==0xF0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF0) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1436
 *             cycles += self._bcs(arg, immediate)
 *         elif opcode==0xF0:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1437
 *         elif opcode==0xF0:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._beq(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1439
 *             arg = data[0]
 * 
 *             cycles += self._beq(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x24:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_beq(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1435
 * 
 *             cycles += self._bcs(arg, immediate)
 *         elif opcode==0xF0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1440
 * 
 *             cycles += self._beq(arg, immediate)
 *         elif opcode==0x24:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x24) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1441
 *             cycles += self._beq(arg, immediate)
 *         elif opcode==0x24:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._bit(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1442
 *         elif opcode==0x24:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x2C:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1443
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._bit(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x2C:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bit(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1440
 * 
 *             cycles += self._beq(arg, immediate)
 *         elif opcode==0x24:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1444
 *             arg = data[0]
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x2C:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x2C) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1445
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x2C:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._bit(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1446
 *         elif opcode==0x2C:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x30:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1447
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._bit(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x30:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bit(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1444
 *             arg = data[0]
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x2C:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1448
 *             arg = self._from_le(data)
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x30:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x30) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1449
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x30:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1450
 *         elif opcode==0x30:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bmi(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1452
 *             arg = data[0]
 * 
 *             cycles += self._bmi(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD0:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bmi(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1448
 *             arg = self._from_le(data)
 *             cycles += self._bit(arg, immediate)
 *         elif opcode==0x30:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1453
 * 
 *             cycles += self._bmi(arg, immediate)
 *         elif opcode==0xD0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD0) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1454
 *             cycles += self._bmi(arg, immediate)
 *         elif opcode==0xD0:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1455
 *         elif opcode==0xD0:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bne(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1457
 *             arg = data[0]
 * 
 *             cycles += self._bne(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x10:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bne(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1453
 * 
 *             cycles += self._bmi(arg, immediate)
 *         elif opcode==0xD0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1458
 * 
 *             cycles += self._bne(arg, immediate)
 *         elif opcode==0x10:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x10) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1459
 *             cycles += self._bne(arg, immediate)
 *         elif opcode==0x10:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1460
 *         elif opcode==0x10:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bpl(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1462
 *             arg = data[0]
 * 
 *             cycles += self._bpl(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x00:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bpl(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1458
 * 
 *             cycles += self._bne(arg, immediate)
 *         elif opcode==0x10:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1463
 * 
 *             cycles += self._bpl(arg, immediate)
 *         elif opcode==0x00:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x00) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1464
 *             cycles += self._bpl(arg, immediate)
 *         elif opcode==0x00:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._brk(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1465
 *         elif opcode==0x00:
 *             cycles = 7
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._brk(arg, immediate)
 *         elif opcode==0x50:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1466
 *             cycles = 7
 *             arg = ARG_NONE
 *             cycles += self._brk(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x50:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_brk(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1463
 * 
 *             cycles += self._bpl(arg, immediate)
 *         elif opcode==0x00:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1467
 *             arg = ARG_NONE
 *             cycles += self._brk(arg, immediate)
 *         elif opcode==0x50:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x50) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1468
 *             cycles += self._brk(arg, immediate)
 *         elif opcode==0x50:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1469
 *         elif opcode==0x50:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bvc(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1471
 *             arg = data[0]
 * 
 *             cycles += self._bvc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x70:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bvc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1467
 *             arg = ARG_NONE
 *             cycles += self._brk(arg, immediate)
 *         elif opcode==0x50:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1472
 * 
 *             cycles += self._bvc(arg, immediate)
 *         elif opcode==0x70:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x70) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1473
 *             cycles += self._bvc(arg, immediate)
 *         elif opcode==0x70:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 * 
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1474
 *         elif opcode==0x70:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 * 
 *             cycles += self._bvs(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1476
 *             arg = data[0]
 * 
 *             cycles += self._bvs(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x18:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_bvs(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1472
 * 
 *             cycles += self._bvc(arg, immediate)
 *         elif opcode==0x70:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1477
 * 
 *             cycles += self._bvs(arg, immediate)
 *         elif opcode==0x18:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x18) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1478
 *             cycles += self._bvs(arg, immediate)
 *         elif opcode==0x18:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._clc(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1479
 *         elif opcode==0x18:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._clc(arg, immediate)
 *         elif opcode==0xD8:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1480
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._clc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD8:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_clc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1477
 * 
 *             cycles += self._bvs(arg, immediate)
 *         elif opcode==0x18:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1481
 *             arg = ARG_NONE
 *             cycles += self._clc(arg, immediate)
 *         elif opcode==0xD8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD8) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1482
 *             cycles += self._clc(arg, immediate)
 *         elif opcode==0xD8:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._cld(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1483
 *         elif opcode==0xD8:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._cld(arg, immediate)
 *         elif opcode==0x58:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1484
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._cld(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x58:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cld(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1481
 *             arg = ARG_NONE
 *             cycles += self._clc(arg, immediate)
 *         elif opcode==0xD8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1485
 *             arg = ARG_NONE
 *             cycles += self._cld(arg, immediate)
 *         elif opcode==0x58:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x58) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1486
 *             cycles += self._cld(arg, immediate)
 *         elif opcode==0x58:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._cli(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1487
 *         elif opcode==0x58:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._cli(arg, immediate)
 *         elif opcode==0xB8:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1488
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._cli(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB8:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cli(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1485
 *             arg = ARG_NONE
 *             cycles += self._cld(arg, immediate)
 *         elif opcode==0x58:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1489
 *             arg = ARG_NONE
 *             cycles += self._cli(arg, immediate)
 *         elif opcode==0xB8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB8) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1490
 *             cycles += self._cli(arg, immediate)
 *         elif opcode==0xB8:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._clv(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1491
 *         elif opcode==0xB8:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._clv(arg, immediate)
 *         elif opcode==0xC9:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1492
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._clv(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC9:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_clv(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1489
 *             arg = ARG_NONE
 *             cycles += self._cli(arg, immediate)
 *         elif opcode==0xB8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1493
 *             arg = ARG_NONE
 *             cycles += self._clv(arg, immediate)
 *         elif opcode==0xC9:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC9) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1494
 *             cycles += self._clv(arg, immediate)
 *         elif opcode==0xC9:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1495
 *         elif opcode==0xC9:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1496
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC5:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1497
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC5:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1493
 *             arg = ARG_NONE
 *             cycles += self._clv(arg, immediate)
 *         elif opcode==0xC9:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1498
 *             immediate = True
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC5:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC5) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1499
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC5:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1500
 *         elif opcode==0xC5:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD5:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1501
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD5:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1498
 *             immediate = True
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC5:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1502
 *             arg = data[0]
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD5:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD5) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1503
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD5:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1504
 *         elif opcode==0xD5:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xCD:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1505
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xCD:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1502
 *             arg = data[0]
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD5:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1506
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xCD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xCD) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1507
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xCD:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1508
 *         elif opcode==0xCD:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xDD:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1509
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xDD:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1506
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xCD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1510
 *             arg = self._from_le(data)
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xDD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xDD) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1511
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xDD:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1512
 *         elif opcode==0xDD:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1513
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD9:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1514
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD9:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1510
 *             arg = self._from_le(data)
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xDD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1515
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD9:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD9) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1516
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD9:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1517
 *         elif opcode==0xD9:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1518
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC1:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1519
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC1:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1515
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD9:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1520
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC1:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1521
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC1:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1522
 *         elif opcode==0xC1:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD1:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1523
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD1:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1520
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xC1:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1524
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD1:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1525
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD1:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1526
 *         elif opcode==0xD1:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1527
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1528
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xE0:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1529
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE0:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1524
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xD1:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1530
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xE0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE0) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1531
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xE0:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1532
 *         elif opcode==0xE0:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._cpx(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1533
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xE4:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1534
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._cpx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE4:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cpx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1530
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._cmp(arg, immediate)
 *         elif opcode==0xE0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1535
 *             immediate = True
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xE4:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE4) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1536
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xE4:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._cpx(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1537
 *         elif opcode==0xE4:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xEC:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1538
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._cpx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xEC:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cpx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1535
 *             immediate = True
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xE4:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1539
 *             arg = data[0]
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xEC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xEC) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1540
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xEC:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._cpx(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1541
 *         elif opcode==0xEC:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xC0:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1542
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._cpx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC0:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cpx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1539
 *             arg = data[0]
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xEC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1543
 *             arg = self._from_le(data)
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xC0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC0) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1544
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xC0:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1545
 *         elif opcode==0xC0:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._cpy(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1546
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC4:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1547
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._cpy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC4:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cpy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1543
 *             arg = self._from_le(data)
 *             cycles += self._cpx(arg, immediate)
 *         elif opcode==0xC0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1548
 *             immediate = True
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC4:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC4) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1549
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC4:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._cpy(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1550
 *         elif opcode==0xC4:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xCC:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1551
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._cpy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xCC:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cpy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1548
 *             immediate = True
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC4:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1552
 *             arg = data[0]
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xCC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xCC) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1553
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xCC:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._cpy(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1554
 *         elif opcode==0xCC:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC6:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1555
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._cpy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC6:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_cpy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1552
 *             arg = data[0]
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xCC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1556
 *             arg = self._from_le(data)
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC6:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC6) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1557
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC6:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._dec(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1558
 *         elif opcode==0xC6:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xD6:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1559
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._dec(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD6:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_dec(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1556
 *             arg = self._from_le(data)
 *             cycles += self._cpy(arg, immediate)
 *         elif opcode==0xC6:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1560
 *             arg = data[0]
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xD6:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD6) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1561
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xD6:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dec(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1562
 *         elif opcode==0xD6:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCE:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1563
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dec(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xCE:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_dec(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1560
 *             arg = data[0]
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xD6:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1564
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCE:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xCE) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1565
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCE:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._dec(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1566
 *         elif opcode==0xCE:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xDE:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1567
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._dec(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xDE:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_dec(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1564
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCE:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1568
 *             arg = self._from_le(data)
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xDE:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xDE) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1569
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xDE:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dec(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1570
 *         elif opcode==0xDE:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCA:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1571
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dec(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xCA:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_dec(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1568
 *             arg = self._from_le(data)
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xDE:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1572
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xCA) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1573
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCA:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._dex(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1574
 *         elif opcode==0xCA:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._dex(arg, immediate)
 *         elif opcode==0x88:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1575
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._dex(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x88:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_dex(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1572
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dec(arg, immediate)
 *         elif opcode==0xCA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1576
 *             arg = ARG_NONE
 *             cycles += self._dex(arg, immediate)
 *         elif opcode==0x88:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x88) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1577
 *             cycles += self._dex(arg, immediate)
 *         elif opcode==0x88:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._dey(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1578
 *         elif opcode==0x88:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._dey(arg, immediate)
 *         elif opcode==0x49:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1579
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._dey(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x49:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_dey(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1576
 *             arg = ARG_NONE
 *             cycles += self._dex(arg, immediate)
 *         elif opcode==0x88:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1580
 *             arg = ARG_NONE
 *             cycles += self._dey(arg, immediate)
 *         elif opcode==0x49:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x49) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1581
 *             cycles += self._dey(arg, immediate)
 *         elif opcode==0x49:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1582
 *         elif opcode==0x49:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1583
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x45:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1584
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x45:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1580
 *             arg = ARG_NONE
 *             cycles += self._dey(arg, immediate)
 *         elif opcode==0x49:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1585
 *             immediate = True
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x45:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x45) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1586
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x45:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1587
 *         elif opcode==0x45:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x55:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1588
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x55:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1585
 *             immediate = True
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x45:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1589
 *             arg = data[0]
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x55:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x55) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1590
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x55:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1591
 *         elif opcode==0x55:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x4D:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1592
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x4D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1589
 *             arg = data[0]
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x55:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1593
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x4D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x4D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1594
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x4D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1595
 *         elif opcode==0x4D:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x5D:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1596
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x5D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1593
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x4D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1597
 *             arg = self._from_le(data)
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x5D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x5D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1598
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x5D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1599
 *         elif opcode==0x5D:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1600
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x59:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1601
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x59:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1597
 *             arg = self._from_le(data)
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x5D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1602
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x59:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x59) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1603
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x59:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1604
 *         elif opcode==0x59:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1605
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x41:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1606
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x41:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1602
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x59:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1607
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x41:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x41) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1608
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x41:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1609
 *         elif opcode==0x41:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x51:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1610
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x51:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1607
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x41:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1611
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x51:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x51) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1612
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x51:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1613
 *         elif opcode==0x51:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1614
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1615
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0xE6:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1616
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE6:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_eor(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1611
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0x51:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1617
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0xE6:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE6) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1618
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0xE6:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._inc(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1619
 *         elif opcode==0xE6:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xF6:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1620
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._inc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF6:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_inc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1617
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._eor(arg, immediate)
 *         elif opcode==0xE6:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1621
 *             arg = data[0]
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xF6:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF6) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1622
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xF6:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._inc(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1623
 *         elif opcode==0xF6:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xEE:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1624
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._inc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xEE:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_inc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1621
 *             arg = data[0]
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xF6:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1625
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xEE:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xEE) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1626
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xEE:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._inc(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1627
 *         elif opcode==0xEE:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xFE:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1628
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._inc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xFE:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_inc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1625
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xEE:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1629
 *             arg = self._from_le(data)
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xFE:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xFE) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1630
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xFE:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._inc(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1631
 *         elif opcode==0xFE:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xE8:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1632
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._inc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE8:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_inc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1629
 *             arg = self._from_le(data)
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xFE:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1633
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xE8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE8) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1634
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xE8:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._inx(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1635
 *         elif opcode==0xE8:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._inx(arg, immediate)
 *         elif opcode==0xC8:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1636
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._inx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC8:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_inx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1633
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._inc(arg, immediate)
 *         elif opcode==0xE8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1637
 *             arg = ARG_NONE
 *             cycles += self._inx(arg, immediate)
 *         elif opcode==0xC8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC8) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1638
 *             cycles += self._inx(arg, immediate)
 *         elif opcode==0xC8:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._iny(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1639
 *         elif opcode==0xC8:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._iny(arg, immediate)
 *         elif opcode==0x4C:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1640
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._iny(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x4C:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_iny(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1637
 *             arg = ARG_NONE
 *             cycles += self._inx(arg, immediate)
 *         elif opcode==0xC8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1641
 *             arg = ARG_NONE
 *             cycles += self._iny(arg, immediate)
 *         elif opcode==0x4C:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x4C) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1642
 *             cycles += self._iny(arg, immediate)
 *         elif opcode==0x4C:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._jmp(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1643
 *         elif opcode==0x4C:
 *             cycles = 3
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x6C:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1644
 *             cycles = 3
 *             arg = self._from_le(data)
 *             cycles += self._jmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x6C:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1641
 *             arg = ARG_NONE
 *             cycles += self._iny(arg, immediate)
 *         elif opcode==0x4C:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1645
 *             arg = self._from_le(data)
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x6C:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x6C) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1646
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x6C:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)
 *             cycles += self._jmp(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1647
 *         elif opcode==0x6C:
 *             cycles = 5
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x20:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data), 1);

    /* "nes/cycore/mos6502.pyx":1648
 *             cycles = 5
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)
 *             cycles += self._jmp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x20:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jmp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1645
 *             arg = self._from_le(data)
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x6C:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1649
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x20:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x20) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1650
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x20:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._jsr(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1651
 *         elif opcode==0x20:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._jsr(arg, immediate)
 *         elif opcode==0xA9:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1652
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._jsr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA9:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_jsr(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1649
 *             arg = self._read_word(self._from_le(data), wrap_at_page=True)
 *             cycles += self._jmp(arg, immediate)
 *         elif opcode==0x20:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1653
 *             arg = self._from_le(data)
 *             cycles += self._jsr(arg, immediate)
 *         elif opcode==0xA9:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA9) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1654
 *             cycles += self._jsr(arg, immediate)
 *         elif opcode==0xA9:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1655
 *         elif opcode==0xA9:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1656
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA5:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1657
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA5:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1653
 *             arg = self._from_le(data)
 *             cycles += self._jsr(arg, immediate)
 *         elif opcode==0xA9:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1658
 *             immediate = True
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA5:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA5) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1659
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA5:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1660
 *         elif opcode==0xA5:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB5:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1661
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB5:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1658
 *             immediate = True
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA5:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1662
 *             arg = data[0]
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB5:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB5) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1663
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB5:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1664
 *         elif opcode==0xB5:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xAD:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1665
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xAD:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1662
 *             arg = data[0]
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB5:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1666
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xAD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xAD) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1667
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xAD:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1668
 *         elif opcode==0xAD:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xBD:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1669
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xBD:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1666
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xAD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1670
 *             arg = self._from_le(data)
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xBD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xBD) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1671
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xBD:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1672
 *         elif opcode==0xBD:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1673
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB9:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1674
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB9:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1670
 *             arg = self._from_le(data)
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xBD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1675
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB9:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB9) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1676
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB9:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1677
 *         elif opcode==0xB9:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1678
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA1:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1679
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA1:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1675
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB9:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1680
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA1:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1681
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA1:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1682
 *         elif opcode==0xA1:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB1:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1683
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB1:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1680
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA1:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1684
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB1:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1685
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB1:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1686
 *         elif opcode==0xB1:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1687
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1688
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA2:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1689
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA2:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lda(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1684
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xB1:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1690
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA2:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA2) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1691
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA2:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1692
 *         elif opcode==0xA2:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._ldx(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1693
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA6:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1694
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._ldx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA6:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1690
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lda(arg, immediate)
 *         elif opcode==0xA2:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1695
 *             immediate = True
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA6:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA6) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1696
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA6:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._ldx(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1697
 *         elif opcode==0xA6:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xB6:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1698
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._ldx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB6:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1695
 *             immediate = True
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA6:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1699
 *             arg = data[0]
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xB6:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB6) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1700
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xB6:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._ldx(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1701
 *         elif opcode==0xB6:
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xAE:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->Y) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1702
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._ldx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xAE:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1699
 *             arg = data[0]
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xB6:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1703
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xAE:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xAE) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1704
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xAE:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._ldx(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1705
 *         elif opcode==0xAE:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xBE:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1706
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._ldx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xBE:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1703
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xAE:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1707
 *             arg = self._from_le(data)
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xBE:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xBE) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1708
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xBE:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1709
 *         elif opcode==0xBE:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ldx(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1710
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA0:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1711
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ldx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA0:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1707
 *             arg = self._from_le(data)
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xBE:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1712
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA0) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1713
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA0:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1714
 *         elif opcode==0xA0:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._ldy(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1715
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xA4:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1716
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._ldy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA4:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1712
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ldx(arg, immediate)
 *         elif opcode==0xA0:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1717
 *             immediate = True
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xA4:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA4) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1718
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xA4:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._ldy(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1719
 *         elif opcode==0xA4:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xB4:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1720
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._ldy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB4:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1717
 *             immediate = True
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xA4:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1721
 *             arg = data[0]
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xB4:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB4) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1722
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xB4:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ldy(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1723
 *         elif opcode==0xB4:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xAC:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1724
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ldy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xAC:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1721
 *             arg = data[0]
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xB4:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1725
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xAC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xAC) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1726
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xAC:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._ldy(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1727
 *         elif opcode==0xAC:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xBC:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1728
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._ldy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xBC:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1725
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xAC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1729
 *             arg = self._from_le(data)
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xBC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xBC) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1730
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xBC:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1731
 *         elif opcode==0xBC:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ldy(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1732
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0x4A:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1733
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ldy(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x4A:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ldy(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1729
 *             arg = self._from_le(data)
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0xBC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1734
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0x4A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x4A) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1735
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0x4A:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._lsr(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1736
 *         elif opcode==0x4A:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x46:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1737
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._lsr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x46:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1734
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ldy(arg, immediate)
 *         elif opcode==0x4A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1738
 *             arg = ARG_NONE
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x46:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x46) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1739
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x46:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._lsr(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1740
 *         elif opcode==0x46:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x56:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1741
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._lsr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x56:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1738
 *             arg = ARG_NONE
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x46:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1742
 *             arg = data[0]
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x56:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x56) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1743
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x56:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lsr(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1744
 *         elif opcode==0x56:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x4E:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1745
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lsr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x4E:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1742
 *             arg = data[0]
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x56:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1746
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x4E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x4E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1747
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x4E:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._lsr(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1748
 *         elif opcode==0x4E:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x5E:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1749
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._lsr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x5E:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1746
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x4E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1750
 *             arg = self._from_le(data)
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x5E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x5E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1751
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x5E:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._lsr(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1752
 *         elif opcode==0x5E:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1753
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._lsr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_lsr(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1750
 *             arg = self._from_le(data)
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0x5E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1754
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  switch (__pyx_v_opcode) {
    case 0xEA:
    case 0x1A:
    case 0x3A:
    case 0x5A:
    case 0x7A:
    case 0xDA:
    case 0xFA:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1755
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._nop(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1756
 *         elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._nop(arg, immediate)
 *         elif opcode==0x09:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1757
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._nop(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x09:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_nop(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1754
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._lsr(arg, immediate)
 *         elif opcode==0xEA or opcode==0x1A or opcode==0x3A or opcode==0x5A or opcode==0x7A or opcode==0xDA or opcode==0xFA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1758
 *             arg = ARG_NONE
 *             cycles += self._nop(arg, immediate)
 *         elif opcode==0x09:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x09) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1759
 *             cycles += self._nop(arg, immediate)
 *         elif opcode==0x09:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1760
 *         elif opcode==0x09:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1761
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x05:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1762
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x05:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1758
 *             arg = ARG_NONE
 *             cycles += self._nop(arg, immediate)
 *         elif opcode==0x09:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1763
 *             immediate = True
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x05:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x05) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1764
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x05:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1765
 *         elif opcode==0x05:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x15:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1766
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x15:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1763
 *             immediate = True
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x05:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1767
 *             arg = data[0]
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x15:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x15) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1768
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x15:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1769
 *         elif opcode==0x15:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x0D:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1770
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x0D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1767
 *             arg = data[0]
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x15:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1771
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x0D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x0D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1772
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x0D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1773
 *         elif opcode==0x0D:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x1D:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1774
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x1D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1771
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x0D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1775
 *             arg = self._from_le(data)
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x1D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x1D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1776
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x1D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1777
 *         elif opcode==0x1D:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1778
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x19:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1779
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x19:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1775
 *             arg = self._from_le(data)
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x1D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1780
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x19:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x19) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1781
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x19:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1782
 *         elif opcode==0x19:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1783
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x01:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1784
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x01:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1780
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x19:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1785
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x01:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x01) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1786
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x01:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1787
 *         elif opcode==0x01:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x11:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1788
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x11:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1785
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x01:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1789
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x11:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x11) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1790
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x11:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1791
 *         elif opcode==0x11:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1792
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1793
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x48:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1794
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x48:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ora(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1789
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x11:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1795
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x48:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x48) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1796
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x48:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._pha(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1797
 *         elif opcode==0x48:
 *             cycles = 3
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._pha(arg, immediate)
 *         elif opcode==0x08:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1798
 *             cycles = 3
 *             arg = ARG_NONE
 *             cycles += self._pha(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x08:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_pha(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1795
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._ora(arg, immediate)
 *         elif opcode==0x48:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1799
 *             arg = ARG_NONE
 *             cycles += self._pha(arg, immediate)
 *         elif opcode==0x08:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x08) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1800
 *             cycles += self._pha(arg, immediate)
 *         elif opcode==0x08:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._php(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1801
 *         elif opcode==0x08:
 *             cycles = 3
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._php(arg, immediate)
 *         elif opcode==0x68:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1802
 *             cycles = 3
 *             arg = ARG_NONE
 *             cycles += self._php(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x68:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_php(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1799
 *             arg = ARG_NONE
 *             cycles += self._pha(arg, immediate)
 *         elif opcode==0x08:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1803
 *             arg = ARG_NONE
 *             cycles += self._php(arg, immediate)
 *         elif opcode==0x68:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x68) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1804
 *             cycles += self._php(arg, immediate)
 *         elif opcode==0x68:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._pla(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1805
 *         elif opcode==0x68:
 *             cycles = 4
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._pla(arg, immediate)
 *         elif opcode==0x28:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1806
 *             cycles = 4
 *             arg = ARG_NONE
 *             cycles += self._pla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x28:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_pla(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1803
 *             arg = ARG_NONE
 *             cycles += self._php(arg, immediate)
 *         elif opcode==0x68:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1807
 *             arg = ARG_NONE
 *             cycles += self._pla(arg, immediate)
 *         elif opcode==0x28:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x28) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1808
 *             cycles += self._pla(arg, immediate)
 *         elif opcode==0x28:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._plp(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1809
 *         elif opcode==0x28:
 *             cycles = 4
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._plp(arg, immediate)
 *         elif opcode==0x2A:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1810
 *             cycles = 4
 *             arg = ARG_NONE
 *             cycles += self._plp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x2A:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_plp(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1807
 *             arg = ARG_NONE
 *             cycles += self._pla(arg, immediate)
 *         elif opcode==0x28:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1811
 *             arg = ARG_NONE
 *             cycles += self._plp(arg, immediate)
 *         elif opcode==0x2A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x2A) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1812
 *             cycles += self._plp(arg, immediate)
 *         elif opcode==0x2A:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._rol(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1813
 *         elif opcode==0x2A:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x26:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1814
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._rol(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x26:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rol(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1811
 *             arg = ARG_NONE
 *             cycles += self._plp(arg, immediate)
 *         elif opcode==0x2A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1815
 *             arg = ARG_NONE
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x26:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x26) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1816
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x26:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._rol(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1817
 *         elif opcode==0x26:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x36:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1818
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._rol(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x36:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rol(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1815
 *             arg = ARG_NONE
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x26:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1819
 *             arg = data[0]
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x36:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x36) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1820
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x36:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rol(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1821
 *         elif opcode==0x36:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x2E:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1822
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rol(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x2E:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rol(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1819
 *             arg = data[0]
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x36:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1823
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x2E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x2E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1824
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x2E:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._rol(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1825
 *         elif opcode==0x2E:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x3E:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1826
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._rol(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x3E:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rol(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1823
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x2E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1827
 *             arg = self._from_le(data)
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x3E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x3E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1828
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x3E:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rol(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1829
 *         elif opcode==0x3E:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x6A:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1830
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rol(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x6A:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rol(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1827
 *             arg = self._from_le(data)
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x3E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1831
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x6A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x6A) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1832
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x6A:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._ror(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1833
 *         elif opcode==0x6A:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x66:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1834
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._ror(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x66:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1831
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rol(arg, immediate)
 *         elif opcode==0x6A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1835
 *             arg = ARG_NONE
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x66:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x66) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1836
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x66:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._ror(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1837
 *         elif opcode==0x66:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x76:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1838
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._ror(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x76:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1835
 *             arg = ARG_NONE
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x66:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1839
 *             arg = data[0]
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x76:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x76) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1840
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x76:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ror(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1841
 *         elif opcode==0x76:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x6E:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1842
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ror(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x6E:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1839
 *             arg = data[0]
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x76:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1843
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x6E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x6E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1844
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x6E:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._ror(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1845
 *         elif opcode==0x6E:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x7E:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1846
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._ror(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x7E:
 *             cycles = 7
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1843
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x6E:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1847
 *             arg = self._from_le(data)
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x7E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x7E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1848
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x7E:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._ror(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":1849
 *         elif opcode==0x7E:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x40:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1850
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._ror(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x40:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_ror(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1847
 *             arg = self._from_le(data)
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x7E:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1851
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x40:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x40) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1852
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x40:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._rti(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1853
 *         elif opcode==0x40:
 *             cycles = 6
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._rti(arg, immediate)
 *         elif opcode==0x60:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1854
 *             cycles = 6
 *             arg = ARG_NONE
 *             cycles += self._rti(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x60:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rti(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1851
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._ror(arg, immediate)
 *         elif opcode==0x40:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1855
 *             arg = ARG_NONE
 *             cycles += self._rti(arg, immediate)
 *         elif opcode==0x60:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x60) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1856
 *             cycles += self._rti(arg, immediate)
 *         elif opcode==0x60:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._rts(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1857
 *         elif opcode==0x60:
 *             cycles = 6
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._rts(arg, immediate)
 *         elif opcode==0xE9 or opcode==0xEB:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1858
 *             cycles = 6
 *             arg = ARG_NONE
 *             cycles += self._rts(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE9 or opcode==0xEB:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_rts(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1855
 *             arg = ARG_NONE
 *             cycles += self._rti(arg, immediate)
 *         elif opcode==0x60:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1859
 *             arg = ARG_NONE
 *             cycles += self._rts(arg, immediate)
 *         elif opcode==0xE9 or opcode==0xEB:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  switch (__pyx_v_opcode) {
    case 0xE9:
    case 0xEB:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1860
 *             cycles += self._rts(arg, immediate)
 *         elif opcode==0xE9 or opcode==0xEB:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1861
 *         elif opcode==0xE9 or opcode==0xEB:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1862
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE5:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1863
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE5:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1859
 *             arg = ARG_NONE
 *             cycles += self._rts(arg, immediate)
 *         elif opcode==0xE9 or opcode==0xEB:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1864
 *             immediate = True
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE5:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE5) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1865
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE5:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1866
 *         elif opcode==0xE5:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF5:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1867
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF5:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1864
 *             immediate = True
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE5:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1868
 *             arg = data[0]
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF5:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF5) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1869
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF5:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1870
 *         elif opcode==0xF5:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xED:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1871
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xED:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1868
 *             arg = data[0]
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF5:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1872
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xED:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xED) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1873
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xED:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1874
 *         elif opcode==0xED:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xFD:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1875
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xFD:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1872
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xED:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1876
 *             arg = self._from_le(data)
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xFD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xFD) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1877
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xFD:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1878
 *         elif opcode==0xFD:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1879
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF9:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1880
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF9:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1876
 *             arg = self._from_le(data)
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xFD:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1881
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF9:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF9) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1882
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF9:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1883
 *         elif opcode==0xF9:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1884
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE1:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1885
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE1:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1881
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF9:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1886
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE1:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1887
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE1:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1888
 *         elif opcode==0xE1:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF1:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1889
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF1:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1886
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xE1:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1890
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF1:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF1) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1891
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF1:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1892
 *         elif opcode==0xF1:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1893
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1894
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0x38:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":1895
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x38:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sbc(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1890
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0xF1:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1896
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0x38:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x38) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1897
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0x38:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._sec(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1898
 *         elif opcode==0x38:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._sec(arg, immediate)
 *         elif opcode==0xF8:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1899
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._sec(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF8:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sec(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1896
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._sbc(arg, immediate)
 *         elif opcode==0x38:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1900
 *             arg = ARG_NONE
 *             cycles += self._sec(arg, immediate)
 *         elif opcode==0xF8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF8) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1901
 *             cycles += self._sec(arg, immediate)
 *         elif opcode==0xF8:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._sed(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1902
 *         elif opcode==0xF8:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._sed(arg, immediate)
 *         elif opcode==0x78:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1903
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._sed(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x78:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sed(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1900
 *             arg = ARG_NONE
 *             cycles += self._sec(arg, immediate)
 *         elif opcode==0xF8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1904
 *             arg = ARG_NONE
 *             cycles += self._sed(arg, immediate)
 *         elif opcode==0x78:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x78) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1905
 *             cycles += self._sed(arg, immediate)
 *         elif opcode==0x78:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._sei(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1906
 *         elif opcode==0x78:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._sei(arg, immediate)
 *         elif opcode==0x85:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1907
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._sei(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x85:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sei(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1904
 *             arg = ARG_NONE
 *             cycles += self._sed(arg, immediate)
 *         elif opcode==0x78:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1908
 *             arg = ARG_NONE
 *             cycles += self._sei(arg, immediate)
 *         elif opcode==0x85:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x85) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1909
 *             cycles += self._sei(arg, immediate)
 *         elif opcode==0x85:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1910
 *         elif opcode==0x85:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x95:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1911
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x95:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1908
 *             arg = ARG_NONE
 *             cycles += self._sei(arg, immediate)
 *         elif opcode==0x85:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1912
 *             arg = data[0]
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x95:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x95) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1913
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x95:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1914
 *         elif opcode==0x95:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x8D:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1915
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x8D:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1912
 *             arg = data[0]
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x95:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1916
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x8D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x8D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1917
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x8D:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1918
 *         elif opcode==0x8D:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x9D:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1919
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x9D:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1916
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x8D:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1920
 *             arg = self._from_le(data)
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x9D:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x9D) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1921
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x9D:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1922
 *         elif opcode==0x9D:
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x99:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1923
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x99:
 *             cycles = 5
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1920
 *             arg = self._from_le(data)
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x9D:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1924
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x99:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x99) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1925
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x99:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":1926
 *         elif opcode==0x99:
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x81:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1927
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x81:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1924
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x99:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1928
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x81:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x81) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1929
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x81:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1930
 *         elif opcode==0x81:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x91:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":1931
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x91:
 *             cycles = 6
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1928
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x81:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1932
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x91:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x91) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1933
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x91:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":1934
 *         elif opcode==0x91:
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":1935
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x86:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":1936
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x86:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sta(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1932
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x91:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1937
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x86:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x86) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1938
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x86:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._stx(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1939
 *         elif opcode==0x86:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x96:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1940
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._stx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x96:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_stx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1937
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sta(arg, immediate)
 *         elif opcode==0x86:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1941
 *             arg = data[0]
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x96:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x96) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1942
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x96:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._stx(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1943
 *         elif opcode==0x96:
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x8E:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->Y) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1944
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._stx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x8E:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_stx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1941
 *             arg = data[0]
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x96:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1945
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x8E:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x8E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1946
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x8E:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._stx(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1947
 *         elif opcode==0x8E:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x84:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1948
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._stx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x84:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_stx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1945
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x8E:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1949
 *             arg = self._from_le(data)
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x84:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x84) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1950
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x84:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._sty(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1951
 *         elif opcode==0x84:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x94:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1952
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._sty(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x94:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sty(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1949
 *             arg = self._from_le(data)
 *             cycles += self._stx(arg, immediate)
 *         elif opcode==0x84:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1953
 *             arg = data[0]
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x94:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x94) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1954
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x94:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sty(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1955
 *         elif opcode==0x94:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x8C:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1956
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sty(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x8C:
 *             cycles = 4
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sty(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1953
 *             arg = data[0]
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x94:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1957
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x8C:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x8C) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1958
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x8C:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._sty(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1959
 *         elif opcode==0x8C:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0xAA:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":1960
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._sty(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xAA:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_sty(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1957
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0x8C:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1961
 *             arg = self._from_le(data)
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0xAA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xAA) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1962
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0xAA:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._tax(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1963
 *         elif opcode==0xAA:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._tax(arg, immediate)
 *         elif opcode==0xA8:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1964
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._tax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA8:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_tax(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1961
 *             arg = self._from_le(data)
 *             cycles += self._sty(arg, immediate)
 *         elif opcode==0xAA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1965
 *             arg = ARG_NONE
 *             cycles += self._tax(arg, immediate)
 *         elif opcode==0xA8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA8) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1966
 *             cycles += self._tax(arg, immediate)
 *         elif opcode==0xA8:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._tay(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1967
 *         elif opcode==0xA8:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._tay(arg, immediate)
 *         elif opcode==0xBA:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1968
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._tay(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xBA:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_tay(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1965
 *             arg = ARG_NONE
 *             cycles += self._tax(arg, immediate)
 *         elif opcode==0xA8:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1969
 *             arg = ARG_NONE
 *             cycles += self._tay(arg, immediate)
 *         elif opcode==0xBA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xBA) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1970
 *             cycles += self._tay(arg, immediate)
 *         elif opcode==0xBA:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._tsx(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1971
 *         elif opcode==0xBA:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._tsx(arg, immediate)
 *         elif opcode==0x8A:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1972
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._tsx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x8A:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_tsx(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1969
 *             arg = ARG_NONE
 *             cycles += self._tay(arg, immediate)
 *         elif opcode==0xBA:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1973
 *             arg = ARG_NONE
 *             cycles += self._tsx(arg, immediate)
 *         elif opcode==0x8A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x8A) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1974
 *             cycles += self._tsx(arg, immediate)
 *         elif opcode==0x8A:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._txa(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1975
 *         elif opcode==0x8A:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._txa(arg, immediate)
 *         elif opcode==0x9A:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1976
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._txa(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x9A:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_txa(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1973
 *             arg = ARG_NONE
 *             cycles += self._tsx(arg, immediate)
 *         elif opcode==0x8A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1977
 *             arg = ARG_NONE
 *             cycles += self._txa(arg, immediate)
 *         elif opcode==0x9A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x9A) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1978
 *             cycles += self._txa(arg, immediate)
 *         elif opcode==0x9A:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._txs(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1979
 *         elif opcode==0x9A:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._txs(arg, immediate)
 *         elif opcode==0x98:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1980
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._txs(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x98:
 *             cycles = 2
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_txs(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1977
 *             arg = ARG_NONE
 *             cycles += self._txa(arg, immediate)
 *         elif opcode==0x9A:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1981
 *             arg = ARG_NONE
 *             cycles += self._txs(arg, immediate)
 *         elif opcode==0x98:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x98) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1982
 *             cycles += self._txs(arg, immediate)
 *         elif opcode==0x98:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._tya(arg, immediate)
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1983
 *         elif opcode==0x98:
 *             cycles = 2
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._tya(arg, immediate)
 *         elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":1984
 *             cycles = 2
 *             arg = ARG_NONE
 *             cycles += self._tya(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:
 *             cycles = 3
 */
    __pyx_v_cycles = (__pyx_v_cycles + ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_tya(__pyx_v_self, __pyx_v_arg, __pyx_v_immediate));

    /* "nes/cycore/mos6502.pyx":1981
 *             arg = ARG_NONE
 *             cycles += self._txs(arg, immediate)
 *         elif opcode==0x98:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1985
 *             arg = ARG_NONE
 *             cycles += self._tya(arg, immediate)
 *         elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  switch (__pyx_v_opcode) {
    case 0x04:
    case 0x14:
    case 0x44:
    case 0x64:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1986
 *             cycles += self._tya(arg, immediate)
 *         elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._dop(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":1987
 *         elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1988
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._dop(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1988, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1988, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1988, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1988, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1988, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":1985
 *             arg = ARG_NONE
 *             cycles += self._tya(arg, immediate)
 *         elif opcode==0x04 or opcode==0x14 or opcode==0x44 or opcode==0x64:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1989
 *             arg = data[0]
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
  switch (__pyx_v_opcode) {
    case 0x14:
    case 0x34:
    case 0x54:
    case 0x74:
    case 0xD4:
    case 0xF4:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1990
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dop(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":1991
 *         elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":1992
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dop(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:
 *             cycles = 2
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dop); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 1992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1992, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1992, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1992, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1992, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1992, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1992, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":1989
 *             arg = data[0]
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x14 or opcode==0x34 or opcode==0x54 or opcode==0x74 or opcode==0xD4 or opcode==0xF4:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1993
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  switch (__pyx_v_opcode) {
    case 0x80:
    case 0x82:
    case 0x89:
    case 0xC2:
    case 0xE2:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1994
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":1995
 *         elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._dop(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":1996
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x0C:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":1997
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._dop(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x0C:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dop); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1997, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1997, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 1997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 1997, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 1997, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":1993
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x80 or opcode==0x82 or opcode==0x89 or opcode==0xC2 or opcode==0xE2:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":1998
 *             immediate = True
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x0C:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x0C) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":1999
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x0C:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._top(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2000
 *         elif opcode==0x0C:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2001
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._top(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:
 *             cycles = 4
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_top); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2001, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2001, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2001, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":1998
 *             immediate = True
 *             cycles += self._dop(arg, immediate)
 *         elif opcode==0x0C:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2002
 *             arg = self._from_le(data)
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  switch (__pyx_v_opcode) {
    case 0x1C:
    case 0x3C:
    case 0x5C:
    case 0x7C:
    case 0xDC:
    case 0xFC:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2003
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2004
 *         elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._top(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2005
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->X) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":2006
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._top(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:
 *             cycles = 1
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_top); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2006, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2006, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2006, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2006, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2006, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2002
 *             arg = self._from_le(data)
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x1C or opcode==0x3C or opcode==0x5C or opcode==0x7C or opcode==0xDC or opcode==0xFC:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2007
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:             # <<<<<<<<<<<<<<
 *             cycles = 1
 *             arg = ARG_NONE
 */
  switch (__pyx_v_opcode) {
    case 0x02:
    case 0x12:
    case 0x22:
    case 0x32:
    case 0x42:
    case 0x52:
    case 0x62:
    case 0x72:
    case 0x92:
    case 0xB2:
    case 0xD2:
    case 0xF2:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2008
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:
 *             cycles = 1             # <<<<<<<<<<<<<<
 *             arg = ARG_NONE
 *             cycles += self._kil(arg, immediate)
 */
    __pyx_v_cycles = 1;

    /* "nes/cycore/mos6502.pyx":2009
 *         elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:
 *             cycles = 1
 *             arg = ARG_NONE             # <<<<<<<<<<<<<<
 *             cycles += self._kil(arg, immediate)
 *         elif opcode==0xA7:
 */
    __pyx_v_arg = __pyx_e_3nes_6cycore_7mos6502_ARG_NONE;

    /* "nes/cycore/mos6502.pyx":2010
 *             cycles = 1
 *             arg = ARG_NONE
 *             cycles += self._kil(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA7:
 *             cycles = 3
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_kil); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2010, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2010, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2010, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2007
 *             cycles += 1 if (data[LO_BYTE] + self.X) > 0xFF else 0
 *             cycles += self._top(arg, immediate)
 *         elif opcode==0x02 or opcode==0x12 or opcode==0x22 or opcode==0x32 or opcode==0x42 or opcode==0x52 or opcode==0x62 or opcode==0x72 or opcode==0x92 or opcode==0xB2 or opcode==0xD2 or opcode==0xF2:             # <<<<<<<<<<<<<<
 *             cycles = 1
 *             arg = ARG_NONE
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2011
 *             arg = ARG_NONE
 *             cycles += self._kil(arg, immediate)
 *         elif opcode==0xA7:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA7) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2012
 *             cycles += self._kil(arg, immediate)
 *         elif opcode==0xA7:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._lax(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":2013
 *         elif opcode==0xA7:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB7:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2014
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._lax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB7:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lax); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2014, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2014, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2014, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2014, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2011
 *             arg = ARG_NONE
 *             cycles += self._kil(arg, immediate)
 *         elif opcode==0xA7:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2015
 *             arg = data[0]
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB7:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB7) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2016
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB7:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._lax(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2017
 *         elif opcode==0xB7:
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xAF:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->Y) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2018
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._lax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xAF:
 *             cycles = 4
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lax); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2018, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2018, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2018, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2015
 *             arg = data[0]
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB7:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2019
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xAF:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xAF) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2020
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xAF:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._lax(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2021
 *         elif opcode==0xAF:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xBF:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2022
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._lax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xBF:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lax); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2022, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2022, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2022, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2022, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2019
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xAF:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2023
 *             arg = self._from_le(data)
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xBF:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xBF) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2024
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xBF:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2025
 *         elif opcode==0xBF:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2026
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xA3:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":2027
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xA3:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lax); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2027, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2027, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2027, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2027, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2023
 *             arg = self._from_le(data)
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xBF:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2028
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xA3:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xA3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2029
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xA3:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lax(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2030
 *         elif opcode==0xA3:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB3:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2031
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xB3:
 *             cycles = 5
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lax); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2031, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2031, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2031, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2031, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2031, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2028
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xA3:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2032
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB3:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xB3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2033
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB3:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2034
 *         elif opcode==0xB3:
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2035
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2036
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0x87:
 */
    if (((((__pyx_v_m & 0xFF) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":2037
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x87:
 *             cycles = 3
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lax); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2037, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2037, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2037, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2037, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2032
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0xB3:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2038
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0x87:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x87) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2039
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0x87:
 *             cycles = 3             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._aax(arg, immediate)
 */
    __pyx_v_cycles = 3;

    /* "nes/cycore/mos6502.pyx":2040
 *         elif opcode==0x87:
 *             cycles = 3
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x97:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2041
 *             cycles = 3
 *             arg = data[0]
 *             cycles += self._aax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x97:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_aax); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2041, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2041, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2041, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2041, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2038
 *             cycles += 1 if ((m & 0xFF) + self.Y) > 0xFF else 0
 *             cycles += self._lax(arg, immediate)
 *         elif opcode==0x87:             # <<<<<<<<<<<<<<
 *             cycles = 3
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2042
 *             arg = data[0]
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x97:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x97) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2043
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x97:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._aax(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2044
 *         elif opcode==0x97:
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x83:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->Y) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2045
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._aax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x83:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_aax); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2045, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2045, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2045, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2045, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2045, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2042
 *             arg = data[0]
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x97:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (data[0] + self.Y) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2046
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x83:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x83) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2047
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x83:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._aax(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2048
 *         elif opcode==0x83:
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x8F:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2049
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._aax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x8F:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_aax); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2049, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2049, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2049, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2049, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2046
 *             arg = (data[0] + self.Y) & 0xFF
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x83:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2050
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x8F:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x8F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2051
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x8F:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._aax(arg, immediate)
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2052
 *         elif opcode==0x8F:
 *             cycles = 4
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0xC7:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2053
 *             cycles = 4
 *             arg = self._from_le(data)
 *             cycles += self._aax(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC7:
 *             cycles = 5
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_aax); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2053, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2053, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2053, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2053, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2053, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2050
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0x8F:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2054
 *             arg = self._from_le(data)
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0xC7:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC7) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2055
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0xC7:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2056
 *         elif opcode==0xC7:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD7:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2057
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD7:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2057, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2057, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2057, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2057, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2054
 *             arg = self._from_le(data)
 *             cycles += self._aax(arg, immediate)
 *         elif opcode==0xC7:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2058
 *             arg = data[0]
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD7:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD7) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2059
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD7:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2060
 *         elif opcode==0xD7:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xCF:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2061
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xCF:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2061, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2061, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2061, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2061, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2061, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2058
 *             arg = data[0]
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD7:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2062
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xCF:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xCF) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2063
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xCF:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2064
 *         elif opcode==0xCF:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDF:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2065
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xDF:
 *             cycles = 7
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2065, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2065, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2065, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2065, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2062
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xCF:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2066
 *             arg = self._from_le(data)
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDF:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xDF) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2067
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDF:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2068
 *         elif opcode==0xDF:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDB:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2069
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xDB:
 *             cycles = 7
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2069, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2069, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2069, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2069, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2066
 *             arg = self._from_le(data)
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDF:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2070
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDB:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xDB) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2071
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDB:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2072
 *         elif opcode==0xDB:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xC3:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2073
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xC3:
 *             cycles = 8
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2073, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2073, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2073, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2073, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2073, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2070
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xDB:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2074
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xC3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xC3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2075
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xC3:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2076
 *         elif opcode==0xC3:
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD3:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2077
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xD3:
 *             cycles = 8
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2077, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2077, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2077, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2077, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2077, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2077, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2074
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xC3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2078
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xD3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2079
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD3:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2080
 *         elif opcode==0xD3:
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2081
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xE7:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2082
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE7:
 *             cycles = 5
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_dcp); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2082, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2082, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2082, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2082, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2082, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2078
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xD3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2083
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xE7:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE7) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2084
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xE7:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2085
 *         elif opcode==0xE7:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF7:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2086
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF7:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2086, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2086, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2086, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2086, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2083
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._dcp(arg, immediate)
 *         elif opcode==0xE7:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2087
 *             arg = data[0]
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF7:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF7) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2088
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF7:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2089
 *         elif opcode==0xF7:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xEF:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2090
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xEF:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2090, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2090, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2090, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2090, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2090, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2090, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2087
 *             arg = data[0]
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF7:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2091
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xEF:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xEF) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2092
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xEF:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2093
 *         elif opcode==0xEF:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFF:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2094
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xFF:
 *             cycles = 7
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2094, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2094, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2094, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2094, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2091
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xEF:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2095
 *             arg = self._from_le(data)
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFF:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xFF) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2096
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFF:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2097
 *         elif opcode==0xFF:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFB:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2098
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xFB:
 *             cycles = 7
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2098, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2098, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2098, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2095
 *             arg = self._from_le(data)
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFF:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2099
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFB:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xFB) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2100
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFB:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2101
 *         elif opcode==0xFB:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xE3:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2102
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xE3:
 *             cycles = 8
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2102, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2102, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2102, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2102, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2102, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2099
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xFB:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2103
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xE3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xE3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2104
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xE3:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2105
 *         elif opcode==0xE3:
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF3:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2106
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xF3:
 *             cycles = 8
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2106, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2106, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2106, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2106, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2103
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xE3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2107
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xF3) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2108
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF3:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2109
 *         elif opcode==0xF3:
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2110
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0x07:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2111
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x07:
 *             cycles = 5
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_isc); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2111, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2111, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2111, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2111, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2111, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2107
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0xF3:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2112
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0x07:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x07) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2113
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0x07:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2114
 *         elif opcode==0x07:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x17:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2115
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x17:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2115, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2115, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2115, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2115, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2112
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._isc(arg, immediate)
 *         elif opcode==0x07:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2116
 *             arg = data[0]
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x17:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x17) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2117
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x17:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2118
 *         elif opcode==0x17:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x0F:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2119
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x0F:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2119, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2119, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2119, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2119, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2116
 *             arg = data[0]
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x17:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2120
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x0F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x0F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2121
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x0F:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2122
 *         elif opcode==0x0F:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1F:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2123
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x1F:
 *             cycles = 7
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2123, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2123, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2123, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2123, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2120
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x0F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2124
 *             arg = self._from_le(data)
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x1F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2125
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1F:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2126
 *         elif opcode==0x1F:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1B:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2127
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x1B:
 *             cycles = 7
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2127, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2127, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2127, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2127, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2124
 *             arg = self._from_le(data)
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2128
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x1B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2129
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1B:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2130
 *         elif opcode==0x1B:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x03:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2131
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x03:
 *             cycles = 8
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2131, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2131, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2131, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2131, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2128
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x1B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2132
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x03:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x03) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2133
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x03:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2134
 *         elif opcode==0x03:
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x13:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2135
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x13:
 *             cycles = 8
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2135, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2135, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2135, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2135, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2132
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x03:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2136
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x13:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x13) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2137
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x13:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2138
 *         elif opcode==0x13:
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2139
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x27:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2140
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x27:
 *             cycles = 5
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_slo); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2140, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2140, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2140, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2140, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2136
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x13:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2141
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x27:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x27) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2142
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x27:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2143
 *         elif opcode==0x27:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x37:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2144
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x37:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2144, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2144, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2144, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2144, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2144, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2141
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._slo(arg, immediate)
 *         elif opcode==0x27:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2145
 *             arg = data[0]
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x37:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x37) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2146
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x37:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2147
 *         elif opcode==0x37:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x2F:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2148
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x2F:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2148, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2148, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2145
 *             arg = data[0]
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x37:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2149
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x2F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x2F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2150
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x2F:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2151
 *         elif opcode==0x2F:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3F:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2152
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x3F:
 *             cycles = 7
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2152, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2152, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2149
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x2F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2153
 *             arg = self._from_le(data)
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x3F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2154
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3F:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2155
 *         elif opcode==0x3F:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3B:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2156
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x3B:
 *             cycles = 7
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2156, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2156, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2156, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2156, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2153
 *             arg = self._from_le(data)
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2157
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x3B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2158
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3B:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2159
 *         elif opcode==0x3B:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x23:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2160
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x23:
 *             cycles = 8
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2160, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2160, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2160, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2160, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2157
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x3B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2161
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x23:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x23) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2162
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x23:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2163
 *         elif opcode==0x23:
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x33:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2164
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x33:
 *             cycles = 8
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2164, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2164, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2161
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x23:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2165
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x33:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x33) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2166
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x33:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2167
 *         elif opcode==0x33:
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2168
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x47:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2169
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x47:
 *             cycles = 5
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rla); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2169, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2169, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2169, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2165
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x33:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2170
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x47:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x47) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2171
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x47:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2172
 *         elif opcode==0x47:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x57:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2173
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x57:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2173, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2173, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2173, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2173, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2170
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rla(arg, immediate)
 *         elif opcode==0x47:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2174
 *             arg = data[0]
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x57:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x57) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2175
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x57:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2176
 *         elif opcode==0x57:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x4F:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2177
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x4F:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2177, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2177, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2177, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2177, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2177, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2174
 *             arg = data[0]
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x57:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2178
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x4F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x4F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2179
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x4F:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2180
 *         elif opcode==0x4F:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5F:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2181
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x5F:
 *             cycles = 7
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2181, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2181, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2181, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2181, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2178
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x4F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2182
 *             arg = self._from_le(data)
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x5F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2183
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5F:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2184
 *         elif opcode==0x5F:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5B:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2185
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x5B:
 *             cycles = 7
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2185, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2185, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2185, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2182
 *             arg = self._from_le(data)
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2186
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x5B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2187
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5B:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2188
 *         elif opcode==0x5B:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x43:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2189
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x43:
 *             cycles = 8
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2189, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2189, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2186
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x5B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2190
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x43:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x43) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2191
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x43:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2192
 *         elif opcode==0x43:
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x53:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2193
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x53:
 *             cycles = 8
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2193, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2193, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2193, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2193, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2190
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x43:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2194
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x53:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x53) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2195
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x53:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2196
 *         elif opcode==0x53:
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2197
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x67:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2198
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x67:
 *             cycles = 5
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sre); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2198, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2198, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2198, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2198, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2194
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x53:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2199
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x67:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x67) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2200
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x67:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2201
 *         elif opcode==0x67:
 *             cycles = 5
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x77:
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2202
 *             cycles = 5
 *             arg = data[0]
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x77:
 *             cycles = 6
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2202, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2202, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2202, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2202, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2199
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._sre(arg, immediate)
 *         elif opcode==0x67:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2203
 *             arg = data[0]
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x77:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x77) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2204
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x77:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2205
 *         elif opcode==0x77:
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6F:
 */
    __pyx_v_arg = (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF);

    /* "nes/cycore/mos6502.pyx":2206
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x6F:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2206, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2206, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2206, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2203
 *             arg = data[0]
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x77:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = (data[0] + self.X) & 0xFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2207
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x6F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2208
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6F:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             arg = self._from_le(data)
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2209
 *         elif opcode==0x6F:
 *             cycles = 6
 *             arg = self._from_le(data)             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7F:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data);

    /* "nes/cycore/mos6502.pyx":2210
 *             cycles = 6
 *             arg = self._from_le(data)
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x7F:
 *             cycles = 7
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2210, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2210, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2210, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2210, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2207
 *             arg = (data[0] + self.X) & 0xFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6F:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             arg = self._from_le(data)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2211
 *             arg = self._from_le(data)
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x7F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2212
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7F:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2213
 *         elif opcode==0x7F:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7B:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2214
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x7B:
 *             cycles = 7
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2214, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2214, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2214, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2214, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2211
 *             arg = self._from_le(data)
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7F:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2215
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x7B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2216
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7B:
 *             cycles = 7             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_cycles = 7;

    /* "nes/cycore/mos6502.pyx":2217
 *         elif opcode==0x7B:
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x63:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2218
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x63:
 *             cycles = 8
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2218, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2218, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2218, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2215
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x7B:             # <<<<<<<<<<<<<<
 *             cycles = 7
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2219
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x63:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x63) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2220
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x63:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2221
 *         elif opcode==0x63:
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x73:
 */
    __pyx_v_arg = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (((__pyx_v_data[0]) + __pyx_v_self->X) & 0xFF), 1);

    /* "nes/cycore/mos6502.pyx":2222
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x73:
 *             cycles = 8
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2222, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2222, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2222, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2222, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2219
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x63:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2223
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x73:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x73) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2224
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x73:
 *             cycles = 8             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 8;

    /* "nes/cycore/mos6502.pyx":2225
 *         elif opcode==0x73:
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2226
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6B:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2227
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x6B:
 *             cycles = 2
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_rra); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2227, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2223
 *             arg = self._read_word((data[0] + self.X) & 0xFF, wrap_at_page=True)
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x73:             # <<<<<<<<<<<<<<
 *             cycles = 8
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2228
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x6B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2229
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6B:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":2230
 *         elif opcode==0x6B:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._arr(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2231
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._arr(arg, immediate)
 *         elif opcode==0x4B:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":2232
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._arr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x4B:
 *             cycles = 2
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_arr); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2232, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2232, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2232, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2228
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._rra(arg, immediate)
 *         elif opcode==0x6B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2233
 *             immediate = True
 *             cycles += self._arr(arg, immediate)
 *         elif opcode==0x4B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x4B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2234
 *             cycles += self._arr(arg, immediate)
 *         elif opcode==0x4B:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":2235
 *         elif opcode==0x4B:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._asr(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2236
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._asr(arg, immediate)
 *         elif opcode==0xAB:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":2237
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._asr(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xAB:
 *             cycles = 2
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_asr); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2237, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2237, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2237, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2233
 *             immediate = True
 *             cycles += self._arr(arg, immediate)
 *         elif opcode==0x4B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2238
 *             immediate = True
 *             cycles += self._asr(arg, immediate)
 *         elif opcode==0xAB:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xAB) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2239
 *             cycles += self._asr(arg, immediate)
 *         elif opcode==0xAB:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":2240
 *         elif opcode==0xAB:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._atx(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2241
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._atx(arg, immediate)
 *         elif opcode==0x0B or opcode==0x2B:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":2242
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._atx(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x0B or opcode==0x2B:
 *             cycles = 2
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_atx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2242, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2242, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2242, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2242, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2242, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2238
 *             immediate = True
 *             cycles += self._asr(arg, immediate)
 *         elif opcode==0xAB:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2243
 *             immediate = True
 *             cycles += self._atx(arg, immediate)
 *         elif opcode==0x0B or opcode==0x2B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  switch (__pyx_v_opcode) {
    case 0x0B:
    case 0x2B:
    __pyx_t_1 = 1;
    break;
    default:
    __pyx_t_1 = 0;
    break;
  }
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2244
 *             cycles += self._atx(arg, immediate)
 *         elif opcode==0x0B or opcode==0x2B:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":2245
 *         elif opcode==0x0B or opcode==0x2B:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._aac(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2246
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._aac(arg, immediate)
 *         elif opcode==0x9F:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":2247
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._aac(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x9F:
 *             cycles = 5
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_aac); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2247, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2247, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2247, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2247, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2243
 *             immediate = True
 *             cycles += self._atx(arg, immediate)
 *         elif opcode==0x0B or opcode==0x2B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2248
 *             immediate = True
 *             cycles += self._aac(arg, immediate)
 *         elif opcode==0x9F:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x9F) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2249
 *             cycles += self._aac(arg, immediate)
 *         elif opcode==0x9F:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2250
 *         elif opcode==0x9F:
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0x93:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2251
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x93:
 *             cycles = 6
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_axa); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2251, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2251, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2251, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2251, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2248
 *             immediate = True
 *             cycles += self._aac(arg, immediate)
 *         elif opcode==0x9F:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2252
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0x93:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x93) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2253
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0x93:
 *             cycles = 6             # <<<<<<<<<<<<<<
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 */
    __pyx_v_cycles = 6;

    /* "nes/cycore/mos6502.pyx":2254
 *         elif opcode==0x93:
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)             # <<<<<<<<<<<<<<
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)
 */
    __pyx_v_m = ((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_read_word(__pyx_v_self, (__pyx_v_data[0]), 1);

    /* "nes/cycore/mos6502.pyx":2255
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0xCB:
 */
    __pyx_v_arg = ((__pyx_v_m + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2256
 *             m = self._read_word(data[0], wrap_at_page=True)
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xCB:
 *             cycles = 2
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_axa); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2256, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2256, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2252
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0x93:             # <<<<<<<<<<<<<<
 *             cycles = 6
 *             m = self._read_word(data[0], wrap_at_page=True)
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2257
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0xCB:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xCB) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2258
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0xCB:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":2259
 *         elif opcode==0xCB:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._axs(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2260
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._axs(arg, immediate)
 *         elif opcode==0xBB:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":2261
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._axs(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0xBB:
 *             cycles = 4
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_axs); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2261, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2261, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2261, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2261, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2261, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2257
 *             arg = (m + self.Y) & 0xFFFF
 *             cycles += self._axa(arg, immediate)
 *         elif opcode==0xCB:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2262
 *             immediate = True
 *             cycles += self._axs(arg, immediate)
 *         elif opcode==0xBB:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0xBB) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2263
 *             cycles += self._axs(arg, immediate)
 *         elif opcode==0xBB:
 *             cycles = 4             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 */
    __pyx_v_cycles = 4;

    /* "nes/cycore/mos6502.pyx":2264
 *         elif opcode==0xBB:
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lar(arg, immediate)
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2265
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0             # <<<<<<<<<<<<<<
 *             cycles += self._lar(arg, immediate)
 *         elif opcode==0x9E:
 */
    if (((((__pyx_v_data[__pyx_e_3nes_6cycore_7mos6502_LO_BYTE]) + __pyx_v_self->Y) > 0xFF) != 0)) {
      __pyx_t_2 = 1;
    } else {
      __pyx_t_2 = 0;
    }
    __pyx_v_cycles = (__pyx_v_cycles + __pyx_t_2);

    /* "nes/cycore/mos6502.pyx":2266
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lar(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x9E:
 *             cycles = 5
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_lar); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2266, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2266, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2266, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2266, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2262
 *             immediate = True
 *             cycles += self._axs(arg, immediate)
 *         elif opcode==0xBB:             # <<<<<<<<<<<<<<
 *             cycles = 4
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2267
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lar(arg, immediate)
 *         elif opcode==0x9E:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x9E) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2268
 *             cycles += self._lar(arg, immediate)
 *         elif opcode==0x9E:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sxa(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2269
 *         elif opcode==0x9E:
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sxa(arg, immediate)
 *         elif opcode==0x9C:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2270
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sxa(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x9C:
 *             cycles = 5
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sxa); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2270, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2270, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2270, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2270, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2267
 *             cycles += 1 if (data[LO_BYTE] + self.Y) > 0xFF else 0
 *             cycles += self._lar(arg, immediate)
 *         elif opcode==0x9E:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2271
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sxa(arg, immediate)
 *         elif opcode==0x9C:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x9C) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2272
 *             cycles += self._sxa(arg, immediate)
 *         elif opcode==0x9C:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sya(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2273
 *         elif opcode==0x9C:
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._sya(arg, immediate)
 *         elif opcode==0x8B:
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->X) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2274
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sya(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x8B:
 *             cycles = 2
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_sya); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_6 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2274, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_6, __pyx_t_10, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2274, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_6) {
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_6); __pyx_t_6 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_10);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_9, __pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_10 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2274, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2274, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2271
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._sxa(arg, immediate)
 *         elif opcode==0x9C:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2275
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sya(arg, immediate)
 *         elif opcode==0x8B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x8B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2276
 *             cycles += self._sya(arg, immediate)
 *         elif opcode==0x8B:
 *             cycles = 2             # <<<<<<<<<<<<<<
 *             arg = data[0]
 *             immediate = True
 */
    __pyx_v_cycles = 2;

    /* "nes/cycore/mos6502.pyx":2277
 *         elif opcode==0x8B:
 *             cycles = 2
 *             arg = data[0]             # <<<<<<<<<<<<<<
 *             immediate = True
 *             cycles += self._xaa(arg, immediate)
 */
    __pyx_v_arg = (__pyx_v_data[0]);

    /* "nes/cycore/mos6502.pyx":2278
 *             cycles = 2
 *             arg = data[0]
 *             immediate = True             # <<<<<<<<<<<<<<
 *             cycles += self._xaa(arg, immediate)
 *         elif opcode==0x9B:
 */
    __pyx_v_immediate = 1;

    /* "nes/cycore/mos6502.pyx":2279
 *             arg = data[0]
 *             immediate = True
 *             cycles += self._xaa(arg, immediate)             # <<<<<<<<<<<<<<
 *         elif opcode==0x9B:
 *             cycles = 5
 */
    __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_xaa); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_10 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[3] = {__pyx_t_10, __pyx_t_8, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      if (__pyx_t_10) {
        __Pyx_GIVEREF(__pyx_t_10); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_10); __pyx_t_10 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_8);
      PyTuple_SET_ITEM(__pyx_t_6, 0+__pyx_t_9, __pyx_t_8);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_6, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_8 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2279, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = PyNumber_InPlaceAdd(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2279, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2275
 *             arg = (self._from_le(data) + self.X) & 0xFFFF
 *             cycles += self._sya(arg, immediate)
 *         elif opcode==0x8B:             # <<<<<<<<<<<<<<
 *             cycles = 2
 *             arg = data[0]
 */
    goto __pyx_L3;
  }

  /* "nes/cycore/mos6502.pyx":2280
 *             immediate = True
 *             cycles += self._xaa(arg, immediate)
 *         elif opcode==0x9B:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  __pyx_t_1 = ((__pyx_v_opcode == 0x9B) != 0);
  if (__pyx_t_1) {

    /* "nes/cycore/mos6502.pyx":2281
 *             cycles += self._xaa(arg, immediate)
 *         elif opcode==0x9B:
 *             cycles = 5             # <<<<<<<<<<<<<<
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._xas(arg, immediate)
 */
    __pyx_v_cycles = 5;

    /* "nes/cycore/mos6502.pyx":2282
 *         elif opcode==0x9B:
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF             # <<<<<<<<<<<<<<
 *             cycles += self._xas(arg, immediate)
 * 
 */
    __pyx_v_arg = ((((struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self->__pyx_vtab)->_from_le(__pyx_v_self, __pyx_v_data) + __pyx_v_self->Y) & 0xFFFF);

    /* "nes/cycore/mos6502.pyx":2283
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 *             cycles += self._xas(arg, immediate)             # <<<<<<<<<<<<<<
 * 
 *         return cycles
 */
    __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_cycles); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_xas); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_arg); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_immediate); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __pyx_t_8 = NULL;
    __pyx_t_9 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_3))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_3);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_3, function);
        __pyx_t_9 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2283, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_3)) {
      PyObject *__pyx_temp[3] = {__pyx_t_8, __pyx_t_6, __pyx_t_7};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_3, __pyx_temp+1-__pyx_t_9, 2+__pyx_t_9); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2283, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    } else
    #endif
    {
      __pyx_t_10 = PyTuple_New(2+__pyx_t_9); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 2283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      if (__pyx_t_8) {
        __Pyx_GIVEREF(__pyx_t_8); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_8); __pyx_t_8 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 0+__pyx_t_9, __pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_7);
      PyTuple_SET_ITEM(__pyx_t_10, 1+__pyx_t_9, __pyx_t_7);
      __pyx_t_6 = 0;
      __pyx_t_7 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_3, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 2283, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = PyNumber_InPlaceAdd(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_9 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_9 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 2283, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_cycles = __pyx_t_9;

    /* "nes/cycore/mos6502.pyx":2280
 *             immediate = True
 *             cycles += self._xaa(arg, immediate)
 *         elif opcode==0x9B:             # <<<<<<<<<<<<<<
 *             cycles = 5
 *             arg = (self._from_le(data) + self.Y) & 0xFFFF
 */
  }
  __pyx_L3:;

  /* "nes/cycore/mos6502.pyx":2285
 *             cycles += self._xas(arg, immediate)
 * 
 *         return cycles             # <<<<<<<<<<<<<<
 */
  __pyx_r = __pyx_v_cycles;
  goto __pyx_L0;

  /* "nes/cycore/mos6502.pyx":1325
 *         return 0
 * 
 *     cdef int run_instr(self, unsigned char opcode, unsigned char data[2]):             # <<<<<<<<<<<<<<
 *         # *********** AUTOGENERATED BY meta.py - DO NOT EDIT DIRECTLY ***********
 *         cdef int cycles=0, arg, immediate=False, m
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_WriteUnraisable("nes.cycore.mos6502.MOS6502.run_instr", __pyx_clineno, __pyx_lineno, __pyx_filename, 1, 0);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_57__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_57__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_56__reduce_cython__(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_56__reduce_cython__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  PyObject *__pyx_t_13 = NULL;
  PyObject *__pyx_t_14 = NULL;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  int __pyx_t_19;
  int __pyx_t_20;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.A, self.C, self.D, self.I, self.N, self.PC, self.SP, self.V, self.X, self.Y, self.Z, self.aax_sets_flags, self.cycles_since_reset, self.instr_size_bytes, self.instructions, self.memory, self.stack_underflow_causes_exception, self.undocumented_support_level)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->A); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->C); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_self->D); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_self->I); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_self->N); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_t_6 = __Pyx_PyInt_From_int(__pyx_v_self->PC); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_7 = __Pyx_PyInt_From_int(__pyx_v_self->SP); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyInt_From_int(__pyx_v_self->V); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->X); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_10 = __Pyx_PyInt_From_unsigned_char(__pyx_v_self->Y); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_10);
  __pyx_t_11 = __Pyx_PyInt_From_int(__pyx_v_self->Z); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __pyx_t_12 = __Pyx_PyInt_From_int(__pyx_v_self->aax_sets_flags); if (unlikely(!__pyx_t_12)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_12);
  __pyx_t_13 = __Pyx_PyInt_From_PY_LONG_LONG(__pyx_v_self->cycles_since_reset); if (unlikely(!__pyx_t_13)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_13);
  __pyx_t_14 = __Pyx_carray_to_py_int(__pyx_v_self->instr_size_bytes, 0x100); if (unlikely(!__pyx_t_14)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_14);
  __pyx_t_15 = __Pyx_PyInt_From_int(__pyx_v_self->stack_underflow_causes_exception); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_15);
  __pyx_t_16 = __Pyx_PyInt_From_int(__pyx_v_self->undocumented_support_level); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_16);
  __pyx_t_17 = PyTuple_New(18); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_17, 3, __pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_5);
  PyTuple_SET_ITEM(__pyx_t_17, 4, __pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_6);
  PyTuple_SET_ITEM(__pyx_t_17, 5, __pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_7);
  PyTuple_SET_ITEM(__pyx_t_17, 6, __pyx_t_7);
  __Pyx_GIVEREF(__pyx_t_8);
  PyTuple_SET_ITEM(__pyx_t_17, 7, __pyx_t_8);
  __Pyx_GIVEREF(__pyx_t_9);
  PyTuple_SET_ITEM(__pyx_t_17, 8, __pyx_t_9);
  __Pyx_GIVEREF(__pyx_t_10);
  PyTuple_SET_ITEM(__pyx_t_17, 9, __pyx_t_10);
  __Pyx_GIVEREF(__pyx_t_11);
  PyTuple_SET_ITEM(__pyx_t_17, 10, __pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_12);
  PyTuple_SET_ITEM(__pyx_t_17, 11, __pyx_t_12);
  __Pyx_GIVEREF(__pyx_t_13);
  PyTuple_SET_ITEM(__pyx_t_17, 12, __pyx_t_13);
  __Pyx_GIVEREF(__pyx_t_14);
  PyTuple_SET_ITEM(__pyx_t_17, 13, __pyx_t_14);
  __Pyx_INCREF(__pyx_v_self->instructions);
  __Pyx_GIVEREF(__pyx_v_self->instructions);
  PyTuple_SET_ITEM(__pyx_t_17, 14, __pyx_v_self->instructions);
  __Pyx_INCREF(((PyObject *)__pyx_v_self->memory));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self->memory));
  PyTuple_SET_ITEM(__pyx_t_17, 15, ((PyObject *)__pyx_v_self->memory));
  __Pyx_GIVEREF(__pyx_t_15);
  PyTuple_SET_ITEM(__pyx_t_17, 16, __pyx_t_15);
  __Pyx_GIVEREF(__pyx_t_16);
  PyTuple_SET_ITEM(__pyx_t_17, 17, __pyx_t_16);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_t_5 = 0;
  __pyx_t_6 = 0;
  __pyx_t_7 = 0;
  __pyx_t_8 = 0;
  __pyx_t_9 = 0;
  __pyx_t_10 = 0;
  __pyx_t_11 = 0;
  __pyx_t_12 = 0;
  __pyx_t_13 = 0;
  __pyx_t_14 = 0;
  __pyx_t_15 = 0;
  __pyx_t_16 = 0;
  __pyx_v_state = ((PyObject*)__pyx_t_17);
  __pyx_t_17 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.A, self.C, self.D, self.I, self.N, self.PC, self.SP, self.V, self.X, self.Y, self.Z, self.aax_sets_flags, self.cycles_since_reset, self.instr_size_bytes, self.instructions, self.memory, self.stack_underflow_causes_exception, self.undocumented_support_level)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_17 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_17);
  __pyx_v__dict = __pyx_t_17;
  __pyx_t_17 = 0;

  /* "(tree fragment)":7
 *     state = (self.A, self.C, self.D, self.I, self.N, self.PC, self.SP, self.V, self.X, self.Y, self.Z, self.aax_sets_flags, self.cycles_since_reset, self.instr_size_bytes, self.instructions, self.memory, self.stack_underflow_causes_exception, self.undocumented_support_level)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_18 = (__pyx_v__dict != Py_None);
  __pyx_t_19 = (__pyx_t_18 != 0);
  if (__pyx_t_19) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_17 = PyTuple_New(1); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_17, 0, __pyx_v__dict);
    __pyx_t_16 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_17); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_DECREF(__pyx_t_17); __pyx_t_17 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_16));
    __pyx_t_16 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.instructions is not None or self.memory is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.A, self.C, self.D, self.I, self.N, self.PC, self.SP, self.V, self.X, self.Y, self.Z, self.aax_sets_flags, self.cycles_since_reset, self.instr_size_bytes, self.instructions, self.memory, self.stack_underflow_causes_exception, self.undocumented_support_level)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.instructions is not None or self.memory is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, None), state
 */
  /*else*/ {
    __pyx_t_18 = (__pyx_v_self->instructions != Py_None);
    __pyx_t_20 = (__pyx_t_18 != 0);
    if (!__pyx_t_20) {
    } else {
      __pyx_t_19 = __pyx_t_20;
      goto __pyx_L4_bool_binop_done;
    }
    __pyx_t_20 = (((PyObject *)__pyx_v_self->memory) != Py_None);
    __pyx_t_18 = (__pyx_t_20 != 0);
    __pyx_t_19 = __pyx_t_18;
    __pyx_L4_bool_binop_done:;
    __pyx_v_use_setstate = __pyx_t_19;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.instructions is not None or self.memory is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, None), state
 *     else:
 */
  __pyx_t_19 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_19) {

    /* "(tree fragment)":13
 *         use_setstate = self.instructions is not None or self.memory is not None
 *     if use_setstate:
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_16, __pyx_n_s_pyx_unpickle_MOS6502); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __pyx_t_17 = PyTuple_New(3); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_17, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_26325635);
    __Pyx_GIVEREF(__pyx_int_26325635);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_int_26325635);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_17, 2, Py_None);
    __pyx_t_15 = PyTuple_New(3); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_16);
    PyTuple_SET_ITEM(__pyx_t_15, 0, __pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_17);
    PyTuple_SET_ITEM(__pyx_t_15, 1, __pyx_t_17);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_15, 2, __pyx_v_state);
    __pyx_t_16 = 0;
    __pyx_t_17 = 0;
    __pyx_r = __pyx_t_15;
    __pyx_t_15 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.instructions is not None or self.memory is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, None), state
 *     else:
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MOS6502__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_15, __pyx_n_s_pyx_unpickle_MOS6502); if (unlikely(!__pyx_t_15)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_15);
    __pyx_t_17 = PyTuple_New(3); if (unlikely(!__pyx_t_17)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_17);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_17, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_26325635);
    __Pyx_GIVEREF(__pyx_int_26325635);
    PyTuple_SET_ITEM(__pyx_t_17, 1, __pyx_int_26325635);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_17, 2, __pyx_v_state);
    __pyx_t_16 = PyTuple_New(2); if (unlikely(!__pyx_t_16)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_16);
    __Pyx_GIVEREF(__pyx_t_15);
    PyTuple_SET_ITEM(__pyx_t_16, 0, __pyx_t_15);
    __Pyx_GIVEREF(__pyx_t_17);
    PyTuple_SET_ITEM(__pyx_t_16, 1, __pyx_t_17);
    __pyx_t_15 = 0;
    __pyx_t_17 = 0;
    __pyx_r = __pyx_t_16;
    __pyx_t_16 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_XDECREF(__pyx_t_12);
  __Pyx_XDECREF(__pyx_t_13);
  __Pyx_XDECREF(__pyx_t_14);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MOS6502__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_59__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_59__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502_7MOS6502_58__setstate_cython__(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502_7MOS6502_58__setstate_cython__(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_MOS6502__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_f_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_MOS6502, (type(self), 0x191b283, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_MOS6502__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("nes.cycore.mos6502.MOS6502.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __pyx_unpickle_MOS6502(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_1__pyx_unpickle_MOS6502(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_3nes_6cycore_7mos6502_1__pyx_unpickle_MOS6502 = {"__pyx_unpickle_MOS6502", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_1__pyx_unpickle_MOS6502, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_3nes_6cycore_7mos6502_1__pyx_unpickle_MOS6502(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_MOS6502 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MOS6502", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MOS6502", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_MOS6502") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_MOS6502", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("nes.cycore.mos6502.__pyx_unpickle_MOS6502", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MOS6502", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x191b283:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0x191b283) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x191b283:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)
 *     __pyx_result = MOS6502.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0x191b283:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = MOS6502.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0x19, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0x191b283:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)
 *     __pyx_result = MOS6502.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_ptype_3nes_6cycore_7mos6502_MOS6502), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)
 *     __pyx_result = MOS6502.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = MOS6502.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_f_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502__set_state(((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0x191b283 = (A, C, D, I, N, PC, SP, V, X, Y, Z, aax_sets_flags, cycles_since_reset, instr_size_bytes, instructions, memory, stack_underflow_causes_exception, undocumented_support_level))" % __pyx_checksum)
 *     __pyx_result = MOS6502.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MOS6502(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("nes.cycore.mos6502.__pyx_unpickle_MOS6502", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]
 *     if len(__pyx_state) > 18 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_f_3nes_6cycore_7mos6502___pyx_unpickle_MOS6502__set_state(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  unsigned char __pyx_t_1;
  int __pyx_t_2;
  PY_LONG_LONG __pyx_t_3;
  int __pyx_t_4[0x100];
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_MOS6502__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 18 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[18])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_char(PyTuple_GET_ITEM(__pyx_v___pyx_state, 0)); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->A = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 1)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->C = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 2)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->D = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 3)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->I = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 4)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->N = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 5)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->PC = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 6)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->SP = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 7)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->V = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_char(PyTuple_GET_ITEM(__pyx_v___pyx_state, 8)); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->X = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_PyInt_As_unsigned_char(PyTuple_GET_ITEM(__pyx_v___pyx_state, 9)); if (unlikely((__pyx_t_1 == (unsigned char)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->Y = __pyx_t_1;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 10)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->Z = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 11)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->aax_sets_flags = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_3 = __Pyx_PyInt_As_PY_LONG_LONG(PyTuple_GET_ITEM(__pyx_v___pyx_state, 12)); if (unlikely((__pyx_t_3 == (PY_LONG_LONG)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->cycles_since_reset = __pyx_t_3;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (unlikely(__Pyx_carray_from_py_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 13), __pyx_t_4, 0x100) < 0)) __PYX_ERR(1, 12, __pyx_L1_error)
  if (unlikely((0x100) != (0x100))) {
    PyErr_Format(PyExc_ValueError, "Assignment to slice of wrong length, expected %" CYTHON_FORMAT_SSIZE_T "d, got %" CYTHON_FORMAT_SSIZE_T "d", (Py_ssize_t)(0x100), (Py_ssize_t)(0x100));
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  memcpy(&(__pyx_v___pyx_result->instr_size_bytes[0]), __pyx_t_4, sizeof(__pyx_v___pyx_result->instr_size_bytes[0]) * (0x100));
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_5 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 14);
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v___pyx_result->instructions);
  __Pyx_DECREF(__pyx_v___pyx_result->instructions);
  __pyx_v___pyx_result->instructions = __pyx_t_5;
  __pyx_t_5 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  if (!(likely(((PyTuple_GET_ITEM(__pyx_v___pyx_state, 15)) == Py_None) || likely(__Pyx_TypeTest(PyTuple_GET_ITEM(__pyx_v___pyx_state, 15), __pyx_ptype_3nes_6cycore_6memory_NESMappedRAM))))) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_t_5 = PyTuple_GET_ITEM(__pyx_v___pyx_state, 15);
  __Pyx_INCREF(__pyx_t_5);
  __Pyx_GIVEREF(__pyx_t_5);
  __Pyx_GOTREF(__pyx_v___pyx_result->memory);
  __Pyx_DECREF(((PyObject *)__pyx_v___pyx_result->memory));
  __pyx_v___pyx_result->memory = ((struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *)__pyx_t_5);
  __pyx_t_5 = 0;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 16)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->stack_underflow_causes_exception = __pyx_t_2;
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_2 = __Pyx_PyInt_As_int(PyTuple_GET_ITEM(__pyx_v___pyx_state, 17)); if (unlikely((__pyx_t_2 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 12, __pyx_L1_error)
  __pyx_v___pyx_result->undocumented_support_level = __pyx_t_2;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]
 *     if len(__pyx_state) > 18 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[18])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_7 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_7 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_8 = ((__pyx_t_7 > 18) != 0);
  if (__pyx_t_8) {
  } else {
    __pyx_t_6 = __pyx_t_8;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_8 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_8 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_9 = (__pyx_t_8 != 0);
  __pyx_t_6 = __pyx_t_9;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_6) {

    /* "(tree fragment)":14
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]
 *     if len(__pyx_state) > 18 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[18])             # <<<<<<<<<<<<<<
 */
    __pyx_t_10 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_11 = __Pyx_PyObject_GetAttrStr(__pyx_t_10, __pyx_n_s_update); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_11);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_10 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_11))) {
      __pyx_t_10 = PyMethod_GET_SELF(__pyx_t_11);
      if (likely(__pyx_t_10)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_11);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_11, function);
      }
    }
    __pyx_t_5 = (__pyx_t_10) ? __Pyx_PyObject_Call2Args(__pyx_t_11, __pyx_t_10, PyTuple_GET_ITEM(__pyx_v___pyx_state, 18)) : __Pyx_PyObject_CallOneArg(__pyx_t_11, PyTuple_GET_ITEM(__pyx_v___pyx_state, 18));
    __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
    if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]
 *     if len(__pyx_state) > 18 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[18])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_MOS6502__set_state(<MOS6502> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_MOS6502__set_state(MOS6502 __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.A = __pyx_state[0]; __pyx_result.C = __pyx_state[1]; __pyx_result.D = __pyx_state[2]; __pyx_result.I = __pyx_state[3]; __pyx_result.N = __pyx_state[4]; __pyx_result.PC = __pyx_state[5]; __pyx_result.SP = __pyx_state[6]; __pyx_result.V = __pyx_state[7]; __pyx_result.X = __pyx_state[8]; __pyx_result.Y = __pyx_state[9]; __pyx_result.Z = __pyx_state[10]; __pyx_result.aax_sets_flags = __pyx_state[11]; __pyx_result.cycles_since_reset = __pyx_state[12]; __pyx_result.instr_size_bytes = __pyx_state[13]; __pyx_result.instructions = __pyx_state[14]; __pyx_result.memory = __pyx_state[15]; __pyx_result.stack_underflow_causes_exception = __pyx_state[16]; __pyx_result.undocumented_support_level = __pyx_state[17]
 *     if len(__pyx_state) > 18 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("nes.cycore.mos6502.__pyx_unpickle_MOS6502__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "carray.to_py":112
 * 
 * @cname("__Pyx_carray_to_py_int")
 * cdef inline list __Pyx_carray_to_py_int(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

static CYTHON_INLINE PyObject *__Pyx_carray_to_py_int(int *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_l = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_to_py_int", 0);

  /* "carray.to_py":115
 *     cdef size_t i
 *     cdef object value
 *     l = PyList_New(length)             # <<<<<<<<<<<<<<
 *     for i in range(<size_t>length):
 *         value = v[i]
 */
  __pyx_t_1 = PyList_New(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_l = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "carray.to_py":116
 *     cdef object value
 *     l = PyList_New(length)
 *     for i in range(<size_t>length):             # <<<<<<<<<<<<<<
 *         value = v[i]
 *         Py_INCREF(value)
 */
  __pyx_t_2 = ((size_t)__pyx_v_length);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "carray.to_py":117
 *     l = PyList_New(length)
 *     for i in range(<size_t>length):
 *         value = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)
 */
    __pyx_t_1 = __Pyx_PyInt_From_int((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "carray.to_py":118
 *     for i in range(<size_t>length):
 *         value = v[i]
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyList_SET_ITEM(l, i, value)
 *     return l
 */
    Py_INCREF(__pyx_v_value);

    /* "carray.to_py":119
 *         value = v[i]
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)             # <<<<<<<<<<<<<<
 *     return l
 * 
 */
    PyList_SET_ITEM(__pyx_v_l, __pyx_v_i, __pyx_v_value);
  }

  /* "carray.to_py":120
 *         Py_INCREF(value)
 *         PyList_SET_ITEM(l, i, value)
 *     return l             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_l);
  __pyx_r = __pyx_v_l;
  goto __pyx_L0;

  /* "carray.to_py":112
 * 
 * @cname("__Pyx_carray_to_py_int")
 * cdef inline list __Pyx_carray_to_py_int(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("carray.to_py.__Pyx_carray_to_py_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_l);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "carray.to_py":124
 * 
 * @cname("__Pyx_carray_to_tuple_int")
 * cdef inline tuple __Pyx_carray_to_tuple_int(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

static CYTHON_INLINE PyObject *__Pyx_carray_to_tuple_int(int *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  size_t __pyx_v_i;
  PyObject *__pyx_v_value = 0;
  PyObject *__pyx_v_t = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  size_t __pyx_t_2;
  size_t __pyx_t_3;
  size_t __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_to_tuple_int", 0);

  /* "carray.to_py":127
 *     cdef size_t i
 *     cdef object value
 *     t = PyTuple_New(length)             # <<<<<<<<<<<<<<
 *     for i in range(<size_t>length):
 *         value = v[i]
 */
  __pyx_t_1 = PyTuple_New(__pyx_v_length); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 127, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_t = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "carray.to_py":128
 *     cdef object value
 *     t = PyTuple_New(length)
 *     for i in range(<size_t>length):             # <<<<<<<<<<<<<<
 *         value = v[i]
 *         Py_INCREF(value)
 */
  __pyx_t_2 = ((size_t)__pyx_v_length);
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "carray.to_py":129
 *     t = PyTuple_New(length)
 *     for i in range(<size_t>length):
 *         value = v[i]             # <<<<<<<<<<<<<<
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)
 */
    __pyx_t_1 = __Pyx_PyInt_From_int((__pyx_v_v[__pyx_v_i])); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 129, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "carray.to_py":130
 *     for i in range(<size_t>length):
 *         value = v[i]
 *         Py_INCREF(value)             # <<<<<<<<<<<<<<
 *         PyTuple_SET_ITEM(t, i, value)
 *     return t
 */
    Py_INCREF(__pyx_v_value);

    /* "carray.to_py":131
 *         value = v[i]
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)             # <<<<<<<<<<<<<<
 *     return t
 */
    PyTuple_SET_ITEM(__pyx_v_t, __pyx_v_i, __pyx_v_value);
  }

  /* "carray.to_py":132
 *         Py_INCREF(value)
 *         PyTuple_SET_ITEM(t, i, value)
 *     return t             # <<<<<<<<<<<<<<
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_t);
  __pyx_r = __pyx_v_t;
  goto __pyx_L0;

  /* "carray.to_py":124
 * 
 * @cname("__Pyx_carray_to_tuple_int")
 * cdef inline tuple __Pyx_carray_to_tuple_int(base_type *v, Py_ssize_t length):             # <<<<<<<<<<<<<<
 *     cdef size_t i
 *     cdef object value
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("carray.to_py.__Pyx_carray_to_tuple_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_int")
 * cdef int __Pyx_carray_from_py_int(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

static int __Pyx_carray_from_py_int(PyObject *__pyx_v_o, int *__pyx_v_v, Py_ssize_t __pyx_v_length) {
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_v_item = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  char const *__pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_carray_from_py_int", 0);

  /* "carray.from_py":78
 * @cname("__Pyx_carray_from_py_int")
 * cdef int __Pyx_carray_from_py_int(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length             # <<<<<<<<<<<<<<
 *     try:
 *         i = len(o)
 */
  __pyx_v_i = __pyx_v_length;

  /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_int(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "carray.from_py":80
 *     cdef Py_ssize_t i = length
 *     try:
 *         i = len(o)             # <<<<<<<<<<<<<<
 *     except (TypeError, OverflowError):
 *         pass
 */
      __pyx_t_4 = PyObject_Length(__pyx_v_o); if (unlikely(__pyx_t_4 == ((Py_ssize_t)-1))) __PYX_ERR(1, 80, __pyx_L3_error)
      __pyx_v_i = __pyx_t_4;

      /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_int(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L8_try_end;
    __pyx_L3_error:;

    /* "carray.from_py":81
 *     try:
 *         i = len(o)
 *     except (TypeError, OverflowError):             # <<<<<<<<<<<<<<
 *         pass
 *     if i == length:
 */
    __pyx_t_5 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError) || __Pyx_PyErr_ExceptionMatches(__pyx_builtin_OverflowError);
    if (__pyx_t_5) {
      __Pyx_ErrRestore(0,0,0);
      goto __pyx_L4_exception_handled;
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "carray.from_py":79
 * cdef int __Pyx_carray_from_py_int(object o, base_type *v, Py_ssize_t length) except -1:
 *     cdef Py_ssize_t i = length
 *     try:             # <<<<<<<<<<<<<<
 *         i = len(o)
 *     except (TypeError, OverflowError):
 */
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L4_exception_handled:;
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    __pyx_L8_try_end:;
  }

  /* "carray.from_py":83
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */
  __pyx_t_6 = ((__pyx_v_i == __pyx_v_length) != 0);
  if (__pyx_t_6) {

    /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    __pyx_t_4 = 0;
    if (likely(PyList_CheckExact(__pyx_v_o)) || PyTuple_CheckExact(__pyx_v_o)) {
      __pyx_t_7 = __pyx_v_o; __Pyx_INCREF(__pyx_t_7); __pyx_t_8 = 0;
      __pyx_t_9 = NULL;
    } else {
      __pyx_t_8 = -1; __pyx_t_7 = PyObject_GetIter(__pyx_v_o); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 84, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_9 = Py_TYPE(__pyx_t_7)->tp_iternext; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 84, __pyx_L1_error)
    }
    for (;;) {
      if (likely(!__pyx_t_9)) {
        if (likely(PyList_CheckExact(__pyx_t_7))) {
          if (__pyx_t_8 >= PyList_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyList_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 84, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 84, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        } else {
          if (__pyx_t_8 >= PyTuple_GET_SIZE(__pyx_t_7)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_10 = PyTuple_GET_ITEM(__pyx_t_7, __pyx_t_8); __Pyx_INCREF(__pyx_t_10); __pyx_t_8++; if (unlikely(0 < 0)) __PYX_ERR(1, 84, __pyx_L1_error)
          #else
          __pyx_t_10 = PySequence_ITEM(__pyx_t_7, __pyx_t_8); __pyx_t_8++; if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 84, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          #endif
        }
      } else {
        __pyx_t_10 = __pyx_t_9(__pyx_t_7);
        if (unlikely(!__pyx_t_10)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(1, 84, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_10);
      }
      __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_10);
      __pyx_t_10 = 0;
      __pyx_v_i = __pyx_t_4;
      __pyx_t_4 = (__pyx_t_4 + 1);

      /* "carray.from_py":85
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */
      __pyx_t_6 = ((__pyx_v_i >= __pyx_v_length) != 0);
      if (__pyx_t_6) {

        /* "carray.from_py":86
 *         for i, item in enumerate(o):
 *             if i >= length:
 *                 break             # <<<<<<<<<<<<<<
 *             v[i] = item
 *         else:
 */
        goto __pyx_L11_break;

        /* "carray.from_py":85
 *     if i == length:
 *         for i, item in enumerate(o):
 *             if i >= length:             # <<<<<<<<<<<<<<
 *                 break
 *             v[i] = item
 */
      }

      /* "carray.from_py":87
 *             if i >= length:
 *                 break
 *             v[i] = item             # <<<<<<<<<<<<<<
 *         else:
 *             i += 1  # convert index to length
 */
      __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_v_item); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 87, __pyx_L1_error)
      (__pyx_v_v[__pyx_v_i]) = __pyx_t_5;

      /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    }
    /*else*/ {

      /* "carray.from_py":89
 *             v[i] = item
 *         else:
 *             i += 1  # convert index to length             # <<<<<<<<<<<<<<
 *             if i == length:
 *                 return 0
 */
      __pyx_v_i = (__pyx_v_i + 1);

      /* "carray.from_py":90
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */
      __pyx_t_6 = ((__pyx_v_i == __pyx_v_length) != 0);
      if (__pyx_t_6) {

        /* "carray.from_py":91
 *             i += 1  # convert index to length
 *             if i == length:
 *                 return 0             # <<<<<<<<<<<<<<
 * 
 *     PyErr_Format(
 */
        __pyx_r = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        goto __pyx_L0;

        /* "carray.from_py":90
 *         else:
 *             i += 1  # convert index to length
 *             if i == length:             # <<<<<<<<<<<<<<
 *                 return 0
 * 
 */
      }
    }

    /* "carray.from_py":84
 *         pass
 *     if i == length:
 *         for i, item in enumerate(o):             # <<<<<<<<<<<<<<
 *             if i >= length:
 *                 break
 */
    __pyx_L11_break:;
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

    /* "carray.from_py":83
 *     except (TypeError, OverflowError):
 *         pass
 *     if i == length:             # <<<<<<<<<<<<<<
 *         for i, item in enumerate(o):
 *             if i >= length:
 */
  }

  /* "carray.from_py":96
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 *          if i >= length else             # <<<<<<<<<<<<<<
 *          "not enough values found during array assignment, expected %zd, got %zd"),
 *         length, i)
 */
  if (((__pyx_v_i >= __pyx_v_length) != 0)) {
    __pyx_t_11 = ((char const *)"too many values found during array assignment, expected %zd");
  } else {
    __pyx_t_11 = ((char const *)"not enough values found during array assignment, expected %zd, got %zd");
  }

  /* "carray.from_py":93
 *                 return 0
 * 
 *     PyErr_Format(             # <<<<<<<<<<<<<<
 *         IndexError,
 *         ("too many values found during array assignment, expected %zd"
 */
  __pyx_t_7 = PyErr_Format(__pyx_builtin_IndexError, __pyx_t_11, __pyx_v_length, __pyx_v_i); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "carray.from_py":77
 * 
 * @cname("__Pyx_carray_from_py_int")
 * cdef int __Pyx_carray_from_py_int(object o, base_type *v, Py_ssize_t length) except -1:             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t i = length
 *     try:
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("carray.from_py.__Pyx_carray_from_py_int", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

/* Python wrapper */
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_array___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shape = 0;
  Py_ssize_t __pyx_v_itemsize;
  PyObject *__pyx_v_format = 0;
  PyObject *__pyx_v_mode = 0;
  int __pyx_v_allocate_buffer;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shape,&__pyx_n_s_itemsize,&__pyx_n_s_format,&__pyx_n_s_mode,&__pyx_n_s_allocate_buffer,0};
    PyObject* values[5] = {0,0,0,0,0};
    values[3] = ((PyObject *)__pyx_n_s_c);
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_shape)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_itemsize)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 1); __PYX_ERR(1, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_format)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, 2); __PYX_ERR(1, 122, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  3:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_mode);
          if (value) { values[3] = value; kw_args--; }
        }
        CYTHON_FALLTHROUGH;
        case  4:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_allocate_buffer);
          if (value) { values[4] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 122, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        CYTHON_FALLTHROUGH;
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        CYTHON_FALLTHROUGH;
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_shape = ((PyObject*)values[0]);
    __pyx_v_itemsize = __Pyx_PyIndex_AsSsize_t(values[1]); if (unlikely((__pyx_v_itemsize == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 122, __pyx_L3_error)
    __pyx_v_format = values[2];
    __pyx_v_mode = values[3];
    if (values[4]) {
      __pyx_v_allocate_buffer = __Pyx_PyObject_IsTrue(values[4]); if (unlikely((__pyx_v_allocate_buffer == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 123, __pyx_L3_error)
    } else {

      /* "View.MemoryView":123
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,
 *                   mode="c", bint allocate_buffer=True):             # <<<<<<<<<<<<<<
 * 
 *         cdef int idx
 */
      __pyx_v_allocate_buffer = ((int)1);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 3, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 122, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  if (unlikely(!__Pyx_ArgTypeTest(((PyObject *)__pyx_v_shape), (&PyTuple_Type), 1, "shape", 1))) __PYX_ERR(1, 122, __pyx_L1_error)
  if (unlikely(((PyObject *)__pyx_v_format) == Py_None)) {
    PyErr_Format(PyExc_TypeError, "Argument '%.200s' must not be None", "format"); __PYX_ERR(1, 122, __pyx_L1_error)
  }
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(((struct __pyx_array_obj *)__pyx_v_self), __pyx_v_shape, __pyx_v_itemsize, __pyx_v_format, __pyx_v_mode, __pyx_v_allocate_buffer);

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array___cinit__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, PyObject *__pyx_v_format, PyObject *__pyx_v_mode, int __pyx_v_allocate_buffer) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_dim;
  PyObject **__pyx_v_p;
  char __pyx_v_order;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  int __pyx_t_8;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);
  __Pyx_INCREF(__pyx_v_format);

  /* "View.MemoryView":129
 *         cdef PyObject **p
 * 
 *         self.ndim = <int> len(shape)             # <<<<<<<<<<<<<<
 *         self.itemsize = itemsize
 * 
 */
  if (unlikely(__pyx_v_shape == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 129, __pyx_L1_error)
  }
  __pyx_t_1 = PyTuple_GET_SIZE(__pyx_v_shape); if (unlikely(__pyx_t_1 == ((Py_ssize_t)-1))) __PYX_ERR(1, 129, __pyx_L1_error)
  __pyx_v_self->ndim = ((int)__pyx_t_1);

  /* "View.MemoryView":130
 * 
 *         self.ndim = <int> len(shape)
 *         self.itemsize = itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not self.ndim:
 */
  __pyx_v_self->itemsize = __pyx_v_itemsize;

  /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_self->ndim != 0)) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__12, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 133, __pyx_L1_error)

    /* "View.MemoryView":132
 *         self.itemsize = itemsize
 * 
 *         if not self.ndim:             # <<<<<<<<<<<<<<
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 */
  }

  /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  __pyx_t_2 = ((__pyx_v_itemsize <= 0) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 136, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 136, __pyx_L1_error)

    /* "View.MemoryView":135
 *             raise ValueError("Empty shape tuple for cython.array")
 * 
 *         if itemsize <= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 */
  }

  /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  __pyx_t_2 = PyBytes_Check(__pyx_v_format); 
  __pyx_t_4 = ((!(__pyx_t_2 != 0)) != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":139
 * 
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')             # <<<<<<<<<<<<<<
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format
 */
    __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_format, __pyx_n_s_encode); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_5, function);
      }
    }
    __pyx_t_3 = (__pyx_t_6) ? __Pyx_PyObject_Call2Args(__pyx_t_5, __pyx_t_6, __pyx_n_s_ASCII) : __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_n_s_ASCII);
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF_SET(__pyx_v_format, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":138
 *             raise ValueError("itemsize <= 0 for cython.array")
 * 
 *         if not isinstance(format, bytes):             # <<<<<<<<<<<<<<
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 */
  }

  /* "View.MemoryView":140
 *         if not isinstance(format, bytes):
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string             # <<<<<<<<<<<<<<
 *         self.format = self._format
 * 
 */
  if (!(likely(PyBytes_CheckExact(__pyx_v_format))||((__pyx_v_format) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_v_format)->tp_name), 0))) __PYX_ERR(1, 140, __pyx_L1_error)
  __pyx_t_3 = __pyx_v_format;
  __Pyx_INCREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->_format);
  __Pyx_DECREF(__pyx_v_self->_format);
  __pyx_v_self->_format = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":141
 *             format = format.encode('ASCII')
 *         self._format = format  # keep a reference to the byte string
 *         self.format = self._format             # <<<<<<<<<<<<<<
 * 
 * 
 */
  if (unlikely(__pyx_v_self->_format == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "expected bytes, NoneType found");
    __PYX_ERR(1, 141, __pyx_L1_error)
  }
  __pyx_t_7 = __Pyx_PyBytes_AsWritableString(__pyx_v_self->_format); if (unlikely((!__pyx_t_7) && PyErr_Occurred())) __PYX_ERR(1, 141, __pyx_L1_error)
  __pyx_v_self->format = __pyx_t_7;

  /* "View.MemoryView":144
 * 
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)             # <<<<<<<<<<<<<<
 *         self._strides = self._shape + self.ndim
 * 
 */
  __pyx_v_self->_shape = ((Py_ssize_t *)PyObject_Malloc((((sizeof(Py_ssize_t)) * __pyx_v_self->ndim) * 2)));

  /* "View.MemoryView":145
 * 
 *         self._shape = <Py_ssize_t *> PyObject_Malloc(sizeof(Py_ssize_t)*self.ndim*2)
 *         self._strides = self._shape + self.ndim             # <<<<<<<<<<<<<<
 * 
 *         if not self._shape:
 */
  __pyx_v_self->_strides = (__pyx_v_self->_shape + __pyx_v_self->ndim);

  /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  __pyx_t_4 = ((!(__pyx_v_self->_shape != 0)) != 0);
  if (unlikely(__pyx_t_4)) {

    /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 148, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 148, __pyx_L1_error)

    /* "View.MemoryView":147
 *         self._strides = self._shape + self.ndim
 * 
 *         if not self._shape:             # <<<<<<<<<<<<<<
 *             raise MemoryError("unable to allocate shape and strides.")
 * 
 */
  }

  /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  __pyx_t_8 = 0;
  __pyx_t_3 = __pyx_v_shape; __Pyx_INCREF(__pyx_t_3); __pyx_t_1 = 0;
  for (;;) {
    if (__pyx_t_1 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_1); __Pyx_INCREF(__pyx_t_5); __pyx_t_1++; if (unlikely(0 < 0)) __PYX_ERR(1, 151, __pyx_L1_error)
    #else
    __pyx_t_5 = PySequence_ITEM(__pyx_t_3, __pyx_t_1); __pyx_t_1++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_5); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 151, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_9;
    __pyx_v_idx = __pyx_t_8;
    __pyx_t_8 = (__pyx_t_8 + 1);

    /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    __pyx_t_4 = ((__pyx_v_dim <= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":153
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))             # <<<<<<<<<<<<<<
 *             self._shape[idx] = dim
 * 
 */
      __pyx_t_5 = __Pyx_PyInt_From_int(__pyx_v_idx); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_10 = PyTuple_New(2); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_GIVEREF(__pyx_t_5);
      PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_10, 1, __pyx_t_6);
      __pyx_t_5 = 0;
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyString_Format(__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_t_10); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_6); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 153, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 153, __pyx_L1_error)

      /* "View.MemoryView":152
 * 
 *         for idx, dim in enumerate(shape):
 *             if dim <= 0:             # <<<<<<<<<<<<<<
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim
 */
    }

    /* "View.MemoryView":154
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 *             self._shape[idx] = dim             # <<<<<<<<<<<<<<
 * 
 *         cdef char order
 */
    (__pyx_v_self->_shape[__pyx_v_idx]) = __pyx_v_dim;

    /* "View.MemoryView":151
 * 
 * 
 *         for idx, dim in enumerate(shape):             # <<<<<<<<<<<<<<
 *             if dim <= 0:
 *                 raise ValueError("Invalid shape in axis %d: %d." % (idx, dim))
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_fortran, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 157, __pyx_L1_error)
  if (__pyx_t_4) {

    /* "View.MemoryView":158
 *         cdef char order
 *         if mode == 'fortran':
 *             order = b'F'             # <<<<<<<<<<<<<<
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 */
    __pyx_v_order = 'F';

    /* "View.MemoryView":159
 *         if mode == 'fortran':
 *             order = b'F'
 *             self.mode = u'fortran'             # <<<<<<<<<<<<<<
 *         elif mode == 'c':
 *             order = b'C'
 */
    __Pyx_INCREF(__pyx_n_u_fortran);
    __Pyx_GIVEREF(__pyx_n_u_fortran);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_fortran;

    /* "View.MemoryView":157
 * 
 *         cdef char order
 *         if mode == 'fortran':             # <<<<<<<<<<<<<<
 *             order = b'F'
 *             self.mode = u'fortran'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
  __pyx_t_4 = (__Pyx_PyString_Equals(__pyx_v_mode, __pyx_n_s_c, Py_EQ)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(1, 160, __pyx_L1_error)
  if (likely(__pyx_t_4)) {

    /* "View.MemoryView":161
 *             self.mode = u'fortran'
 *         elif mode == 'c':
 *             order = b'C'             # <<<<<<<<<<<<<<
 *             self.mode = u'c'
 *         else:
 */
    __pyx_v_order = 'C';

    /* "View.MemoryView":162
 *         elif mode == 'c':
 *             order = b'C'
 *             self.mode = u'c'             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 */
    __Pyx_INCREF(__pyx_n_u_c);
    __Pyx_GIVEREF(__pyx_n_u_c);
    __Pyx_GOTREF(__pyx_v_self->mode);
    __Pyx_DECREF(__pyx_v_self->mode);
    __pyx_v_self->mode = __pyx_n_u_c;

    /* "View.MemoryView":160
 *             order = b'F'
 *             self.mode = u'fortran'
 *         elif mode == 'c':             # <<<<<<<<<<<<<<
 *             order = b'C'
 *             self.mode = u'c'
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":164
 *             self.mode = u'c'
 *         else:
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)             # <<<<<<<<<<<<<<
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,
 */
  /*else*/ {
    __pyx_t_3 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_v_mode); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_10 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 164, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_Raise(__pyx_t_10, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __PYX_ERR(1, 164, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "View.MemoryView":166
 *             raise ValueError("Invalid mode, expected 'c' or 'fortran', got %s" % mode)
 * 
 *         self.len = fill_contig_strides_array(self._shape, self._strides,             # <<<<<<<<<<<<<<
 *                                              itemsize, self.ndim, order)
 * 
 */
  __pyx_v_self->len = __pyx_fill_contig_strides_array(__pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_itemsize, __pyx_v_self->ndim, __pyx_v_order);

  /* "View.MemoryView":169
 *                                              itemsize, self.ndim, order)
 * 
 *         self.free_data = allocate_buffer             # <<<<<<<<<<<<<<
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:
 */
  __pyx_v_self->free_data = __pyx_v_allocate_buffer;

  /* "View.MemoryView":170
 * 
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'             # <<<<<<<<<<<<<<
 *         if allocate_buffer:
 * 
 */
  __pyx_t_10 = PyObject_RichCompare(__pyx_v_format, __pyx_n_b_O, Py_EQ); __Pyx_XGOTREF(__pyx_t_10); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 170, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_10); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 170, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
  __pyx_v_self->dtype_is_object = __pyx_t_4;

  /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_4 = (__pyx_v_allocate_buffer != 0);
  if (__pyx_t_4) {

    /* "View.MemoryView":174
 * 
 * 
 *             self.data = <char *>malloc(self.len)             # <<<<<<<<<<<<<<
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")
 */
    __pyx_v_self->data = ((char *)malloc(__pyx_v_self->len));

    /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    __pyx_t_4 = ((!(__pyx_v_self->data != 0)) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
      __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_MemoryError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(1, 176, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_Raise(__pyx_t_10, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __PYX_ERR(1, 176, __pyx_L1_error)

      /* "View.MemoryView":175
 * 
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:             # <<<<<<<<<<<<<<
 *                 raise MemoryError("unable to allocate array data.")
 * 
 */
    }

    /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    __pyx_t_4 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":179
 * 
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data             # <<<<<<<<<<<<<<
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 */
      __pyx_v_p = ((PyObject **)__pyx_v_self->data);

      /* "View.MemoryView":180
 *             if self.dtype_is_object:
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):             # <<<<<<<<<<<<<<
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)
 */
      if (unlikely(__pyx_v_itemsize == 0)) {
        PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
        __PYX_ERR(1, 180, __pyx_L1_error)
      }
      else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_self->len))) {
        PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
        __PYX_ERR(1, 180, __pyx_L1_error)
      }
      __pyx_t_1 = (__pyx_v_self->len / __pyx_v_itemsize);
      __pyx_t_9 = __pyx_t_1;
      for (__pyx_t_11 = 0; __pyx_t_11 < __pyx_t_9; __pyx_t_11+=1) {
        __pyx_v_i = __pyx_t_11;

        /* "View.MemoryView":181
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None             # <<<<<<<<<<<<<<
 *                     Py_INCREF(Py_None)
 * 
 */
        (__pyx_v_p[__pyx_v_i]) = Py_None;

        /* "View.MemoryView":182
 *                 for i in range(self.len / itemsize):
 *                     p[i] = Py_None
 *                     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
        Py_INCREF(Py_None);
      }

      /* "View.MemoryView":178
 *                 raise MemoryError("unable to allocate array data.")
 * 
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 p = <PyObject **> self.data
 *                 for i in range(self.len / itemsize):
 */
    }

    /* "View.MemoryView":171
 *         self.free_data = allocate_buffer
 *         self.dtype_is_object = format == b'O'
 *         if allocate_buffer:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":122
 *         cdef bint dtype_is_object
 * 
 *     def __cinit__(array self, tuple shape, Py_ssize_t itemsize, format not None,             # <<<<<<<<<<<<<<
 *                   mode="c", bint allocate_buffer=True):
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.array.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_format);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_array_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(((struct __pyx_array_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_2__getbuffer__(struct __pyx_array_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_bufmode;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  char *__pyx_t_4;
  Py_ssize_t __pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":186
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1             # <<<<<<<<<<<<<<
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
  __pyx_v_bufmode = -1;

  /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
  __pyx_t_1 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_c, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 187, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":188
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 */
    __pyx_v_bufmode = (PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":187
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         cdef int bufmode = -1
 *         if self.mode == u"c":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  __pyx_t_2 = (__Pyx_PyUnicode_Equals(__pyx_v_self->mode, __pyx_n_u_fortran, Py_EQ)); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 189, __pyx_L1_error)
  __pyx_t_1 = (__pyx_t_2 != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":190
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS             # <<<<<<<<<<<<<<
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 */
    __pyx_v_bufmode = (PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS);

    /* "View.MemoryView":189
 *         if self.mode == u"c":
 *             bufmode = PyBUF_C_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         elif self.mode == u"fortran":             # <<<<<<<<<<<<<<
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  __pyx_t_1 = ((!((__pyx_v_flags & __pyx_v_bufmode) != 0)) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 192, __pyx_L1_error)

    /* "View.MemoryView":191
 *         elif self.mode == u"fortran":
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):             # <<<<<<<<<<<<<<
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 */
  }

  /* "View.MemoryView":193
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data             # <<<<<<<<<<<<<<
 *         info.len = self.len
 *         info.ndim = self.ndim
 */
  __pyx_t_4 = __pyx_v_self->data;
  __pyx_v_info->buf = __pyx_t_4;

  /* "View.MemoryView":194
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")
 *         info.buf = self.data
 *         info.len = self.len             # <<<<<<<<<<<<<<
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 */
  __pyx_t_5 = __pyx_v_self->len;
  __pyx_v_info->len = __pyx_t_5;

  /* "View.MemoryView":195
 *         info.buf = self.data
 *         info.len = self.len
 *         info.ndim = self.ndim             # <<<<<<<<<<<<<<
 *         info.shape = self._shape
 *         info.strides = self._strides
 */
  __pyx_t_6 = __pyx_v_self->ndim;
  __pyx_v_info->ndim = __pyx_t_6;

  /* "View.MemoryView":196
 *         info.len = self.len
 *         info.ndim = self.ndim
 *         info.shape = self._shape             # <<<<<<<<<<<<<<
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 */
  __pyx_t_7 = __pyx_v_self->_shape;
  __pyx_v_info->shape = __pyx_t_7;

  /* "View.MemoryView":197
 *         info.ndim = self.ndim
 *         info.shape = self._shape
 *         info.strides = self._strides             # <<<<<<<<<<<<<<
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 */
  __pyx_t_7 = __pyx_v_self->_strides;
  __pyx_v_info->strides = __pyx_t_7;

  /* "View.MemoryView":198
 *         info.shape = self._shape
 *         info.strides = self._strides
 *         info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *         info.itemsize = self.itemsize
 *         info.readonly = 0
 */
  __pyx_v_info->suboffsets = NULL;

  /* "View.MemoryView":199
 *         info.strides = self._strides
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize             # <<<<<<<<<<<<<<
 *         info.readonly = 0
 * 
 */
  __pyx_t_5 = __pyx_v_self->itemsize;
  __pyx_v_info->itemsize = __pyx_t_5;

  /* "View.MemoryView":200
 *         info.suboffsets = NULL
 *         info.itemsize = self.itemsize
 *         info.readonly = 0             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  __pyx_v_info->readonly = 0;

  /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":203
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_4 = __pyx_v_self->format;
    __pyx_v_info->format = __pyx_t_4;

    /* "View.MemoryView":202
 *         info.readonly = 0
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.format
 *         else:
 */
    goto __pyx_L5;
  }

  /* "View.MemoryView":205
 *             info.format = self.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.obj = self
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L5:;

  /* "View.MemoryView":207
 *             info.format = NULL
 * 
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":185
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         cdef int bufmode = -1
 *         if self.mode == u"c":
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

/* Python wrapper */
static void __pyx_array___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_array___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_array___pyx_pf_15View_dot_MemoryView_5array_4__dealloc__(struct __pyx_array_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
  __pyx_t_1 = ((__pyx_v_self->callback_free_data != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":213
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)             # <<<<<<<<<<<<<<
 *         elif self.free_data:
 *             if self.dtype_is_object:
 */
    __pyx_v_self->callback_free_data(__pyx_v_self->data);

    /* "View.MemoryView":212
 * 
 *     def __dealloc__(array self):
 *         if self.callback_free_data != NULL:             # <<<<<<<<<<<<<<
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  __pyx_t_1 = (__pyx_v_self->free_data != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    __pyx_t_1 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":216
 *         elif self.free_data:
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,             # <<<<<<<<<<<<<<
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_self->data, __pyx_v_self->_shape, __pyx_v_self->_strides, __pyx_v_self->ndim, 0);

      /* "View.MemoryView":215
 *             self.callback_free_data(self.data)
 *         elif self.free_data:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 */
    }

    /* "View.MemoryView":218
 *                 refcount_objects_in_slice(self.data, self._shape,
 *                                           self._strides, self.ndim, False)
 *             free(self.data)             # <<<<<<<<<<<<<<
 *         PyObject_Free(self._shape)
 * 
 */
    free(__pyx_v_self->data);

    /* "View.MemoryView":214
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 *         elif self.free_data:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 refcount_objects_in_slice(self.data, self._shape,
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":219
 *                                           self._strides, self.ndim, False)
 *             free(self.data)
 *         PyObject_Free(self._shape)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  PyObject_Free(__pyx_v_self->_shape);

  /* "View.MemoryView":211
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")
 * 
 *     def __dealloc__(array self):             # <<<<<<<<<<<<<<
 *         if self.callback_free_data != NULL:
 *             self.callback_free_data(self.data)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_5array_7memview___get__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_5array_7memview___get__(struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":223
 *     @property
 *     def memview(self):
 *         return self.get_memview()             # <<<<<<<<<<<<<<
 * 
 *     @cname('get_memview')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = ((struct __pyx_vtabstruct_array *)__pyx_v_self->__pyx_vtab)->get_memview(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":222
 * 
 *     @property
 *     def memview(self):             # <<<<<<<<<<<<<<
 *         return self.get_memview()
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.memview.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

static PyObject *__pyx_array_get_memview(struct __pyx_array_obj *__pyx_v_self) {
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_memview", 0);

  /* "View.MemoryView":227
 *     @cname('get_memview')
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE             # <<<<<<<<<<<<<<
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 */
  __pyx_v_flags = ((PyBUF_ANY_CONTIGUOUS | PyBUF_FORMAT) | PyBUF_WRITABLE);

  /* "View.MemoryView":228
 *     cdef get_memview(self):
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  PyTuple_SET_ITEM(__pyx_t_3, 0, ((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":226
 * 
 *     @cname('get_memview')
 *     cdef get_memview(self):             # <<<<<<<<<<<<<<
 *         flags =  PyBUF_ANY_CONTIGUOUS|PyBUF_FORMAT|PyBUF_WRITABLE
 *         return  memoryview(self, flags, self.dtype_is_object)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.array.get_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

/* Python wrapper */
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_array___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_array___pyx_pf_15View_dot_MemoryView_5array_6__len__(struct __pyx_array_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":231
 * 
 *     def __len__(self):
 *         return self._shape[0]             # <<<<<<<<<<<<<<
 * 
 *     def __getattr__(self, attr):
 */
  __pyx_r = (__pyx_v_self->_shape[0]);
  goto __pyx_L0;

  /* "View.MemoryView":230
 *         return  memoryview(self, flags, self.dtype_is_object)
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         return self._shape[0]
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr); /*proto*/
static PyObject *__pyx_array___getattr__(PyObject *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getattr__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_attr));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_8__getattr__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_attr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getattr__", 0);

  /* "View.MemoryView":234
 * 
 *     def __getattr__(self, attr):
 *         return getattr(self.memview, attr)             # <<<<<<<<<<<<<<
 * 
 *     def __getitem__(self, item):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetAttr(__pyx_t_1, __pyx_v_attr); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 234, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":233
 *         return self._shape[0]
 * 
 *     def __getattr__(self, attr):             # <<<<<<<<<<<<<<
 *         return getattr(self.memview, attr)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getattr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

/* Python wrapper */
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item); /*proto*/
static PyObject *__pyx_array___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_array___pyx_pf_15View_dot_MemoryView_5array_10__getitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":237
 * 
 *     def __getitem__(self, item):
 *         return self.memview[item]             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(self, item, value):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetItem(__pyx_t_1, __pyx_v_item); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":236
 *         return getattr(self.memview, attr)
 * 
 *     def __getitem__(self, item):             # <<<<<<<<<<<<<<
 *         return self.memview[item]
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.array.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

/* Python wrapper */
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value); /*proto*/
static int __pyx_array___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v_item), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_array___pyx_pf_15View_dot_MemoryView_5array_12__setitem__(struct __pyx_array_obj *__pyx_v_self, PyObject *__pyx_v_item, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);

  /* "View.MemoryView":240
 * 
 *     def __setitem__(self, item, value):
 *         self.memview[item] = value             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_memview); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (unlikely(PyObject_SetItem(__pyx_t_1, __pyx_v_item, __pyx_v_value) < 0)) __PYX_ERR(1, 240, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "View.MemoryView":239
 *         return self.memview[item]
 * 
 *     def __setitem__(self, item, value):             # <<<<<<<<<<<<<<
 *         self.memview[item] = value
 * 
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_array_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array___reduce_cython__(((struct __pyx_array_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array___reduce_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_array_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_array_2__setstate_cython__(((struct __pyx_array_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_array_2__setstate_cython__(CYTHON_UNUSED struct __pyx_array_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.array.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

static struct __pyx_array_obj *__pyx_array_new(PyObject *__pyx_v_shape, Py_ssize_t __pyx_v_itemsize, char *__pyx_v_format, char *__pyx_v_mode, char *__pyx_v_buf) {
  struct __pyx_array_obj *__pyx_v_result = 0;
  struct __pyx_array_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("array_cwrapper", 0);

  /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_buf == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":249
 * 
 *     if buf == NULL:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))             # <<<<<<<<<<<<<<
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(4); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 3, __pyx_t_4);
    __pyx_t_2 = 0;
    __pyx_t_3 = 0;
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_5, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 249, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":248
 *     cdef array result
 * 
 *     if buf == NULL:             # <<<<<<<<<<<<<<
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
  /*else*/ {
    __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_itemsize); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = __Pyx_PyBytes_FromString(__pyx_v_format); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_mode, 0, strlen(__pyx_v_mode), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(4); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_2, 2, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 3, __pyx_t_3);
    __pyx_t_4 = 0;
    __pyx_t_5 = 0;
    __pyx_t_3 = 0;

    /* "View.MemoryView":252
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)             # <<<<<<<<<<<<<<
 *         result.data = buf
 * 
 */
    __pyx_t_3 = __Pyx_PyDict_NewPresized(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 252, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (PyDict_SetItem(__pyx_t_3, __pyx_n_s_allocate_buffer, Py_False) < 0) __PYX_ERR(1, 252, __pyx_L1_error)

    /* "View.MemoryView":251
 *         result = array(shape, itemsize, format, mode.decode('ASCII'))
 *     else:
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),             # <<<<<<<<<<<<<<
 *                        allocate_buffer=False)
 *         result.data = buf
 */
    __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_array_type), __pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 251, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result = ((struct __pyx_array_obj *)__pyx_t_5);
    __pyx_t_5 = 0;

    /* "View.MemoryView":253
 *         result = array(shape, itemsize, format, mode.decode('ASCII'),
 *                        allocate_buffer=False)
 *         result.data = buf             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
    __pyx_v_result->data = __pyx_v_buf;
  }
  __pyx_L3:;

  /* "View.MemoryView":255
 *         result.data = buf
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(((PyObject *)__pyx_r));
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":244
 * 
 * @cname("__pyx_array_new")
 * cdef array array_cwrapper(tuple shape, Py_ssize_t itemsize, char *format,             # <<<<<<<<<<<<<<
 *                           char *mode, char *buf):
 *     cdef array result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.array_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

/* Python wrapper */
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_MemviewEnum___init__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_name = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_name,0};
    PyObject* values[1] = {0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_name)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__init__") < 0)) __PYX_ERR(1, 281, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 1) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
    }
    __pyx_v_name = values[0];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__init__", 1, 1, 1, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 281, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.Enum.__init__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), __pyx_v_name);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum___init__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v_name) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__init__", 0);

  /* "View.MemoryView":282
 *     cdef object name
 *     def __init__(self, name):
 *         self.name = name             # <<<<<<<<<<<<<<
 *     def __repr__(self):
 *         return self.name
 */
  __Pyx_INCREF(__pyx_v_name);
  __Pyx_GIVEREF(__pyx_v_name);
  __Pyx_GOTREF(__pyx_v_self->name);
  __Pyx_DECREF(__pyx_v_self->name);
  __pyx_v_self->name = __pyx_v_name;

  /* "View.MemoryView":281
 * cdef class Enum(object):
 *     cdef object name
 *     def __init__(self, name):             # <<<<<<<<<<<<<<
 *         self.name = name
 *     def __repr__(self):
 */

  /* function exit code */
  __pyx_r = 0;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

/* Python wrapper */
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_MemviewEnum___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_MemviewEnum___pyx_pf_15View_dot_MemoryView_4Enum_2__repr__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":284
 *         self.name = name
 *     def __repr__(self):
 *         return self.name             # <<<<<<<<<<<<<<
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->name);
  __pyx_r = __pyx_v_self->name;
  goto __pyx_L0;

  /* "View.MemoryView":283
 *     def __init__(self, name):
 *         self.name = name
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return self.name
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum___reduce_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum___reduce_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self) {
  PyObject *__pyx_v_state = 0;
  PyObject *__pyx_v__dict = 0;
  int __pyx_v_use_setstate;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":5
 *     cdef object _dict
 *     cdef bint use_setstate
 *     state = (self.name,)             # <<<<<<<<<<<<<<
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 */
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_v_self->name);
  __Pyx_GIVEREF(__pyx_v_self->name);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v_self->name);
  __pyx_v_state = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "(tree fragment)":6
 *     cdef bint use_setstate
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)             # <<<<<<<<<<<<<<
 *     if _dict is not None:
 *         state += (_dict,)
 */
  __pyx_t_1 = __Pyx_GetAttr3(((PyObject *)__pyx_v_self), __pyx_n_s_dict, Py_None); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 6, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v__dict = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
  __pyx_t_2 = (__pyx_v__dict != Py_None);
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":8
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:
 *         state += (_dict,)             # <<<<<<<<<<<<<<
 *         use_setstate = True
 *     else:
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_v__dict);
    __Pyx_GIVEREF(__pyx_v__dict);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_v__dict);
    __pyx_t_4 = PyNumber_InPlaceAdd(__pyx_v_state, __pyx_t_1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 8, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_state, ((PyObject*)__pyx_t_4));
    __pyx_t_4 = 0;

    /* "(tree fragment)":9
 *     if _dict is not None:
 *         state += (_dict,)
 *         use_setstate = True             # <<<<<<<<<<<<<<
 *     else:
 *         use_setstate = self.name is not None
 */
    __pyx_v_use_setstate = 1;

    /* "(tree fragment)":7
 *     state = (self.name,)
 *     _dict = getattr(self, '__dict__', None)
 *     if _dict is not None:             # <<<<<<<<<<<<<<
 *         state += (_dict,)
 *         use_setstate = True
 */
    goto __pyx_L3;
  }

  /* "(tree fragment)":11
 *         use_setstate = True
 *     else:
 *         use_setstate = self.name is not None             # <<<<<<<<<<<<<<
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_self->name != Py_None);
    __pyx_v_use_setstate = __pyx_t_3;
  }
  __pyx_L3:;

  /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  __pyx_t_3 = (__pyx_v_use_setstate != 0);
  if (__pyx_t_3) {

    /* "(tree fragment)":13
 *         use_setstate = self.name is not None
 *     if use_setstate:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state             # <<<<<<<<<<<<<<
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_4, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    PyTuple_SET_ITEM(__pyx_t_1, 2, Py_None);
    __pyx_t_5 = PyTuple_New(3); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 13, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_5, 1, __pyx_t_1);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_5, 2, __pyx_v_state);
    __pyx_t_4 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_5;
    __pyx_t_5 = 0;
    goto __pyx_L0;

    /* "(tree fragment)":12
 *     else:
 *         use_setstate = self.name is not None
 *     if use_setstate:             # <<<<<<<<<<<<<<
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 */
  }

  /* "(tree fragment)":15
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, None), state
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_GetModuleGlobalName(__pyx_t_5, __pyx_n_s_pyx_unpickle_Enum); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_1 = PyTuple_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_GIVEREF(((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    PyTuple_SET_ITEM(__pyx_t_1, 0, ((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))));
    __Pyx_INCREF(__pyx_int_184977713);
    __Pyx_GIVEREF(__pyx_int_184977713);
    PyTuple_SET_ITEM(__pyx_t_1, 1, __pyx_int_184977713);
    __Pyx_INCREF(__pyx_v_state);
    __Pyx_GIVEREF(__pyx_v_state);
    PyTuple_SET_ITEM(__pyx_t_1, 2, __pyx_v_state);
    __pyx_t_4 = PyTuple_New(2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 15, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_5);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_1);
    __pyx_t_5 = 0;
    __pyx_t_1 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;
  }

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     cdef tuple state
 *     cdef object _dict
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.Enum.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_state);
  __Pyx_XDECREF(__pyx_v__dict);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_MemviewEnum_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_MemviewEnum_2__setstate_cython__(((struct __pyx_MemviewEnum_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_MemviewEnum_2__setstate_cython__(struct __pyx_MemviewEnum_obj *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":17
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)             # <<<<<<<<<<<<<<
 */
  if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 17, __pyx_L1_error)
  __pyx_t_1 = __pyx_unpickle_Enum__set_state(__pyx_v_self, ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "(tree fragment)":16
 *     else:
 *         return __pyx_unpickle_Enum, (type(self), 0xb068931, state)
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_unpickle_Enum__set_state(self, __pyx_state)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.Enum.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

static void *__pyx_align_pointer(void *__pyx_v_memory, size_t __pyx_v_alignment) {
  Py_intptr_t __pyx_v_aligned_p;
  size_t __pyx_v_offset;
  void *__pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":300
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory             # <<<<<<<<<<<<<<
 *     cdef size_t offset
 * 
 */
  __pyx_v_aligned_p = ((Py_intptr_t)__pyx_v_memory);

  /* "View.MemoryView":304
 * 
 *     with cython.cdivision(True):
 *         offset = aligned_p % alignment             # <<<<<<<<<<<<<<
 * 
 *     if offset > 0:
 */
  __pyx_v_offset = (__pyx_v_aligned_p % __pyx_v_alignment);

  /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  __pyx_t_1 = ((__pyx_v_offset > 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":307
 * 
 *     if offset > 0:
 *         aligned_p += alignment - offset             # <<<<<<<<<<<<<<
 * 
 *     return <void *> aligned_p
 */
    __pyx_v_aligned_p = (__pyx_v_aligned_p + (__pyx_v_alignment - __pyx_v_offset));

    /* "View.MemoryView":306
 *         offset = aligned_p % alignment
 * 
 *     if offset > 0:             # <<<<<<<<<<<<<<
 *         aligned_p += alignment - offset
 * 
 */
  }

  /* "View.MemoryView":309
 *         aligned_p += alignment - offset
 * 
 *     return <void *> aligned_p             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = ((void *)__pyx_v_aligned_p);
  goto __pyx_L0;

  /* "View.MemoryView":298
 * 
 * @cname('__pyx_align_pointer')
 * cdef void *align_pointer(void *memory, size_t alignment) nogil:             # <<<<<<<<<<<<<<
 *     "Align pointer memory on a given boundary"
 *     cdef Py_intptr_t aligned_p = <Py_intptr_t> memory
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

/* Python wrapper */
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static int __pyx_memoryview___cinit__(PyObject *__pyx_v_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_obj = 0;
  int __pyx_v_flags;
  int __pyx_v_dtype_is_object;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__cinit__ (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_obj,&__pyx_n_s_flags,&__pyx_n_s_dtype_is_object,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_obj)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_flags)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, 1); __PYX_ERR(1, 345, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (kw_args > 0) {
          PyObject* value = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_dtype_is_object);
          if (value) { values[2] = value; kw_args--; }
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__cinit__") < 0)) __PYX_ERR(1, 345, __pyx_L3_error)
      }
    } else {
      switch (PyTuple_GET_SIZE(__pyx_args)) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        break;
        default: goto __pyx_L5_argtuple_error;
      }
    }
    __pyx_v_obj = values[0];
    __pyx_v_flags = __Pyx_PyInt_As_int(values[1]); if (unlikely((__pyx_v_flags == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 345, __pyx_L3_error)
    if (values[2]) {
      __pyx_v_dtype_is_object = __Pyx_PyObject_IsTrue(values[2]); if (unlikely((__pyx_v_dtype_is_object == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 345, __pyx_L3_error)
    } else {
      __pyx_v_dtype_is_object = ((int)0);
    }
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__cinit__", 0, 2, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return -1;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_obj, __pyx_v_flags, __pyx_v_dtype_is_object);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview___cinit__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj, int __pyx_v_flags, int __pyx_v_dtype_is_object) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__cinit__", 0);

  /* "View.MemoryView":346
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj             # <<<<<<<<<<<<<<
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 */
  __Pyx_INCREF(__pyx_v_obj);
  __Pyx_GIVEREF(__pyx_v_obj);
  __Pyx_GOTREF(__pyx_v_self->obj);
  __Pyx_DECREF(__pyx_v_self->obj);
  __pyx_v_self->obj = __pyx_v_obj;

  /* "View.MemoryView":347
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):
 *         self.obj = obj
 *         self.flags = flags             # <<<<<<<<<<<<<<
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 */
  __pyx_v_self->flags = __pyx_v_flags;

  /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  __pyx_t_2 = (((PyObject *)Py_TYPE(((PyObject *)__pyx_v_self))) == ((PyObject *)__pyx_memoryview_type));
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (!__pyx_t_3) {
  } else {
    __pyx_t_1 = __pyx_t_3;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_3 = (__pyx_v_obj != Py_None);
  __pyx_t_2 = (__pyx_t_3 != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_1) {

    /* "View.MemoryView":349
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)             # <<<<<<<<<<<<<<
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 */
    __pyx_t_4 = __Pyx_GetBuffer(__pyx_v_obj, (&__pyx_v_self->view), __pyx_v_flags); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 349, __pyx_L1_error)

    /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    __pyx_t_1 = ((((PyObject *)__pyx_v_self->view.obj) == NULL) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":351
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None             # <<<<<<<<<<<<<<
 *                 Py_INCREF(Py_None)
 * 
 */
      ((Py_buffer *)(&__pyx_v_self->view))->obj = Py_None;

      /* "View.MemoryView":352
 *             if <PyObject *> self.view.obj == NULL:
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         global __pyx_memoryview_thread_locks_used
 */
      Py_INCREF(Py_None);

      /* "View.MemoryView":350
 *         if type(self) is memoryview or obj is not None:
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:             # <<<<<<<<<<<<<<
 *                 (<__pyx_buffer *> &self.view).obj = Py_None
 *                 Py_INCREF(Py_None)
 */
    }

    /* "View.MemoryView":348
 *         self.obj = obj
 *         self.flags = flags
 *         if type(self) is memoryview or obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_GetBuffer(obj, &self.view, flags)
 *             if <PyObject *> self.view.obj == NULL:
 */
  }

  /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  __pyx_t_1 = ((__pyx_memoryview_thread_locks_used < 8) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":356
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]             # <<<<<<<<<<<<<<
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 */
    __pyx_v_self->lock = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);

    /* "View.MemoryView":357
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1             # <<<<<<<<<<<<<<
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 */
    __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used + 1);

    /* "View.MemoryView":355
 * 
 *         global __pyx_memoryview_thread_locks_used
 *         if __pyx_memoryview_thread_locks_used < THREAD_LOCKS_PREALLOCATED:             # <<<<<<<<<<<<<<
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 */
  }

  /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":359
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()             # <<<<<<<<<<<<<<
 *             if self.lock is NULL:
 *                 raise MemoryError
 */
    __pyx_v_self->lock = PyThread_allocate_lock();

    /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    __pyx_t_1 = ((__pyx_v_self->lock == NULL) != 0);
    if (unlikely(__pyx_t_1)) {

      /* "View.MemoryView":361
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 361, __pyx_L1_error)

      /* "View.MemoryView":360
 *         if self.lock is NULL:
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 * 
 */
    }

    /* "View.MemoryView":358
 *             self.lock = __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]
 *             __pyx_memoryview_thread_locks_used += 1
 *         if self.lock is NULL:             # <<<<<<<<<<<<<<
 *             self.lock = PyThread_allocate_lock()
 *             if self.lock is NULL:
 */
  }

  /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":364
 * 
 *         if flags & PyBUF_FORMAT:
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')             # <<<<<<<<<<<<<<
 *         else:
 *             self.dtype_is_object = dtype_is_object
 */
    __pyx_t_2 = (((__pyx_v_self->view.format[0]) == 'O') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L11_bool_binop_done;
    }
    __pyx_t_2 = (((__pyx_v_self->view.format[1]) == '\x00') != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L11_bool_binop_done:;
    __pyx_v_self->dtype_is_object = __pyx_t_1;

    /* "View.MemoryView":363
 *                 raise MemoryError
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 */
    goto __pyx_L10;
  }

  /* "View.MemoryView":366
 *             self.dtype_is_object = (self.view.format[0] == b'O' and self.view.format[1] == b'\0')
 *         else:
 *             self.dtype_is_object = dtype_is_object             # <<<<<<<<<<<<<<
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 */
  /*else*/ {
    __pyx_v_self->dtype_is_object = __pyx_v_dtype_is_object;
  }
  __pyx_L10:;

  /* "View.MemoryView":368
 *             self.dtype_is_object = dtype_is_object
 * 
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(             # <<<<<<<<<<<<<<
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL
 */
  __pyx_v_self->acquisition_count_aligned_p = ((__pyx_atomic_int *)__pyx_align_pointer(((void *)(&(__pyx_v_self->acquisition_count[0]))), (sizeof(__pyx_atomic_int))));

  /* "View.MemoryView":370
 *         self.acquisition_count_aligned_p = <__pyx_atomic_int *> align_pointer(
 *                   <void *> &self.acquisition_count[0], sizeof(__pyx_atomic_int))
 *         self.typeinfo = NULL             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(memoryview self):
 */
  __pyx_v_self->typeinfo = NULL;

  /* "View.MemoryView":345
 *     cdef __Pyx_TypeInfo *typeinfo
 * 
 *     def __cinit__(memoryview self, object obj, int flags, bint dtype_is_object=False):             # <<<<<<<<<<<<<<
 *         self.obj = obj
 *         self.flags = flags
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_AddTraceback("View.MemoryView.memoryview.__cinit__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

/* Python wrapper */
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryview___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_2__dealloc__(struct __pyx_memoryview_obj *__pyx_v_self) {
  int __pyx_v_i;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyThread_type_lock __pyx_t_6;
  PyThread_type_lock __pyx_t_7;
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
  __pyx_t_1 = (__pyx_v_self->obj != Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":374
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)             # <<<<<<<<<<<<<<
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 */
    __Pyx_ReleaseBuffer((&__pyx_v_self->view));

    /* "View.MemoryView":373
 * 
 *     def __dealloc__(memoryview self):
 *         if self.obj is not None:             # <<<<<<<<<<<<<<
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  __pyx_t_2 = ((((Py_buffer *)(&__pyx_v_self->view))->obj == Py_None) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":377
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL             # <<<<<<<<<<<<<<
 *             Py_DECREF(Py_None)
 * 
 */
    ((Py_buffer *)(&__pyx_v_self->view))->obj = NULL;

    /* "View.MemoryView":378
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 *             Py_DECREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *         cdef int i
 */
    Py_DECREF(Py_None);

    /* "View.MemoryView":375
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 *         elif (<__pyx_buffer *> &self.view).obj == Py_None:             # <<<<<<<<<<<<<<
 * 
 *             (<__pyx_buffer *> &self.view).obj = NULL
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  __pyx_t_2 = ((__pyx_v_self->lock != NULL) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":383
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):             # <<<<<<<<<<<<<<
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 */
    __pyx_t_3 = __pyx_memoryview_thread_locks_used;
    __pyx_t_4 = __pyx_t_3;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      __pyx_t_2 = (((__pyx_memoryview_thread_locks[__pyx_v_i]) == __pyx_v_self->lock) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":385
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1             # <<<<<<<<<<<<<<
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 */
        __pyx_memoryview_thread_locks_used = (__pyx_memoryview_thread_locks_used - 1);

        /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        __pyx_t_2 = ((__pyx_v_i != __pyx_memoryview_thread_locks_used) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":388
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])             # <<<<<<<<<<<<<<
 *                     break
 *             else:
 */
          __pyx_t_6 = (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]);
          __pyx_t_7 = (__pyx_memoryview_thread_locks[__pyx_v_i]);

          /* "View.MemoryView":387
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (             # <<<<<<<<<<<<<<
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break
 */
          (__pyx_memoryview_thread_locks[__pyx_v_i]) = __pyx_t_6;
          (__pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used]) = __pyx_t_7;

          /* "View.MemoryView":386
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:             # <<<<<<<<<<<<<<
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 */
        }

        /* "View.MemoryView":389
 *                         __pyx_memoryview_thread_locks[i], __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used] = (
 *                             __pyx_memoryview_thread_locks[__pyx_memoryview_thread_locks_used], __pyx_memoryview_thread_locks[i])
 *                     break             # <<<<<<<<<<<<<<
 *             else:
 *                 PyThread_free_lock(self.lock)
 */
        goto __pyx_L6_break;

        /* "View.MemoryView":384
 *         if self.lock != NULL:
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:             # <<<<<<<<<<<<<<
 *                     __pyx_memoryview_thread_locks_used -= 1
 *                     if i != __pyx_memoryview_thread_locks_used:
 */
      }
    }
    /*else*/ {

      /* "View.MemoryView":391
 *                     break
 *             else:
 *                 PyThread_free_lock(self.lock)             # <<<<<<<<<<<<<<
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 */
      PyThread_free_lock(__pyx_v_self->lock);
    }
    __pyx_L6_break:;

    /* "View.MemoryView":382
 *         cdef int i
 *         global __pyx_memoryview_thread_locks_used
 *         if self.lock != NULL:             # <<<<<<<<<<<<<<
 *             for i in range(__pyx_memoryview_thread_locks_used):
 *                 if __pyx_memoryview_thread_locks[i] is self.lock:
 */
  }

  /* "View.MemoryView":372
 *         self.typeinfo = NULL
 * 
 *     def __dealloc__(memoryview self):             # <<<<<<<<<<<<<<
 *         if self.obj is not None:
 *             __Pyx_ReleaseBuffer(&self.view)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

static char *__pyx_memoryview_get_item_pointer(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  Py_ssize_t __pyx_v_dim;
  char *__pyx_v_itemp;
  PyObject *__pyx_v_idx = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  PyObject *(*__pyx_t_4)(PyObject *);
  PyObject *__pyx_t_5 = NULL;
  Py_ssize_t __pyx_t_6;
  char *__pyx_t_7;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_item_pointer", 0);

  /* "View.MemoryView":395
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf             # <<<<<<<<<<<<<<
 * 
 *         for dim, idx in enumerate(index):
 */
  __pyx_v_itemp = ((char *)__pyx_v_self->view.buf);

  /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  __pyx_t_1 = 0;
  if (likely(PyList_CheckExact(__pyx_v_index)) || PyTuple_CheckExact(__pyx_v_index)) {
    __pyx_t_2 = __pyx_v_index; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    __pyx_t_4 = NULL;
  } else {
    __pyx_t_3 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 397, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 397, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_4)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      } else {
        if (__pyx_t_3 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_5 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(1, 397, __pyx_L1_error)
        #else
        __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 397, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        #endif
      }
    } else {
      __pyx_t_5 = __pyx_t_4(__pyx_t_2);
      if (unlikely(!__pyx_t_5)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 397, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_5);
    }
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_5);
    __pyx_t_5 = 0;
    __pyx_v_dim = __pyx_t_1;
    __pyx_t_1 = (__pyx_t_1 + 1);

    /* "View.MemoryView":398
 * 
 *         for dim, idx in enumerate(index):
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)             # <<<<<<<<<<<<<<
 * 
 *         return itemp
 */
    __pyx_t_6 = __Pyx_PyIndex_AsSsize_t(__pyx_v_idx); if (unlikely((__pyx_t_6 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 398, __pyx_L1_error)
    __pyx_t_7 = __pyx_pybuffer_index((&__pyx_v_self->view), __pyx_v_itemp, __pyx_t_6, __pyx_v_dim); if (unlikely(__pyx_t_7 == ((char *)NULL))) __PYX_ERR(1, 398, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_7;

    /* "View.MemoryView":397
 *         cdef char *itemp = <char *> self.view.buf
 * 
 *         for dim, idx in enumerate(index):             # <<<<<<<<<<<<<<
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":400
 *             itemp = pybuffer_index(&self.view, itemp, idx, dim)
 * 
 *         return itemp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_itemp;
  goto __pyx_L0;

  /* "View.MemoryView":393
 *                 PyThread_free_lock(self.lock)
 * 
 *     cdef char *get_item_pointer(memoryview self, object index) except NULL:             # <<<<<<<<<<<<<<
 *         cdef Py_ssize_t dim
 *         cdef char *itemp = <char *> self.view.buf
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.get_item_pointer", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index); /*proto*/
static PyObject *__pyx_memoryview___getitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_4__getitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_indices = NULL;
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  char *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__getitem__", 0);

  /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  __pyx_t_1 = (__pyx_v_index == __pyx_builtin_Ellipsis);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":405
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:
 *             return self             # <<<<<<<<<<<<<<
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __pyx_r = ((PyObject *)__pyx_v_self);
    goto __pyx_L0;

    /* "View.MemoryView":404
 * 
 *     def __getitem__(memoryview self, object index):
 *         if index is Ellipsis:             # <<<<<<<<<<<<<<
 *             return self
 * 
 */
  }

  /* "View.MemoryView":407
 *             return self
 * 
 *         have_slices, indices = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         cdef char *itemp
 */
  __pyx_t_3 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 407, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (likely(__pyx_t_3 != Py_None)) {
    PyObject* sequence = __pyx_t_3;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 407, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_t_5);
    #else
    __pyx_t_4 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    #endif
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 407, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_v_indices = __pyx_t_5;
  __pyx_t_5 = 0;

  /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  __pyx_t_2 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_2 < 0)) __PYX_ERR(1, 410, __pyx_L1_error)
  if (__pyx_t_2) {

    /* "View.MemoryView":411
 *         cdef char *itemp
 *         if have_slices:
 *             return memview_slice(self, indices)             # <<<<<<<<<<<<<<
 *         else:
 *             itemp = self.get_item_pointer(indices)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((PyObject *)__pyx_memview_slice(__pyx_v_self, __pyx_v_indices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 411, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":410
 * 
 *         cdef char *itemp
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             return memview_slice(self, indices)
 *         else:
 */
  }

  /* "View.MemoryView":413
 *             return memview_slice(self, indices)
 *         else:
 *             itemp = self.get_item_pointer(indices)             # <<<<<<<<<<<<<<
 *             return self.convert_item_to_object(itemp)
 * 
 */
  /*else*/ {
    __pyx_t_6 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_indices); if (unlikely(__pyx_t_6 == ((char *)NULL))) __PYX_ERR(1, 413, __pyx_L1_error)
    __pyx_v_itemp = __pyx_t_6;

    /* "View.MemoryView":414
 *         else:
 *             itemp = self.get_item_pointer(indices)
 *             return self.convert_item_to_object(itemp)             # <<<<<<<<<<<<<<
 * 
 *     def __setitem__(memoryview self, object index, object value):
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->convert_item_to_object(__pyx_v_self, __pyx_v_itemp); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 414, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":403
 * 
 * 
 *     def __getitem__(memoryview self, object index):             # <<<<<<<<<<<<<<
 *         if index is Ellipsis:
 *             return self
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_indices);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

/* Python wrapper */
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value); /*proto*/
static int __pyx_memoryview___setitem__(PyObject *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setitem__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v_index), ((PyObject *)__pyx_v_value));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_6__setitem__(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_have_slices = NULL;
  PyObject *__pyx_v_obj = NULL;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setitem__", 0);
  __Pyx_INCREF(__pyx_v_index);

  /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  __pyx_t_1 = (__pyx_v_self->view.readonly != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 418, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 418, __pyx_L1_error)

    /* "View.MemoryView":417
 * 
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":420
 *             raise TypeError("Cannot assign to read-only memoryview")
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *         if have_slices:
 */
  __pyx_t_2 = _unellipsify(__pyx_v_index, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(__pyx_t_2 != Py_None)) {
    PyObject* sequence = __pyx_t_2;
    Py_ssize_t size = __Pyx_PySequence_SIZE(sequence);
    if (unlikely(size != 2)) {
      if (size > 2) __Pyx_RaiseTooManyValuesError(2);
      else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
      __PYX_ERR(1, 420, __pyx_L1_error)
    }
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
    __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_t_4);
    #else
    __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    #endif
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  } else {
    __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 420, __pyx_L1_error)
  }
  __pyx_v_have_slices = __pyx_t_3;
  __pyx_t_3 = 0;
  __Pyx_DECREF_SET(__pyx_v_index, __pyx_t_4);
  __pyx_t_4 = 0;

  /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
  __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_have_slices); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 422, __pyx_L1_error)
  if (__pyx_t_1) {

    /* "View.MemoryView":423
 * 
 *         if have_slices:
 *             obj = self.is_slice(value)             # <<<<<<<<<<<<<<
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)
 */
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->is_slice(__pyx_v_self, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 423, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_obj = __pyx_t_2;
    __pyx_t_2 = 0;

    /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_obj); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 424, __pyx_L1_error)
    if (__pyx_t_1) {

      /* "View.MemoryView":425
 *             obj = self.is_slice(value)
 *             if obj:
 *                 self.setitem_slice_assignment(self[index], obj)             # <<<<<<<<<<<<<<
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)
 */
      __pyx_t_2 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_4 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assignment(__pyx_v_self, __pyx_t_2, __pyx_v_obj); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 425, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

      /* "View.MemoryView":424
 *         if have_slices:
 *             obj = self.is_slice(value)
 *             if obj:             # <<<<<<<<<<<<<<
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":427
 *                 self.setitem_slice_assignment(self[index], obj)
 *             else:
 *                 self.setitem_slice_assign_scalar(self[index], value)             # <<<<<<<<<<<<<<
 *         else:
 *             self.setitem_indexed(index, value)
 */
    /*else*/ {
      __pyx_t_4 = __Pyx_PyObject_GetItem(((PyObject *)__pyx_v_self), __pyx_v_index); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      if (!(likely(((__pyx_t_4) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_4, __pyx_memoryview_type))))) __PYX_ERR(1, 427, __pyx_L1_error)
      __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_slice_assign_scalar(__pyx_v_self, ((struct __pyx_memoryview_obj *)__pyx_t_4), __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 427, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __pyx_L5:;

    /* "View.MemoryView":422
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 * 
 *         if have_slices:             # <<<<<<<<<<<<<<
 *             obj = self.is_slice(value)
 *             if obj:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":429
 *                 self.setitem_slice_assign_scalar(self[index], value)
 *         else:
 *             self.setitem_indexed(index, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef is_slice(self, obj):
 */
  /*else*/ {
    __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->setitem_indexed(__pyx_v_self, __pyx_v_index, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L4:;

  /* "View.MemoryView":416
 *             return self.convert_item_to_object(itemp)
 * 
 *     def __setitem__(memoryview self, object index, object value):             # <<<<<<<<<<<<<<
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setitem__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_have_slices);
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

static PyObject *__pyx_memoryview_is_slice(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_obj) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_slice", 0);
  __Pyx_INCREF(__pyx_v_obj);

  /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_obj, __pyx_memoryview_type); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
    {
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __Pyx_ExceptionSave(&__pyx_t_3, &__pyx_t_4, &__pyx_t_5);
      __Pyx_XGOTREF(__pyx_t_3);
      __Pyx_XGOTREF(__pyx_t_4);
      __Pyx_XGOTREF(__pyx_t_5);
      /*try:*/ {

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_6 = __Pyx_PyInt_From_int(((__pyx_v_self->flags & (~PyBUF_WRITABLE)) | PyBUF_ANY_CONTIGUOUS)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":435
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)             # <<<<<<<<<<<<<<
 *             except TypeError:
 *                 return None
 */
        __pyx_t_7 = __Pyx_PyBool_FromLong(__pyx_v_self->dtype_is_object); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 435, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);

        /* "View.MemoryView":434
 *         if not isinstance(obj, memoryview):
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,             # <<<<<<<<<<<<<<
 *                                  self.dtype_is_object)
 *             except TypeError:
 */
        __pyx_t_8 = PyTuple_New(3); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_v_obj);
        __Pyx_GIVEREF(__pyx_v_obj);
        PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_v_obj);
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 1, __pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_7);
        PyTuple_SET_ITEM(__pyx_t_8, 2, __pyx_t_7);
        __pyx_t_6 = 0;
        __pyx_t_7 = 0;
        __pyx_t_7 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_8, NULL); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 434, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_DECREF_SET(__pyx_v_obj, __pyx_t_7);
        __pyx_t_7 = 0;

        /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      }
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L9_try_end;
      __pyx_L4_error:;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;

      /* "View.MemoryView":436
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 *             except TypeError:             # <<<<<<<<<<<<<<
 *                 return None
 * 
 */
      __pyx_t_9 = __Pyx_PyErr_ExceptionMatches(__pyx_builtin_TypeError);
      if (__pyx_t_9) {
        __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
        if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_6) < 0) __PYX_ERR(1, 436, __pyx_L6_except_error)
        __Pyx_GOTREF(__pyx_t_7);
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GOTREF(__pyx_t_6);

        /* "View.MemoryView":437
 *                                  self.dtype_is_object)
 *             except TypeError:
 *                 return None             # <<<<<<<<<<<<<<
 * 
 *         return obj
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_r = Py_None; __Pyx_INCREF(Py_None);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L7_except_return;
      }
      goto __pyx_L6_except_error;
      __pyx_L6_except_error:;

      /* "View.MemoryView":433
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):
 *             try:             # <<<<<<<<<<<<<<
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 *                                  self.dtype_is_object)
 */
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L1_error;
      __pyx_L7_except_return:;
      __Pyx_XGIVEREF(__pyx_t_3);
      __Pyx_XGIVEREF(__pyx_t_4);
      __Pyx_XGIVEREF(__pyx_t_5);
      __Pyx_ExceptionReset(__pyx_t_3, __pyx_t_4, __pyx_t_5);
      goto __pyx_L0;
      __pyx_L9_try_end:;
    }

    /* "View.MemoryView":432
 * 
 *     cdef is_slice(self, obj):
 *         if not isinstance(obj, memoryview):             # <<<<<<<<<<<<<<
 *             try:
 *                 obj = memoryview(obj, self.flags & ~PyBUF_WRITABLE | PyBUF_ANY_CONTIGUOUS,
 */
  }

  /* "View.MemoryView":439
 *                 return None
 * 
 *         return obj             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assignment(self, dst, src):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_obj);
  __pyx_r = __pyx_v_obj;
  goto __pyx_L0;

  /* "View.MemoryView":431
 *             self.setitem_indexed(index, value)
 * 
 *     cdef is_slice(self, obj):             # <<<<<<<<<<<<<<
 *         if not isinstance(obj, memoryview):
 *             try:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_obj);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

static PyObject *__pyx_memoryview_setitem_slice_assignment(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_dst, PyObject *__pyx_v_src) {
  __Pyx_memviewslice __pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_src_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  __Pyx_memviewslice *__pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assignment", 0);

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  if (!(likely(((__pyx_v_src) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_src, __pyx_memoryview_type))))) __PYX_ERR(1, 445, __pyx_L1_error)
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_src), (&__pyx_v_src_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 445, __pyx_L1_error)

  /* "View.MemoryView":446
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],             # <<<<<<<<<<<<<<
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 */
  if (!(likely(((__pyx_v_dst) == Py_None) || likely(__Pyx_TypeTest(__pyx_v_dst, __pyx_memoryview_type))))) __PYX_ERR(1, 446, __pyx_L1_error)
  __pyx_t_2 = __pyx_memoryview_get_slice_from_memoryview(((struct __pyx_memoryview_obj *)__pyx_v_dst), (&__pyx_v_dst_slice)); if (unlikely(__pyx_t_2 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 446, __pyx_L1_error)

  /* "View.MemoryView":447
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 */
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_src, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_4 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_PyObject_GetAttrStr(__pyx_v_dst, __pyx_n_s_ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 447, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":445
 *         cdef __Pyx_memviewslice src_slice
 * 
 *         memoryview_copy_contents(get_slice_from_memview(src, &src_slice)[0],             # <<<<<<<<<<<<<<
 *                                  get_slice_from_memview(dst, &dst_slice)[0],
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 */
  __pyx_t_6 = __pyx_memoryview_copy_contents((__pyx_t_1[0]), (__pyx_t_2[0]), __pyx_t_4, __pyx_t_5, __pyx_v_self->dtype_is_object); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 445, __pyx_L1_error)

  /* "View.MemoryView":441
 *         return obj
 * 
 *     cdef setitem_slice_assignment(self, dst, src):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice dst_slice
 *         cdef __Pyx_memviewslice src_slice
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

static PyObject *__pyx_memoryview_setitem_slice_assign_scalar(struct __pyx_memoryview_obj *__pyx_v_self, struct __pyx_memoryview_obj *__pyx_v_dst, PyObject *__pyx_v_value) {
  int __pyx_v_array[0x80];
  void *__pyx_v_tmp;
  void *__pyx_v_item;
  __Pyx_memviewslice *__pyx_v_dst_slice;
  __Pyx_memviewslice __pyx_v_tmp_slice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_slice_assign_scalar", 0);

  /* "View.MemoryView":451
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):
 *         cdef int array[128]
 *         cdef void *tmp = NULL             # <<<<<<<<<<<<<<
 *         cdef void *item
 * 
 */
  __pyx_v_tmp = NULL;

  /* "View.MemoryView":456
 *         cdef __Pyx_memviewslice *dst_slice
 *         cdef __Pyx_memviewslice tmp_slice
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)             # <<<<<<<<<<<<<<
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_dst, (&__pyx_v_tmp_slice)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 456, __pyx_L1_error)
  __pyx_v_dst_slice = __pyx_t_1;

  /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
  __pyx_t_2 = ((((size_t)__pyx_v_self->view.itemsize) > (sizeof(__pyx_v_array))) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":459
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)             # <<<<<<<<<<<<<<
 *             if tmp == NULL:
 *                 raise MemoryError
 */
    __pyx_v_tmp = PyMem_Malloc(__pyx_v_self->view.itemsize);

    /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    __pyx_t_2 = ((__pyx_v_tmp == NULL) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":461
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 *                 raise MemoryError             # <<<<<<<<<<<<<<
 *             item = tmp
 *         else:
 */
      PyErr_NoMemory(); __PYX_ERR(1, 461, __pyx_L1_error)

      /* "View.MemoryView":460
 *         if <size_t>self.view.itemsize > sizeof(array):
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:             # <<<<<<<<<<<<<<
 *                 raise MemoryError
 *             item = tmp
 */
    }

    /* "View.MemoryView":462
 *             if tmp == NULL:
 *                 raise MemoryError
 *             item = tmp             # <<<<<<<<<<<<<<
 *         else:
 *             item = <void *> array
 */
    __pyx_v_item = __pyx_v_tmp;

    /* "View.MemoryView":458
 *         dst_slice = get_slice_from_memview(dst, &tmp_slice)
 * 
 *         if <size_t>self.view.itemsize > sizeof(array):             # <<<<<<<<<<<<<<
 *             tmp = PyMem_Malloc(self.view.itemsize)
 *             if tmp == NULL:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":464
 *             item = tmp
 *         else:
 *             item = <void *> array             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  /*else*/ {
    __pyx_v_item = ((void *)__pyx_v_array);
  }
  __pyx_L3:;

  /* "View.MemoryView":466
 *             item = <void *> array
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value
 */
  /*try:*/ {

    /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
    __pyx_t_2 = (__pyx_v_self->dtype_is_object != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":468
 *         try:
 *             if self.dtype_is_object:
 *                 (<PyObject **> item)[0] = <PyObject *> value             # <<<<<<<<<<<<<<
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)
 */
      (((PyObject **)__pyx_v_item)[0]) = ((PyObject *)__pyx_v_value);

      /* "View.MemoryView":467
 * 
 *         try:
 *             if self.dtype_is_object:             # <<<<<<<<<<<<<<
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 */
      goto __pyx_L8;
    }

    /* "View.MemoryView":470
 *                 (<PyObject **> item)[0] = <PyObject *> value
 *             else:
 *                 self.assign_item_from_object(<char *> item, value)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    /*else*/ {
      __pyx_t_3 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, ((char *)__pyx_v_item), __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 470, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    }
    __pyx_L8:;

    /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    __pyx_t_2 = ((__pyx_v_self->view.suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":475
 * 
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)             # <<<<<<<<<<<<<<
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 *                                 item, self.dtype_is_object)
 */
      __pyx_t_3 = assert_direct_dimensions(__pyx_v_self->view.suboffsets, __pyx_v_self->view.ndim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 475, __pyx_L6_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

      /* "View.MemoryView":474
 * 
 * 
 *             if self.view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,
 */
    }

    /* "View.MemoryView":476
 *             if self.view.suboffsets != NULL:
 *                 assert_direct_dimensions(self.view.suboffsets, self.view.ndim)
 *             slice_assign_scalar(dst_slice, dst.view.ndim, self.view.itemsize,             # <<<<<<<<<<<<<<
 *                                 item, self.dtype_is_object)
 *         finally:
 */
    __pyx_memoryview_slice_assign_scalar(__pyx_v_dst_slice, __pyx_v_dst->view.ndim, __pyx_v_self->view.itemsize, __pyx_v_item, __pyx_v_self->dtype_is_object);
  }

  /* "View.MemoryView":479
 *                                 item, self.dtype_is_object)
 *         finally:
 *             PyMem_Free(tmp)             # <<<<<<<<<<<<<<
 * 
 *     cdef setitem_indexed(self, index, value):
 */
  /*finally:*/ {
    /*normal exit:*/{
      PyMem_Free(__pyx_v_tmp);
      goto __pyx_L7;
    }
    __pyx_L6_error:;
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __Pyx_PyThreadState_assign
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        PyMem_Free(__pyx_v_tmp);
      }
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
    }
    __pyx_L7:;
  }

  /* "View.MemoryView":449
 *                                  src.ndim, dst.ndim, self.dtype_is_object)
 * 
 *     cdef setitem_slice_assign_scalar(self, memoryview dst, value):             # <<<<<<<<<<<<<<
 *         cdef int array[128]
 *         cdef void *tmp = NULL
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_slice_assign_scalar", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

static PyObject *__pyx_memoryview_setitem_indexed(struct __pyx_memoryview_obj *__pyx_v_self, PyObject *__pyx_v_index, PyObject *__pyx_v_value) {
  char *__pyx_v_itemp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  char *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("setitem_indexed", 0);

  /* "View.MemoryView":482
 * 
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)             # <<<<<<<<<<<<<<
 *         self.assign_item_from_object(itemp, value)
 * 
 */
  __pyx_t_1 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->get_item_pointer(__pyx_v_self, __pyx_v_index); if (unlikely(__pyx_t_1 == ((char *)NULL))) __PYX_ERR(1, 482, __pyx_L1_error)
  __pyx_v_itemp = __pyx_t_1;

  /* "View.MemoryView":483
 *     cdef setitem_indexed(self, index, value):
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_memoryview *)__pyx_v_self->__pyx_vtab)->assign_item_from_object(__pyx_v_self, __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 483, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":481
 *             PyMem_Free(tmp)
 * 
 *     cdef setitem_indexed(self, index, value):             # <<<<<<<<<<<<<<
 *         cdef char *itemp = self.get_item_pointer(index)
 *         self.assign_item_from_object(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.setitem_indexed", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_convert_item_to_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_v_struct = NULL;
  PyObject *__pyx_v_bytesitem = 0;
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  int __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  size_t __pyx_t_10;
  int __pyx_t_11;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":488
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef bytes bytesitem
 * 
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 488, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":491
 *         cdef bytes bytesitem
 * 
 *         bytesitem = itemp[:self.view.itemsize]             # <<<<<<<<<<<<<<
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 */
  __pyx_t_1 = __Pyx_PyBytes_FromStringAndSize(__pyx_v_itemp + 0, __pyx_v_self->view.itemsize - 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 491, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_bytesitem = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_2, &__pyx_t_3, &__pyx_t_4);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_4);
    /*try:*/ {

      /* "View.MemoryView":493
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)             # <<<<<<<<<<<<<<
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_unpack); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 493, __pyx_L3_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = NULL;
      __pyx_t_8 = 0;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_5))) {
        __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_5);
        if (likely(__pyx_t_7)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
          __Pyx_INCREF(__pyx_t_7);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_5, function);
          __pyx_t_8 = 1;
        }
      }
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
        PyObject *__pyx_temp[3] = {__pyx_t_7, __pyx_t_6, __pyx_v_bytesitem};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_9 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_9);
        if (__pyx_t_7) {
          __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_7); __pyx_t_7 = NULL;
        }
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_9, 0+__pyx_t_8, __pyx_t_6);
        __Pyx_INCREF(__pyx_v_bytesitem);
        __Pyx_GIVEREF(__pyx_v_bytesitem);
        PyTuple_SET_ITEM(__pyx_t_9, 1+__pyx_t_8, __pyx_v_bytesitem);
        __pyx_t_6 = 0;
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 493, __pyx_L3_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      }
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_v_result = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    }

    /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
    /*else:*/ {
      __pyx_t_10 = strlen(__pyx_v_self->view.format); 
      __pyx_t_11 = ((__pyx_t_10 == 1) != 0);
      if (__pyx_t_11) {

        /* "View.MemoryView":498
 *         else:
 *             if len(self.view.format) == 1:
 *                 return result[0]             # <<<<<<<<<<<<<<
 *             return result
 * 
 */
        __Pyx_XDECREF(__pyx_r);
        __pyx_t_1 = __Pyx_GetItemInt(__pyx_v_result, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 498, __pyx_L5_except_error)
        __Pyx_GOTREF(__pyx_t_1);
        __pyx_r = __pyx_t_1;
        __pyx_t_1 = 0;
        goto __pyx_L6_except_return;

        /* "View.MemoryView":497
 *             raise ValueError("Unable to convert item to object")
 *         else:
 *             if len(self.view.format) == 1:             # <<<<<<<<<<<<<<
 *                 return result[0]
 *             return result
 */
      }

      /* "View.MemoryView":499
 *             if len(self.view.format) == 1:
 *                 return result[0]
 *             return result             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_v_result);
      __pyx_r = __pyx_v_result;
      goto __pyx_L6_except_return;
    }
    __pyx_L3_error:;
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;

    /* "View.MemoryView":494
 *         try:
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:             # <<<<<<<<<<<<<<
 *             raise ValueError("Unable to convert item to object")
 *         else:
 */
    __Pyx_ErrFetch(&__pyx_t_1, &__pyx_t_5, &__pyx_t_9);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_error); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 494, __pyx_L5_except_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyErr_GivenExceptionMatches(__pyx_t_1, __pyx_t_6);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_ErrRestore(__pyx_t_1, __pyx_t_5, __pyx_t_9);
    __pyx_t_1 = 0; __pyx_t_5 = 0; __pyx_t_9 = 0;
    if (__pyx_t_8) {
      __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_9, &__pyx_t_5, &__pyx_t_1) < 0) __PYX_ERR(1, 494, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_9);
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_1);

      /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 495, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 495, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "View.MemoryView":492
 * 
 *         bytesitem = itemp[:self.view.itemsize]
 *         try:             # <<<<<<<<<<<<<<
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 */
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L1_error;
    __pyx_L6_except_return:;
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_XGIVEREF(__pyx_t_4);
    __Pyx_ExceptionReset(__pyx_t_2, __pyx_t_3, __pyx_t_4);
    goto __pyx_L0;
  }

  /* "View.MemoryView":485
 *         self.assign_item_from_object(itemp, value)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memoryview.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesitem);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

static PyObject *__pyx_memoryview_assign_item_from_object(struct __pyx_memoryview_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_v_struct = NULL;
  char __pyx_v_c;
  PyObject *__pyx_v_bytesvalue = 0;
  Py_ssize_t __pyx_v_i;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_t_3;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  Py_ssize_t __pyx_t_9;
  PyObject *__pyx_t_10 = NULL;
  char *__pyx_t_11;
  char *__pyx_t_12;
  char *__pyx_t_13;
  char *__pyx_t_14;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":504
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 *         import struct             # <<<<<<<<<<<<<<
 *         cdef char c
 *         cdef bytes bytesvalue
 */
  __pyx_t_1 = __Pyx_Import(__pyx_n_s_struct, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 504, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_struct = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
  __pyx_t_2 = PyTuple_Check(__pyx_v_value); 
  __pyx_t_3 = (__pyx_t_2 != 0);
  if (__pyx_t_3) {

    /* "View.MemoryView":510
 * 
 *         if isinstance(value, tuple):
 *             bytesvalue = struct.pack(self.view.format, *value)             # <<<<<<<<<<<<<<
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = PyTuple_New(1); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PySequence_Tuple(__pyx_v_value); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = PyNumber_Add(__pyx_t_5, __pyx_t_4); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_6, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 510, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 510, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "View.MemoryView":509
 *         cdef Py_ssize_t i
 * 
 *         if isinstance(value, tuple):             # <<<<<<<<<<<<<<
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":512
 *             bytesvalue = struct.pack(self.view.format, *value)
 *         else:
 *             bytesvalue = struct.pack(self.view.format, value)             # <<<<<<<<<<<<<<
 * 
 *         for i, c in enumerate(bytesvalue):
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_struct, __pyx_n_s_pack); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyBytes_FromString(__pyx_v_self->view.format); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_5 = NULL;
    __pyx_t_7 = 0;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_7 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_t_1, __pyx_v_value};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_7, 2+__pyx_t_7); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    } else
    #endif
    {
      __pyx_t_8 = PyTuple_New(2+__pyx_t_7); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_GIVEREF(__pyx_t_1);
      PyTuple_SET_ITEM(__pyx_t_8, 0+__pyx_t_7, __pyx_t_1);
      __Pyx_INCREF(__pyx_v_value);
      __Pyx_GIVEREF(__pyx_v_value);
      PyTuple_SET_ITEM(__pyx_t_8, 1+__pyx_t_7, __pyx_v_value);
      __pyx_t_1 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_8, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 512, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (!(likely(PyBytes_CheckExact(__pyx_t_4))||((__pyx_t_4) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "bytes", Py_TYPE(__pyx_t_4)->tp_name), 0))) __PYX_ERR(1, 512, __pyx_L1_error)
    __pyx_v_bytesvalue = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
  __pyx_t_9 = 0;
  if (unlikely(__pyx_v_bytesvalue == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' is not iterable");
    __PYX_ERR(1, 514, __pyx_L1_error)
  }
  __Pyx_INCREF(__pyx_v_bytesvalue);
  __pyx_t_10 = __pyx_v_bytesvalue;
  __pyx_t_12 = PyBytes_AS_STRING(__pyx_t_10);
  __pyx_t_13 = (__pyx_t_12 + PyBytes_GET_SIZE(__pyx_t_10));
  for (__pyx_t_14 = __pyx_t_12; __pyx_t_14 < __pyx_t_13; __pyx_t_14++) {
    __pyx_t_11 = __pyx_t_14;
    __pyx_v_c = (__pyx_t_11[0]);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    __pyx_v_i = __pyx_t_9;

    /* "View.MemoryView":514
 *             bytesvalue = struct.pack(self.view.format, value)
 * 
 *         for i, c in enumerate(bytesvalue):             # <<<<<<<<<<<<<<
 *             itemp[i] = c
 * 
 */
    __pyx_t_9 = (__pyx_t_9 + 1);

    /* "View.MemoryView":515
 * 
 *         for i, c in enumerate(bytesvalue):
 *             itemp[i] = c             # <<<<<<<<<<<<<<
 * 
 *     @cname('getbuffer')
 */
    (__pyx_v_itemp[__pyx_v_i]) = __pyx_v_c;
  }
  __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;

  /* "View.MemoryView":501
 *             return result
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         """Only used if instantiated manually by the user, or if Cython doesn't
 *         know how to convert the type"""
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("View.MemoryView.memoryview.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_struct);
  __Pyx_XDECREF(__pyx_v_bytesvalue);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_memoryview_getbuffer(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_8__getbuffer__(struct __pyx_memoryview_obj *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  char *__pyx_t_5;
  void *__pyx_t_6;
  int __pyx_t_7;
  Py_ssize_t __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  if (__pyx_v_info == NULL) {
    PyErr_SetString(PyExc_BufferError, "PyObject_GetBuffer: view==NULL argument is obsolete");
    return -1;
  }
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(__pyx_v_info->obj);

  /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  __pyx_t_2 = ((__pyx_v_flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_2 = (__pyx_v_self->view.readonly != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L4_bool_binop_done:;
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 520, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 520, __pyx_L1_error)

    /* "View.MemoryView":519
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:             # <<<<<<<<<<<<<<
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 */
  }

  /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_ND) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":523
 * 
 *         if flags & PyBUF_ND:
 *             info.shape = self.view.shape             # <<<<<<<<<<<<<<
 *         else:
 *             info.shape = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.shape;
    __pyx_v_info->shape = __pyx_t_4;

    /* "View.MemoryView":522
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 * 
 *         if flags & PyBUF_ND:             # <<<<<<<<<<<<<<
 *             info.shape = self.view.shape
 *         else:
 */
    goto __pyx_L6;
  }

  /* "View.MemoryView":525
 *             info.shape = self.view.shape
 *         else:
 *             info.shape = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_STRIDES:
 */
  /*else*/ {
    __pyx_v_info->shape = NULL;
  }
  __pyx_L6:;

  /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_STRIDES) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":528
 * 
 *         if flags & PyBUF_STRIDES:
 *             info.strides = self.view.strides             # <<<<<<<<<<<<<<
 *         else:
 *             info.strides = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.strides;
    __pyx_v_info->strides = __pyx_t_4;

    /* "View.MemoryView":527
 *             info.shape = NULL
 * 
 *         if flags & PyBUF_STRIDES:             # <<<<<<<<<<<<<<
 *             info.strides = self.view.strides
 *         else:
 */
    goto __pyx_L7;
  }

  /* "View.MemoryView":530
 *             info.strides = self.view.strides
 *         else:
 *             info.strides = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_INDIRECT:
 */
  /*else*/ {
    __pyx_v_info->strides = NULL;
  }
  __pyx_L7:;

  /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_INDIRECT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":533
 * 
 *         if flags & PyBUF_INDIRECT:
 *             info.suboffsets = self.view.suboffsets             # <<<<<<<<<<<<<<
 *         else:
 *             info.suboffsets = NULL
 */
    __pyx_t_4 = __pyx_v_self->view.suboffsets;
    __pyx_v_info->suboffsets = __pyx_t_4;

    /* "View.MemoryView":532
 *             info.strides = NULL
 * 
 *         if flags & PyBUF_INDIRECT:             # <<<<<<<<<<<<<<
 *             info.suboffsets = self.view.suboffsets
 *         else:
 */
    goto __pyx_L8;
  }

  /* "View.MemoryView":535
 *             info.suboffsets = self.view.suboffsets
 *         else:
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_FORMAT:
 */
  /*else*/ {
    __pyx_v_info->suboffsets = NULL;
  }
  __pyx_L8:;

  /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_flags & PyBUF_FORMAT) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":538
 * 
 *         if flags & PyBUF_FORMAT:
 *             info.format = self.view.format             # <<<<<<<<<<<<<<
 *         else:
 *             info.format = NULL
 */
    __pyx_t_5 = __pyx_v_self->view.format;
    __pyx_v_info->format = __pyx_t_5;

    /* "View.MemoryView":537
 *             info.suboffsets = NULL
 * 
 *         if flags & PyBUF_FORMAT:             # <<<<<<<<<<<<<<
 *             info.format = self.view.format
 *         else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":540
 *             info.format = self.view.format
 *         else:
 *             info.format = NULL             # <<<<<<<<<<<<<<
 * 
 *         info.buf = self.view.buf
 */
  /*else*/ {
    __pyx_v_info->format = NULL;
  }
  __pyx_L9:;

  /* "View.MemoryView":542
 *             info.format = NULL
 * 
 *         info.buf = self.view.buf             # <<<<<<<<<<<<<<
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 */
  __pyx_t_6 = __pyx_v_self->view.buf;
  __pyx_v_info->buf = __pyx_t_6;

  /* "View.MemoryView":543
 * 
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim             # <<<<<<<<<<<<<<
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 */
  __pyx_t_7 = __pyx_v_self->view.ndim;
  __pyx_v_info->ndim = __pyx_t_7;

  /* "View.MemoryView":544
 *         info.buf = self.view.buf
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize             # <<<<<<<<<<<<<<
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 */
  __pyx_t_8 = __pyx_v_self->view.itemsize;
  __pyx_v_info->itemsize = __pyx_t_8;

  /* "View.MemoryView":545
 *         info.ndim = self.view.ndim
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len             # <<<<<<<<<<<<<<
 *         info.readonly = self.view.readonly
 *         info.obj = self
 */
  __pyx_t_8 = __pyx_v_self->view.len;
  __pyx_v_info->len = __pyx_t_8;

  /* "View.MemoryView":546
 *         info.itemsize = self.view.itemsize
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly             # <<<<<<<<<<<<<<
 *         info.obj = self
 * 
 */
  __pyx_t_1 = __pyx_v_self->view.readonly;
  __pyx_v_info->readonly = __pyx_t_1;

  /* "View.MemoryView":547
 *         info.len = self.view.len
 *         info.readonly = self.view.readonly
 *         info.obj = self             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
  __Pyx_GOTREF(__pyx_v_info->obj);
  __Pyx_DECREF(__pyx_v_info->obj);
  __pyx_v_info->obj = ((PyObject *)__pyx_v_self);

  /* "View.MemoryView":518
 * 
 *     @cname('getbuffer')
 *     def __getbuffer__(self, Py_buffer *info, int flags):             # <<<<<<<<<<<<<<
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = 0;
  }
  __pyx_L2:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_1T___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":554
 *     @property
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&result.from_slice)
 *         return result
 */
  __pyx_t_1 = __pyx_memoryview_copy_object(__pyx_v_self); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 554, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (!(likely(((__pyx_t_1) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_1, __pyx_memoryviewslice_type))))) __PYX_ERR(1, 554, __pyx_L1_error)
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "View.MemoryView":555
 *     def T(self):
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)             # <<<<<<<<<<<<<<
 *         return result
 * 
 */
  __pyx_t_2 = __pyx_memslice_transpose((&__pyx_v_result->from_slice)); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 555, __pyx_L1_error)

  /* "View.MemoryView":556
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 *         return result             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":553
 * 
 *     @property
 *     def T(self):             # <<<<<<<<<<<<<<
 *         cdef _memoryviewslice result = memoryview_copy(self)
 *         transpose_memslice(&result.from_slice)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.T.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4base___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":560
 *     @property
 *     def base(self):
 *         return self.obj             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->obj);
  __pyx_r = __pyx_v_self->obj;
  goto __pyx_L0;

  /* "View.MemoryView":559
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.obj
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_5shape___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_length;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":564
 *     @property
 *     def shape(self):
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_3 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
  for (__pyx_t_4 = __pyx_v_self->view.shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_length = (__pyx_t_2[0]);
    __pyx_t_5 = PyInt_FromSsize_t(__pyx_v_length); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_5))) __PYX_ERR(1, 564, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }
  __pyx_t_5 = PyList_AsTuple(((PyObject*)__pyx_t_1)); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 564, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":563
 * 
 *     @property
 *     def shape(self):             # <<<<<<<<<<<<<<
 *         return tuple([length for length in self.view.shape[:self.view.ndim]])
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview.shape.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_7strides___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_stride;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  __pyx_t_1 = ((__pyx_v_self->view.strides == NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 570, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 570, __pyx_L1_error)

    /* "View.MemoryView":568
 *     @property
 *     def strides(self):
 *         if self.view.strides == NULL:             # <<<<<<<<<<<<<<
 * 
 *             raise ValueError("Buffer view does not expose strides")
 */
  }

  /* "View.MemoryView":572
 *             raise ValueError("Buffer view does not expose strides")
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_v_self->view.strides + __pyx_v_self->view.ndim);
  for (__pyx_t_5 = __pyx_v_self->view.strides; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
    __pyx_t_3 = __pyx_t_5;
    __pyx_v_stride = (__pyx_t_3[0]);
    __pyx_t_6 = PyInt_FromSsize_t(__pyx_v_stride); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_2, (PyObject*)__pyx_t_6))) __PYX_ERR(1, 572, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }
  __pyx_t_6 = PyList_AsTuple(((PyObject*)__pyx_t_2)); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 572, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_6;
  __pyx_t_6 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":567
 * 
 *     @property
 *     def strides(self):             # <<<<<<<<<<<<<<
 *         if self.view.strides == NULL:
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.strides.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_10suboffsets___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.suboffsets == NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_Multiply(__pyx_tuple__23, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 577, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":576
 *     @property
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:             # <<<<<<<<<<<<<<
 *             return (-1,) * self.view.ndim
 * 
 */
  }

  /* "View.MemoryView":579
 *             return (-1,) * self.view.ndim
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = (__pyx_v_self->view.suboffsets + __pyx_v_self->view.ndim);
  for (__pyx_t_6 = __pyx_v_self->view.suboffsets; __pyx_t_6 < __pyx_t_5; __pyx_t_6++) {
    __pyx_t_4 = __pyx_t_6;
    __pyx_v_suboffset = (__pyx_t_4[0]);
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_suboffset); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_3, (PyObject*)__pyx_t_2))) __PYX_ERR(1, 579, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_t_2 = PyList_AsTuple(((PyObject*)__pyx_t_3)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":575
 * 
 *     @property
 *     def suboffsets(self):             # <<<<<<<<<<<<<<
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.suboffsets.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4ndim___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":583
 *     @property
 *     def ndim(self):
 *         return self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_self->view.ndim); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":582
 * 
 *     @property
 *     def ndim(self):             # <<<<<<<<<<<<<<
 *         return self.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.ndim.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_8itemsize___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":587
 *     @property
 *     def itemsize(self):
 *         return self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 587, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":586
 * 
 *     @property
 *     def itemsize(self):             # <<<<<<<<<<<<<<
 *         return self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.itemsize.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_6nbytes___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":591
 *     @property
 *     def nbytes(self):
 *         return self.size * self.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_size); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_self->view.itemsize); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Multiply(__pyx_t_1, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 591, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":590
 * 
 *     @property
 *     def nbytes(self):             # <<<<<<<<<<<<<<
 *         return self.size * self.view.itemsize
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.nbytes.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_10memoryview_4size___get__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_v_result = NULL;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;
  Py_ssize_t *__pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  __pyx_t_1 = (__pyx_v_self->_size == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":596
 *     def size(self):
 *         if self._size is None:
 *             result = 1             # <<<<<<<<<<<<<<
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 */
    __Pyx_INCREF(__pyx_int_1);
    __pyx_v_result = __pyx_int_1;

    /* "View.MemoryView":598
 *             result = 1
 * 
 *             for length in self.view.shape[:self.view.ndim]:             # <<<<<<<<<<<<<<
 *                 result *= length
 * 
 */
    __pyx_t_4 = (__pyx_v_self->view.shape + __pyx_v_self->view.ndim);
    for (__pyx_t_5 = __pyx_v_self->view.shape; __pyx_t_5 < __pyx_t_4; __pyx_t_5++) {
      __pyx_t_3 = __pyx_t_5;
      __pyx_t_6 = PyInt_FromSsize_t((__pyx_t_3[0])); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 598, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_6);
      __pyx_t_6 = 0;

      /* "View.MemoryView":599
 * 
 *             for length in self.view.shape[:self.view.ndim]:
 *                 result *= length             # <<<<<<<<<<<<<<
 * 
 *             self._size = result
 */
      __pyx_t_6 = PyNumber_InPlaceMultiply(__pyx_v_result, __pyx_v_length); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 599, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF_SET(__pyx_v_result, __pyx_t_6);
      __pyx_t_6 = 0;
    }

    /* "View.MemoryView":601
 *                 result *= length
 * 
 *             self._size = result             # <<<<<<<<<<<<<<
 * 
 *         return self._size
 */
    __Pyx_INCREF(__pyx_v_result);
    __Pyx_GIVEREF(__pyx_v_result);
    __Pyx_GOTREF(__pyx_v_self->_size);
    __Pyx_DECREF(__pyx_v_self->_size);
    __pyx_v_self->_size = __pyx_v_result;

    /* "View.MemoryView":595
 *     @property
 *     def size(self):
 *         if self._size is None:             # <<<<<<<<<<<<<<
 *             result = 1
 * 
 */
  }

  /* "View.MemoryView":603
 *             self._size = result
 * 
 *         return self._size             # <<<<<<<<<<<<<<
 * 
 *     def __len__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->_size);
  __pyx_r = __pyx_v_self->_size;
  goto __pyx_L0;

  /* "View.MemoryView":594
 * 
 *     @property
 *     def size(self):             # <<<<<<<<<<<<<<
 *         if self._size is None:
 *             result = 1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("View.MemoryView.memoryview.size.__get__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

/* Python wrapper */
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self); /*proto*/
static Py_ssize_t __pyx_memoryview___len__(PyObject *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__len__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static Py_ssize_t __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_10__len__(struct __pyx_memoryview_obj *__pyx_v_self) {
  Py_ssize_t __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__len__", 0);

  /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  __pyx_t_1 = ((__pyx_v_self->view.ndim >= 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":607
 *     def __len__(self):
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]             # <<<<<<<<<<<<<<
 * 
 *         return 0
 */
    __pyx_r = (__pyx_v_self->view.shape[0]);
    goto __pyx_L0;

    /* "View.MemoryView":606
 * 
 *     def __len__(self):
 *         if self.view.ndim >= 1:             # <<<<<<<<<<<<<<
 *             return self.view.shape[0]
 * 
 */
  }

  /* "View.MemoryView":609
 *             return self.view.shape[0]
 * 
 *         return 0             # <<<<<<<<<<<<<<
 * 
 *     def __repr__(self):
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":605
 *         return self._size
 * 
 *     def __len__(self):             # <<<<<<<<<<<<<<
 *         if self.view.ndim >= 1:
 *             return self.view.shape[0]
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___repr__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__repr__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_12__repr__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__repr__", 0);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":613
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))             # <<<<<<<<<<<<<<
 * 
 *     def __str__(self):
 */
  __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_builtin_id, ((PyObject *)__pyx_v_self)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 613, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);

  /* "View.MemoryView":612
 * 
 *     def __repr__(self):
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,             # <<<<<<<<<<<<<<
 *                                                id(self))
 * 
 */
  __pyx_t_3 = PyTuple_New(2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 612, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":611
 *         return 0
 * 
 *     def __repr__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r at 0x%x>" % (self.base.__class__.__name__,
 *                                                id(self))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__repr__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

/* Python wrapper */
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_memoryview___str__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__str__ (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_14__str__(struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__str__", 0);

  /* "View.MemoryView":616
 * 
 *     def __str__(self):
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_base); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_class); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_MemoryView_of_r_object, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 616, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":615
 *                                                id(self))
 * 
 *     def __str__(self):             # <<<<<<<<<<<<<<
 *         return "<MemoryView of %r object>" % (self.base.__class__.__name__,)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__str__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_c_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_c_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_16is_c_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_c_contig", 0);

  /* "View.MemoryView":622
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 622, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":623
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def is_f_contig(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'C', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 623, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":619
 * 
 * 
 *     def is_c_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_c_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_is_f_contig(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("is_f_contig (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_18is_f_contig(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice *__pyx_v_mslice;
  __Pyx_memviewslice __pyx_v_tmp;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice *__pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("is_f_contig", 0);

  /* "View.MemoryView":628
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)             # <<<<<<<<<<<<<<
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 */
  __pyx_t_1 = __pyx_memoryview_get_slice_from_memoryview(__pyx_v_self, (&__pyx_v_tmp)); if (unlikely(__pyx_t_1 == ((__Pyx_memviewslice *)NULL))) __PYX_ERR(1, 628, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":629
 *         cdef __Pyx_memviewslice tmp
 *         mslice = get_slice_from_memview(self, &tmp)
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)             # <<<<<<<<<<<<<<
 * 
 *     def copy(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_memviewslice_is_contig((__pyx_v_mslice[0]), 'F', __pyx_v_self->view.ndim)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 629, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":625
 *         return slice_is_contig(mslice[0], 'C', self.view.ndim)
 * 
 *     def is_f_contig(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice *mslice
 *         cdef __Pyx_memviewslice tmp
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.is_f_contig", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_20copy(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_mslice;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy", 0);

  /* "View.MemoryView":633
 *     def copy(self):
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &mslice)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_F_CONTIGUOUS));

  /* "View.MemoryView":635
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 * 
 *         slice_copy(self, &mslice)             # <<<<<<<<<<<<<<
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,
 *                                    self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_mslice));

  /* "View.MemoryView":636
 * 
 *         slice_copy(self, &mslice)
 *         mslice = slice_copy_contig(&mslice, "c", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                    self.view.itemsize,
 *                                    flags|PyBUF_C_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_mslice), ((char *)"c"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_C_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 636, __pyx_L1_error)
  __pyx_v_mslice = __pyx_t_1;

  /* "View.MemoryView":641
 *                                    self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &mslice)             # <<<<<<<<<<<<<<
 * 
 *     def copy_fortran(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_mslice)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 641, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":631
 *         return slice_is_contig(mslice[0], 'F', self.view.ndim)
 * 
 *     def copy(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice mslice
 *         cdef int flags = self.flags & ~PyBUF_F_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

/* Python wrapper */
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_memoryview_copy_fortran(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("copy_fortran (wrapper)", 0);
  __pyx_r = __pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_memoryview___pyx_pf_15View_dot_MemoryView_10memoryview_22copy_fortran(struct __pyx_memoryview_obj *__pyx_v_self) {
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  int __pyx_v_flags;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_memviewslice __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("copy_fortran", 0);

  /* "View.MemoryView":645
 *     def copy_fortran(self):
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS             # <<<<<<<<<<<<<<
 * 
 *         slice_copy(self, &src)
 */
  __pyx_v_flags = (__pyx_v_self->flags & (~PyBUF_C_CONTIGUOUS));

  /* "View.MemoryView":647
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 * 
 *         slice_copy(self, &src)             # <<<<<<<<<<<<<<
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,
 *                                 self.view.itemsize,
 */
  __pyx_memoryview_slice_copy(__pyx_v_self, (&__pyx_v_src));

  /* "View.MemoryView":648
 * 
 *         slice_copy(self, &src)
 *         dst = slice_copy_contig(&src, "fortran", self.view.ndim,             # <<<<<<<<<<<<<<
 *                                 self.view.itemsize,
 *                                 flags|PyBUF_F_CONTIGUOUS,
 */
  __pyx_t_1 = __pyx_memoryview_copy_new_contig((&__pyx_v_src), ((char *)"fortran"), __pyx_v_self->view.ndim, __pyx_v_self->view.itemsize, (__pyx_v_flags | PyBUF_F_CONTIGUOUS), __pyx_v_self->dtype_is_object); if (unlikely(PyErr_Occurred())) __PYX_ERR(1, 648, __pyx_L1_error)
  __pyx_v_dst = __pyx_t_1;

  /* "View.MemoryView":653
 *                                 self.dtype_is_object)
 * 
 *         return memoryview_copy_from_slice(self, &dst)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_2 = __pyx_memoryview_copy_object_from_slice(__pyx_v_self, (&__pyx_v_dst)); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 653, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_r = __pyx_t_2;
  __pyx_t_2 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":643
 *         return memoryview_copy_from_slice(self, &mslice)
 * 
 *     def copy_fortran(self):             # <<<<<<<<<<<<<<
 *         cdef __Pyx_memviewslice src, dst
 *         cdef int flags = self.flags & ~PyBUF_C_CONTIGUOUS
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView.memoryview.copy_fortran", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview___reduce_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__24, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryview_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryview_2__setstate_cython__(((struct __pyx_memoryview_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryview_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryview_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

static PyObject *__pyx_memoryview_new(PyObject *__pyx_v_o, int __pyx_v_flags, int __pyx_v_dtype_is_object, __Pyx_TypeInfo *__pyx_v_typeinfo) {
  struct __pyx_memoryview_obj *__pyx_v_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_cwrapper", 0);

  /* "View.MemoryView":658
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)             # <<<<<<<<<<<<<<
 *     result.typeinfo = typeinfo
 *     return result
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_flags); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_v_o);
  __Pyx_GIVEREF(__pyx_v_o);
  PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_o);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryview_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 658, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryview_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":659
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo             # <<<<<<<<<<<<<<
 *     return result
 * 
 */
  __pyx_v_result->typeinfo = __pyx_v_typeinfo;

  /* "View.MemoryView":660
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_check')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":657
 * 
 * @cname('__pyx_memoryview_new')
 * cdef memoryview_cwrapper(object o, int flags, bint dtype_is_object, __Pyx_TypeInfo *typeinfo):             # <<<<<<<<<<<<<<
 *     cdef memoryview result = memoryview(o, flags, dtype_is_object)
 *     result.typeinfo = typeinfo
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_cwrapper", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

static CYTHON_INLINE int __pyx_memoryview_check(PyObject *__pyx_v_o) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("memoryview_check", 0);

  /* "View.MemoryView":664
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):
 *     return isinstance(o, memoryview)             # <<<<<<<<<<<<<<
 * 
 * cdef tuple _unellipsify(object index, int ndim):
 */
  __pyx_t_1 = __Pyx_TypeCheck(__pyx_v_o, __pyx_memoryview_type); 
  __pyx_r = __pyx_t_1;
  goto __pyx_L0;

  /* "View.MemoryView":663
 * 
 * @cname('__pyx_memoryview_check')
 * cdef inline bint memoryview_check(object o):             # <<<<<<<<<<<<<<
 *     return isinstance(o, memoryview)
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

static PyObject *_unellipsify(PyObject *__pyx_v_index, int __pyx_v_ndim) {
  PyObject *__pyx_v_tup = NULL;
  PyObject *__pyx_v_result = NULL;
  int __pyx_v_have_slices;
  int __pyx_v_seen_ellipsis;
  CYTHON_UNUSED PyObject *__pyx_v_idx = NULL;
  PyObject *__pyx_v_item = NULL;
  Py_ssize_t __pyx_v_nslices;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  Py_ssize_t __pyx_t_8;
  int __pyx_t_9;
  int __pyx_t_10;
  PyObject *__pyx_t_11 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("_unellipsify", 0);

  /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
  __pyx_t_1 = PyTuple_Check(__pyx_v_index); 
  __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":672
 *     """
 *     if not isinstance(index, tuple):
 *         tup = (index,)             # <<<<<<<<<<<<<<
 *     else:
 *         tup = index
 */
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 672, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_index);
    __Pyx_GIVEREF(__pyx_v_index);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_v_index);
    __pyx_v_tup = __pyx_t_3;
    __pyx_t_3 = 0;

    /* "View.MemoryView":671
 *     full slices.
 *     """
 *     if not isinstance(index, tuple):             # <<<<<<<<<<<<<<
 *         tup = (index,)
 *     else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":674
 *         tup = (index,)
 *     else:
 *         tup = index             # <<<<<<<<<<<<<<
 * 
 *     result = []
 */
  /*else*/ {
    __Pyx_INCREF(__pyx_v_index);
    __pyx_v_tup = __pyx_v_index;
  }
  __pyx_L3:;

  /* "View.MemoryView":676
 *         tup = index
 * 
 *     result = []             # <<<<<<<<<<<<<<
 *     have_slices = False
 *     seen_ellipsis = False
 */
  __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 676, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_v_result = ((PyObject*)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "View.MemoryView":677
 * 
 *     result = []
 *     have_slices = False             # <<<<<<<<<<<<<<
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 */
  __pyx_v_have_slices = 0;

  /* "View.MemoryView":678
 *     result = []
 *     have_slices = False
 *     seen_ellipsis = False             # <<<<<<<<<<<<<<
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 */
  __pyx_v_seen_ellipsis = 0;

  /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  __Pyx_INCREF(__pyx_int_0);
  __pyx_t_3 = __pyx_int_0;
  if (likely(PyList_CheckExact(__pyx_v_tup)) || PyTuple_CheckExact(__pyx_v_tup)) {
    __pyx_t_4 = __pyx_v_tup; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_v_tup); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 679, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_4))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_7 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_7); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(1, 679, __pyx_L1_error)
        #else
        __pyx_t_7 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
      }
    } else {
      __pyx_t_7 = __pyx_t_6(__pyx_t_4);
      if (unlikely(!__pyx_t_7)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 679, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_7);
    }
    __Pyx_XDECREF_SET(__pyx_v_item, __pyx_t_7);
    __pyx_t_7 = 0;
    __Pyx_INCREF(__pyx_t_3);
    __Pyx_XDECREF_SET(__pyx_v_idx, __pyx_t_3);
    __pyx_t_7 = __Pyx_PyInt_AddObjC(__pyx_t_3, __pyx_int_1, 1, 0, 0); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 679, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_3);
    __pyx_t_3 = __pyx_t_7;
    __pyx_t_7 = 0;

    /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
    __pyx_t_2 = (__pyx_v_item == __pyx_builtin_Ellipsis);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
      __pyx_t_1 = ((!(__pyx_v_seen_ellipsis != 0)) != 0);
      if (__pyx_t_1) {

        /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
        __pyx_t_8 = PyObject_Length(__pyx_v_tup); if (unlikely(__pyx_t_8 == ((Py_ssize_t)-1))) __PYX_ERR(1, 682, __pyx_L1_error)
        __pyx_t_7 = PyList_New(1 * ((((__pyx_v_ndim - __pyx_t_8) + 1)<0) ? 0:((__pyx_v_ndim - __pyx_t_8) + 1))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 682, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        { Py_ssize_t __pyx_temp;
          for (__pyx_temp=0; __pyx_temp < ((__pyx_v_ndim - __pyx_t_8) + 1); __pyx_temp++) {
            __Pyx_INCREF(__pyx_slice__26);
            __Pyx_GIVEREF(__pyx_slice__26);
            PyList_SET_ITEM(__pyx_t_7, __pyx_temp, __pyx_slice__26);
          }
        }
        __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_7); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 682, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

        /* "View.MemoryView":683
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True             # <<<<<<<<<<<<<<
 *             else:
 *                 result.append(slice(None))
 */
        __pyx_v_seen_ellipsis = 1;

        /* "View.MemoryView":681
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:
 *             if not seen_ellipsis:             # <<<<<<<<<<<<<<
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 *                 seen_ellipsis = True
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":685
 *                 seen_ellipsis = True
 *             else:
 *                 result.append(slice(None))             # <<<<<<<<<<<<<<
 *             have_slices = True
 *         else:
 */
      /*else*/ {
        __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_slice__26); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 685, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":686
 *             else:
 *                 result.append(slice(None))
 *             have_slices = True             # <<<<<<<<<<<<<<
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 */
      __pyx_v_have_slices = 1;

      /* "View.MemoryView":680
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):
 *         if item is Ellipsis:             # <<<<<<<<<<<<<<
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
    /*else*/ {
      __pyx_t_2 = PySlice_Check(__pyx_v_item); 
      __pyx_t_10 = ((!(__pyx_t_2 != 0)) != 0);
      if (__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = ((!(PyIndex_Check(__pyx_v_item) != 0)) != 0);
      __pyx_t_1 = __pyx_t_10;
      __pyx_L9_bool_binop_done:;
      if (unlikely(__pyx_t_1)) {

        /* "View.MemoryView":689
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):
 *                 raise TypeError("Cannot index with type '%s'" % type(item))             # <<<<<<<<<<<<<<
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 */
        __pyx_t_7 = __Pyx_PyString_FormatSafe(__pyx_kp_s_Cannot_index_with_type_s, ((PyObject *)Py_TYPE(__pyx_v_item))); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_7);
        __pyx_t_11 = __Pyx_PyObject_CallOneArg(__pyx_builtin_TypeError, __pyx_t_7); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 689, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_Raise(__pyx_t_11, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __PYX_ERR(1, 689, __pyx_L1_error)

        /* "View.MemoryView":688
 *             have_slices = True
 *         else:
 *             if not isinstance(item, slice) and not PyIndex_Check(item):             # <<<<<<<<<<<<<<
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 */
      }

      /* "View.MemoryView":691
 *                 raise TypeError("Cannot index with type '%s'" % type(item))
 * 
 *             have_slices = have_slices or isinstance(item, slice)             # <<<<<<<<<<<<<<
 *             result.append(item)
 * 
 */
      __pyx_t_10 = (__pyx_v_have_slices != 0);
      if (!__pyx_t_10) {
      } else {
        __pyx_t_1 = __pyx_t_10;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = PySlice_Check(__pyx_v_item); 
      __pyx_t_2 = (__pyx_t_10 != 0);
      __pyx_t_1 = __pyx_t_2;
      __pyx_L11_bool_binop_done:;
      __pyx_v_have_slices = __pyx_t_1;

      /* "View.MemoryView":692
 * 
 *             have_slices = have_slices or isinstance(item, slice)
 *             result.append(item)             # <<<<<<<<<<<<<<
 * 
 *     nslices = ndim - len(result)
 */
      __pyx_t_9 = __Pyx_PyList_Append(__pyx_v_result, __pyx_v_item); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 692, __pyx_L1_error)
    }
    __pyx_L6:;

    /* "View.MemoryView":679
 *     have_slices = False
 *     seen_ellipsis = False
 *     for idx, item in enumerate(tup):             # <<<<<<<<<<<<<<
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 */
  }
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":694
 *             result.append(item)
 * 
 *     nslices = ndim - len(result)             # <<<<<<<<<<<<<<
 *     if nslices:
 *         result.extend([slice(None)] * nslices)
 */
  __pyx_t_5 = PyList_GET_SIZE(__pyx_v_result); if (unlikely(__pyx_t_5 == ((Py_ssize_t)-1))) __PYX_ERR(1, 694, __pyx_L1_error)
  __pyx_v_nslices = (__pyx_v_ndim - __pyx_t_5);

  /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  __pyx_t_1 = (__pyx_v_nslices != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":696
 *     nslices = ndim - len(result)
 *     if nslices:
 *         result.extend([slice(None)] * nslices)             # <<<<<<<<<<<<<<
 * 
 *     return have_slices or nslices, tuple(result)
 */
    __pyx_t_3 = PyList_New(1 * ((__pyx_v_nslices<0) ? 0:__pyx_v_nslices)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 696, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    { Py_ssize_t __pyx_temp;
      for (__pyx_temp=0; __pyx_temp < __pyx_v_nslices; __pyx_temp++) {
        __Pyx_INCREF(__pyx_slice__26);
        __Pyx_GIVEREF(__pyx_slice__26);
        PyList_SET_ITEM(__pyx_t_3, __pyx_temp, __pyx_slice__26);
      }
    }
    __pyx_t_9 = __Pyx_PyList_Extend(__pyx_v_result, __pyx_t_3); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 696, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "View.MemoryView":695
 * 
 *     nslices = ndim - len(result)
 *     if nslices:             # <<<<<<<<<<<<<<
 *         result.extend([slice(None)] * nslices)
 * 
 */
  }

  /* "View.MemoryView":698
 *         result.extend([slice(None)] * nslices)
 * 
 *     return have_slices or nslices, tuple(result)             # <<<<<<<<<<<<<<
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 */
  __Pyx_XDECREF(__pyx_r);
  if (!__pyx_v_have_slices) {
  } else {
    __pyx_t_4 = __Pyx_PyBool_FromLong(__pyx_v_have_slices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L14_bool_binop_done;
  }
  __pyx_t_4 = PyInt_FromSsize_t(__pyx_v_nslices); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_3 = __pyx_t_4;
  __pyx_t_4 = 0;
  __pyx_L14_bool_binop_done:;
  __pyx_t_4 = PyList_AsTuple(__pyx_v_result); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_11 = PyTuple_New(2); if (unlikely(!__pyx_t_11)) __PYX_ERR(1, 698, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_11);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_11, 0, __pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_4);
  PyTuple_SET_ITEM(__pyx_t_11, 1, __pyx_t_4);
  __pyx_t_3 = 0;
  __pyx_t_4 = 0;
  __pyx_r = ((PyObject*)__pyx_t_11);
  __pyx_t_11 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":666
 *     return isinstance(o, memoryview)
 * 
 * cdef tuple _unellipsify(object index, int ndim):             # <<<<<<<<<<<<<<
 *     """
 *     Replace all ellipses with full slices and fill incomplete indices with
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("View.MemoryView._unellipsify", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_tup);
  __Pyx_XDECREF(__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_idx);
  __Pyx_XDECREF(__pyx_v_item);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

static PyObject *assert_direct_dimensions(Py_ssize_t *__pyx_v_suboffsets, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assert_direct_dimensions", 0);

  /* "View.MemoryView":701
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")
 */
  __pyx_t_2 = (__pyx_v_suboffsets + __pyx_v_ndim);
  for (__pyx_t_3 = __pyx_v_suboffsets; __pyx_t_3 < __pyx_t_2; __pyx_t_3++) {
    __pyx_t_1 = __pyx_t_3;
    __pyx_v_suboffset = (__pyx_t_1[0]);

    /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    __pyx_t_4 = ((__pyx_v_suboffset >= 0) != 0);
    if (unlikely(__pyx_t_4)) {

      /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 703, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_Raise(__pyx_t_5, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __PYX_ERR(1, 703, __pyx_L1_error)

      /* "View.MemoryView":702
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             raise ValueError("Indirect dimensions not supported")
 * 
 */
    }
  }

  /* "View.MemoryView":700
 *     return have_slices or nslices, tuple(result)
 * 
 * cdef assert_direct_dimensions(Py_ssize_t *suboffsets, int ndim):             # <<<<<<<<<<<<<<
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.assert_direct_dimensions", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

static struct __pyx_memoryview_obj *__pyx_memview_slice(struct __pyx_memoryview_obj *__pyx_v_memview, PyObject *__pyx_v_indices) {
  int __pyx_v_new_ndim;
  int __pyx_v_suboffset_dim;
  int __pyx_v_dim;
  __Pyx_memviewslice __pyx_v_src;
  __Pyx_memviewslice __pyx_v_dst;
  __Pyx_memviewslice *__pyx_v_p_src;
  struct __pyx_memoryviewslice_obj *__pyx_v_memviewsliceobj = 0;
  __Pyx_memviewslice *__pyx_v_p_dst;
  int *__pyx_v_p_suboffset_dim;
  Py_ssize_t __pyx_v_start;
  Py_ssize_t __pyx_v_stop;
  Py_ssize_t __pyx_v_step;
  int __pyx_v_have_start;
  int __pyx_v_have_stop;
  int __pyx_v_have_step;
  PyObject *__pyx_v_index = NULL;
  struct __pyx_memoryview_obj *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  struct __pyx_memoryview_obj *__pyx_t_4;
  char *__pyx_t_5;
  int __pyx_t_6;
  Py_ssize_t __pyx_t_7;
  PyObject *(*__pyx_t_8)(PyObject *);
  PyObject *__pyx_t_9 = NULL;
  Py_ssize_t __pyx_t_10;
  int __pyx_t_11;
  Py_ssize_t __pyx_t_12;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memview_slice", 0);

  /* "View.MemoryView":711
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim             # <<<<<<<<<<<<<<
 *     cdef bint negative_step
 *     cdef __Pyx_memviewslice src, dst
 */
  __pyx_v_new_ndim = 0;
  __pyx_v_suboffset_dim = -1;

  /* "View.MemoryView":718
 * 
 * 
 *     memset(&dst, 0, sizeof(dst))             # <<<<<<<<<<<<<<
 * 
 *     cdef _memoryviewslice memviewsliceobj
 */
  (void)(memset((&__pyx_v_dst), 0, (sizeof(__pyx_v_dst))));

  /* "View.MemoryView":722
 *     cdef _memoryviewslice memviewsliceobj
 * 
 *     assert memview.view.ndim > 0             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
  #ifndef CYTHON_WITHOUT_ASSERTIONS
  if (unlikely(!Py_OptimizeFlag)) {
    if (unlikely(!((__pyx_v_memview->view.ndim > 0) != 0))) {
      PyErr_SetNone(PyExc_AssertionError);
      __PYX_ERR(1, 722, __pyx_L1_error)
    }
  }
  #endif

  /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":725
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview             # <<<<<<<<<<<<<<
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 725, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_memviewsliceobj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":726
 *     if isinstance(memview, _memoryviewslice):
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, &src)
 */
    __pyx_v_p_src = (&__pyx_v_memviewsliceobj->from_slice);

    /* "View.MemoryView":724
 *     assert memview.view.ndim > 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         memviewsliceobj = memview
 *         p_src = &memviewsliceobj.from_slice
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":728
 *         p_src = &memviewsliceobj.from_slice
 *     else:
 *         slice_copy(memview, &src)             # <<<<<<<<<<<<<<
 *         p_src = &src
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_src));

    /* "View.MemoryView":729
 *     else:
 *         slice_copy(memview, &src)
 *         p_src = &src             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_p_src = (&__pyx_v_src);
  }
  __pyx_L3:;

  /* "View.MemoryView":735
 * 
 * 
 *     dst.memview = p_src.memview             # <<<<<<<<<<<<<<
 *     dst.data = p_src.data
 * 
 */
  __pyx_t_4 = __pyx_v_p_src->memview;
  __pyx_v_dst.memview = __pyx_t_4;

  /* "View.MemoryView":736
 * 
 *     dst.memview = p_src.memview
 *     dst.data = p_src.data             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_v_p_src->data;
  __pyx_v_dst.data = __pyx_t_5;

  /* "View.MemoryView":741
 * 
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst             # <<<<<<<<<<<<<<
 *     cdef int *p_suboffset_dim = &suboffset_dim
 *     cdef Py_ssize_t start, stop, step
 */
  __pyx_v_p_dst = (&__pyx_v_dst);

  /* "View.MemoryView":742
 * 
 *     cdef __Pyx_memviewslice *p_dst = &dst
 *     cdef int *p_suboffset_dim = &suboffset_dim             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t start, stop, step
 *     cdef bint have_start, have_stop, have_step
 */
  __pyx_v_p_suboffset_dim = (&__pyx_v_suboffset_dim);

  /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  __pyx_t_6 = 0;
  if (likely(PyList_CheckExact(__pyx_v_indices)) || PyTuple_CheckExact(__pyx_v_indices)) {
    __pyx_t_3 = __pyx_v_indices; __Pyx_INCREF(__pyx_t_3); __pyx_t_7 = 0;
    __pyx_t_8 = NULL;
  } else {
    __pyx_t_7 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_v_indices); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 746, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 746, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_8)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_7 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      } else {
        if (__pyx_t_7 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_9 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_7); __Pyx_INCREF(__pyx_t_9); __pyx_t_7++; if (unlikely(0 < 0)) __PYX_ERR(1, 746, __pyx_L1_error)
        #else
        __pyx_t_9 = PySequence_ITEM(__pyx_t_3, __pyx_t_7); __pyx_t_7++; if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 746, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_9);
        #endif
      }
    } else {
      __pyx_t_9 = __pyx_t_8(__pyx_t_3);
      if (unlikely(!__pyx_t_9)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(1, 746, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_9);
    }
    __Pyx_XDECREF_SET(__pyx_v_index, __pyx_t_9);
    __pyx_t_9 = 0;
    __pyx_v_dim = __pyx_t_6;
    __pyx_t_6 = (__pyx_t_6 + 1);

    /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
    __pyx_t_2 = (PyIndex_Check(__pyx_v_index) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":751
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 *                 index, 0, 0, # start, stop, step             # <<<<<<<<<<<<<<
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 */
      __pyx_t_10 = __Pyx_PyIndex_AsSsize_t(__pyx_v_index); if (unlikely((__pyx_t_10 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 751, __pyx_L1_error)

      /* "View.MemoryView":748
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_t_10, 0, 0, 0, 0, 0, 0); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 748, __pyx_L1_error)

      /* "View.MemoryView":747
 * 
 *     for dim, index in enumerate(indices):
 *         if PyIndex_Check(index):             # <<<<<<<<<<<<<<
 *             slice_memviewslice(
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
    __pyx_t_2 = (__pyx_v_index == Py_None);
    __pyx_t_1 = (__pyx_t_2 != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":755
 *                 False)
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1             # <<<<<<<<<<<<<<
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 */
      (__pyx_v_p_dst->shape[__pyx_v_new_ndim]) = 1;

      /* "View.MemoryView":756
 *         elif index is None:
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0             # <<<<<<<<<<<<<<
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1
 */
      (__pyx_v_p_dst->strides[__pyx_v_new_ndim]) = 0;

      /* "View.MemoryView":757
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1             # <<<<<<<<<<<<<<
 *             new_ndim += 1
 *         else:
 */
      (__pyx_v_p_dst->suboffsets[__pyx_v_new_ndim]) = -1L;

      /* "View.MemoryView":758
 *             p_dst.strides[new_ndim] = 0
 *             p_dst.suboffsets[new_ndim] = -1
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 *         else:
 *             start = index.start or 0
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);

      /* "View.MemoryView":754
 *                 0, 0, 0, # have_{start,stop,step}
 *                 False)
 *         elif index is None:             # <<<<<<<<<<<<<<
 *             p_dst.shape[new_ndim] = 1
 *             p_dst.strides[new_ndim] = 0
 */
      goto __pyx_L6;
    }

    /* "View.MemoryView":760
 *             new_ndim += 1
 *         else:
 *             start = index.start or 0             # <<<<<<<<<<<<<<
 *             stop = index.stop or 0
 *             step = index.step or 0
 */
    /*else*/ {
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 760, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 760, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 760, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L7_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L7_bool_binop_done:;
      __pyx_v_start = __pyx_t_10;

      /* "View.MemoryView":761
 *         else:
 *             start = index.start or 0
 *             stop = index.stop or 0             # <<<<<<<<<<<<<<
 *             step = index.step or 0
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 761, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 761, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 761, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L9_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L9_bool_binop_done:;
      __pyx_v_stop = __pyx_t_10;

      /* "View.MemoryView":762
 *             start = index.start or 0
 *             stop = index.stop or 0
 *             step = index.step or 0             # <<<<<<<<<<<<<<
 * 
 *             have_start = index.start is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 762, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_9); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(1, 762, __pyx_L1_error)
      if (!__pyx_t_1) {
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      } else {
        __pyx_t_12 = __Pyx_PyIndex_AsSsize_t(__pyx_t_9); if (unlikely((__pyx_t_12 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 762, __pyx_L1_error)
        __pyx_t_10 = __pyx_t_12;
        __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
        goto __pyx_L11_bool_binop_done;
      }
      __pyx_t_10 = 0;
      __pyx_L11_bool_binop_done:;
      __pyx_v_step = __pyx_t_10;

      /* "View.MemoryView":764
 *             step = index.step or 0
 * 
 *             have_start = index.start is not None             # <<<<<<<<<<<<<<
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_start); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 764, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_start = __pyx_t_1;

      /* "View.MemoryView":765
 * 
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None             # <<<<<<<<<<<<<<
 *             have_step = index.step is not None
 * 
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_stop); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 765, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_stop = __pyx_t_1;

      /* "View.MemoryView":766
 *             have_start = index.start is not None
 *             have_stop = index.stop is not None
 *             have_step = index.step is not None             # <<<<<<<<<<<<<<
 * 
 *             slice_memviewslice(
 */
      __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_v_index, __pyx_n_s_step); if (unlikely(!__pyx_t_9)) __PYX_ERR(1, 766, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      __pyx_t_1 = (__pyx_t_9 != Py_None);
      __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_v_have_step = __pyx_t_1;

      /* "View.MemoryView":768
 *             have_step = index.step is not None
 * 
 *             slice_memviewslice(             # <<<<<<<<<<<<<<
 *                 p_dst, p_src.shape[dim], p_src.strides[dim], p_src.suboffsets[dim],
 *                 dim, new_ndim, p_suboffset_dim,
 */
      __pyx_t_11 = __pyx_memoryview_slice_memviewslice(__pyx_v_p_dst, (__pyx_v_p_src->shape[__pyx_v_dim]), (__pyx_v_p_src->strides[__pyx_v_dim]), (__pyx_v_p_src->suboffsets[__pyx_v_dim]), __pyx_v_dim, __pyx_v_new_ndim, __pyx_v_p_suboffset_dim, __pyx_v_start, __pyx_v_stop, __pyx_v_step, __pyx_v_have_start, __pyx_v_have_stop, __pyx_v_have_step, 1); if (unlikely(__pyx_t_11 == ((int)-1))) __PYX_ERR(1, 768, __pyx_L1_error)

      /* "View.MemoryView":774
 *                 have_start, have_stop, have_step,
 *                 True)
 *             new_ndim += 1             # <<<<<<<<<<<<<<
 * 
 *     if isinstance(memview, _memoryviewslice):
 */
      __pyx_v_new_ndim = (__pyx_v_new_ndim + 1);
    }
    __pyx_L6:;

    /* "View.MemoryView":746
 *     cdef bint have_start, have_stop, have_step
 * 
 *     for dim, index in enumerate(indices):             # <<<<<<<<<<<<<<
 *         if PyIndex_Check(index):
 *             slice_memviewslice(
 */
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":778
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_dtype_func,
 *                                     memview.dtype_is_object)
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 778, __pyx_L1_error) }

    /* "View.MemoryView":779
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 *     else:
 */
    if (unlikely(!__pyx_v_memviewsliceobj)) { __Pyx_RaiseUnboundLocalError("memviewsliceobj"); __PYX_ERR(1, 779, __pyx_L1_error) }

    /* "View.MemoryView":777
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         return memoryview_fromslice(dst, new_ndim,             # <<<<<<<<<<<<<<
 *                                     memviewsliceobj.to_object_func,
 *                                     memviewsliceobj.to_dtype_func,
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, __pyx_v_memviewsliceobj->to_object_func, __pyx_v_memviewsliceobj->to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 777, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 777, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":776
 *             new_ndim += 1
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         return memoryview_fromslice(dst, new_ndim,
 *                                     memviewsliceobj.to_object_func,
 */
  }

  /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(((PyObject *)__pyx_r));

    /* "View.MemoryView":783
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,
 *                                     memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_fromslice(__pyx_v_dst, __pyx_v_new_ndim, NULL, NULL, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 782, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);

    /* "View.MemoryView":782
 *                                     memview.dtype_is_object)
 *     else:
 *         return memoryview_fromslice(dst, new_ndim, NULL, NULL,             # <<<<<<<<<<<<<<
 *                                     memview.dtype_is_object)
 * 
 */
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_memoryview_type))))) __PYX_ERR(1, 782, __pyx_L1_error)
    __pyx_r = ((struct __pyx_memoryview_obj *)__pyx_t_3);
    __pyx_t_3 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":710
 * 
 * @cname('__pyx_memview_slice')
 * cdef memoryview memview_slice(memoryview memview, object indices):             # <<<<<<<<<<<<<<
 *     cdef int new_ndim = 0, suboffset_dim = -1, dim
 *     cdef bint negative_step
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_AddTraceback("View.MemoryView.memview_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_memviewsliceobj);
  __Pyx_XDECREF(__pyx_v_index);
  __Pyx_XGIVEREF((PyObject *)__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

static int __pyx_memoryview_slice_memviewslice(__Pyx_memviewslice *__pyx_v_dst, Py_ssize_t __pyx_v_shape, Py_ssize_t __pyx_v_stride, Py_ssize_t __pyx_v_suboffset, int __pyx_v_dim, int __pyx_v_new_ndim, int *__pyx_v_suboffset_dim, Py_ssize_t __pyx_v_start, Py_ssize_t __pyx_v_stop, Py_ssize_t __pyx_v_step, int __pyx_v_have_start, int __pyx_v_have_stop, int __pyx_v_have_step, int __pyx_v_is_slice) {
  Py_ssize_t __pyx_v_new_shape;
  int __pyx_v_negative_step;
  int __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
  __pyx_t_1 = ((!(__pyx_v_is_slice != 0)) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    __pyx_t_1 = ((__pyx_v_start < 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":830
 * 
 *         if start < 0:
 *             start += shape             # <<<<<<<<<<<<<<
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 */
      __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

      /* "View.MemoryView":829
 *     if not is_slice:
 * 
 *         if start < 0:             # <<<<<<<<<<<<<<
 *             start += shape
 *         if not 0 <= start < shape:
 */
    }

    /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    __pyx_t_1 = (0 <= __pyx_v_start);
    if (__pyx_t_1) {
      __pyx_t_1 = (__pyx_v_start < __pyx_v_shape);
    }
    __pyx_t_2 = ((!(__pyx_t_1 != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":832
 *             start += shape
 *         if not 0 <= start < shape:
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)             # <<<<<<<<<<<<<<
 *     else:
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"Index out of bounds (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 832, __pyx_L1_error)

      /* "View.MemoryView":831
 *         if start < 0:
 *             start += shape
 *         if not 0 <= start < shape:             # <<<<<<<<<<<<<<
 *             _err_dim(IndexError, "Index out of bounds (axis %d)", dim)
 *     else:
 */
    }

    /* "View.MemoryView":827
 *     cdef bint negative_step
 * 
 *     if not is_slice:             # <<<<<<<<<<<<<<
 * 
 *         if start < 0:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":835
 *     else:
 * 
 *         negative_step = have_step != 0 and step < 0             # <<<<<<<<<<<<<<
 * 
 *         if have_step and step == 0:
 */
  /*else*/ {
    __pyx_t_1 = ((__pyx_v_have_step != 0) != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step < 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L6_bool_binop_done:;
    __pyx_v_negative_step = __pyx_t_2;

    /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    __pyx_t_1 = (__pyx_v_have_step != 0);
    if (__pyx_t_1) {
    } else {
      __pyx_t_2 = __pyx_t_1;
      goto __pyx_L9_bool_binop_done;
    }
    __pyx_t_1 = ((__pyx_v_step == 0) != 0);
    __pyx_t_2 = __pyx_t_1;
    __pyx_L9_bool_binop_done:;
    if (__pyx_t_2) {

      /* "View.MemoryView":838
 * 
 *         if have_step and step == 0:
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Step may not be zero (axis %d)"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 838, __pyx_L1_error)

      /* "View.MemoryView":837
 *         negative_step = have_step != 0 and step < 0
 * 
 *         if have_step and step == 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Step may not be zero (axis %d)", dim)
 * 
 */
    }

    /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
    __pyx_t_2 = (__pyx_v_have_start != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
      __pyx_t_2 = ((__pyx_v_start < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":843
 *         if have_start:
 *             if start < 0:
 *                 start += shape             # <<<<<<<<<<<<<<
 *                 if start < 0:
 *                     start = 0
 */
        __pyx_v_start = (__pyx_v_start + __pyx_v_shape);

        /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        __pyx_t_2 = ((__pyx_v_start < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":845
 *                 start += shape
 *                 if start < 0:
 *                     start = 0             # <<<<<<<<<<<<<<
 *             elif start >= shape:
 *                 if negative_step:
 */
          __pyx_v_start = 0;

          /* "View.MemoryView":844
 *             if start < 0:
 *                 start += shape
 *                 if start < 0:             # <<<<<<<<<<<<<<
 *                     start = 0
 *             elif start >= shape:
 */
        }

        /* "View.MemoryView":842
 * 
 *         if have_start:
 *             if start < 0:             # <<<<<<<<<<<<<<
 *                 start += shape
 *                 if start < 0:
 */
        goto __pyx_L12;
      }

      /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      __pyx_t_2 = ((__pyx_v_start >= __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
        __pyx_t_2 = (__pyx_v_negative_step != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":848
 *             elif start >= shape:
 *                 if negative_step:
 *                     start = shape - 1             # <<<<<<<<<<<<<<
 *                 else:
 *                     start = shape
 */
          __pyx_v_start = (__pyx_v_shape - 1);

          /* "View.MemoryView":847
 *                     start = 0
 *             elif start >= shape:
 *                 if negative_step:             # <<<<<<<<<<<<<<
 *                     start = shape - 1
 *                 else:
 */
          goto __pyx_L14;
        }

        /* "View.MemoryView":850
 *                     start = shape - 1
 *                 else:
 *                     start = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        /*else*/ {
          __pyx_v_start = __pyx_v_shape;
        }
        __pyx_L14:;

        /* "View.MemoryView":846
 *                 if start < 0:
 *                     start = 0
 *             elif start >= shape:             # <<<<<<<<<<<<<<
 *                 if negative_step:
 *                     start = shape - 1
 */
      }
      __pyx_L12:;

      /* "View.MemoryView":841
 * 
 * 
 *         if have_start:             # <<<<<<<<<<<<<<
 *             if start < 0:
 *                 start += shape
 */
      goto __pyx_L11;
    }

    /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":853
 *         else:
 *             if negative_step:
 *                 start = shape - 1             # <<<<<<<<<<<<<<
 *             else:
 *                 start = 0
 */
        __pyx_v_start = (__pyx_v_shape - 1);

        /* "View.MemoryView":852
 *                     start = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 start = shape - 1
 *             else:
 */
        goto __pyx_L15;
      }

      /* "View.MemoryView":855
 *                 start = shape - 1
 *             else:
 *                 start = 0             # <<<<<<<<<<<<<<
 * 
 *         if have_stop:
 */
      /*else*/ {
        __pyx_v_start = 0;
      }
      __pyx_L15:;
    }
    __pyx_L11:;

    /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
    __pyx_t_2 = (__pyx_v_have_stop != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
      __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":859
 *         if have_stop:
 *             if stop < 0:
 *                 stop += shape             # <<<<<<<<<<<<<<
 *                 if stop < 0:
 *                     stop = 0
 */
        __pyx_v_stop = (__pyx_v_stop + __pyx_v_shape);

        /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        __pyx_t_2 = ((__pyx_v_stop < 0) != 0);
        if (__pyx_t_2) {

          /* "View.MemoryView":861
 *                 stop += shape
 *                 if stop < 0:
 *                     stop = 0             # <<<<<<<<<<<<<<
 *             elif stop > shape:
 *                 stop = shape
 */
          __pyx_v_stop = 0;

          /* "View.MemoryView":860
 *             if stop < 0:
 *                 stop += shape
 *                 if stop < 0:             # <<<<<<<<<<<<<<
 *                     stop = 0
 *             elif stop > shape:
 */
        }

        /* "View.MemoryView":858
 * 
 *         if have_stop:
 *             if stop < 0:             # <<<<<<<<<<<<<<
 *                 stop += shape
 *                 if stop < 0:
 */
        goto __pyx_L17;
      }

      /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      __pyx_t_2 = ((__pyx_v_stop > __pyx_v_shape) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":863
 *                     stop = 0
 *             elif stop > shape:
 *                 stop = shape             # <<<<<<<<<<<<<<
 *         else:
 *             if negative_step:
 */
        __pyx_v_stop = __pyx_v_shape;

        /* "View.MemoryView":862
 *                 if stop < 0:
 *                     stop = 0
 *             elif stop > shape:             # <<<<<<<<<<<<<<
 *                 stop = shape
 *         else:
 */
      }
      __pyx_L17:;

      /* "View.MemoryView":857
 *                 start = 0
 * 
 *         if have_stop:             # <<<<<<<<<<<<<<
 *             if stop < 0:
 *                 stop += shape
 */
      goto __pyx_L16;
    }

    /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
    /*else*/ {
      __pyx_t_2 = (__pyx_v_negative_step != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":866
 *         else:
 *             if negative_step:
 *                 stop = -1             # <<<<<<<<<<<<<<
 *             else:
 *                 stop = shape
 */
        __pyx_v_stop = -1L;

        /* "View.MemoryView":865
 *                 stop = shape
 *         else:
 *             if negative_step:             # <<<<<<<<<<<<<<
 *                 stop = -1
 *             else:
 */
        goto __pyx_L19;
      }

      /* "View.MemoryView":868
 *                 stop = -1
 *             else:
 *                 stop = shape             # <<<<<<<<<<<<<<
 * 
 *         if not have_step:
 */
      /*else*/ {
        __pyx_v_stop = __pyx_v_shape;
      }
      __pyx_L19:;
    }
    __pyx_L16:;

    /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    __pyx_t_2 = ((!(__pyx_v_have_step != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":871
 * 
 *         if not have_step:
 *             step = 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_step = 1;

      /* "View.MemoryView":870
 *                 stop = shape
 * 
 *         if not have_step:             # <<<<<<<<<<<<<<
 *             step = 1
 * 
 */
    }

    /* "View.MemoryView":875
 * 
 *         with cython.cdivision(True):
 *             new_shape = (stop - start) // step             # <<<<<<<<<<<<<<
 * 
 *             if (stop - start) - step * new_shape:
 */
    __pyx_v_new_shape = ((__pyx_v_stop - __pyx_v_start) / __pyx_v_step);

    /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    __pyx_t_2 = (((__pyx_v_stop - __pyx_v_start) - (__pyx_v_step * __pyx_v_new_shape)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":878
 * 
 *             if (stop - start) - step * new_shape:
 *                 new_shape += 1             # <<<<<<<<<<<<<<
 * 
 *         if new_shape < 0:
 */
      __pyx_v_new_shape = (__pyx_v_new_shape + 1);

      /* "View.MemoryView":877
 *             new_shape = (stop - start) // step
 * 
 *             if (stop - start) - step * new_shape:             # <<<<<<<<<<<<<<
 *                 new_shape += 1
 * 
 */
    }

    /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    __pyx_t_2 = ((__pyx_v_new_shape < 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":881
 * 
 *         if new_shape < 0:
 *             new_shape = 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_new_shape = 0;

      /* "View.MemoryView":880
 *                 new_shape += 1
 * 
 *         if new_shape < 0:             # <<<<<<<<<<<<<<
 *             new_shape = 0
 * 
 */
    }

    /* "View.MemoryView":884
 * 
 * 
 *         dst.strides[new_ndim] = stride * step             # <<<<<<<<<<<<<<
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset
 */
    (__pyx_v_dst->strides[__pyx_v_new_ndim]) = (__pyx_v_stride * __pyx_v_step);

    /* "View.MemoryView":885
 * 
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape             # <<<<<<<<<<<<<<
 *         dst.suboffsets[new_ndim] = suboffset
 * 
 */
    (__pyx_v_dst->shape[__pyx_v_new_ndim]) = __pyx_v_new_shape;

    /* "View.MemoryView":886
 *         dst.strides[new_ndim] = stride * step
 *         dst.shape[new_ndim] = new_shape
 *         dst.suboffsets[new_ndim] = suboffset             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_dst->suboffsets[__pyx_v_new_ndim]) = __pyx_v_suboffset;
  }
  __pyx_L3:;

  /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
  __pyx_t_2 = (((__pyx_v_suboffset_dim[0]) < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":890
 * 
 *     if suboffset_dim[0] < 0:
 *         dst.data += start * stride             # <<<<<<<<<<<<<<
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 */
    __pyx_v_dst->data = (__pyx_v_dst->data + (__pyx_v_start * __pyx_v_stride));

    /* "View.MemoryView":889
 * 
 * 
 *     if suboffset_dim[0] < 0:             # <<<<<<<<<<<<<<
 *         dst.data += start * stride
 *     else:
 */
    goto __pyx_L23;
  }

  /* "View.MemoryView":892
 *         dst.data += start * stride
 *     else:
 *         dst.suboffsets[suboffset_dim[0]] += start * stride             # <<<<<<<<<<<<<<
 * 
 *     if suboffset >= 0:
 */
  /*else*/ {
    __pyx_t_3 = (__pyx_v_suboffset_dim[0]);
    (__pyx_v_dst->suboffsets[__pyx_t_3]) = ((__pyx_v_dst->suboffsets[__pyx_t_3]) + (__pyx_v_start * __pyx_v_stride));
  }
  __pyx_L23:;

  /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
    __pyx_t_2 = ((!(__pyx_v_is_slice != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
      __pyx_t_2 = ((__pyx_v_new_ndim == 0) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":897
 *         if not is_slice:
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 */
        __pyx_v_dst->data = ((((char **)__pyx_v_dst->data)[0]) + __pyx_v_suboffset);

        /* "View.MemoryView":896
 *     if suboffset >= 0:
 *         if not is_slice:
 *             if new_ndim == 0:             # <<<<<<<<<<<<<<
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 */
        goto __pyx_L26;
      }

      /* "View.MemoryView":899
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "             # <<<<<<<<<<<<<<
 *                                      "must be indexed and not sliced", dim)
 *         else:
 */
      /*else*/ {

        /* "View.MemoryView":900
 *             else:
 *                 _err_dim(IndexError, "All dimensions preceding dimension %d "
 *                                      "must be indexed and not sliced", dim)             # <<<<<<<<<<<<<<
 *         else:
 *             suboffset_dim[0] = new_ndim
 */
        __pyx_t_3 = __pyx_memoryview_err_dim(__pyx_builtin_IndexError, ((char *)"All dimensions preceding dimension %d must be indexed and not sliced"), __pyx_v_dim); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 899, __pyx_L1_error)
      }
      __pyx_L26:;

      /* "View.MemoryView":895
 * 
 *     if suboffset >= 0:
 *         if not is_slice:             # <<<<<<<<<<<<<<
 *             if new_ndim == 0:
 *                 dst.data = (<char **> dst.data)[0] + suboffset
 */
      goto __pyx_L25;
    }

    /* "View.MemoryView":902
 *                                      "must be indexed and not sliced", dim)
 *         else:
 *             suboffset_dim[0] = new_ndim             # <<<<<<<<<<<<<<
 * 
 *     return 0
 */
    /*else*/ {
      (__pyx_v_suboffset_dim[0]) = __pyx_v_new_ndim;
    }
    __pyx_L25:;

    /* "View.MemoryView":894
 *         dst.suboffsets[suboffset_dim[0]] += start * stride
 * 
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         if not is_slice:
 *             if new_ndim == 0:
 */
  }

  /* "View.MemoryView":904
 *             suboffset_dim[0] = new_ndim
 * 
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":807
 * 
 * @cname('__pyx_memoryview_slice_memviewslice')
 * cdef int slice_memviewslice(             # <<<<<<<<<<<<<<
 *         __Pyx_memviewslice *dst,
 *         Py_ssize_t shape, Py_ssize_t stride, Py_ssize_t suboffset,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.slice_memviewslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

static char *__pyx_pybuffer_index(Py_buffer *__pyx_v_view, char *__pyx_v_bufp, Py_ssize_t __pyx_v_index, Py_ssize_t __pyx_v_dim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_suboffset;
  Py_ssize_t __pyx_v_itemsize;
  char *__pyx_v_resultp;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("pybuffer_index", 0);

  /* "View.MemoryView":912
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t itemsize = view.itemsize
 *     cdef char *resultp
 */
  __pyx_v_suboffset = -1L;

  /* "View.MemoryView":913
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 *     cdef Py_ssize_t itemsize = view.itemsize             # <<<<<<<<<<<<<<
 *     cdef char *resultp
 * 
 */
  __pyx_t_1 = __pyx_v_view->itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
  __pyx_t_2 = ((__pyx_v_view->ndim == 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":917
 * 
 *     if view.ndim == 0:
 *         shape = view.len / itemsize             # <<<<<<<<<<<<<<
 *         stride = itemsize
 *     else:
 */
    if (unlikely(__pyx_v_itemsize == 0)) {
      PyErr_SetString(PyExc_ZeroDivisionError, "integer division or modulo by zero");
      __PYX_ERR(1, 917, __pyx_L1_error)
    }
    else if (sizeof(Py_ssize_t) == sizeof(long) && (!(((Py_ssize_t)-1) > 0)) && unlikely(__pyx_v_itemsize == (Py_ssize_t)-1)  && unlikely(UNARY_NEG_WOULD_OVERFLOW(__pyx_v_view->len))) {
      PyErr_SetString(PyExc_OverflowError, "value too large to perform division");
      __PYX_ERR(1, 917, __pyx_L1_error)
    }
    __pyx_v_shape = (__pyx_v_view->len / __pyx_v_itemsize);

    /* "View.MemoryView":918
 *     if view.ndim == 0:
 *         shape = view.len / itemsize
 *         stride = itemsize             # <<<<<<<<<<<<<<
 *     else:
 *         shape = view.shape[dim]
 */
    __pyx_v_stride = __pyx_v_itemsize;

    /* "View.MemoryView":916
 *     cdef char *resultp
 * 
 *     if view.ndim == 0:             # <<<<<<<<<<<<<<
 *         shape = view.len / itemsize
 *         stride = itemsize
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":920
 *         stride = itemsize
 *     else:
 *         shape = view.shape[dim]             # <<<<<<<<<<<<<<
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 */
  /*else*/ {
    __pyx_v_shape = (__pyx_v_view->shape[__pyx_v_dim]);

    /* "View.MemoryView":921
 *     else:
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]             # <<<<<<<<<<<<<<
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]
 */
    __pyx_v_stride = (__pyx_v_view->strides[__pyx_v_dim]);

    /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    __pyx_t_2 = ((__pyx_v_view->suboffsets != NULL) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":923
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:
 *             suboffset = view.suboffsets[dim]             # <<<<<<<<<<<<<<
 * 
 *     if index < 0:
 */
      __pyx_v_suboffset = (__pyx_v_view->suboffsets[__pyx_v_dim]);

      /* "View.MemoryView":922
 *         shape = view.shape[dim]
 *         stride = view.strides[dim]
 *         if view.suboffsets != NULL:             # <<<<<<<<<<<<<<
 *             suboffset = view.suboffsets[dim]
 * 
 */
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  __pyx_t_2 = ((__pyx_v_index < 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":926
 * 
 *     if index < 0:
 *         index += view.shape[dim]             # <<<<<<<<<<<<<<
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 */
    __pyx_v_index = (__pyx_v_index + (__pyx_v_view->shape[__pyx_v_dim]));

    /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    __pyx_t_2 = ((__pyx_v_index < 0) != 0);
    if (unlikely(__pyx_t_2)) {

      /* "View.MemoryView":928
 *         index += view.shape[dim]
 *         if index < 0:
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     if index >= shape:
 */
      __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 928, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 928, __pyx_L1_error)

      /* "View.MemoryView":927
 *     if index < 0:
 *         index += view.shape[dim]
 *         if index < 0:             # <<<<<<<<<<<<<<
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
    }

    /* "View.MemoryView":925
 *             suboffset = view.suboffsets[dim]
 * 
 *     if index < 0:             # <<<<<<<<<<<<<<
 *         index += view.shape[dim]
 *         if index < 0:
 */
  }

  /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_index >= __pyx_v_shape) != 0);
  if (unlikely(__pyx_t_2)) {

    /* "View.MemoryView":931
 * 
 *     if index >= shape:
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)             # <<<<<<<<<<<<<<
 * 
 *     resultp = bufp + index * stride
 */
    __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_builtin_IndexError, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 931, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 931, __pyx_L1_error)

    /* "View.MemoryView":930
 *             raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     if index >= shape:             # <<<<<<<<<<<<<<
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 */
  }

  /* "View.MemoryView":933
 *         raise IndexError("Out of bounds on buffer access (axis %d)" % dim)
 * 
 *     resultp = bufp + index * stride             # <<<<<<<<<<<<<<
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset
 */
  __pyx_v_resultp = (__pyx_v_bufp + (__pyx_v_index * __pyx_v_stride));

  /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  __pyx_t_2 = ((__pyx_v_suboffset >= 0) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":935
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:
 *         resultp = (<char **> resultp)[0] + suboffset             # <<<<<<<<<<<<<<
 * 
 *     return resultp
 */
    __pyx_v_resultp = ((((char **)__pyx_v_resultp)[0]) + __pyx_v_suboffset);

    /* "View.MemoryView":934
 * 
 *     resultp = bufp + index * stride
 *     if suboffset >= 0:             # <<<<<<<<<<<<<<
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 */
  }

  /* "View.MemoryView":937
 *         resultp = (<char **> resultp)[0] + suboffset
 * 
 *     return resultp             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_resultp;
  goto __pyx_L0;

  /* "View.MemoryView":910
 * 
 * @cname('__pyx_pybuffer_index')
 * cdef char *pybuffer_index(Py_buffer *view, char *bufp, Py_ssize_t index,             # <<<<<<<<<<<<<<
 *                           Py_ssize_t dim) except NULL:
 *     cdef Py_ssize_t shape, stride, suboffset = -1
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView.pybuffer_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

static int __pyx_memslice_transpose(__Pyx_memviewslice *__pyx_v_memslice) {
  int __pyx_v_ndim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  int __pyx_v_i;
  int __pyx_v_j;
  int __pyx_r;
  int __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  long __pyx_t_3;
  long __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;
  int __pyx_t_7;
  int __pyx_t_8;
  int __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":944
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:
 *     cdef int ndim = memslice.memview.view.ndim             # <<<<<<<<<<<<<<
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 */
  __pyx_t_1 = __pyx_v_memslice->memview->view.ndim;
  __pyx_v_ndim = __pyx_t_1;

  /* "View.MemoryView":946
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 *     cdef Py_ssize_t *shape = memslice.shape             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t *strides = memslice.strides
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->shape;
  __pyx_v_shape = __pyx_t_2;

  /* "View.MemoryView":947
 * 
 *     cdef Py_ssize_t *shape = memslice.shape
 *     cdef Py_ssize_t *strides = memslice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_v_memslice->strides;
  __pyx_v_strides = __pyx_t_2;

  /* "View.MemoryView":951
 * 
 *     cdef int i, j
 *     for i in range(ndim / 2):             # <<<<<<<<<<<<<<
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 */
  __pyx_t_3 = (__pyx_v_ndim / 2);
  __pyx_t_4 = __pyx_t_3;
  for (__pyx_t_1 = 0; __pyx_t_1 < __pyx_t_4; __pyx_t_1+=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":952
 *     cdef int i, j
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i             # <<<<<<<<<<<<<<
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]
 */
    __pyx_v_j = ((__pyx_v_ndim - 1) - __pyx_v_i);

    /* "View.MemoryView":953
 *     for i in range(ndim / 2):
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]             # <<<<<<<<<<<<<<
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 */
    __pyx_t_5 = (__pyx_v_strides[__pyx_v_j]);
    __pyx_t_6 = (__pyx_v_strides[__pyx_v_i]);
    (__pyx_v_strides[__pyx_v_i]) = __pyx_t_5;
    (__pyx_v_strides[__pyx_v_j]) = __pyx_t_6;

    /* "View.MemoryView":954
 *         j = ndim - 1 - i
 *         strides[i], strides[j] = strides[j], strides[i]
 *         shape[i], shape[j] = shape[j], shape[i]             # <<<<<<<<<<<<<<
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 */
    __pyx_t_6 = (__pyx_v_shape[__pyx_v_j]);
    __pyx_t_5 = (__pyx_v_shape[__pyx_v_i]);
    (__pyx_v_shape[__pyx_v_i]) = __pyx_t_6;
    (__pyx_v_shape[__pyx_v_j]) = __pyx_t_5;

    /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_i]) >= 0) != 0);
    if (!__pyx_t_8) {
    } else {
      __pyx_t_7 = __pyx_t_8;
      goto __pyx_L6_bool_binop_done;
    }
    __pyx_t_8 = (((__pyx_v_memslice->suboffsets[__pyx_v_j]) >= 0) != 0);
    __pyx_t_7 = __pyx_t_8;
    __pyx_L6_bool_binop_done:;
    if (__pyx_t_7) {

      /* "View.MemoryView":957
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")             # <<<<<<<<<<<<<<
 * 
 *     return 1
 */
      __pyx_t_9 = __pyx_memoryview_err(__pyx_builtin_ValueError, ((char *)"Cannot transpose memoryview with indirect dimensions")); if (unlikely(__pyx_t_9 == ((int)-1))) __PYX_ERR(1, 957, __pyx_L1_error)

      /* "View.MemoryView":956
 *         shape[i], shape[j] = shape[j], shape[i]
 * 
 *         if memslice.suboffsets[i] >= 0 or memslice.suboffsets[j] >= 0:             # <<<<<<<<<<<<<<
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 */
    }
  }

  /* "View.MemoryView":959
 *             _err(ValueError, "Cannot transpose memoryview with indirect dimensions")
 * 
 *     return 1             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = 1;
  goto __pyx_L0;

  /* "View.MemoryView":943
 * 
 * @cname('__pyx_memslice_transpose')
 * cdef int transpose_memslice(__Pyx_memviewslice *memslice) nogil except 0:             # <<<<<<<<<<<<<<
 *     cdef int ndim = memslice.memview.view.ndim
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.transpose_memslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = 0;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

/* Python wrapper */
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self); /*proto*/
static void __pyx_memoryviewslice___dealloc__(PyObject *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__ (wrapper)", 0);
  __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_memoryviewslice___pyx_pf_15View_dot_MemoryView_16_memoryviewslice___dealloc__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__dealloc__", 0);

  /* "View.MemoryView":977
 * 
 *     def __dealloc__(self):
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)             # <<<<<<<<<<<<<<
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 */
  __PYX_XDEC_MEMVIEW((&__pyx_v_self->from_slice), 1);

  /* "View.MemoryView":976
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     def __dealloc__(self):             # <<<<<<<<<<<<<<
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

static PyObject *__pyx_memoryviewslice_convert_item_to_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("convert_item_to_object", 0);

  /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_object_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":981
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)             # <<<<<<<<<<<<<<
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_v_self->to_object_func(__pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 981, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;

    /* "View.MemoryView":980
 * 
 *     cdef convert_item_to_object(self, char *itemp):
 *         if self.to_object_func != NULL:             # <<<<<<<<<<<<<<
 *             return self.to_object_func(itemp)
 *         else:
 */
  }

  /* "View.MemoryView":983
 *             return self.to_object_func(itemp)
 *         else:
 *             return memoryview.convert_item_to_object(self, itemp)             # <<<<<<<<<<<<<<
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_2 = __pyx_memoryview_convert_item_to_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 983, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L0;
  }

  /* "View.MemoryView":979
 *         __PYX_XDEC_MEMVIEW(&self.from_slice, 1)
 * 
 *     cdef convert_item_to_object(self, char *itemp):             # <<<<<<<<<<<<<<
 *         if self.to_object_func != NULL:
 *             return self.to_object_func(itemp)
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.convert_item_to_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

static PyObject *__pyx_memoryviewslice_assign_item_from_object(struct __pyx_memoryviewslice_obj *__pyx_v_self, char *__pyx_v_itemp, PyObject *__pyx_v_value) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("assign_item_from_object", 0);

  /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
  __pyx_t_1 = ((__pyx_v_self->to_dtype_func != NULL) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":987
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)             # <<<<<<<<<<<<<<
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)
 */
    __pyx_t_2 = __pyx_v_self->to_dtype_func(__pyx_v_itemp, __pyx_v_value); if (unlikely(__pyx_t_2 == ((int)0))) __PYX_ERR(1, 987, __pyx_L1_error)

    /* "View.MemoryView":986
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):
 *         if self.to_dtype_func != NULL:             # <<<<<<<<<<<<<<
 *             self.to_dtype_func(itemp, value)
 *         else:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":989
 *             self.to_dtype_func(itemp, value)
 *         else:
 *             memoryview.assign_item_from_object(self, itemp, value)             # <<<<<<<<<<<<<<
 * 
 *     @property
 */
  /*else*/ {
    __pyx_t_3 = __pyx_memoryview_assign_item_from_object(((struct __pyx_memoryview_obj *)__pyx_v_self), __pyx_v_itemp, __pyx_v_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 989, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  }
  __pyx_L3:;

  /* "View.MemoryView":985
 *             return memoryview.convert_item_to_object(self, itemp)
 * 
 *     cdef assign_item_from_object(self, char *itemp, object value):             # <<<<<<<<<<<<<<
 *         if self.to_dtype_func != NULL:
 *             self.to_dtype_func(itemp, value)
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.assign_item_from_object", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__ (wrapper)", 0);
  __pyx_r = __pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView_16_memoryviewslice_4base___get__(struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__get__", 0);

  /* "View.MemoryView":993
 *     @property
 *     def base(self):
 *         return self.from_object             # <<<<<<<<<<<<<<
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_self->from_object);
  __pyx_r = __pyx_v_self->from_object;
  goto __pyx_L0;

  /* "View.MemoryView":992
 * 
 *     @property
 *     def base(self):             # <<<<<<<<<<<<<<
 *         return self.from_object
 * 
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_1__reduce_cython__(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__reduce_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice___reduce_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice___reduce_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__reduce_cython__", 0);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__28, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 2, __pyx_L1_error)

  /* "(tree fragment)":1
 * def __reduce_cython__(self):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__reduce_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

/* Python wrapper */
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state); /*proto*/
static PyObject *__pyx_pw___pyx_memoryviewslice_3__setstate_cython__(PyObject *__pyx_v_self, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__setstate_cython__ (wrapper)", 0);
  __pyx_r = __pyx_pf___pyx_memoryviewslice_2__setstate_cython__(((struct __pyx_memoryviewslice_obj *)__pyx_v_self), ((PyObject *)__pyx_v___pyx_state));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf___pyx_memoryviewslice_2__setstate_cython__(CYTHON_UNUSED struct __pyx_memoryviewslice_obj *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__setstate_cython__", 0);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_TypeError, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 4, __pyx_L1_error)

  /* "(tree fragment)":3
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):             # <<<<<<<<<<<<<<
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView._memoryviewslice.__setstate_cython__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

static PyObject *__pyx_memoryview_fromslice(__Pyx_memviewslice __pyx_v_memviewslice, int __pyx_v_ndim, PyObject *(*__pyx_v_to_object_func)(char *), int (*__pyx_v_to_dtype_func)(char *, PyObject *), int __pyx_v_dtype_is_object) {
  struct __pyx_memoryviewslice_obj *__pyx_v_result = 0;
  Py_ssize_t __pyx_v_suboffset;
  PyObject *__pyx_v_length = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_TypeInfo *__pyx_t_4;
  Py_buffer __pyx_t_5;
  Py_ssize_t *__pyx_t_6;
  Py_ssize_t *__pyx_t_7;
  Py_ssize_t *__pyx_t_8;
  Py_ssize_t __pyx_t_9;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_fromslice", 0);

  /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  __pyx_t_1 = ((((PyObject *)__pyx_v_memviewslice.memview) == Py_None) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1008
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:
 *         return None             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_r = Py_None; __Pyx_INCREF(Py_None);
    goto __pyx_L0;

    /* "View.MemoryView":1007
 *     cdef _memoryviewslice result
 * 
 *     if <PyObject *> memviewslice.memview == Py_None:             # <<<<<<<<<<<<<<
 *         return None
 * 
 */
  }

  /* "View.MemoryView":1013
 * 
 * 
 *     result = _memoryviewslice(None, 0, dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 *     result.from_slice = memviewslice
 */
  __pyx_t_2 = __Pyx_PyBool_FromLong(__pyx_v_dtype_is_object); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyTuple_New(3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  PyTuple_SET_ITEM(__pyx_t_3, 0, Py_None);
  __Pyx_INCREF(__pyx_int_0);
  __Pyx_GIVEREF(__pyx_int_0);
  PyTuple_SET_ITEM(__pyx_t_3, 1, __pyx_int_0);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_3, 2, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_memoryviewslice_type), __pyx_t_3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1013, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_result = ((struct __pyx_memoryviewslice_obj *)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":1015
 *     result = _memoryviewslice(None, 0, dtype_is_object)
 * 
 *     result.from_slice = memviewslice             # <<<<<<<<<<<<<<
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 */
  __pyx_v_result->from_slice = __pyx_v_memviewslice;

  /* "View.MemoryView":1016
 * 
 *     result.from_slice = memviewslice
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)             # <<<<<<<<<<<<<<
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 */
  __PYX_INC_MEMVIEW((&__pyx_v_memviewslice), 1);

  /* "View.MemoryView":1018
 *     __PYX_INC_MEMVIEW(&memviewslice, 1)
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base             # <<<<<<<<<<<<<<
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_memviewslice.memview), __pyx_n_s_base); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1018, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_result->from_object);
  __Pyx_DECREF(__pyx_v_result->from_object);
  __pyx_v_result->from_object = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "View.MemoryView":1019
 * 
 *     result.from_object = (<memoryview> memviewslice.memview).base
 *     result.typeinfo = memviewslice.memview.typeinfo             # <<<<<<<<<<<<<<
 * 
 *     result.view = memviewslice.memview.view
 */
  __pyx_t_4 = __pyx_v_memviewslice.memview->typeinfo;
  __pyx_v_result->__pyx_base.typeinfo = __pyx_t_4;

  /* "View.MemoryView":1021
 *     result.typeinfo = memviewslice.memview.typeinfo
 * 
 *     result.view = memviewslice.memview.view             # <<<<<<<<<<<<<<
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 */
  __pyx_t_5 = __pyx_v_memviewslice.memview->view;
  __pyx_v_result->__pyx_base.view = __pyx_t_5;

  /* "View.MemoryView":1022
 * 
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data             # <<<<<<<<<<<<<<
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 */
  __pyx_v_result->__pyx_base.view.buf = ((void *)__pyx_v_memviewslice.data);

  /* "View.MemoryView":1023
 *     result.view = memviewslice.memview.view
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim             # <<<<<<<<<<<<<<
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)
 */
  __pyx_v_result->__pyx_base.view.ndim = __pyx_v_ndim;

  /* "View.MemoryView":1024
 *     result.view.buf = <void *> memviewslice.data
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None             # <<<<<<<<<<<<<<
 *     Py_INCREF(Py_None)
 * 
 */
  ((Py_buffer *)(&__pyx_v_result->__pyx_base.view))->obj = Py_None;

  /* "View.MemoryView":1025
 *     result.view.ndim = ndim
 *     (<__pyx_buffer *> &result.view).obj = Py_None
 *     Py_INCREF(Py_None)             # <<<<<<<<<<<<<<
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 */
  Py_INCREF(Py_None);

  /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
  __pyx_t_1 = ((((struct __pyx_memoryview_obj *)__pyx_v_memviewslice.memview)->flags & PyBUF_WRITABLE) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1028
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:
 *         result.flags = PyBUF_RECORDS             # <<<<<<<<<<<<<<
 *     else:
 *         result.flags = PyBUF_RECORDS_RO
 */
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS;

    /* "View.MemoryView":1027
 *     Py_INCREF(Py_None)
 * 
 *     if (<memoryview>memviewslice.memview).flags & PyBUF_WRITABLE:             # <<<<<<<<<<<<<<
 *         result.flags = PyBUF_RECORDS
 *     else:
 */
    goto __pyx_L4;
  }

  /* "View.MemoryView":1030
 *         result.flags = PyBUF_RECORDS
 *     else:
 *         result.flags = PyBUF_RECORDS_RO             # <<<<<<<<<<<<<<
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 */
  /*else*/ {
    __pyx_v_result->__pyx_base.flags = PyBUF_RECORDS_RO;
  }
  __pyx_L4:;

  /* "View.MemoryView":1032
 *         result.flags = PyBUF_RECORDS_RO
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape             # <<<<<<<<<<<<<<
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides
 * 
 */
  __pyx_v_result->__pyx_base.view.shape = ((Py_ssize_t *)__pyx_v_result->from_slice.shape);

  /* "View.MemoryView":1033
 * 
 *     result.view.shape = <Py_ssize_t *> result.from_slice.shape
 *     result.view.strides = <Py_ssize_t *> result.from_slice.strides             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_result->__pyx_base.view.strides = ((Py_ssize_t *)__pyx_v_result->from_slice.strides);

  /* "View.MemoryView":1036
 * 
 * 
 *     result.view.suboffsets = NULL             # <<<<<<<<<<<<<<
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 */
  __pyx_v_result->__pyx_base.view.suboffsets = NULL;

  /* "View.MemoryView":1037
 * 
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:             # <<<<<<<<<<<<<<
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 */
  __pyx_t_7 = (__pyx_v_result->from_slice.suboffsets + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->from_slice.suboffsets; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_v_suboffset = (__pyx_t_6[0]);

    /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    __pyx_t_1 = ((__pyx_v_suboffset >= 0) != 0);
    if (__pyx_t_1) {

      /* "View.MemoryView":1039
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_result->__pyx_base.view.suboffsets = ((Py_ssize_t *)__pyx_v_result->from_slice.suboffsets);

      /* "View.MemoryView":1040
 *         if suboffset >= 0:
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break             # <<<<<<<<<<<<<<
 * 
 *     result.view.len = result.view.itemsize
 */
      goto __pyx_L6_break;

      /* "View.MemoryView":1038
 *     result.view.suboffsets = NULL
 *     for suboffset in result.from_slice.suboffsets[:ndim]:
 *         if suboffset >= 0:             # <<<<<<<<<<<<<<
 *             result.view.suboffsets = <Py_ssize_t *> result.from_slice.suboffsets
 *             break
 */
    }
  }
  __pyx_L6_break:;

  /* "View.MemoryView":1042
 *             break
 * 
 *     result.view.len = result.view.itemsize             # <<<<<<<<<<<<<<
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length
 */
  __pyx_t_9 = __pyx_v_result->__pyx_base.view.itemsize;
  __pyx_v_result->__pyx_base.view.len = __pyx_t_9;

  /* "View.MemoryView":1043
 * 
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         result.view.len *= length
 * 
 */
  __pyx_t_7 = (__pyx_v_result->__pyx_base.view.shape + __pyx_v_ndim);
  for (__pyx_t_8 = __pyx_v_result->__pyx_base.view.shape; __pyx_t_8 < __pyx_t_7; __pyx_t_8++) {
    __pyx_t_6 = __pyx_t_8;
    __pyx_t_2 = PyInt_FromSsize_t((__pyx_t_6[0])); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1043, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_XDECREF_SET(__pyx_v_length, __pyx_t_2);
    __pyx_t_2 = 0;

    /* "View.MemoryView":1044
 *     result.view.len = result.view.itemsize
 *     for length in result.view.shape[:ndim]:
 *         result.view.len *= length             # <<<<<<<<<<<<<<
 * 
 *     result.to_object_func = to_object_func
 */
    __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_result->__pyx_base.view.len); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = PyNumber_InPlaceMultiply(__pyx_t_2, __pyx_v_length); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_9 = __Pyx_PyIndex_AsSsize_t(__pyx_t_3); if (unlikely((__pyx_t_9 == (Py_ssize_t)-1) && PyErr_Occurred())) __PYX_ERR(1, 1044, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_v_result->__pyx_base.view.len = __pyx_t_9;
  }

  /* "View.MemoryView":1046
 *         result.view.len *= length
 * 
 *     result.to_object_func = to_object_func             # <<<<<<<<<<<<<<
 *     result.to_dtype_func = to_dtype_func
 * 
 */
  __pyx_v_result->to_object_func = __pyx_v_to_object_func;

  /* "View.MemoryView":1047
 * 
 *     result.to_object_func = to_object_func
 *     result.to_dtype_func = to_dtype_func             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  __pyx_v_result->to_dtype_func = __pyx_v_to_dtype_func;

  /* "View.MemoryView":1049
 *     result.to_dtype_func = to_dtype_func
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_result));
  __pyx_r = ((PyObject *)__pyx_v_result);
  goto __pyx_L0;

  /* "View.MemoryView":999
 * 
 * @cname('__pyx_memoryview_fromslice')
 * cdef memoryview_fromslice(__Pyx_memviewslice memviewslice,             # <<<<<<<<<<<<<<
 *                           int ndim,
 *                           object (*to_object_func)(char *),
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.memoryview_fromslice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_result);
  __Pyx_XDECREF(__pyx_v_length);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

static __Pyx_memviewslice *__pyx_memoryview_get_slice_from_memoryview(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_mslice) {
  struct __pyx_memoryviewslice_obj *__pyx_v_obj = 0;
  __Pyx_memviewslice *__pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("get_slice_from_memview", 0);

  /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1056
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview             # <<<<<<<<<<<<<<
 *         return &obj.from_slice
 *     else:
 */
    if (!(likely(((((PyObject *)__pyx_v_memview)) == Py_None) || likely(__Pyx_TypeTest(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type))))) __PYX_ERR(1, 1056, __pyx_L1_error)
    __pyx_t_3 = ((PyObject *)__pyx_v_memview);
    __Pyx_INCREF(__pyx_t_3);
    __pyx_v_obj = ((struct __pyx_memoryviewslice_obj *)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "View.MemoryView":1057
 *     if isinstance(memview, _memoryviewslice):
 *         obj = memview
 *         return &obj.from_slice             # <<<<<<<<<<<<<<
 *     else:
 *         slice_copy(memview, mslice)
 */
    __pyx_r = (&__pyx_v_obj->from_slice);
    goto __pyx_L0;

    /* "View.MemoryView":1055
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         obj = memview
 *         return &obj.from_slice
 */
  }

  /* "View.MemoryView":1059
 *         return &obj.from_slice
 *     else:
 *         slice_copy(memview, mslice)             # <<<<<<<<<<<<<<
 *         return mslice
 * 
 */
  /*else*/ {
    __pyx_memoryview_slice_copy(__pyx_v_memview, __pyx_v_mslice);

    /* "View.MemoryView":1060
 *     else:
 *         slice_copy(memview, mslice)
 *         return mslice             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_slice_copy')
 */
    __pyx_r = __pyx_v_mslice;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1052
 * 
 * @cname('__pyx_memoryview_get_slice_from_memoryview')
 * cdef __Pyx_memviewslice *get_slice_from_memview(memoryview memview,             # <<<<<<<<<<<<<<
 *                                                    __Pyx_memviewslice *mslice) except NULL:
 *     cdef _memoryviewslice obj
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("View.MemoryView.get_slice_from_memview", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_obj);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

static void __pyx_memoryview_slice_copy(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_dst) {
  int __pyx_v_dim;
  Py_ssize_t *__pyx_v_shape;
  Py_ssize_t *__pyx_v_strides;
  Py_ssize_t *__pyx_v_suboffsets;
  __Pyx_RefNannyDeclarations
  Py_ssize_t *__pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  __Pyx_RefNannySetupContext("slice_copy", 0);

  /* "View.MemoryView":1067
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 * 
 *     shape = memview.view.shape             # <<<<<<<<<<<<<<
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets
 */
  __pyx_t_1 = __pyx_v_memview->view.shape;
  __pyx_v_shape = __pyx_t_1;

  /* "View.MemoryView":1068
 * 
 *     shape = memview.view.shape
 *     strides = memview.view.strides             # <<<<<<<<<<<<<<
 *     suboffsets = memview.view.suboffsets
 * 
 */
  __pyx_t_1 = __pyx_v_memview->view.strides;
  __pyx_v_strides = __pyx_t_1;

  /* "View.MemoryView":1069
 *     shape = memview.view.shape
 *     strides = memview.view.strides
 *     suboffsets = memview.view.suboffsets             # <<<<<<<<<<<<<<
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 */
  __pyx_t_1 = __pyx_v_memview->view.suboffsets;
  __pyx_v_suboffsets = __pyx_t_1;

  /* "View.MemoryView":1071
 *     suboffsets = memview.view.suboffsets
 * 
 *     dst.memview = <__pyx_memoryview *> memview             # <<<<<<<<<<<<<<
 *     dst.data = <char *> memview.view.buf
 * 
 */
  __pyx_v_dst->memview = ((struct __pyx_memoryview_obj *)__pyx_v_memview);

  /* "View.MemoryView":1072
 * 
 *     dst.memview = <__pyx_memoryview *> memview
 *     dst.data = <char *> memview.view.buf             # <<<<<<<<<<<<<<
 * 
 *     for dim in range(memview.view.ndim):
 */
  __pyx_v_dst->data = ((char *)__pyx_v_memview->view.buf);

  /* "View.MemoryView":1074
 *     dst.data = <char *> memview.view.buf
 * 
 *     for dim in range(memview.view.ndim):             # <<<<<<<<<<<<<<
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 */
  __pyx_t_2 = __pyx_v_memview->view.ndim;
  __pyx_t_3 = __pyx_t_2;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_dim = __pyx_t_4;

    /* "View.MemoryView":1075
 * 
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]             # <<<<<<<<<<<<<<
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 */
    (__pyx_v_dst->shape[__pyx_v_dim]) = (__pyx_v_shape[__pyx_v_dim]);

    /* "View.MemoryView":1076
 *     for dim in range(memview.view.ndim):
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]             # <<<<<<<<<<<<<<
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1
 * 
 */
    (__pyx_v_dst->strides[__pyx_v_dim]) = (__pyx_v_strides[__pyx_v_dim]);

    /* "View.MemoryView":1077
 *         dst.shape[dim] = shape[dim]
 *         dst.strides[dim] = strides[dim]
 *         dst.suboffsets[dim] = suboffsets[dim] if suboffsets else -1             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object')
 */
    if ((__pyx_v_suboffsets != 0)) {
      __pyx_t_5 = (__pyx_v_suboffsets[__pyx_v_dim]);
    } else {
      __pyx_t_5 = -1L;
    }
    (__pyx_v_dst->suboffsets[__pyx_v_dim]) = __pyx_t_5;
  }

  /* "View.MemoryView":1063
 * 
 * @cname('__pyx_memoryview_slice_copy')
 * cdef void slice_copy(memoryview memview, __Pyx_memviewslice *dst):             # <<<<<<<<<<<<<<
 *     cdef int dim
 *     cdef (Py_ssize_t*) shape, strides, suboffsets
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

static PyObject *__pyx_memoryview_copy_object(struct __pyx_memoryview_obj *__pyx_v_memview) {
  __Pyx_memviewslice __pyx_v_memviewslice;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy", 0);

  /* "View.MemoryView":1083
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)             # <<<<<<<<<<<<<<
 *     return memoryview_copy_from_slice(memview, &memviewslice)
 * 
 */
  __pyx_memoryview_slice_copy(__pyx_v_memview, (&__pyx_v_memviewslice));

  /* "View.MemoryView":1084
 *     cdef __Pyx_memviewslice memviewslice
 *     slice_copy(memview, &memviewslice)
 *     return memoryview_copy_from_slice(memview, &memviewslice)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_memoryview_copy_object_from_slice(__pyx_v_memview, (&__pyx_v_memviewslice)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1084, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1080
 * 
 * @cname('__pyx_memoryview_copy_object')
 * cdef memoryview_copy(memoryview memview):             # <<<<<<<<<<<<<<
 *     "Create a new memoryview object"
 *     cdef __Pyx_memviewslice memviewslice
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

static PyObject *__pyx_memoryview_copy_object_from_slice(struct __pyx_memoryview_obj *__pyx_v_memview, __Pyx_memviewslice *__pyx_v_memviewslice) {
  PyObject *(*__pyx_v_to_object_func)(char *);
  int (*__pyx_v_to_dtype_func)(char *, PyObject *);
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *(*__pyx_t_3)(char *);
  int (*__pyx_t_4)(char *, PyObject *);
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("memoryview_copy_from_slice", 0);

  /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
  __pyx_t_1 = __Pyx_TypeCheck(((PyObject *)__pyx_v_memview), __pyx_memoryviewslice_type); 
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1095
 * 
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func             # <<<<<<<<<<<<<<
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 */
    __pyx_t_3 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_object_func;
    __pyx_v_to_object_func = __pyx_t_3;

    /* "View.MemoryView":1096
 *     if isinstance(memview, _memoryviewslice):
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func             # <<<<<<<<<<<<<<
 *     else:
 *         to_object_func = NULL
 */
    __pyx_t_4 = ((struct __pyx_memoryviewslice_obj *)__pyx_v_memview)->to_dtype_func;
    __pyx_v_to_dtype_func = __pyx_t_4;

    /* "View.MemoryView":1094
 *     cdef int (*to_dtype_func)(char *, object) except 0
 * 
 *     if isinstance(memview, _memoryviewslice):             # <<<<<<<<<<<<<<
 *         to_object_func = (<_memoryviewslice> memview).to_object_func
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1098
 *         to_dtype_func = (<_memoryviewslice> memview).to_dtype_func
 *     else:
 *         to_object_func = NULL             # <<<<<<<<<<<<<<
 *         to_dtype_func = NULL
 * 
 */
  /*else*/ {
    __pyx_v_to_object_func = NULL;

    /* "View.MemoryView":1099
 *     else:
 *         to_object_func = NULL
 *         to_dtype_func = NULL             # <<<<<<<<<<<<<<
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 */
    __pyx_v_to_dtype_func = NULL;
  }
  __pyx_L3:;

  /* "View.MemoryView":1101
 *         to_dtype_func = NULL
 * 
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,             # <<<<<<<<<<<<<<
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)
 */
  __Pyx_XDECREF(__pyx_r);

  /* "View.MemoryView":1103
 *     return memoryview_fromslice(memviewslice[0], memview.view.ndim,
 *                                 to_object_func, to_dtype_func,
 *                                 memview.dtype_is_object)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_5 = __pyx_memoryview_fromslice((__pyx_v_memviewslice[0]), __pyx_v_memview->view.ndim, __pyx_v_to_object_func, __pyx_v_to_dtype_func, __pyx_v_memview->dtype_is_object); if (unlikely(!__pyx_t_5)) __PYX_ERR(1, 1101, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __pyx_r = __pyx_t_5;
  __pyx_t_5 = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1087
 * 
 * @cname('__pyx_memoryview_copy_object_from_slice')
 * cdef memoryview_copy_from_slice(memoryview memview, __Pyx_memviewslice *memviewslice):             # <<<<<<<<<<<<<<
 *     """
 *     Create a new memoryview object from a given memoryview object and slice.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.memoryview_copy_from_slice", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

static Py_ssize_t abs_py_ssize_t(Py_ssize_t __pyx_v_arg) {
  Py_ssize_t __pyx_r;
  int __pyx_t_1;

  /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arg < 0) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1111
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:
 *         return -arg             # <<<<<<<<<<<<<<
 *     else:
 *         return arg
 */
    __pyx_r = (-__pyx_v_arg);
    goto __pyx_L0;

    /* "View.MemoryView":1110
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:
 *     if arg < 0:             # <<<<<<<<<<<<<<
 *         return -arg
 *     else:
 */
  }

  /* "View.MemoryView":1113
 *         return -arg
 *     else:
 *         return arg             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_get_best_slice_order')
 */
  /*else*/ {
    __pyx_r = __pyx_v_arg;
    goto __pyx_L0;
  }

  /* "View.MemoryView":1109
 * 
 * 
 * cdef Py_ssize_t abs_py_ssize_t(Py_ssize_t arg) nogil:             # <<<<<<<<<<<<<<
 *     if arg < 0:
 *         return -arg
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

static char __pyx_get_best_slice_order(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim) {
  int __pyx_v_i;
  Py_ssize_t __pyx_v_c_stride;
  Py_ssize_t __pyx_v_f_stride;
  char __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1121
 *     """
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t f_stride = 0
 * 
 */
  __pyx_v_c_stride = 0;

  /* "View.MemoryView":1122
 *     cdef int i
 *     cdef Py_ssize_t c_stride = 0
 *     cdef Py_ssize_t f_stride = 0             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_f_stride = 0;

  /* "View.MemoryView":1124
 *     cdef Py_ssize_t f_stride = 0
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1126
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_c_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1127
 *         if mslice.shape[i] > 1:
 *             c_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
      goto __pyx_L4_break;

      /* "View.MemoryView":1125
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             c_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L4_break:;

  /* "View.MemoryView":1129
 *             break
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 */
  __pyx_t_1 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_1;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    __pyx_t_2 = (((__pyx_v_mslice->shape[__pyx_v_i]) > 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1131
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]             # <<<<<<<<<<<<<<
 *             break
 * 
 */
      __pyx_v_f_stride = (__pyx_v_mslice->strides[__pyx_v_i]);

      /* "View.MemoryView":1132
 *         if mslice.shape[i] > 1:
 *             f_stride = mslice.strides[i]
 *             break             # <<<<<<<<<<<<<<
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 */
      goto __pyx_L7_break;

      /* "View.MemoryView":1130
 * 
 *     for i in range(ndim):
 *         if mslice.shape[i] > 1:             # <<<<<<<<<<<<<<
 *             f_stride = mslice.strides[i]
 *             break
 */
    }
  }
  __pyx_L7_break:;

  /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  __pyx_t_2 = ((abs_py_ssize_t(__pyx_v_c_stride) <= abs_py_ssize_t(__pyx_v_f_stride)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1135
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):
 *         return 'C'             # <<<<<<<<<<<<<<
 *     else:
 *         return 'F'
 */
    __pyx_r = 'C';
    goto __pyx_L0;

    /* "View.MemoryView":1134
 *             break
 * 
 *     if abs_py_ssize_t(c_stride) <= abs_py_ssize_t(f_stride):             # <<<<<<<<<<<<<<
 *         return 'C'
 *     else:
 */
  }

  /* "View.MemoryView":1137
 *         return 'C'
 *     else:
 *         return 'F'             # <<<<<<<<<<<<<<
 * 
 * @cython.cdivision(True)
 */
  /*else*/ {
    __pyx_r = 'F';
    goto __pyx_L0;
  }

  /* "View.MemoryView":1116
 * 
 * @cname('__pyx_get_best_slice_order')
 * cdef char get_best_order(__Pyx_memviewslice *mslice, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     """
 *     Figure out the best memory access order for a given slice.
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

static void _copy_strided_to_strided(char *__pyx_v_src_data, Py_ssize_t *__pyx_v_src_strides, char *__pyx_v_dst_data, Py_ssize_t *__pyx_v_dst_strides, Py_ssize_t *__pyx_v_src_shape, Py_ssize_t *__pyx_v_dst_shape, int __pyx_v_ndim, size_t __pyx_v_itemsize) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  CYTHON_UNUSED Py_ssize_t __pyx_v_src_extent;
  Py_ssize_t __pyx_v_dst_extent;
  Py_ssize_t __pyx_v_src_stride;
  Py_ssize_t __pyx_v_dst_stride;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  Py_ssize_t __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  Py_ssize_t __pyx_t_6;

  /* "View.MemoryView":1147
 * 
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 */
  __pyx_v_src_extent = (__pyx_v_src_shape[0]);

  /* "View.MemoryView":1148
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 */
  __pyx_v_dst_extent = (__pyx_v_dst_shape[0]);

  /* "View.MemoryView":1149
 *     cdef Py_ssize_t src_extent = src_shape[0]
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 */
  __pyx_v_src_stride = (__pyx_v_src_strides[0]);

  /* "View.MemoryView":1150
 *     cdef Py_ssize_t dst_extent = dst_shape[0]
 *     cdef Py_ssize_t src_stride = src_strides[0]
 *     cdef Py_ssize_t dst_stride = dst_strides[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_dst_stride = (__pyx_v_dst_strides[0]);

  /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    __pyx_t_2 = ((__pyx_v_src_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }
    __pyx_t_2 = ((__pyx_v_dst_stride > 0) != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L5_bool_binop_done;
    }

    /* "View.MemoryView":1154
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):             # <<<<<<<<<<<<<<
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 */
    __pyx_t_2 = (((size_t)__pyx_v_src_stride) == __pyx_v_itemsize);
    if (__pyx_t_2) {
      __pyx_t_2 = (__pyx_v_itemsize == ((size_t)__pyx_v_dst_stride));
    }
    __pyx_t_3 = (__pyx_t_2 != 0);
    __pyx_t_1 = __pyx_t_3;
    __pyx_L5_bool_binop_done:;

    /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
    if (__pyx_t_1) {

      /* "View.MemoryView":1155
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)             # <<<<<<<<<<<<<<
 *        else:
 *            for i in range(dst_extent):
 */
      (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, (__pyx_v_itemsize * __pyx_v_dst_extent)));

      /* "View.MemoryView":1153
 * 
 *     if ndim == 1:
 *        if (src_stride > 0 and dst_stride > 0 and             # <<<<<<<<<<<<<<
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 */
      goto __pyx_L4;
    }

    /* "View.MemoryView":1157
 *            memcpy(dst_data, src_data, itemsize * dst_extent)
 *        else:
 *            for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 */
    /*else*/ {
      __pyx_t_4 = __pyx_v_dst_extent;
      __pyx_t_5 = __pyx_t_4;
      for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
        __pyx_v_i = __pyx_t_6;

        /* "View.MemoryView":1158
 *        else:
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)             # <<<<<<<<<<<<<<
 *                src_data += src_stride
 *                dst_data += dst_stride
 */
        (void)(memcpy(__pyx_v_dst_data, __pyx_v_src_data, __pyx_v_itemsize));

        /* "View.MemoryView":1159
 *            for i in range(dst_extent):
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride             # <<<<<<<<<<<<<<
 *                dst_data += dst_stride
 *     else:
 */
        __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

        /* "View.MemoryView":1160
 *                memcpy(dst_data, src_data, itemsize)
 *                src_data += src_stride
 *                dst_data += dst_stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(dst_extent):
 */
        __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
      }
    }
    __pyx_L4:;

    /* "View.MemoryView":1152
 *     cdef Py_ssize_t dst_stride = dst_strides[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *        if (src_stride > 0 and dst_stride > 0 and
 *            <size_t> src_stride == itemsize == <size_t> dst_stride):
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1162
 *                dst_data += dst_stride
 *     else:
 *         for i in range(dst_extent):             # <<<<<<<<<<<<<<
 *             _copy_strided_to_strided(src_data, src_strides + 1,
 *                                      dst_data, dst_strides + 1,
 */
  /*else*/ {
    __pyx_t_4 = __pyx_v_dst_extent;
    __pyx_t_5 = __pyx_t_4;
    for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
      __pyx_v_i = __pyx_t_6;

      /* "View.MemoryView":1163
 *     else:
 *         for i in range(dst_extent):
 *             _copy_strided_to_strided(src_data, src_strides + 1,             # <<<<<<<<<<<<<<
 *                                      dst_data, dst_strides + 1,
 *                                      src_shape + 1, dst_shape + 1,
 */
      _copy_strided_to_strided(__pyx_v_src_data, (__pyx_v_src_strides + 1), __pyx_v_dst_data, (__pyx_v_dst_strides + 1), (__pyx_v_src_shape + 1), (__pyx_v_dst_shape + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize);

      /* "View.MemoryView":1167
 *                                      src_shape + 1, dst_shape + 1,
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride             # <<<<<<<<<<<<<<
 *             dst_data += dst_stride
 * 
 */
      __pyx_v_src_data = (__pyx_v_src_data + __pyx_v_src_stride);

      /* "View.MemoryView":1168
 *                                      ndim - 1, itemsize)
 *             src_data += src_stride
 *             dst_data += dst_stride             # <<<<<<<<<<<<<<
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,
 */
      __pyx_v_dst_data = (__pyx_v_dst_data + __pyx_v_dst_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1140
 * 
 * @cython.cdivision(True)
 * cdef void _copy_strided_to_strided(char *src_data, Py_ssize_t *src_strides,             # <<<<<<<<<<<<<<
 *                                    char *dst_data, Py_ssize_t *dst_strides,
 *                                    Py_ssize_t *src_shape, Py_ssize_t *dst_shape,
 */

  /* function exit code */
}

/* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

static void copy_strided_to_strided(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize) {

  /* "View.MemoryView":1173
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 *     _copy_strided_to_strided(src.data, src.strides, dst.data, dst.strides,             # <<<<<<<<<<<<<<
 *                              src.shape, dst.shape, ndim, itemsize)
 * 
 */
  _copy_strided_to_strided(__pyx_v_src->data, __pyx_v_src->strides, __pyx_v_dst->data, __pyx_v_dst->strides, __pyx_v_src->shape, __pyx_v_dst->shape, __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1170
 *             dst_data += dst_stride
 * 
 * cdef void copy_strided_to_strided(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice *dst,
 *                                   int ndim, size_t itemsize) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

static Py_ssize_t __pyx_memoryview_slice_get_size(__Pyx_memviewslice *__pyx_v_src, int __pyx_v_ndim) {
  Py_ssize_t __pyx_v_shape;
  Py_ssize_t __pyx_v_size;
  Py_ssize_t __pyx_r;
  Py_ssize_t __pyx_t_1;
  Py_ssize_t *__pyx_t_2;
  Py_ssize_t *__pyx_t_3;
  Py_ssize_t *__pyx_t_4;

  /* "View.MemoryView":1179
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 * 
 *     for shape in src.shape[:ndim]:
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_size = __pyx_t_1;

  /* "View.MemoryView":1181
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 * 
 *     for shape in src.shape[:ndim]:             # <<<<<<<<<<<<<<
 *         size *= shape
 * 
 */
  __pyx_t_3 = (__pyx_v_src->shape + __pyx_v_ndim);
  for (__pyx_t_4 = __pyx_v_src->shape; __pyx_t_4 < __pyx_t_3; __pyx_t_4++) {
    __pyx_t_2 = __pyx_t_4;
    __pyx_v_shape = (__pyx_t_2[0]);

    /* "View.MemoryView":1182
 * 
 *     for shape in src.shape[:ndim]:
 *         size *= shape             # <<<<<<<<<<<<<<
 * 
 *     return size
 */
    __pyx_v_size = (__pyx_v_size * __pyx_v_shape);
  }

  /* "View.MemoryView":1184
 *         size *= shape
 * 
 *     return size             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_fill_contig_strides_array')
 */
  __pyx_r = __pyx_v_size;
  goto __pyx_L0;

  /* "View.MemoryView":1177
 * 
 * @cname('__pyx_memoryview_slice_get_size')
 * cdef Py_ssize_t slice_get_size(__Pyx_memviewslice *src, int ndim) nogil:             # <<<<<<<<<<<<<<
 *     "Return the size of the memory occupied by the slice in number of bytes"
 *     cdef Py_ssize_t shape, size = src.memview.view.itemsize
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

static Py_ssize_t __pyx_fill_contig_strides_array(Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, Py_ssize_t __pyx_v_stride, int __pyx_v_ndim, char __pyx_v_order) {
  int __pyx_v_idx;
  Py_ssize_t __pyx_r;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;

  /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
  __pyx_t_1 = ((__pyx_v_order == 'F') != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1197
 * 
 *     if order == 'F':
 *         for idx in range(ndim):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
    __pyx_t_2 = __pyx_v_ndim;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_idx = __pyx_t_4;

      /* "View.MemoryView":1198
 *     if order == 'F':
 *         for idx in range(ndim):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 *     else:
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1199
 *         for idx in range(ndim):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }

    /* "View.MemoryView":1196
 *     cdef int idx
 * 
 *     if order == 'F':             # <<<<<<<<<<<<<<
 *         for idx in range(ndim):
 *             strides[idx] = stride
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1201
 *             stride *= shape[idx]
 *     else:
 *         for idx in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *             strides[idx] = stride
 *             stride *= shape[idx]
 */
  /*else*/ {
    for (__pyx_t_2 = (__pyx_v_ndim - 1); __pyx_t_2 > -1; __pyx_t_2-=1) {
      __pyx_v_idx = __pyx_t_2;

      /* "View.MemoryView":1202
 *     else:
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride             # <<<<<<<<<<<<<<
 *             stride *= shape[idx]
 * 
 */
      (__pyx_v_strides[__pyx_v_idx]) = __pyx_v_stride;

      /* "View.MemoryView":1203
 *         for idx in range(ndim - 1, -1, -1):
 *             strides[idx] = stride
 *             stride *= shape[idx]             # <<<<<<<<<<<<<<
 * 
 *     return stride
 */
      __pyx_v_stride = (__pyx_v_stride * (__pyx_v_shape[__pyx_v_idx]));
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1205
 *             stride *= shape[idx]
 * 
 *     return stride             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 */
  __pyx_r = __pyx_v_stride;
  goto __pyx_L0;

  /* "View.MemoryView":1187
 * 
 * @cname('__pyx_fill_contig_strides_array')
 * cdef Py_ssize_t fill_contig_strides_array(             # <<<<<<<<<<<<<<
 *                 Py_ssize_t *shape, Py_ssize_t *strides, Py_ssize_t stride,
 *                 int ndim, char order) nogil:
 */

  /* function exit code */
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

static void *__pyx_memoryview_copy_data_to_temp(__Pyx_memviewslice *__pyx_v_src, __Pyx_memviewslice *__pyx_v_tmpslice, char __pyx_v_order, int __pyx_v_ndim) {
  int __pyx_v_i;
  void *__pyx_v_result;
  size_t __pyx_v_itemsize;
  size_t __pyx_v_size;
  void *__pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  struct __pyx_memoryview_obj *__pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1219
 *     cdef void *result
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 */
  __pyx_t_1 = __pyx_v_src->memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1220
 * 
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef size_t size = slice_get_size(src, ndim)             # <<<<<<<<<<<<<<
 * 
 *     result = malloc(size)
 */
  __pyx_v_size = __pyx_memoryview_slice_get_size(__pyx_v_src, __pyx_v_ndim);

  /* "View.MemoryView":1222
 *     cdef size_t size = slice_get_size(src, ndim)
 * 
 *     result = malloc(size)             # <<<<<<<<<<<<<<
 *     if not result:
 *         _err(MemoryError, NULL)
 */
  __pyx_v_result = malloc(__pyx_v_size);

  /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_result != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1224
 *     result = malloc(size)
 *     if not result:
 *         _err(MemoryError, NULL)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_3 = __pyx_memoryview_err(__pyx_builtin_MemoryError, NULL); if (unlikely(__pyx_t_3 == ((int)-1))) __PYX_ERR(1, 1224, __pyx_L1_error)

    /* "View.MemoryView":1223
 * 
 *     result = malloc(size)
 *     if not result:             # <<<<<<<<<<<<<<
 *         _err(MemoryError, NULL)
 * 
 */
  }

  /* "View.MemoryView":1227
 * 
 * 
 *     tmpslice.data = <char *> result             # <<<<<<<<<<<<<<
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 */
  __pyx_v_tmpslice->data = ((char *)__pyx_v_result);

  /* "View.MemoryView":1228
 * 
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview             # <<<<<<<<<<<<<<
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 */
  __pyx_t_4 = __pyx_v_src->memview;
  __pyx_v_tmpslice->memview = __pyx_t_4;

  /* "View.MemoryView":1229
 *     tmpslice.data = <char *> result
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1230
 *     tmpslice.memview = src.memview
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]             # <<<<<<<<<<<<<<
 *         tmpslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_tmpslice->shape[__pyx_v_i]) = (__pyx_v_src->shape[__pyx_v_i]);

    /* "View.MemoryView":1231
 *     for i in range(ndim):
 *         tmpslice.shape[i] = src.shape[i]
 *         tmpslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,
 */
    (__pyx_v_tmpslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1233
 *         tmpslice.suboffsets[i] = -1
 * 
 *     fill_contig_strides_array(&tmpslice.shape[0], &tmpslice.strides[0], itemsize,             # <<<<<<<<<<<<<<
 *                               ndim, order)
 * 
 */
  (void)(__pyx_fill_contig_strides_array((&(__pyx_v_tmpslice->shape[0])), (&(__pyx_v_tmpslice->strides[0])), __pyx_v_itemsize, __pyx_v_ndim, __pyx_v_order));

  /* "View.MemoryView":1237
 * 
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0
 */
  __pyx_t_3 = __pyx_v_ndim;
  __pyx_t_5 = __pyx_t_3;
  for (__pyx_t_6 = 0; __pyx_t_6 < __pyx_t_5; __pyx_t_6+=1) {
    __pyx_v_i = __pyx_t_6;

    /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    __pyx_t_2 = (((__pyx_v_tmpslice->shape[__pyx_v_i]) == 1) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1239
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:
 *             tmpslice.strides[i] = 0             # <<<<<<<<<<<<<<
 * 
 *     if slice_is_contig(src[0], order, ndim):
 */
      (__pyx_v_tmpslice->strides[__pyx_v_i]) = 0;

      /* "View.MemoryView":1238
 * 
 *     for i in range(ndim):
 *         if tmpslice.shape[i] == 1:             # <<<<<<<<<<<<<<
 *             tmpslice.strides[i] = 0
 * 
 */
    }
  }

  /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
  __pyx_t_2 = (__pyx_memviewslice_is_contig((__pyx_v_src[0]), __pyx_v_order, __pyx_v_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1242
 * 
 *     if slice_is_contig(src[0], order, ndim):
 *         memcpy(result, src.data, size)             # <<<<<<<<<<<<<<
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 */
    (void)(memcpy(__pyx_v_result, __pyx_v_src->data, __pyx_v_size));

    /* "View.MemoryView":1241
 *             tmpslice.strides[i] = 0
 * 
 *     if slice_is_contig(src[0], order, ndim):             # <<<<<<<<<<<<<<
 *         memcpy(result, src.data, size)
 *     else:
 */
    goto __pyx_L9;
  }

  /* "View.MemoryView":1244
 *         memcpy(result, src.data, size)
 *     else:
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)             # <<<<<<<<<<<<<<
 * 
 *     return result
 */
  /*else*/ {
    copy_strided_to_strided(__pyx_v_src, __pyx_v_tmpslice, __pyx_v_ndim, __pyx_v_itemsize);
  }
  __pyx_L9:;

  /* "View.MemoryView":1246
 *         copy_strided_to_strided(src, tmpslice, ndim, itemsize)
 * 
 *     return result             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_result;
  goto __pyx_L0;

  /* "View.MemoryView":1208
 * 
 * @cname('__pyx_memoryview_copy_data_to_temp')
 * cdef void *copy_data_to_temp(__Pyx_memviewslice *src,             # <<<<<<<<<<<<<<
 *                              __Pyx_memviewslice *tmpslice,
 *                              char order,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.copy_data_to_temp", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = NULL;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

static int __pyx_memoryview_err_extents(int __pyx_v_i, Py_ssize_t __pyx_v_extent1, Py_ssize_t __pyx_v_extent2) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_extents", 0);

  /* "View.MemoryView":1254
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 *                                                         (i, extent1, extent2))             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err_dim')
 */
  __pyx_t_1 = __Pyx_PyInt_From_int(__pyx_v_i); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyInt_FromSsize_t(__pyx_v_extent1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyInt_FromSsize_t(__pyx_v_extent2); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyTuple_New(3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1254, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_3);
  PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_3);
  __pyx_t_1 = 0;
  __pyx_t_2 = 0;
  __pyx_t_3 = 0;

  /* "View.MemoryView":1253
 * cdef int _err_extents(int i, Py_ssize_t extent1,
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %             # <<<<<<<<<<<<<<
 *                                                         (i, extent1, extent2))
 * 
 */
  __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_builtin_ValueError, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_4, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __PYX_ERR(1, 1253, __pyx_L1_error)

  /* "View.MemoryView":1251
 * 
 * @cname('__pyx_memoryview_err_extents')
 * cdef int _err_extents(int i, Py_ssize_t extent1,             # <<<<<<<<<<<<<<
 *                              Py_ssize_t extent2) except -1 with gil:
 *     raise ValueError("got differing extents in dimension %d (got %d and %d)" %
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_extents", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

static int __pyx_memoryview_err_dim(PyObject *__pyx_v_error, char *__pyx_v_msg, int __pyx_v_dim) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err_dim", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1258
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:
 *     raise error(msg.decode('ascii') % dim)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_err')
 */
  __pyx_t_2 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_dim); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyUnicode_Format(__pyx_t_2, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_INCREF(__pyx_v_error);
  __pyx_t_3 = __pyx_v_error; __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_3);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_3, function);
    }
  }
  __pyx_t_1 = (__pyx_t_2) ? __Pyx_PyObject_Call2Args(__pyx_t_3, __pyx_t_2, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4);
  __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 1258, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_Raise(__pyx_t_1, 0, 0, 0);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __PYX_ERR(1, 1258, __pyx_L1_error)

  /* "View.MemoryView":1257
 * 
 * @cname('__pyx_memoryview_err_dim')
 * cdef int _err_dim(object error, char *msg, int dim) except -1 with gil:             # <<<<<<<<<<<<<<
 *     raise error(msg.decode('ascii') % dim)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("View.MemoryView._err_dim", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

static int __pyx_memoryview_err(PyObject *__pyx_v_error, char *__pyx_v_msg) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("_err", 0);
  __Pyx_INCREF(__pyx_v_error);

  /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_msg != NULL) != 0);
  if (unlikely(__pyx_t_1)) {

    /* "View.MemoryView":1263
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))             # <<<<<<<<<<<<<<
 *     else:
 *         raise error
 */
    __pyx_t_3 = __Pyx_decode_c_string(__pyx_v_msg, 0, strlen(__pyx_v_msg), NULL, NULL, PyUnicode_DecodeASCII); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_INCREF(__pyx_v_error);
    __pyx_t_4 = __pyx_v_error; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    __pyx_t_2 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_4, __pyx_t_5, __pyx_t_3) : __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_t_3);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1263, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(1, 1263, __pyx_L1_error)

    /* "View.MemoryView":1262
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:
 *     if msg != NULL:             # <<<<<<<<<<<<<<
 *         raise error(msg.decode('ascii'))
 *     else:
 */
  }

  /* "View.MemoryView":1265
 *         raise error(msg.decode('ascii'))
 *     else:
 *         raise error             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_copy_contents')
 */
  /*else*/ {
    __Pyx_Raise(__pyx_v_error, 0, 0, 0);
    __PYX_ERR(1, 1265, __pyx_L1_error)
  }

  /* "View.MemoryView":1261
 * 
 * @cname('__pyx_memoryview_err')
 * cdef int _err(object error, char *msg) except -1 with gil:             # <<<<<<<<<<<<<<
 *     if msg != NULL:
 *         raise error(msg.decode('ascii'))
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView._err", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __Pyx_XDECREF(__pyx_v_error);
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
  return __pyx_r;
}

/* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

static int __pyx_memoryview_copy_contents(__Pyx_memviewslice __pyx_v_src, __Pyx_memviewslice __pyx_v_dst, int __pyx_v_src_ndim, int __pyx_v_dst_ndim, int __pyx_v_dtype_is_object) {
  void *__pyx_v_tmpdata;
  size_t __pyx_v_itemsize;
  int __pyx_v_i;
  char __pyx_v_order;
  int __pyx_v_broadcasting;
  int __pyx_v_direct_copy;
  __Pyx_memviewslice __pyx_v_tmp;
  int __pyx_v_ndim;
  int __pyx_r;
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  int __pyx_t_6;
  void *__pyx_t_7;
  int __pyx_t_8;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;

  /* "View.MemoryView":1276
 *     Check for overlapping memory and verify the shapes.
 *     """
 *     cdef void *tmpdata = NULL             # <<<<<<<<<<<<<<
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 */
  __pyx_v_tmpdata = NULL;

  /* "View.MemoryView":1277
 *     """
 *     cdef void *tmpdata = NULL
 *     cdef size_t itemsize = src.memview.view.itemsize             # <<<<<<<<<<<<<<
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 */
  __pyx_t_1 = __pyx_v_src.memview->view.itemsize;
  __pyx_v_itemsize = __pyx_t_1;

  /* "View.MemoryView":1279
 *     cdef size_t itemsize = src.memview.view.itemsize
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)             # <<<<<<<<<<<<<<
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False
 */
  __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_src), __pyx_v_src_ndim);

  /* "View.MemoryView":1280
 *     cdef int i
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False             # <<<<<<<<<<<<<<
 *     cdef bint direct_copy = False
 *     cdef __Pyx_memviewslice tmp
 */
  __pyx_v_broadcasting = 0;

  /* "View.MemoryView":1281
 *     cdef char order = get_best_order(&src, src_ndim)
 *     cdef bint broadcasting = False
 *     cdef bint direct_copy = False             # <<<<<<<<<<<<<<
 *     cdef __Pyx_memviewslice tmp
 * 
 */
  __pyx_v_direct_copy = 0;

  /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
  __pyx_t_2 = ((__pyx_v_src_ndim < __pyx_v_dst_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1285
 * 
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_src), __pyx_v_src_ndim, __pyx_v_dst_ndim);

    /* "View.MemoryView":1284
 *     cdef __Pyx_memviewslice tmp
 * 
 *     if src_ndim < dst_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  __pyx_t_2 = ((__pyx_v_dst_ndim < __pyx_v_src_ndim) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1287
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:
 *         broadcast_leading(&dst, dst_ndim, src_ndim)             # <<<<<<<<<<<<<<
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)
 */
    __pyx_memoryview_broadcast_leading((&__pyx_v_dst), __pyx_v_dst_ndim, __pyx_v_src_ndim);

    /* "View.MemoryView":1286
 *     if src_ndim < dst_ndim:
 *         broadcast_leading(&src, src_ndim, dst_ndim)
 *     elif dst_ndim < src_ndim:             # <<<<<<<<<<<<<<
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 */
  }
  __pyx_L3:;

  /* "View.MemoryView":1289
 *         broadcast_leading(&dst, dst_ndim, src_ndim)
 * 
 *     cdef int ndim = max(src_ndim, dst_ndim)             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim):
 */
  __pyx_t_3 = __pyx_v_dst_ndim;
  __pyx_t_4 = __pyx_v_src_ndim;
  if (((__pyx_t_3 > __pyx_t_4) != 0)) {
    __pyx_t_5 = __pyx_t_3;
  } else {
    __pyx_t_5 = __pyx_t_4;
  }
  __pyx_v_ndim = __pyx_t_5;

  /* "View.MemoryView":1291
 *     cdef int ndim = max(src_ndim, dst_ndim)
 * 
 *     for i in range(ndim):             # <<<<<<<<<<<<<<
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 */
  __pyx_t_5 = __pyx_v_ndim;
  __pyx_t_3 = __pyx_t_5;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) != (__pyx_v_dst.shape[__pyx_v_i])) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
      __pyx_t_2 = (((__pyx_v_src.shape[__pyx_v_i]) == 1) != 0);
      if (__pyx_t_2) {

        /* "View.MemoryView":1294
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:
 *                 broadcasting = True             # <<<<<<<<<<<<<<
 *                 src.strides[i] = 0
 *             else:
 */
        __pyx_v_broadcasting = 1;

        /* "View.MemoryView":1295
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 *                 src.strides[i] = 0             # <<<<<<<<<<<<<<
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 */
        (__pyx_v_src.strides[__pyx_v_i]) = 0;

        /* "View.MemoryView":1293
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:
 *             if src.shape[i] == 1:             # <<<<<<<<<<<<<<
 *                 broadcasting = True
 *                 src.strides[i] = 0
 */
        goto __pyx_L7;
      }

      /* "View.MemoryView":1297
 *                 src.strides[i] = 0
 *             else:
 *                 _err_extents(i, dst.shape[i], src.shape[i])             # <<<<<<<<<<<<<<
 * 
 *         if src.suboffsets[i] >= 0:
 */
      /*else*/ {
        __pyx_t_6 = __pyx_memoryview_err_extents(__pyx_v_i, (__pyx_v_dst.shape[__pyx_v_i]), (__pyx_v_src.shape[__pyx_v_i])); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1297, __pyx_L1_error)
      }
      __pyx_L7:;

      /* "View.MemoryView":1292
 * 
 *     for i in range(ndim):
 *         if src.shape[i] != dst.shape[i]:             # <<<<<<<<<<<<<<
 *             if src.shape[i] == 1:
 *                 broadcasting = True
 */
    }

    /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    __pyx_t_2 = (((__pyx_v_src.suboffsets[__pyx_v_i]) >= 0) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1300
 * 
 *         if src.suboffsets[i] >= 0:
 *             _err_dim(ValueError, "Dimension %d is not direct", i)             # <<<<<<<<<<<<<<
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 */
      __pyx_t_6 = __pyx_memoryview_err_dim(__pyx_builtin_ValueError, ((char *)"Dimension %d is not direct"), __pyx_v_i); if (unlikely(__pyx_t_6 == ((int)-1))) __PYX_ERR(1, 1300, __pyx_L1_error)

      /* "View.MemoryView":1299
 *                 _err_extents(i, dst.shape[i], src.shape[i])
 * 
 *         if src.suboffsets[i] >= 0:             # <<<<<<<<<<<<<<
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 */
    }
  }

  /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  __pyx_t_2 = (__pyx_slices_overlap((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    __pyx_t_2 = ((!(__pyx_memviewslice_is_contig(__pyx_v_src, __pyx_v_order, __pyx_v_ndim) != 0)) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1305
 * 
 *         if not slice_is_contig(src, order, ndim):
 *             order = get_best_order(&dst, ndim)             # <<<<<<<<<<<<<<
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 */
      __pyx_v_order = __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim);

      /* "View.MemoryView":1304
 *     if slices_overlap(&src, &dst, ndim, itemsize):
 * 
 *         if not slice_is_contig(src, order, ndim):             # <<<<<<<<<<<<<<
 *             order = get_best_order(&dst, ndim)
 * 
 */
    }

    /* "View.MemoryView":1307
 *             order = get_best_order(&dst, ndim)
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)             # <<<<<<<<<<<<<<
 *         src = tmp
 * 
 */
    __pyx_t_7 = __pyx_memoryview_copy_data_to_temp((&__pyx_v_src), (&__pyx_v_tmp), __pyx_v_order, __pyx_v_ndim); if (unlikely(__pyx_t_7 == ((void *)NULL))) __PYX_ERR(1, 1307, __pyx_L1_error)
    __pyx_v_tmpdata = __pyx_t_7;

    /* "View.MemoryView":1308
 * 
 *         tmpdata = copy_data_to_temp(&src, &tmp, order, ndim)
 *         src = tmp             # <<<<<<<<<<<<<<
 * 
 *     if not broadcasting:
 */
    __pyx_v_src = __pyx_v_tmp;

    /* "View.MemoryView":1302
 *             _err_dim(ValueError, "Dimension %d is not direct", i)
 * 
 *     if slices_overlap(&src, &dst, ndim, itemsize):             # <<<<<<<<<<<<<<
 * 
 *         if not slice_is_contig(src, order, ndim):
 */
  }

  /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = ((!(__pyx_v_broadcasting != 0)) != 0);
  if (__pyx_t_2) {

    /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'C', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1314
 * 
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)             # <<<<<<<<<<<<<<
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'C', __pyx_v_ndim);

      /* "View.MemoryView":1313
 * 
 * 
 *         if slice_is_contig(src, 'C', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 */
      goto __pyx_L12;
    }

    /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    __pyx_t_2 = (__pyx_memviewslice_is_contig(__pyx_v_src, 'F', __pyx_v_ndim) != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1316
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):
 *             direct_copy = slice_is_contig(dst, 'F', ndim)             # <<<<<<<<<<<<<<
 * 
 *         if direct_copy:
 */
      __pyx_v_direct_copy = __pyx_memviewslice_is_contig(__pyx_v_dst, 'F', __pyx_v_ndim);

      /* "View.MemoryView":1315
 *         if slice_is_contig(src, 'C', ndim):
 *             direct_copy = slice_is_contig(dst, 'C', ndim)
 *         elif slice_is_contig(src, 'F', ndim):             # <<<<<<<<<<<<<<
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 */
    }
    __pyx_L12:;

    /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_2 = (__pyx_v_direct_copy != 0);
    if (__pyx_t_2) {

      /* "View.MemoryView":1320
 *         if direct_copy:
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

      /* "View.MemoryView":1321
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))             # <<<<<<<<<<<<<<
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 */
      (void)(memcpy(__pyx_v_dst.data, __pyx_v_src.data, __pyx_memoryview_slice_get_size((&__pyx_v_src), __pyx_v_ndim)));

      /* "View.MemoryView":1322
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 *             free(tmpdata)
 *             return 0
 */
      __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

      /* "View.MemoryView":1323
 *             memcpy(dst.data, src.data, slice_get_size(&src, ndim))
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)             # <<<<<<<<<<<<<<
 *             return 0
 * 
 */
      free(__pyx_v_tmpdata);

      /* "View.MemoryView":1324
 *             refcount_copying(&dst, dtype_is_object, ndim, True)
 *             free(tmpdata)
 *             return 0             # <<<<<<<<<<<<<<
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):
 */
      __pyx_r = 0;
      goto __pyx_L0;

      /* "View.MemoryView":1318
 *             direct_copy = slice_is_contig(dst, 'F', ndim)
 * 
 *         if direct_copy:             # <<<<<<<<<<<<<<
 * 
 *             refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    }

    /* "View.MemoryView":1310
 *         src = tmp
 * 
 *     if not broadcasting:             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = (__pyx_v_order == 'F');
  if (__pyx_t_2) {
    __pyx_t_2 = ('F' == __pyx_get_best_slice_order((&__pyx_v_dst), __pyx_v_ndim));
  }
  __pyx_t_8 = (__pyx_t_2 != 0);
  if (__pyx_t_8) {

    /* "View.MemoryView":1329
 * 
 * 
 *         transpose_memslice(&src)             # <<<<<<<<<<<<<<
 *         transpose_memslice(&dst)
 * 
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_src)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1329, __pyx_L1_error)

    /* "View.MemoryView":1330
 * 
 *         transpose_memslice(&src)
 *         transpose_memslice(&dst)             # <<<<<<<<<<<<<<
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 */
    __pyx_t_5 = __pyx_memslice_transpose((&__pyx_v_dst)); if (unlikely(__pyx_t_5 == ((int)0))) __PYX_ERR(1, 1330, __pyx_L1_error)

    /* "View.MemoryView":1326
 *             return 0
 * 
 *     if order == 'F' == get_best_order(&dst, ndim):             # <<<<<<<<<<<<<<
 * 
 * 
 */
  }

  /* "View.MemoryView":1332
 *         transpose_memslice(&dst)
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1333
 * 
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)             # <<<<<<<<<<<<<<
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 */
  copy_strided_to_strided((&__pyx_v_src), (&__pyx_v_dst), __pyx_v_ndim, __pyx_v_itemsize);

  /* "View.MemoryView":1334
 *     refcount_copying(&dst, dtype_is_object, ndim, False)
 *     copy_strided_to_strided(&src, &dst, ndim, itemsize)
 *     refcount_copying(&dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 *     free(tmpdata)
 */
  __pyx_memoryview_refcount_copying((&__pyx_v_dst), __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1336
 *     refcount_copying(&dst, dtype_is_object, ndim, True)
 * 
 *     free(tmpdata)             # <<<<<<<<<<<<<<
 *     return 0
 * 
 */
  free(__pyx_v_tmpdata);

  /* "View.MemoryView":1337
 * 
 *     free(tmpdata)
 *     return 0             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 */
  __pyx_r = 0;
  goto __pyx_L0;

  /* "View.MemoryView":1268
 * 
 * @cname('__pyx_memoryview_copy_contents')
 * cdef int memoryview_copy_contents(__Pyx_memviewslice src,             # <<<<<<<<<<<<<<
 *                                   __Pyx_memviewslice dst,
 *                                   int src_ndim, int dst_ndim,
 */

  /* function exit code */
  __pyx_L1_error:;
  {
    #ifdef WITH_THREAD
    PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
    #endif
    __Pyx_AddTraceback("View.MemoryView.memoryview_copy_contents", __pyx_clineno, __pyx_lineno, __pyx_filename);
    #ifdef WITH_THREAD
    __Pyx_PyGILState_Release(__pyx_gilstate_save);
    #endif
  }
  __pyx_r = -1;
  __pyx_L0:;
  return __pyx_r;
}

/* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

static void __pyx_memoryview_broadcast_leading(__Pyx_memviewslice *__pyx_v_mslice, int __pyx_v_ndim, int __pyx_v_ndim_other) {
  int __pyx_v_i;
  int __pyx_v_offset;
  int __pyx_t_1;
  int __pyx_t_2;
  int __pyx_t_3;

  /* "View.MemoryView":1344
 *                             int ndim_other) nogil:
 *     cdef int i
 *     cdef int offset = ndim_other - ndim             # <<<<<<<<<<<<<<
 * 
 *     for i in range(ndim - 1, -1, -1):
 */
  __pyx_v_offset = (__pyx_v_ndim_other - __pyx_v_ndim);

  /* "View.MemoryView":1346
 *     cdef int offset = ndim_other - ndim
 * 
 *     for i in range(ndim - 1, -1, -1):             # <<<<<<<<<<<<<<
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 */
  for (__pyx_t_1 = (__pyx_v_ndim - 1); __pyx_t_1 > -1; __pyx_t_1-=1) {
    __pyx_v_i = __pyx_t_1;

    /* "View.MemoryView":1347
 * 
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]             # <<<<<<<<<<<<<<
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 */
    (__pyx_v_mslice->shape[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->shape[__pyx_v_i]);

    /* "View.MemoryView":1348
 *     for i in range(ndim - 1, -1, -1):
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 */
    (__pyx_v_mslice->strides[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->strides[__pyx_v_i]);

    /* "View.MemoryView":1349
 *         mslice.shape[i + offset] = mslice.shape[i]
 *         mslice.strides[i + offset] = mslice.strides[i]
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]             # <<<<<<<<<<<<<<
 * 
 *     for i in range(offset):
 */
    (__pyx_v_mslice->suboffsets[(__pyx_v_i + __pyx_v_offset)]) = (__pyx_v_mslice->suboffsets[__pyx_v_i]);
  }

  /* "View.MemoryView":1351
 *         mslice.suboffsets[i + offset] = mslice.suboffsets[i]
 * 
 *     for i in range(offset):             # <<<<<<<<<<<<<<
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 */
  __pyx_t_1 = __pyx_v_offset;
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1352
 * 
 *     for i in range(offset):
 *         mslice.shape[i] = 1             # <<<<<<<<<<<<<<
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1
 */
    (__pyx_v_mslice->shape[__pyx_v_i]) = 1;

    /* "View.MemoryView":1353
 *     for i in range(offset):
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]             # <<<<<<<<<<<<<<
 *         mslice.suboffsets[i] = -1
 * 
 */
    (__pyx_v_mslice->strides[__pyx_v_i]) = (__pyx_v_mslice->strides[0]);

    /* "View.MemoryView":1354
 *         mslice.shape[i] = 1
 *         mslice.strides[i] = mslice.strides[0]
 *         mslice.suboffsets[i] = -1             # <<<<<<<<<<<<<<
 * 
 * 
 */
    (__pyx_v_mslice->suboffsets[__pyx_v_i]) = -1L;
  }

  /* "View.MemoryView":1340
 * 
 * @cname('__pyx_memoryview_broadcast_leading')
 * cdef void broadcast_leading(__Pyx_memviewslice *mslice,             # <<<<<<<<<<<<<<
 *                             int ndim,
 *                             int ndim_other) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

static void __pyx_memoryview_refcount_copying(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_dtype_is_object, int __pyx_v_ndim, int __pyx_v_inc) {
  int __pyx_t_1;

  /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  __pyx_t_1 = (__pyx_v_dtype_is_object != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1367
 * 
 *     if dtype_is_object:
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,             # <<<<<<<<<<<<<<
 *                                            dst.strides, ndim, inc)
 * 
 */
    __pyx_memoryview_refcount_objects_in_slice_with_gil(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_inc);

    /* "View.MemoryView":1366
 * 
 * 
 *     if dtype_is_object:             # <<<<<<<<<<<<<<
 *         refcount_objects_in_slice_with_gil(dst.data, dst.shape,
 *                                            dst.strides, ndim, inc)
 */
  }

  /* "View.MemoryView":1362
 * 
 * @cname('__pyx_memoryview_refcount_copying')
 * cdef void refcount_copying(__Pyx_memviewslice *dst, bint dtype_is_object,             # <<<<<<<<<<<<<<
 *                            int ndim, bint inc) nogil:
 * 
 */

  /* function exit code */
}

/* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

static void __pyx_memoryview_refcount_objects_in_slice_with_gil(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  __Pyx_RefNannyDeclarations
  #ifdef WITH_THREAD
  PyGILState_STATE __pyx_gilstate_save = __Pyx_PyGILState_Ensure();
  #endif
  __Pyx_RefNannySetupContext("refcount_objects_in_slice_with_gil", 0);

  /* "View.MemoryView":1374
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 *     refcount_objects_in_slice(data, shape, strides, ndim, inc)             # <<<<<<<<<<<<<<
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 */
  __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, __pyx_v_shape, __pyx_v_strides, __pyx_v_ndim, __pyx_v_inc);

  /* "View.MemoryView":1371
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice_with_gil')
 * cdef void refcount_objects_in_slice_with_gil(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                              Py_ssize_t *strides, int ndim,
 *                                              bint inc) with gil:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  #ifdef WITH_THREAD
  __Pyx_PyGILState_Release(__pyx_gilstate_save);
  #endif
}

/* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

static void __pyx_memoryview_refcount_objects_in_slice(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, int __pyx_v_inc) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  __Pyx_RefNannySetupContext("refcount_objects_in_slice", 0);

  /* "View.MemoryView":1381
 *     cdef Py_ssize_t i
 * 
 *     for i in range(shape[0]):             # <<<<<<<<<<<<<<
 *         if ndim == 1:
 *             if inc:
 */
  __pyx_t_1 = (__pyx_v_shape[0]);
  __pyx_t_2 = __pyx_t_1;
  for (__pyx_t_3 = 0; __pyx_t_3 < __pyx_t_2; __pyx_t_3+=1) {
    __pyx_v_i = __pyx_t_3;

    /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
    __pyx_t_4 = ((__pyx_v_ndim == 1) != 0);
    if (__pyx_t_4) {

      /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
      __pyx_t_4 = (__pyx_v_inc != 0);
      if (__pyx_t_4) {

        /* "View.MemoryView":1384
 *         if ndim == 1:
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])
 */
        Py_INCREF((((PyObject **)__pyx_v_data)[0]));

        /* "View.MemoryView":1383
 *     for i in range(shape[0]):
 *         if ndim == 1:
 *             if inc:             # <<<<<<<<<<<<<<
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 */
        goto __pyx_L6;
      }

      /* "View.MemoryView":1386
 *                 Py_INCREF((<PyObject **> data)[0])
 *             else:
 *                 Py_DECREF((<PyObject **> data)[0])             # <<<<<<<<<<<<<<
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 */
      /*else*/ {
        Py_DECREF((((PyObject **)__pyx_v_data)[0]));
      }
      __pyx_L6:;

      /* "View.MemoryView":1382
 * 
 *     for i in range(shape[0]):
 *         if ndim == 1:             # <<<<<<<<<<<<<<
 *             if inc:
 *                 Py_INCREF((<PyObject **> data)[0])
 */
      goto __pyx_L5;
    }

    /* "View.MemoryView":1388
 *                 Py_DECREF((<PyObject **> data)[0])
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                       ndim - 1, inc)
 * 
 */
    /*else*/ {

      /* "View.MemoryView":1389
 *         else:
 *             refcount_objects_in_slice(data, shape + 1, strides + 1,
 *                                       ndim - 1, inc)             # <<<<<<<<<<<<<<
 * 
 *         data += strides[0]
 */
      __pyx_memoryview_refcount_objects_in_slice(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_inc);
    }
    __pyx_L5:;

    /* "View.MemoryView":1391
 *                                       ndim - 1, inc)
 * 
 *         data += strides[0]             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_v_data = (__pyx_v_data + (__pyx_v_strides[0]));
  }

  /* "View.MemoryView":1377
 * 
 * @cname('__pyx_memoryview_refcount_objects_in_slice')
 * cdef void refcount_objects_in_slice(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                                     Py_ssize_t *strides, int ndim, bint inc):
 *     cdef Py_ssize_t i
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

static void __pyx_memoryview_slice_assign_scalar(__Pyx_memviewslice *__pyx_v_dst, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item, int __pyx_v_dtype_is_object) {

  /* "View.MemoryView":1400
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)             # <<<<<<<<<<<<<<
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 0);

  /* "View.MemoryView":1401
 *                               bint dtype_is_object) nogil:
 *     refcount_copying(dst, dtype_is_object, ndim, False)
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,             # <<<<<<<<<<<<<<
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)
 */
  __pyx_memoryview__slice_assign_scalar(__pyx_v_dst->data, __pyx_v_dst->shape, __pyx_v_dst->strides, __pyx_v_ndim, __pyx_v_itemsize, __pyx_v_item);

  /* "View.MemoryView":1403
 *     _slice_assign_scalar(dst.data, dst.shape, dst.strides, ndim,
 *                          itemsize, item)
 *     refcount_copying(dst, dtype_is_object, ndim, True)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_memoryview_refcount_copying(__pyx_v_dst, __pyx_v_dtype_is_object, __pyx_v_ndim, 1);

  /* "View.MemoryView":1397
 * 
 * @cname('__pyx_memoryview_slice_assign_scalar')
 * cdef void slice_assign_scalar(__Pyx_memviewslice *dst, int ndim,             # <<<<<<<<<<<<<<
 *                               size_t itemsize, void *item,
 *                               bint dtype_is_object) nogil:
 */

  /* function exit code */
}

/* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

static void __pyx_memoryview__slice_assign_scalar(char *__pyx_v_data, Py_ssize_t *__pyx_v_shape, Py_ssize_t *__pyx_v_strides, int __pyx_v_ndim, size_t __pyx_v_itemsize, void *__pyx_v_item) {
  CYTHON_UNUSED Py_ssize_t __pyx_v_i;
  Py_ssize_t __pyx_v_stride;
  Py_ssize_t __pyx_v_extent;
  int __pyx_t_1;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  Py_ssize_t __pyx_t_4;

  /* "View.MemoryView":1411
 *                               size_t itemsize, void *item) nogil:
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]             # <<<<<<<<<<<<<<
 *     cdef Py_ssize_t extent = shape[0]
 * 
 */
  __pyx_v_stride = (__pyx_v_strides[0]);

  /* "View.MemoryView":1412
 *     cdef Py_ssize_t i
 *     cdef Py_ssize_t stride = strides[0]
 *     cdef Py_ssize_t extent = shape[0]             # <<<<<<<<<<<<<<
 * 
 *     if ndim == 1:
 */
  __pyx_v_extent = (__pyx_v_shape[0]);

  /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
  __pyx_t_1 = ((__pyx_v_ndim == 1) != 0);
  if (__pyx_t_1) {

    /* "View.MemoryView":1415
 * 
 *     if ndim == 1:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             memcpy(data, item, itemsize)
 *             data += stride
 */
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1416
 *     if ndim == 1:
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)             # <<<<<<<<<<<<<<
 *             data += stride
 *     else:
 */
      (void)(memcpy(__pyx_v_data, __pyx_v_item, __pyx_v_itemsize));

      /* "View.MemoryView":1417
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 *             data += stride             # <<<<<<<<<<<<<<
 *     else:
 *         for i in range(extent):
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }

    /* "View.MemoryView":1414
 *     cdef Py_ssize_t extent = shape[0]
 * 
 *     if ndim == 1:             # <<<<<<<<<<<<<<
 *         for i in range(extent):
 *             memcpy(data, item, itemsize)
 */
    goto __pyx_L3;
  }

  /* "View.MemoryView":1419
 *             data += stride
 *     else:
 *         for i in range(extent):             # <<<<<<<<<<<<<<
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 */
  /*else*/ {
    __pyx_t_2 = __pyx_v_extent;
    __pyx_t_3 = __pyx_t_2;
    for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
      __pyx_v_i = __pyx_t_4;

      /* "View.MemoryView":1420
 *     else:
 *         for i in range(extent):
 *             _slice_assign_scalar(data, shape + 1, strides + 1,             # <<<<<<<<<<<<<<
 *                                 ndim - 1, itemsize, item)
 *             data += stride
 */
      __pyx_memoryview__slice_assign_scalar(__pyx_v_data, (__pyx_v_shape + 1), (__pyx_v_strides + 1), (__pyx_v_ndim - 1), __pyx_v_itemsize, __pyx_v_item);

      /* "View.MemoryView":1422
 *             _slice_assign_scalar(data, shape + 1, strides + 1,
 *                                 ndim - 1, itemsize, item)
 *             data += stride             # <<<<<<<<<<<<<<
 * 
 * 
 */
      __pyx_v_data = (__pyx_v_data + __pyx_v_stride);
    }
  }
  __pyx_L3:;

  /* "View.MemoryView":1407
 * 
 * @cname('__pyx_memoryview__slice_assign_scalar')
 * cdef void _slice_assign_scalar(char *data, Py_ssize_t *shape,             # <<<<<<<<<<<<<<
 *                               Py_ssize_t *strides, int ndim,
 *                               size_t itemsize, void *item) nogil:
 */

  /* function exit code */
}

/* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

/* Python wrapper */
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum = {"__pyx_unpickle_Enum", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_15View_dot_MemoryView_1__pyx_unpickle_Enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v___pyx_type = 0;
  long __pyx_v___pyx_checksum;
  PyObject *__pyx_v___pyx_state = 0;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_pyx_type,&__pyx_n_s_pyx_checksum,&__pyx_n_s_pyx_state,0};
    PyObject* values[3] = {0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        CYTHON_FALLTHROUGH;
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        CYTHON_FALLTHROUGH;
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        CYTHON_FALLTHROUGH;
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_type)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        CYTHON_FALLTHROUGH;
        case  1:
        if (likely((values[1] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_checksum)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 1); __PYX_ERR(1, 1, __pyx_L3_error)
        }
        CYTHON_FALLTHROUGH;
        case  2:
        if (likely((values[2] = __Pyx_PyDict_GetItemStr(__pyx_kwds, __pyx_n_s_pyx_state)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, 2); __PYX_ERR(1, 1, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "__pyx_unpickle_Enum") < 0)) __PYX_ERR(1, 1, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 3) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
    }
    __pyx_v___pyx_type = values[0];
    __pyx_v___pyx_checksum = __Pyx_PyInt_As_long(values[1]); if (unlikely((__pyx_v___pyx_checksum == (long)-1) && PyErr_Occurred())) __PYX_ERR(1, 1, __pyx_L3_error)
    __pyx_v___pyx_state = values[2];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("__pyx_unpickle_Enum", 1, 3, 3, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(1, 1, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(__pyx_self, __pyx_v___pyx_type, __pyx_v___pyx_checksum, __pyx_v___pyx_state);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_15View_dot_MemoryView___pyx_unpickle_Enum(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v___pyx_type, long __pyx_v___pyx_checksum, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_v___pyx_PickleError = 0;
  PyObject *__pyx_v___pyx_result = 0;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum", 0);

  /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  __pyx_t_1 = ((__pyx_v___pyx_checksum != 0xb068931) != 0);
  if (__pyx_t_1) {

    /* "(tree fragment)":5
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError             # <<<<<<<<<<<<<<
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_PickleError);
    __Pyx_GIVEREF(__pyx_n_s_PickleError);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_PickleError);
    __pyx_t_3 = __Pyx_Import(__pyx_n_s_pickle, __pyx_t_2, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_3, __pyx_n_s_PickleError); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 5, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_t_2);
    __pyx_v___pyx_PickleError = __pyx_t_2;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":6
 *     if __pyx_checksum != 0xb068931:
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)             # <<<<<<<<<<<<<<
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 */
    __pyx_t_2 = __Pyx_PyInt_From_long(__pyx_v___pyx_checksum); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_4 = __Pyx_PyString_Format(__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_t_2); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_INCREF(__pyx_v___pyx_PickleError);
    __pyx_t_2 = __pyx_v___pyx_PickleError; __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    __pyx_t_3 = (__pyx_t_5) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_5, __pyx_t_4) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4);
    __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 6, __pyx_L1_error)

    /* "(tree fragment)":4
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 *     if __pyx_checksum != 0xb068931:             # <<<<<<<<<<<<<<
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 */
  }

  /* "(tree fragment)":7
 *         from pickle import PickleError as __pyx_PickleError
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)             # <<<<<<<<<<<<<<
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_MemviewEnum_type), __pyx_n_s_new); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  __pyx_t_3 = (__pyx_t_4) ? __Pyx_PyObject_Call2Args(__pyx_t_2, __pyx_t_4, __pyx_v___pyx_type) : __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v___pyx_type);
  __Pyx_XDECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_v___pyx_result = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  __pyx_t_1 = (__pyx_v___pyx_state != Py_None);
  __pyx_t_6 = (__pyx_t_1 != 0);
  if (__pyx_t_6) {

    /* "(tree fragment)":9
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)             # <<<<<<<<<<<<<<
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 */
    if (!(likely(PyTuple_CheckExact(__pyx_v___pyx_state))||((__pyx_v___pyx_state) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_v___pyx_state)->tp_name), 0))) __PYX_ERR(1, 9, __pyx_L1_error)
    __pyx_t_3 = __pyx_unpickle_Enum__set_state(((struct __pyx_MemviewEnum_obj *)__pyx_v___pyx_result), ((PyObject*)__pyx_v___pyx_state)); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 9, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

    /* "(tree fragment)":8
 *         raise __pyx_PickleError("Incompatible checksums (%s vs 0xb068931 = (name))" % __pyx_checksum)
 *     __pyx_result = Enum.__new__(__pyx_type)
 *     if __pyx_state is not None:             # <<<<<<<<<<<<<<
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 */
  }

  /* "(tree fragment)":10
 *     if __pyx_state is not None:
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result             # <<<<<<<<<<<<<<
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v___pyx_result);
  __pyx_r = __pyx_v___pyx_result;
  goto __pyx_L0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v___pyx_PickleError);
  __Pyx_XDECREF(__pyx_v___pyx_result);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

static PyObject *__pyx_unpickle_Enum__set_state(struct __pyx_MemviewEnum_obj *__pyx_v___pyx_result, PyObject *__pyx_v___pyx_state) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__pyx_unpickle_Enum__set_state", 0);

  /* "(tree fragment)":12
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]             # <<<<<<<<<<<<<<
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
    __PYX_ERR(1, 12, __pyx_L1_error)
  }
  __pyx_t_1 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_1);
  __Pyx_GOTREF(__pyx_v___pyx_result->name);
  __Pyx_DECREF(__pyx_v___pyx_result->name);
  __pyx_v___pyx_result->name = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  if (unlikely(__pyx_v___pyx_state == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "object of type 'NoneType' has no len()");
    __PYX_ERR(1, 13, __pyx_L1_error)
  }
  __pyx_t_3 = PyTuple_GET_SIZE(__pyx_v___pyx_state); if (unlikely(__pyx_t_3 == ((Py_ssize_t)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_4 = ((__pyx_t_3 > 1) != 0);
  if (__pyx_t_4) {
  } else {
    __pyx_t_2 = __pyx_t_4;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_4 = __Pyx_HasAttr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(__pyx_t_4 == ((int)-1))) __PYX_ERR(1, 13, __pyx_L1_error)
  __pyx_t_5 = (__pyx_t_4 != 0);
  __pyx_t_2 = __pyx_t_5;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_2) {

    /* "(tree fragment)":14
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 *         __pyx_result.__dict__.update(__pyx_state[1])             # <<<<<<<<<<<<<<
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v___pyx_result), __pyx_n_s_dict); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_update); if (unlikely(!__pyx_t_7)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_7);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(__pyx_v___pyx_state == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 14, __pyx_L1_error)
    }
    __pyx_t_6 = __Pyx_GetItemInt_Tuple(__pyx_v___pyx_state, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
      __pyx_t_8 = PyMethod_GET_SELF(__pyx_t_7);
      if (likely(__pyx_t_8)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_7, function);
      }
    }
    __pyx_t_1 = (__pyx_t_8) ? __Pyx_PyObject_Call2Args(__pyx_t_7, __pyx_t_8, __pyx_t_6) : __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_6);
    __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 14, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "(tree fragment)":13
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):             # <<<<<<<<<<<<<<
 *         __pyx_result.__dict__.update(__pyx_state[1])
 */
  }

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("View.MemoryView.__pyx_unpickle_Enum__set_state", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_3nes_6cycore_7mos6502_MOS6502 __pyx_vtable_3nes_6cycore_7mos6502_MOS6502;

static PyObject *__pyx_tp_new_3nes_6cycore_7mos6502_MOS6502(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)o);
  p->__pyx_vtab = __pyx_vtabptr_3nes_6cycore_7mos6502_MOS6502;
  p->memory = ((struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *)Py_None); Py_INCREF(Py_None);
  p->instructions = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_3nes_6cycore_7mos6502_MOS6502(PyObject *o) {
  struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *p = (struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->memory);
  Py_CLEAR(p->instructions);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_3nes_6cycore_7mos6502_MOS6502(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *p = (struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)o;
  if (p->memory) {
    e = (*v)(((PyObject *)p->memory), a); if (e) return e;
  }
  if (p->instructions) {
    e = (*v)(p->instructions, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_3nes_6cycore_7mos6502_MOS6502(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *p = (struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *)o;
  tmp = ((PyObject*)p->memory);
  p->memory = ((struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->instructions);
  p->instructions = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyObject *__pyx_getprop_3nes_6cycore_7mos6502_7MOS6502_cycles_since_reset(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_3nes_6cycore_7mos6502_7MOS6502_18cycles_since_reset_1__get__(o);
}

static PyMethodDef __pyx_methods_3nes_6cycore_7mos6502_MOS6502[] = {
  {"reset", (PyCFunction)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_3reset, METH_NOARGS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_2reset},
  {"_make_bytecode_dict", (PyCFunction)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_5_make_bytecode_dict, METH_NOARGS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_4_make_bytecode_dict},
  {"format_instruction", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_7format_instruction, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_6format_instruction},
  {"log_line", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_9log_line, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_8log_line},
  {"set_reset_vector", (PyCFunction)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_11set_reset_vector, METH_O, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_10set_reset_vector},
  {"_dop", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_13_dop, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_12_dop},
  {"_top", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_15_top, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_14_top},
  {"_kil", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_17_kil, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_16_kil},
  {"_lax", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_19_lax, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_18_lax},
  {"_aax", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_21_aax, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_20_aax},
  {"_dcp", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_23_dcp, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_22_dcp},
  {"_isc", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_25_isc, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_24_isc},
  {"_slo", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_27_slo, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_26_slo},
  {"_rla", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_29_rla, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_28_rla},
  {"_rra", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_31_rra, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_30_rra},
  {"_sre", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_33_sre, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_32_sre},
  {"_arr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_35_arr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_34_arr},
  {"_asr", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_37_asr, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_36_asr},
  {"_atx", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_39_atx, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_38_atx},
  {"_aac", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_41_aac, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_40_aac},
  {"_axa", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_43_axa, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_42_axa},
  {"_axs", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_45_axs, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_44_axs},
  {"_lar", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_47_lar, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_46_lar},
  {"_sxa", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_49_sxa, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_48_sxa},
  {"_sya", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_51_sya, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_50_sya},
  {"_xaa", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_53_xaa, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_52_xaa},
  {"_xas", (PyCFunction)(void*)(PyCFunctionWithKeywords)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_55_xas, METH_VARARGS|METH_KEYWORDS, __pyx_doc_3nes_6cycore_7mos6502_7MOS6502_54_xas},
  {"__reduce_cython__", (PyCFunction)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_57__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw_3nes_6cycore_7mos6502_7MOS6502_59__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_3nes_6cycore_7mos6502_MOS6502[] = {
  {(char *)"cycles_since_reset", __pyx_getprop_3nes_6cycore_7mos6502_7MOS6502_cycles_since_reset, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type_3nes_6cycore_7mos6502_MOS6502 = {
  PyVarObject_HEAD_INIT(0, 0)
  "nes.cycore.mos6502.MOS6502", /*tp_name*/
  sizeof(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_3nes_6cycore_7mos6502_MOS6502, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "\n    Software emulator for MOS Technologies 6502 CPU.  The chip in the NES is a modified version of the 6502 that does\n    not support BCD mode (which makes it a bit simpler).\n\n    References:\n       [1] https://www.masswerk.at/6502/6502_instruction_set.html\n       [2] http://www.obelisk.me.uk/6502/addressing.html\n       [3] https://www.csh.rit.edu/~moffitt/docs/6502.html\n\n       [4] V-flag:  http://www.6502.org/tutorials/vflag.html\n       [5] Decimal mode: http://www.6502.org/tutorials/decimal_mode.html\n       [6] PC: http://6502.org/tutorials/6502opcodes.html#PC\n       [7] http://forum.6502.org/viewtopic.php?t=1708\n\n       [8] Assembler online: https://www.masswerk.at/6502/assembler.html\n       [9] CPU startup sequence:  https://www.pagetable.com/?p=410\n       [10] B-flag: http://wiki.nesdev.com/w/index.php/Status_flags#The_B_flag\n       [11] V-flag: http://www.righto.com/2012/12/the-6502-overflow-flag-explained.html\n       [12] JMP indirect 'bug':  http://forum.6502.org/viewtopic.php?f=2&t=6170\n\n       [13] Undocumented opcodes: http://nesdev.com/undocumented_opcodes.txt\n       [14] Undocumented opcodes: http://www.ffd2.com/fridge/docs/6502-NMOS.extra.opcodes\n\n       [15] Amazing timing data: http://www.6502.org/users/andre/petindex/local/64doc.txt\n\n       [16] Undocumented opcodes: http://hitmen.c02.at/files/docs/c64/NoMoreSecrets-NMOS6510UnintendedOpcodes-20162412.pdf\n\n    ", /*tp_doc*/
  __pyx_tp_traverse_3nes_6cycore_7mos6502_MOS6502, /*tp_traverse*/
  __pyx_tp_clear_3nes_6cycore_7mos6502_MOS6502, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_3nes_6cycore_7mos6502_MOS6502, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_3nes_6cycore_7mos6502_MOS6502, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_pw_3nes_6cycore_7mos6502_7MOS6502_1__init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_3nes_6cycore_7mos6502_MOS6502, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_array __pyx_vtable_array;

static PyObject *__pyx_tp_new_array(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_array_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_array_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_array;
  p->mode = ((PyObject*)Py_None); Py_INCREF(Py_None);
  p->_format = ((PyObject*)Py_None); Py_INCREF(Py_None);
  if (unlikely(__pyx_array___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_array(PyObject *o) {
  struct __pyx_array_obj *p = (struct __pyx_array_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_array___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->mode);
  Py_CLEAR(p->_format);
  (*Py_TYPE(o)->tp_free)(o);
}
static PyObject *__pyx_sq_item_array(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_array(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_array___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_tp_getattro_array(PyObject *o, PyObject *n) {
  PyObject *v = __Pyx_PyObject_GenericGetAttr(o, n);
  if (!v && PyErr_ExceptionMatches(PyExc_AttributeError)) {
    PyErr_Clear();
    v = __pyx_array___getattr__(o, n);
  }
  return v;
}

static PyObject *__pyx_getprop___pyx_array_memview(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_5array_7memview_1__get__(o);
}

static PyMethodDef __pyx_methods_array[] = {
  {"__getattr__", (PyCFunction)__pyx_array___getattr__, METH_O|METH_COEXIST, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_array_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_array_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_array[] = {
  {(char *)"memview", __pyx_getprop___pyx_array_memview, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_array = {
  __pyx_array___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_array, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_array = {
  __pyx_array___len__, /*mp_length*/
  __pyx_array___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_array, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_array = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_array_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_array = {
  PyVarObject_HEAD_INIT(0, 0)
  "nes.cycore.mos6502.array", /*tp_name*/
  sizeof(struct __pyx_array_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_array, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_array, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_array, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  __pyx_tp_getattro_array, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_array, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  0, /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_array, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_array, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_array, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyObject *__pyx_tp_new_Enum(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_MemviewEnum_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_MemviewEnum_obj *)o);
  p->name = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_Enum(PyObject *o) {
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->name);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_Enum(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  if (p->name) {
    e = (*v)(p->name, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_Enum(PyObject *o) {
  PyObject* tmp;
  struct __pyx_MemviewEnum_obj *p = (struct __pyx_MemviewEnum_obj *)o;
  tmp = ((PyObject*)p->name);
  p->name = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_Enum[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_MemviewEnum_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_MemviewEnum = {
  PyVarObject_HEAD_INIT(0, 0)
  "nes.cycore.mos6502.Enum", /*tp_name*/
  sizeof(struct __pyx_MemviewEnum_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_Enum, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_MemviewEnum___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_Enum, /*tp_traverse*/
  __pyx_tp_clear_Enum, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_Enum, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  __pyx_MemviewEnum___init__, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_Enum, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct_memoryview __pyx_vtable_memoryview;

static PyObject *__pyx_tp_new_memoryview(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryview_obj *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryview_obj *)o);
  p->__pyx_vtab = __pyx_vtabptr_memoryview;
  p->obj = Py_None; Py_INCREF(Py_None);
  p->_size = Py_None; Py_INCREF(Py_None);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  p->view.obj = NULL;
  if (unlikely(__pyx_memoryview___cinit__(o, a, k) < 0)) goto bad;
  return o;
  bad:
  Py_DECREF(o); o = 0;
  return NULL;
}

static void __pyx_tp_dealloc_memoryview(PyObject *o) {
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryview___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->obj);
  Py_CLEAR(p->_size);
  Py_CLEAR(p->_array_interface);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_memoryview(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  if (p->obj) {
    e = (*v)(p->obj, a); if (e) return e;
  }
  if (p->_size) {
    e = (*v)(p->_size, a); if (e) return e;
  }
  if (p->_array_interface) {
    e = (*v)(p->_array_interface, a); if (e) return e;
  }
  if (p->view.obj) {
    e = (*v)(p->view.obj, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_memoryview(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryview_obj *p = (struct __pyx_memoryview_obj *)o;
  tmp = ((PyObject*)p->obj);
  p->obj = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_size);
  p->_size = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->_array_interface);
  p->_array_interface = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  Py_CLEAR(p->view.obj);
  return 0;
}
static PyObject *__pyx_sq_item_memoryview(PyObject *o, Py_ssize_t i) {
  PyObject *r;
  PyObject *x = PyInt_FromSsize_t(i); if(!x) return 0;
  r = Py_TYPE(o)->tp_as_mapping->mp_subscript(o, x);
  Py_DECREF(x);
  return r;
}

static int __pyx_mp_ass_subscript_memoryview(PyObject *o, PyObject *i, PyObject *v) {
  if (v) {
    return __pyx_memoryview___setitem__(o, i, v);
  }
  else {
    PyErr_Format(PyExc_NotImplementedError,
      "Subscript deletion not supported by %.200s", Py_TYPE(o)->tp_name);
    return -1;
  }
}

static PyObject *__pyx_getprop___pyx_memoryview_T(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_1T_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4base_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_shape(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_5shape_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_strides(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_7strides_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_suboffsets(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_10suboffsets_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_ndim(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4ndim_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_itemsize(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_8itemsize_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_nbytes(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_6nbytes_1__get__(o);
}

static PyObject *__pyx_getprop___pyx_memoryview_size(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_10memoryview_4size_1__get__(o);
}

static PyMethodDef __pyx_methods_memoryview[] = {
  {"is_c_contig", (PyCFunction)__pyx_memoryview_is_c_contig, METH_NOARGS, 0},
  {"is_f_contig", (PyCFunction)__pyx_memoryview_is_f_contig, METH_NOARGS, 0},
  {"copy", (PyCFunction)__pyx_memoryview_copy, METH_NOARGS, 0},
  {"copy_fortran", (PyCFunction)__pyx_memoryview_copy_fortran, METH_NOARGS, 0},
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryview_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets_memoryview[] = {
  {(char *)"T", __pyx_getprop___pyx_memoryview_T, 0, (char *)0, 0},
  {(char *)"base", __pyx_getprop___pyx_memoryview_base, 0, (char *)0, 0},
  {(char *)"shape", __pyx_getprop___pyx_memoryview_shape, 0, (char *)0, 0},
  {(char *)"strides", __pyx_getprop___pyx_memoryview_strides, 0, (char *)0, 0},
  {(char *)"suboffsets", __pyx_getprop___pyx_memoryview_suboffsets, 0, (char *)0, 0},
  {(char *)"ndim", __pyx_getprop___pyx_memoryview_ndim, 0, (char *)0, 0},
  {(char *)"itemsize", __pyx_getprop___pyx_memoryview_itemsize, 0, (char *)0, 0},
  {(char *)"nbytes", __pyx_getprop___pyx_memoryview_nbytes, 0, (char *)0, 0},
  {(char *)"size", __pyx_getprop___pyx_memoryview_size, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PySequenceMethods __pyx_tp_as_sequence_memoryview = {
  __pyx_memoryview___len__, /*sq_length*/
  0, /*sq_concat*/
  0, /*sq_repeat*/
  __pyx_sq_item_memoryview, /*sq_item*/
  0, /*sq_slice*/
  0, /*sq_ass_item*/
  0, /*sq_ass_slice*/
  0, /*sq_contains*/
  0, /*sq_inplace_concat*/
  0, /*sq_inplace_repeat*/
};

static PyMappingMethods __pyx_tp_as_mapping_memoryview = {
  __pyx_memoryview___len__, /*mp_length*/
  __pyx_memoryview___getitem__, /*mp_subscript*/
  __pyx_mp_ass_subscript_memoryview, /*mp_ass_subscript*/
};

static PyBufferProcs __pyx_tp_as_buffer_memoryview = {
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getreadbuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getwritebuffer*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getsegcount*/
  #endif
  #if PY_MAJOR_VERSION < 3
  0, /*bf_getcharbuffer*/
  #endif
  __pyx_memoryview_getbuffer, /*bf_getbuffer*/
  0, /*bf_releasebuffer*/
};

static PyTypeObject __pyx_type___pyx_memoryview = {
  PyVarObject_HEAD_INIT(0, 0)
  "nes.cycore.mos6502.memoryview", /*tp_name*/
  sizeof(struct __pyx_memoryview_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_memoryview, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  __pyx_memoryview___repr__, /*tp_repr*/
  0, /*tp_as_number*/
  &__pyx_tp_as_sequence_memoryview, /*tp_as_sequence*/
  &__pyx_tp_as_mapping_memoryview, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  __pyx_memoryview___str__, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  &__pyx_tp_as_buffer_memoryview, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_memoryview, /*tp_traverse*/
  __pyx_tp_clear_memoryview, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_memoryview, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets_memoryview, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_memoryview, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};
static struct __pyx_vtabstruct__memoryviewslice __pyx_vtable__memoryviewslice;

static PyObject *__pyx_tp_new__memoryviewslice(PyTypeObject *t, PyObject *a, PyObject *k) {
  struct __pyx_memoryviewslice_obj *p;
  PyObject *o = __pyx_tp_new_memoryview(t, a, k);
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_memoryviewslice_obj *)o);
  p->__pyx_base.__pyx_vtab = (struct __pyx_vtabstruct_memoryview*)__pyx_vtabptr__memoryviewslice;
  p->from_object = Py_None; Py_INCREF(Py_None);
  p->from_slice.memview = NULL;
  return o;
}

static void __pyx_tp_dealloc__memoryviewslice(PyObject *o) {
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  #if CYTHON_USE_TP_FINALIZE
  if (unlikely(PyType_HasFeature(Py_TYPE(o), Py_TPFLAGS_HAVE_FINALIZE) && Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  {
    PyObject *etype, *eval, *etb;
    PyErr_Fetch(&etype, &eval, &etb);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) + 1);
    __pyx_memoryviewslice___dealloc__(o);
    __Pyx_SET_REFCNT(o, Py_REFCNT(o) - 1);
    PyErr_Restore(etype, eval, etb);
  }
  Py_CLEAR(p->from_object);
  PyObject_GC_Track(o);
  __pyx_tp_dealloc_memoryview(o);
}

static int __pyx_tp_traverse__memoryviewslice(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  e = __pyx_tp_traverse_memoryview(o, v, a); if (e) return e;
  if (p->from_object) {
    e = (*v)(p->from_object, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear__memoryviewslice(PyObject *o) {
  PyObject* tmp;
  struct __pyx_memoryviewslice_obj *p = (struct __pyx_memoryviewslice_obj *)o;
  __pyx_tp_clear_memoryview(o);
  tmp = ((PyObject*)p->from_object);
  p->from_object = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  __PYX_XDEC_MEMVIEW(&p->from_slice, 1);
  return 0;
}

static PyObject *__pyx_getprop___pyx_memoryviewslice_base(PyObject *o, CYTHON_UNUSED void *x) {
  return __pyx_pw_15View_dot_MemoryView_16_memoryviewslice_4base_1__get__(o);
}

static PyMethodDef __pyx_methods__memoryviewslice[] = {
  {"__reduce_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_1__reduce_cython__, METH_NOARGS, 0},
  {"__setstate_cython__", (PyCFunction)__pyx_pw___pyx_memoryviewslice_3__setstate_cython__, METH_O, 0},
  {0, 0, 0, 0}
};

static struct PyGetSetDef __pyx_getsets__memoryviewslice[] = {
  {(char *)"base", __pyx_getprop___pyx_memoryviewslice_base, 0, (char *)0, 0},
  {0, 0, 0, 0, 0}
};

static PyTypeObject __pyx_type___pyx_memoryviewslice = {
  PyVarObject_HEAD_INIT(0, 0)
  "nes.cycore.mos6502._memoryviewslice", /*tp_name*/
  sizeof(struct __pyx_memoryviewslice_obj), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc__memoryviewslice, /*tp_dealloc*/
  #if PY_VERSION_HEX < 0x030800b4
  0, /*tp_print*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4
  0, /*tp_vectorcall_offset*/
  #endif
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___repr__, /*tp_repr*/
  #else
  0, /*tp_repr*/
  #endif
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  #if CYTHON_COMPILING_IN_PYPY
  __pyx_memoryview___str__, /*tp_str*/
  #else
  0, /*tp_str*/
  #endif
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Internal class for passing memoryview slices to Python", /*tp_doc*/
  __pyx_tp_traverse__memoryviewslice, /*tp_traverse*/
  __pyx_tp_clear__memoryviewslice, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods__memoryviewslice, /*tp_methods*/
  0, /*tp_members*/
  __pyx_getsets__memoryviewslice, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new__memoryviewslice, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b1
  0, /*tp_vectorcall*/
  #endif
  #if PY_VERSION_HEX >= 0x030800b4 && PY_VERSION_HEX < 0x03090000
  0, /*tp_print*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
#if CYTHON_PEP489_MULTI_PHASE_INIT
static PyObject* __pyx_pymod_create(PyObject *spec, PyModuleDef *def); /*proto*/
static int __pyx_pymod_exec_mos6502(PyObject* module); /*proto*/
static PyModuleDef_Slot __pyx_moduledef_slots[] = {
  {Py_mod_create, (void*)__pyx_pymod_create},
  {Py_mod_exec, (void*)__pyx_pymod_exec_mos6502},
  {0, NULL}
};
#endif

static struct PyModuleDef __pyx_moduledef = {
    PyModuleDef_HEAD_INIT,
    "mos6502",
    0, /* m_doc */
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    0, /* m_size */
  #else
    -1, /* m_size */
  #endif
    __pyx_methods /* m_methods */,
  #if CYTHON_PEP489_MULTI_PHASE_INIT
    __pyx_moduledef_slots, /* m_slots */
  #else
    NULL, /* m_reload */
  #endif
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif
#ifndef CYTHON_SMALL_CODE
#if defined(__clang__)
    #define CYTHON_SMALL_CODE
#elif defined(__GNUC__) && (__GNUC__ > 4 || (__GNUC__ == 4 && __GNUC_MINOR__ >= 3))
    #define CYTHON_SMALL_CODE __attribute__((cold))
#else
    #define CYTHON_SMALL_CODE
#endif
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_u_, __pyx_k_, sizeof(__pyx_k_), 0, 1, 0, 0},
  {&__pyx_kp_u_02X_X_02X_04X, __pyx_k_02X_X_02X_04X, sizeof(__pyx_k_02X_X_02X_04X), 0, 1, 0, 0},
  {&__pyx_kp_u_02X_Y_04X_04X, __pyx_k_02X_Y_04X_04X, sizeof(__pyx_k_02X_Y_04X_04X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X, __pyx_k_0_02X, sizeof(__pyx_k_0_02X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X_1_04X, __pyx_k_0_02X_1_04X, sizeof(__pyx_k_0_02X_1_04X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X_2, __pyx_k_0_02X_2, sizeof(__pyx_k_0_02X_2), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X_3, __pyx_k_0_02X_3, sizeof(__pyx_k_0_02X_3), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X_4, __pyx_k_0_02X_4, sizeof(__pyx_k_0_02X_4), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X_X, __pyx_k_0_02X_X, sizeof(__pyx_k_0_02X_X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_02X_Y, __pyx_k_0_02X_Y, sizeof(__pyx_k_0_02X_Y), 0, 1, 0, 0},
  {&__pyx_kp_u_0_04X, __pyx_k_0_04X, sizeof(__pyx_k_0_04X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_04X_2, __pyx_k_0_04X_2, sizeof(__pyx_k_0_04X_2), 0, 1, 0, 0},
  {&__pyx_kp_u_0_04X_3, __pyx_k_0_04X_3, sizeof(__pyx_k_0_04X_3), 0, 1, 0, 0},
  {&__pyx_kp_u_0_04X_X, __pyx_k_0_04X_X, sizeof(__pyx_k_0_04X_X), 0, 1, 0, 0},
  {&__pyx_kp_u_0_04X_Y, __pyx_k_0_04X_Y, sizeof(__pyx_k_0_04X_Y), 0, 1, 0, 0},
  {&__pyx_n_u_A, __pyx_k_A, sizeof(__pyx_k_A), 0, 1, 0, 1},
  {&__pyx_n_s_ABSOLUTE, __pyx_k_ABSOLUTE, sizeof(__pyx_k_ABSOLUTE), 0, 0, 1, 1},
  {&__pyx_n_s_ABSOLUTE_X, __pyx_k_ABSOLUTE_X, sizeof(__pyx_k_ABSOLUTE_X), 0, 0, 1, 1},
  {&__pyx_n_s_ABSOLUTE_Y, __pyx_k_ABSOLUTE_Y, sizeof(__pyx_k_ABSOLUTE_Y), 0, 0, 1, 1},
  {&__pyx_n_s_ACCUMULATOR, __pyx_k_ACCUMULATOR, sizeof(__pyx_k_ACCUMULATOR), 0, 0, 1, 1},
  {&__pyx_n_s_ASCII, __pyx_k_ASCII, sizeof(__pyx_k_ASCII), 0, 0, 1, 1},
  {&__pyx_kp_u_A_02X_X_02X_Y_02X_P_02X_SP_02X_P, __pyx_k_A_02X_X_02X_Y_02X_P_02X_SP_02X_P, sizeof(__pyx_k_A_02X_X_02X_Y_02X_P_02X_SP_02X_P), 0, 1, 0, 0},
  {&__pyx_n_s_AddressModes, __pyx_k_AddressModes, sizeof(__pyx_k_AddressModes), 0, 0, 1, 1},
  {&__pyx_kp_s_Buffer_view_does_not_expose_stri, __pyx_k_Buffer_view_does_not_expose_stri, sizeof(__pyx_k_Buffer_view_does_not_expose_stri), 0, 0, 1, 0},
  {&__pyx_kp_s_Can_only_create_a_buffer_that_is, __pyx_k_Can_only_create_a_buffer_that_is, sizeof(__pyx_k_Can_only_create_a_buffer_that_is), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_assign_to_read_only_memor, __pyx_k_Cannot_assign_to_read_only_memor, sizeof(__pyx_k_Cannot_assign_to_read_only_memor), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_create_writable_memory_vi, __pyx_k_Cannot_create_writable_memory_vi, sizeof(__pyx_k_Cannot_create_writable_memory_vi), 0, 0, 1, 0},
  {&__pyx_kp_s_Cannot_index_with_type_s, __pyx_k_Cannot_index_with_type_s, sizeof(__pyx_k_Cannot_index_with_type_s), 0, 0, 1, 0},
  {&__pyx_n_s_Ellipsis, __pyx_k_Ellipsis, sizeof(__pyx_k_Ellipsis), 0, 0, 1, 1},
  {&__pyx_kp_s_Empty_shape_tuple_for_cython_arr, __pyx_k_Empty_shape_tuple_for_cython_arr, sizeof(__pyx_k_Empty_shape_tuple_for_cython_arr), 0, 0, 1, 0},
  {&__pyx_n_s_IMMEDIATE, __pyx_k_IMMEDIATE, sizeof(__pyx_k_IMMEDIATE), 0, 0, 1, 1},
  {&__pyx_n_s_IMPLIED, __pyx_k_IMPLIED, sizeof(__pyx_k_IMPLIED), 0, 0, 1, 1},
  {&__pyx_n_s_INDIRECT, __pyx_k_INDIRECT, sizeof(__pyx_k_INDIRECT), 0, 0, 1, 1},
  {&__pyx_n_s_INDIRECT_X, __pyx_k_INDIRECT_X, sizeof(__pyx_k_INDIRECT_X), 0, 0, 1, 1},
  {&__pyx_n_s_INDIRECT_Y, __pyx_k_INDIRECT_Y, sizeof(__pyx_k_INDIRECT_Y), 0, 0, 1, 1},
  {&__pyx_n_s_INSTRUCTION_SET, __pyx_k_INSTRUCTION_SET, sizeof(__pyx_k_INSTRUCTION_SET), 0, 0, 1, 1},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0x19, __pyx_k_Incompatible_checksums_s_vs_0x19, sizeof(__pyx_k_Incompatible_checksums_s_vs_0x19), 0, 0, 1, 0},
  {&__pyx_kp_s_Incompatible_checksums_s_vs_0xb0, __pyx_k_Incompatible_checksums_s_vs_0xb0, sizeof(__pyx_k_Incompatible_checksums_s_vs_0xb0), 0, 0, 1, 0},
  {&__pyx_n_s_IndexError, __pyx_k_IndexError, sizeof(__pyx_k_IndexError), 0, 0, 1, 1},
  {&__pyx_kp_s_Indirect_dimensions_not_supporte, __pyx_k_Indirect_dimensions_not_supporte, sizeof(__pyx_k_Indirect_dimensions_not_supporte), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_mode_expected_c_or_fortr, __pyx_k_Invalid_mode_expected_c_or_fortr, sizeof(__pyx_k_Invalid_mode_expected_c_or_fortr), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_shape_in_axis_d_d, __pyx_k_Invalid_shape_in_axis_d_d, sizeof(__pyx_k_Invalid_shape_in_axis_d_d), 0, 0, 1, 0},
  {&__pyx_kp_u_KIL_instruction_Processor_halted, __pyx_k_KIL_instruction_Processor_halted, sizeof(__pyx_k_KIL_instruction_Processor_halted), 0, 1, 0, 0},
  {&__pyx_n_s_MOS6502, __pyx_k_MOS6502, sizeof(__pyx_k_MOS6502), 0, 0, 1, 1},
  {&__pyx_n_s_MemoryError, __pyx_k_MemoryError, sizeof(__pyx_k_MemoryError), 0, 0, 1, 1},
  {&__pyx_kp_s_MemoryView_of_r_at_0x_x, __pyx_k_MemoryView_of_r_at_0x_x, sizeof(__pyx_k_MemoryView_of_r_at_0x_x), 0, 0, 1, 0},
  {&__pyx_kp_s_MemoryView_of_r_object, __pyx_k_MemoryView_of_r_object, sizeof(__pyx_k_MemoryView_of_r_object), 0, 0, 1, 0},
  {&__pyx_n_s_NamedInstruction, __pyx_k_NamedInstruction, sizeof(__pyx_k_NamedInstruction), 0, 0, 1, 1},
  {&__pyx_n_b_O, __pyx_k_O, sizeof(__pyx_k_O), 0, 0, 0, 1},
  {&__pyx_kp_s_Out_of_bounds_on_buffer_access_a, __pyx_k_Out_of_bounds_on_buffer_access_a, sizeof(__pyx_k_Out_of_bounds_on_buffer_access_a), 0, 0, 1, 0},
  {&__pyx_n_s_OverflowError, __pyx_k_OverflowError, sizeof(__pyx_k_OverflowError), 0, 0, 1, 1},
  {&__pyx_n_s_PickleError, __pyx_k_PickleError, sizeof(__pyx_k_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_RELATIVE, __pyx_k_RELATIVE, sizeof(__pyx_k_RELATIVE), 0, 0, 1, 1},
  {&__pyx_kp_u_Stack_underflow, __pyx_k_Stack_underflow, sizeof(__pyx_k_Stack_underflow), 0, 1, 0, 0},
  {&__pyx_n_s_TypeError, __pyx_k_TypeError, sizeof(__pyx_k_TypeError), 0, 0, 1, 1},
  {&__pyx_kp_s_Unable_to_convert_item_to_object, __pyx_k_Unable_to_convert_item_to_object, sizeof(__pyx_k_Unable_to_convert_item_to_object), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_n_s_View_MemoryView, __pyx_k_View_MemoryView, sizeof(__pyx_k_View_MemoryView), 0, 0, 1, 1},
  {&__pyx_n_s_ZEROPAGE, __pyx_k_ZEROPAGE, sizeof(__pyx_k_ZEROPAGE), 0, 0, 1, 1},
  {&__pyx_n_s_ZEROPAGE_X, __pyx_k_ZEROPAGE_X, sizeof(__pyx_k_ZEROPAGE_X), 0, 0, 1, 1},
  {&__pyx_n_s_ZEROPAGE_Y, __pyx_k_ZEROPAGE_Y, sizeof(__pyx_k_ZEROPAGE_Y), 0, 0, 1, 1},
  {&__pyx_n_s__10, __pyx_k__10, sizeof(__pyx_k__10), 0, 0, 1, 1},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_kp_u__3, __pyx_k__3, sizeof(__pyx_k__3), 0, 1, 0, 0},
  {&__pyx_kp_u__5, __pyx_k__5, sizeof(__pyx_k__5), 0, 1, 0, 0},
  {&__pyx_kp_u__6, __pyx_k__6, sizeof(__pyx_k__6), 0, 1, 0, 0},
  {&__pyx_kp_u__7, __pyx_k__7, sizeof(__pyx_k__7), 0, 1, 0, 0},
  {&__pyx_n_s__9, __pyx_k__9, sizeof(__pyx_k__9), 0, 0, 1, 1},
  {&__pyx_n_s_aac, __pyx_k_aac, sizeof(__pyx_k_aac), 0, 0, 1, 1},
  {&__pyx_n_s_aax, __pyx_k_aax, sizeof(__pyx_k_aax), 0, 0, 1, 1},
  {&__pyx_n_s_aax_sets_flags, __pyx_k_aax_sets_flags, sizeof(__pyx_k_aax_sets_flags), 0, 0, 1, 1},
  {&__pyx_n_s_addr, __pyx_k_addr, sizeof(__pyx_k_addr), 0, 0, 1, 1},
  {&__pyx_n_s_allocate_buffer, __pyx_k_allocate_buffer, sizeof(__pyx_k_allocate_buffer), 0, 0, 1, 1},
  {&__pyx_n_s_arg, __pyx_k_arg, sizeof(__pyx_k_arg), 0, 0, 1, 1},
  {&__pyx_n_s_arr, __pyx_k_arr, sizeof(__pyx_k_arr), 0, 0, 1, 1},
  {&__pyx_n_s_asr, __pyx_k_asr, sizeof(__pyx_k_asr), 0, 0, 1, 1},
  {&__pyx_n_s_atx, __pyx_k_atx, sizeof(__pyx_k_atx), 0, 0, 1, 1},
  {&__pyx_n_s_axa, __pyx_k_axa, sizeof(__pyx_k_axa), 0, 0, 1, 1},
  {&__pyx_n_s_axs, __pyx_k_axs, sizeof(__pyx_k_axs), 0, 0, 1, 1},
  {&__pyx_n_s_base, __pyx_k_base, sizeof(__pyx_k_base), 0, 0, 1, 1},
  {&__pyx_n_s_bytecode, __pyx_k_bytecode, sizeof(__pyx_k_bytecode), 0, 0, 1, 1},
  {&__pyx_n_s_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 0, 1, 1},
  {&__pyx_n_u_c, __pyx_k_c, sizeof(__pyx_k_c), 0, 1, 0, 1},
  {&__pyx_n_s_caps, __pyx_k_caps, sizeof(__pyx_k_caps), 0, 0, 1, 1},
  {&__pyx_n_s_class, __pyx_k_class, sizeof(__pyx_k_class), 0, 0, 1, 1},
  {&__pyx_n_s_cline_in_traceback, __pyx_k_cline_in_traceback, sizeof(__pyx_k_cline_in_traceback), 0, 0, 1, 1},
  {&__pyx_kp_s_contiguous_and_direct, __pyx_k_contiguous_and_direct, sizeof(__pyx_k_contiguous_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_contiguous_and_indirect, __pyx_k_contiguous_and_indirect, sizeof(__pyx_k_contiguous_and_indirect), 0, 0, 1, 0},
  {&__pyx_n_s_cycles, __pyx_k_cycles, sizeof(__pyx_k_cycles), 0, 0, 1, 1},
  {&__pyx_n_s_data, __pyx_k_data, sizeof(__pyx_k_data), 0, 0, 1, 1},
  {&__pyx_n_s_dcp, __pyx_k_dcp, sizeof(__pyx_k_dcp), 0, 0, 1, 1},
  {&__pyx_n_s_dict, __pyx_k_dict, sizeof(__pyx_k_dict), 0, 0, 1, 1},
  {&__pyx_n_s_dop, __pyx_k_dop, sizeof(__pyx_k_dop), 0, 0, 1, 1},
  {&__pyx_n_s_dtype_is_object, __pyx_k_dtype_is_object, sizeof(__pyx_k_dtype_is_object), 0, 0, 1, 1},
  {&__pyx_n_s_encode, __pyx_k_encode, sizeof(__pyx_k_encode), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_flags, __pyx_k_flags, sizeof(__pyx_k_flags), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_format_instruction, __pyx_k_format_instruction, sizeof(__pyx_k_format_instruction), 0, 0, 1, 1},
  {&__pyx_n_s_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 0, 1, 1},
  {&__pyx_n_u_fortran, __pyx_k_fortran, sizeof(__pyx_k_fortran), 0, 1, 0, 1},
  {&__pyx_n_s_getstate, __pyx_k_getstate, sizeof(__pyx_k_getstate), 0, 0, 1, 1},
  {&__pyx_kp_s_got_differing_extents_in_dimensi, __pyx_k_got_differing_extents_in_dimensi, sizeof(__pyx_k_got_differing_extents_in_dimensi), 0, 0, 1, 0},
  {&__pyx_n_s_id, __pyx_k_id, sizeof(__pyx_k_id), 0, 0, 1, 1},
  {&__pyx_n_s_immediate, __pyx_k_immediate, sizeof(__pyx_k_immediate), 0, 0, 1, 1},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_instr, __pyx_k_instr, sizeof(__pyx_k_instr), 0, 0, 1, 1},
  {&__pyx_n_s_isc, __pyx_k_isc, sizeof(__pyx_k_isc), 0, 0, 1, 1},
  {&__pyx_n_s_items, __pyx_k_items, sizeof(__pyx_k_items), 0, 0, 1, 1},
  {&__pyx_n_s_itemsize, __pyx_k_itemsize, sizeof(__pyx_k_itemsize), 0, 0, 1, 1},
  {&__pyx_kp_s_itemsize_0_for_cython_array, __pyx_k_itemsize_0_for_cython_array, sizeof(__pyx_k_itemsize_0_for_cython_array), 0, 0, 1, 0},
  {&__pyx_n_s_kil, __pyx_k_kil, sizeof(__pyx_k_kil), 0, 0, 1, 1},
  {&__pyx_n_s_lar, __pyx_k_lar, sizeof(__pyx_k_lar), 0, 0, 1, 1},
  {&__pyx_n_s_lax, __pyx_k_lax, sizeof(__pyx_k_lax), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_make_bytecode_dict, __pyx_k_make_bytecode_dict, sizeof(__pyx_k_make_bytecode_dict), 0, 0, 1, 1},
  {&__pyx_n_s_memory, __pyx_k_memory, sizeof(__pyx_k_memory), 0, 0, 1, 1},
  {&__pyx_n_s_memview, __pyx_k_memview, sizeof(__pyx_k_memview), 0, 0, 1, 1},
  {&__pyx_n_s_mode, __pyx_k_mode, sizeof(__pyx_k_mode), 0, 0, 1, 1},
  {&__pyx_n_s_modes, __pyx_k_modes, sizeof(__pyx_k_modes), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_n_s_ndim, __pyx_k_ndim, sizeof(__pyx_k_ndim), 0, 0, 1, 1},
  {&__pyx_n_s_nes_cycore_mos6502, __pyx_k_nes_cycore_mos6502, sizeof(__pyx_k_nes_cycore_mos6502), 0, 0, 1, 1},
  {&__pyx_n_s_nes_instructions, __pyx_k_nes_instructions, sizeof(__pyx_k_nes_instructions), 0, 0, 1, 1},
  {&__pyx_n_s_new, __pyx_k_new, sizeof(__pyx_k_new), 0, 0, 1, 1},
  {&__pyx_kp_s_no_default___reduce___due_to_non, __pyx_k_no_default___reduce___due_to_non, sizeof(__pyx_k_no_default___reduce___due_to_non), 0, 0, 1, 0},
  {&__pyx_n_s_obj, __pyx_k_obj, sizeof(__pyx_k_obj), 0, 0, 1, 1},
  {&__pyx_n_s_pack, __pyx_k_pack, sizeof(__pyx_k_pack), 0, 0, 1, 1},
  {&__pyx_n_s_pickle, __pyx_k_pickle, sizeof(__pyx_k_pickle), 0, 0, 1, 1},
  {&__pyx_n_s_ppu_line, __pyx_k_ppu_line, sizeof(__pyx_k_ppu_line), 0, 0, 1, 1},
  {&__pyx_n_s_ppu_pixel, __pyx_k_ppu_pixel, sizeof(__pyx_k_ppu_pixel), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_PickleError, __pyx_k_pyx_PickleError, sizeof(__pyx_k_pyx_PickleError), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_checksum, __pyx_k_pyx_checksum, sizeof(__pyx_k_pyx_checksum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_getbuffer, __pyx_k_pyx_getbuffer, sizeof(__pyx_k_pyx_getbuffer), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_result, __pyx_k_pyx_result, sizeof(__pyx_k_pyx_result), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_state, __pyx_k_pyx_state, sizeof(__pyx_k_pyx_state), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_type, __pyx_k_pyx_type, sizeof(__pyx_k_pyx_type), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_Enum, __pyx_k_pyx_unpickle_Enum, sizeof(__pyx_k_pyx_unpickle_Enum), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_unpickle_MOS6502, __pyx_k_pyx_unpickle_MOS6502, sizeof(__pyx_k_pyx_unpickle_MOS6502), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_reduce, __pyx_k_reduce, sizeof(__pyx_k_reduce), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_cython, __pyx_k_reduce_cython, sizeof(__pyx_k_reduce_cython), 0, 0, 1, 1},
  {&__pyx_n_s_reduce_ex, __pyx_k_reduce_ex, sizeof(__pyx_k_reduce_ex), 0, 0, 1, 1},
  {&__pyx_n_s_reset, __pyx_k_reset, sizeof(__pyx_k_reset), 0, 0, 1, 1},
  {&__pyx_n_s_rla, __pyx_k_rla, sizeof(__pyx_k_rla), 0, 0, 1, 1},
  {&__pyx_n_s_rra, __pyx_k_rra, sizeof(__pyx_k_rra), 0, 0, 1, 1},
  {&__pyx_n_s_setstate, __pyx_k_setstate, sizeof(__pyx_k_setstate), 0, 0, 1, 1},
  {&__pyx_n_s_setstate_cython, __pyx_k_setstate_cython, sizeof(__pyx_k_setstate_cython), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_n_s_size_bytes, __pyx_k_size_bytes, sizeof(__pyx_k_size_bytes), 0, 0, 1, 1},
  {&__pyx_n_s_slo, __pyx_k_slo, sizeof(__pyx_k_slo), 0, 0, 1, 1},
  {&__pyx_n_s_sre, __pyx_k_sre, sizeof(__pyx_k_sre), 0, 0, 1, 1},
  {&__pyx_n_s_stack_underflow_causes_exception, __pyx_k_stack_underflow_causes_exception, sizeof(__pyx_k_stack_underflow_causes_exception), 0, 0, 1, 1},
  {&__pyx_n_s_start, __pyx_k_start, sizeof(__pyx_k_start), 0, 0, 1, 1},
  {&__pyx_n_s_step, __pyx_k_step, sizeof(__pyx_k_step), 0, 0, 1, 1},
  {&__pyx_n_s_stop, __pyx_k_stop, sizeof(__pyx_k_stop), 0, 0, 1, 1},
  {&__pyx_kp_s_strided_and_direct, __pyx_k_strided_and_direct, sizeof(__pyx_k_strided_and_direct), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_direct_or_indirect, __pyx_k_strided_and_direct_or_indirect, sizeof(__pyx_k_strided_and_direct_or_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_strided_and_indirect, __pyx_k_strided_and_indirect, sizeof(__pyx_k_strided_and_indirect), 0, 0, 1, 0},
  {&__pyx_kp_s_stringsource, __pyx_k_stringsource, sizeof(__pyx_k_stringsource), 0, 0, 1, 0},
  {&__pyx_n_s_struct, __pyx_k_struct, sizeof(__pyx_k_struct), 0, 0, 1, 1},
  {&__pyx_n_s_sxa, __pyx_k_sxa, sizeof(__pyx_k_sxa), 0, 0, 1, 1},
  {&__pyx_n_s_sya, __pyx_k_sya, sizeof(__pyx_k_sya), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_top, __pyx_k_top, sizeof(__pyx_k_top), 0, 0, 1, 1},
  {&__pyx_kp_s_unable_to_allocate_array_data, __pyx_k_unable_to_allocate_array_data, sizeof(__pyx_k_unable_to_allocate_array_data), 0, 0, 1, 0},
  {&__pyx_kp_s_unable_to_allocate_shape_and_str, __pyx_k_unable_to_allocate_shape_and_str, sizeof(__pyx_k_unable_to_allocate_shape_and_str), 0, 0, 1, 0},
  {&__pyx_n_s_undocumented_support_level, __pyx_k_undocumented_support_level, sizeof(__pyx_k_undocumented_support_level), 0, 0, 1, 1},
  {&__pyx_n_s_unpack, __pyx_k_unpack, sizeof(__pyx_k_unpack), 0, 0, 1, 1},
  {&__pyx_n_s_update, __pyx_k_update, sizeof(__pyx_k_update), 0, 0, 1, 1},
  {&__pyx_n_s_upper, __pyx_k_upper, sizeof(__pyx_k_upper), 0, 0, 1, 1},
  {&__pyx_n_s_xaa, __pyx_k_xaa, sizeof(__pyx_k_xaa), 0, 0, 1, 1},
  {&__pyx_n_s_xas, __pyx_k_xas, sizeof(__pyx_k_xas), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static CYTHON_SMALL_CODE int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 203, __pyx_L1_error)
  __pyx_builtin_OverflowError = __Pyx_GetBuiltinName(__pyx_n_s_OverflowError); if (!__pyx_builtin_OverflowError) __PYX_ERR(0, 341, __pyx_L1_error)
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 1048, __pyx_L1_error)
  __pyx_builtin_TypeError = __Pyx_GetBuiltinName(__pyx_n_s_TypeError); if (!__pyx_builtin_TypeError) __PYX_ERR(1, 81, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(1, 84, __pyx_L1_error)
  __pyx_builtin_IndexError = __Pyx_GetBuiltinName(__pyx_n_s_IndexError); if (!__pyx_builtin_IndexError) __PYX_ERR(1, 94, __pyx_L1_error)
  __pyx_builtin_MemoryError = __Pyx_GetBuiltinName(__pyx_n_s_MemoryError); if (!__pyx_builtin_MemoryError) __PYX_ERR(1, 148, __pyx_L1_error)
  __pyx_builtin_Ellipsis = __Pyx_GetBuiltinName(__pyx_n_s_Ellipsis); if (!__pyx_builtin_Ellipsis) __PYX_ERR(1, 404, __pyx_L1_error)
  __pyx_builtin_id = __Pyx_GetBuiltinName(__pyx_n_s_id); if (!__pyx_builtin_id) __PYX_ERR(1, 613, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "nes/cycore/mos6502.pyx":202
 * 
 *         instr = self.instructions[bytecode]
 *         data = bytearray(2)             # <<<<<<<<<<<<<<
 *         for i in range(instr.size_bytes - 1):
 *             data[i] = self.memory.read(self.PC + 1 + i)
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_int_2); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 202, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "nes/cycore/mos6502.pyx":341
 *         cdef unsigned char v
 *         if self.SP == 0xFF and self.stack_underflow_causes_exception:
 *             raise OverflowError("Stack underflow")             # <<<<<<<<<<<<<<
 *         self.SP = (self.SP + 1) & 0xFF
 *         v = self.memory.read(STACK_PAGE + self.SP)
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_u_Stack_underflow); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 341, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "nes/cycore/mos6502.pyx":1048
 *         """
 *         if self.undocumented_support_level >= 1:
 *             raise ValueError("KIL instruction.  Processor halted.")             # <<<<<<<<<<<<<<
 *         return 0
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_u_KIL_instruction_Processor_halted); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 1048, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "View.MemoryView":133
 * 
 *         if not self.ndim:
 *             raise ValueError("Empty shape tuple for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if itemsize <= 0:
 */
  __pyx_tuple__12 = PyTuple_Pack(1, __pyx_kp_s_Empty_shape_tuple_for_cython_arr); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(1, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "View.MemoryView":136
 * 
 *         if itemsize <= 0:
 *             raise ValueError("itemsize <= 0 for cython.array")             # <<<<<<<<<<<<<<
 * 
 *         if not isinstance(format, bytes):
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_itemsize_0_for_cython_array); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(1, 136, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "View.MemoryView":148
 * 
 *         if not self._shape:
 *             raise MemoryError("unable to allocate shape and strides.")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_shape_and_str); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(1, 148, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "View.MemoryView":176
 *             self.data = <char *>malloc(self.len)
 *             if not self.data:
 *                 raise MemoryError("unable to allocate array data.")             # <<<<<<<<<<<<<<
 * 
 *             if self.dtype_is_object:
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_unable_to_allocate_array_data); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(1, 176, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "View.MemoryView":192
 *             bufmode = PyBUF_F_CONTIGUOUS | PyBUF_ANY_CONTIGUOUS
 *         if not (flags & bufmode):
 *             raise ValueError("Can only create a buffer that is contiguous in memory.")             # <<<<<<<<<<<<<<
 *         info.buf = self.data
 *         info.len = self.len
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_Can_only_create_a_buffer_that_is); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(1, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "View.MemoryView":418
 *     def __setitem__(memoryview self, object index, object value):
 *         if self.view.readonly:
 *             raise TypeError("Cannot assign to read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         have_slices, index = _unellipsify(index, self.view.ndim)
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_Cannot_assign_to_read_only_memor); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(1, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "View.MemoryView":495
 *             result = struct.unpack(self.view.format, bytesitem)
 *         except struct.error:
 *             raise ValueError("Unable to convert item to object")             # <<<<<<<<<<<<<<
 *         else:
 *             if len(self.view.format) == 1:
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_Unable_to_convert_item_to_object); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(1, 495, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "View.MemoryView":520
 *     def __getbuffer__(self, Py_buffer *info, int flags):
 *         if flags & PyBUF_WRITABLE and self.view.readonly:
 *             raise ValueError("Cannot create writable memory view from read-only memoryview")             # <<<<<<<<<<<<<<
 * 
 *         if flags & PyBUF_ND:
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_Cannot_create_writable_memory_vi); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(1, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "View.MemoryView":570
 *         if self.view.strides == NULL:
 * 
 *             raise ValueError("Buffer view does not expose strides")             # <<<<<<<<<<<<<<
 * 
 *         return tuple([stride for stride in self.view.strides[:self.view.ndim]])
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_Buffer_view_does_not_expose_stri); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(1, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "View.MemoryView":577
 *     def suboffsets(self):
 *         if self.view.suboffsets == NULL:
 *             return (-1,) * self.view.ndim             # <<<<<<<<<<<<<<
 * 
 *         return tuple([suboffset for suboffset in self.view.suboffsets[:self.view.ndim]])
 */
  __pyx_tuple__23 = PyTuple_New(1); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(1, 577, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_INCREF(__pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_int_neg_1);
  PyTuple_SET_ITEM(__pyx_tuple__23, 0, __pyx_int_neg_1);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__24 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "View.MemoryView":682
 *         if item is Ellipsis:
 *             if not seen_ellipsis:
 *                 result.extend([slice(None)] * (ndim - len(tup) + 1))             # <<<<<<<<<<<<<<
 *                 seen_ellipsis = True
 *             else:
 */
  __pyx_slice__26 = PySlice_New(Py_None, Py_None, Py_None); if (unlikely(!__pyx_slice__26)) __PYX_ERR(1, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__26);
  __Pyx_GIVEREF(__pyx_slice__26);

  /* "View.MemoryView":703
 *     for suboffset in suboffsets[:ndim]:
 *         if suboffset >= 0:
 *             raise ValueError("Indirect dimensions not supported")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_Indirect_dimensions_not_supporte); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(1, 703, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "(tree fragment)":2
 * def __reduce_cython__(self):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 */
  __pyx_tuple__28 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__28)) __PYX_ERR(1, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__28);
  __Pyx_GIVEREF(__pyx_tuple__28);

  /* "(tree fragment)":4
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")
 * def __setstate_cython__(self, __pyx_state):
 *     raise TypeError("no default __reduce__ due to non-trivial __cinit__")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_kp_s_no_default___reduce___due_to_non); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(1, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "(tree fragment)":1
 * def __pyx_unpickle_MOS6502(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__30 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);
  __pyx_codeobj__31 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__30, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_MOS6502, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__31)) __PYX_ERR(1, 1, __pyx_L1_error)

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct_or_indirect); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_s_strided_and_direct); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_s_strided_and_indirect); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_direct); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_s_contiguous_and_indirect); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_tuple__37 = PyTuple_Pack(5, __pyx_n_s_pyx_type, __pyx_n_s_pyx_checksum, __pyx_n_s_pyx_state, __pyx_n_s_pyx_PickleError, __pyx_n_s_pyx_result); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(3, 0, 5, 0, CO_OPTIMIZED|CO_NEWLOCALS, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_stringsource, __pyx_n_s_pyx_unpickle_Enum, 1, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_238 = PyInt_FromLong(238); if (unlikely(!__pyx_int_238)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_255 = PyInt_FromLong(255); if (unlikely(!__pyx_int_255)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_65280 = PyInt_FromLong(65280L); if (unlikely(!__pyx_int_65280)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_26325635 = PyInt_FromLong(26325635L); if (unlikely(!__pyx_int_26325635)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_184977713 = PyInt_FromLong(184977713L); if (unlikely(!__pyx_int_184977713)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_neg_1 = PyInt_FromLong(-1); if (unlikely(!__pyx_int_neg_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static CYTHON_SMALL_CODE int __Pyx_modinit_global_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_export_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_init_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_type_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_variable_import_code(void); /*proto*/
static CYTHON_SMALL_CODE int __Pyx_modinit_function_import_code(void); /*proto*/

static int __Pyx_modinit_global_init_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_global_init_code", 0);
  /*--- Global init code ---*/
  generic = Py_None; Py_INCREF(Py_None);
  strided = Py_None; Py_INCREF(Py_None);
  indirect = Py_None; Py_INCREF(Py_None);
  contiguous = Py_None; Py_INCREF(Py_None);
  indirect_contiguous = Py_None; Py_INCREF(Py_None);
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_variable_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_export_code", 0);
  /*--- Variable export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_export_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_export_code", 0);
  /*--- Function export code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_type_init_code(void) {
  __Pyx_RefNannyDeclarations
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_init_code", 0);
  /*--- Type init code ---*/
  __pyx_vtabptr_3nes_6cycore_7mos6502_MOS6502 = &__pyx_vtable_3nes_6cycore_7mos6502_MOS6502;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.make_instruction_data_tables = (void (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_make_instruction_data_tables;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.run_instr = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char, unsigned char *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_run_instr;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.dma_pause = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_dma_pause;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._from_le = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__from_le;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._read_word = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__read_word;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.trigger_nmi = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_trigger_nmi;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.trigger_irq = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_trigger_irq;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.run_next_instr = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_run_next_instr;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._status_to_byte = (unsigned char (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__status_to_byte;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._status_from_byte = (void (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__status_from_byte;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.push_stack = (void (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_push_stack;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502.pop_stack = (unsigned char (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *))__pyx_f_3nes_6cycore_7mos6502_7MOS6502_pop_stack;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._neg = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__neg;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._from_2sc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, unsigned char))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__from_2sc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._set_zn = (void (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__set_zn;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._adc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__adc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._and = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__and;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._asl = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__asl;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._jump_relative = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__jump_relative;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bcc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bcc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bcs = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bcs;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._beq = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__beq;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bit = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bit;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bmi = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bmi;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bne = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bne;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bpl = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bpl;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._interrupt = (PyObject *(*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__interrupt;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._brk = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__brk;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bvc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bvc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._bvs = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__bvs;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._clc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__clc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._cld = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__cld;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._cli = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__cli;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._clv = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__clv;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._compare = (void (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__compare;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._cmp = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__cmp;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._cpx = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__cpx;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._cpy = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__cpy;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._dec = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__dec;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._dex = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__dex;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._dey = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__dey;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._eor = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__eor;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._inc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__inc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._inx = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__inx;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._iny = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__iny;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._jmp = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__jmp;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._jsr = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__jsr;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._lda = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__lda;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._ldx = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__ldx;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._ldy = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__ldy;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._lsr = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__lsr;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._nop = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__nop;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._ora = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__ora;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._pha = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__pha;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._php = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__php;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._pla = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__pla;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._plp = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__plp;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._rol = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__rol;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._ror = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__ror;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._rti = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__rti;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._rts = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__rts;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._sbc = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__sbc;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._sec = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__sec;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._sed = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__sed;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._sei = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__sei;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._sta = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__sta;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._stx = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__stx;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._sty = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__sty;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._tax = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__tax;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._tay = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__tay;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._tsx = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__tsx;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._txa = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__txa;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._txs = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__txs;
  __pyx_vtable_3nes_6cycore_7mos6502_MOS6502._tya = (int (*)(struct __pyx_obj_3nes_6cycore_7mos6502_MOS6502 *, int, int))__pyx_f_3nes_6cycore_7mos6502_7MOS6502__tya;
  if (PyType_Ready(&__pyx_type_3nes_6cycore_7mos6502_MOS6502) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type_3nes_6cycore_7mos6502_MOS6502.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type_3nes_6cycore_7mos6502_MOS6502.tp_dictoffset && __pyx_type_3nes_6cycore_7mos6502_MOS6502.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type_3nes_6cycore_7mos6502_MOS6502.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type_3nes_6cycore_7mos6502_MOS6502.tp_dict, __pyx_vtabptr_3nes_6cycore_7mos6502_MOS6502) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  if (PyObject_SetAttr(__pyx_m, __pyx_n_s_MOS6502, (PyObject *)&__pyx_type_3nes_6cycore_7mos6502_MOS6502) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type_3nes_6cycore_7mos6502_MOS6502) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_7mos6502_MOS6502 = &__pyx_type_3nes_6cycore_7mos6502_MOS6502;
  __pyx_vtabptr_array = &__pyx_vtable_array;
  __pyx_vtable_array.get_memview = (PyObject *(*)(struct __pyx_array_obj *))__pyx_array_get_memview;
  if (PyType_Ready(&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_array.tp_print = 0;
  #endif
  if (__Pyx_SetVtable(__pyx_type___pyx_array.tp_dict, __pyx_vtabptr_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_array) < 0) __PYX_ERR(1, 105, __pyx_L1_error)
  __pyx_array_type = &__pyx_type___pyx_array;
  if (PyType_Ready(&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_MemviewEnum.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_MemviewEnum.tp_dictoffset && __pyx_type___pyx_MemviewEnum.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_MemviewEnum.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_MemviewEnum) < 0) __PYX_ERR(1, 279, __pyx_L1_error)
  __pyx_MemviewEnum_type = &__pyx_type___pyx_MemviewEnum;
  __pyx_vtabptr_memoryview = &__pyx_vtable_memoryview;
  __pyx_vtable_memoryview.get_item_pointer = (char *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_get_item_pointer;
  __pyx_vtable_memoryview.is_slice = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_is_slice;
  __pyx_vtable_memoryview.setitem_slice_assignment = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_slice_assignment;
  __pyx_vtable_memoryview.setitem_slice_assign_scalar = (PyObject *(*)(struct __pyx_memoryview_obj *, struct __pyx_memoryview_obj *, PyObject *))__pyx_memoryview_setitem_slice_assign_scalar;
  __pyx_vtable_memoryview.setitem_indexed = (PyObject *(*)(struct __pyx_memoryview_obj *, PyObject *, PyObject *))__pyx_memoryview_setitem_indexed;
  __pyx_vtable_memoryview.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryview_convert_item_to_object;
  __pyx_vtable_memoryview.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryview_assign_item_from_object;
  if (PyType_Ready(&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryview.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryview.tp_dictoffset && __pyx_type___pyx_memoryview.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryview.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryview.tp_dict, __pyx_vtabptr_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryview) < 0) __PYX_ERR(1, 330, __pyx_L1_error)
  __pyx_memoryview_type = &__pyx_type___pyx_memoryview;
  __pyx_vtabptr__memoryviewslice = &__pyx_vtable__memoryviewslice;
  __pyx_vtable__memoryviewslice.__pyx_base = *__pyx_vtabptr_memoryview;
  __pyx_vtable__memoryviewslice.__pyx_base.convert_item_to_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *))__pyx_memoryviewslice_convert_item_to_object;
  __pyx_vtable__memoryviewslice.__pyx_base.assign_item_from_object = (PyObject *(*)(struct __pyx_memoryview_obj *, char *, PyObject *))__pyx_memoryviewslice_assign_item_from_object;
  __pyx_type___pyx_memoryviewslice.tp_base = __pyx_memoryview_type;
  if (PyType_Ready(&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  #if PY_VERSION_HEX < 0x030800B1
  __pyx_type___pyx_memoryviewslice.tp_print = 0;
  #endif
  if ((CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP) && likely(!__pyx_type___pyx_memoryviewslice.tp_dictoffset && __pyx_type___pyx_memoryviewslice.tp_getattro == PyObject_GenericGetAttr)) {
    __pyx_type___pyx_memoryviewslice.tp_getattro = __Pyx_PyObject_GenericGetAttr;
  }
  if (__Pyx_SetVtable(__pyx_type___pyx_memoryviewslice.tp_dict, __pyx_vtabptr__memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  if (__Pyx_setup_reduce((PyObject*)&__pyx_type___pyx_memoryviewslice) < 0) __PYX_ERR(1, 965, __pyx_L1_error)
  __pyx_memoryviewslice_type = &__pyx_type___pyx_memoryviewslice;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_type_import_code(void) {
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannySetupContext("__Pyx_modinit_type_import_code", 0);
  /*--- Type import code ---*/
  __pyx_t_1 = PyImport_ImportModule("nes.cycore.apu"); if (unlikely(!__pyx_t_1)) __PYX_ERR(2, 51, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3nes_6cycore_3apu_APUEnvelope = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "APUEnvelope", sizeof(struct __pyx_obj_3nes_6cycore_3apu_APUEnvelope), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_APUEnvelope) __PYX_ERR(2, 51, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_APUEnvelope = (struct __pyx_vtabstruct_3nes_6cycore_3apu_APUEnvelope*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_APUEnvelope->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_APUEnvelope)) __PYX_ERR(2, 51, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_3apu_APUUnit = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "APUUnit", sizeof(struct __pyx_obj_3nes_6cycore_3apu_APUUnit), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_APUUnit) __PYX_ERR(2, 60, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_APUUnit = (struct __pyx_vtabstruct_3nes_6cycore_3apu_APUUnit*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_APUUnit->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_APUUnit)) __PYX_ERR(2, 60, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_3apu_APUTriangle = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "APUTriangle", sizeof(struct __pyx_obj_3nes_6cycore_3apu_APUTriangle), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_APUTriangle) __PYX_ERR(2, 71, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_APUTriangle = (struct __pyx_vtabstruct_3nes_6cycore_3apu_APUTriangle*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_APUTriangle->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_APUTriangle)) __PYX_ERR(2, 71, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_3apu_APUPulse = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "APUPulse", sizeof(struct __pyx_obj_3nes_6cycore_3apu_APUPulse), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_APUPulse) __PYX_ERR(2, 83, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_APUPulse = (struct __pyx_vtabstruct_3nes_6cycore_3apu_APUPulse*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_APUPulse->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_APUPulse)) __PYX_ERR(2, 83, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_3apu_APUNoise = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "APUNoise", sizeof(struct __pyx_obj_3nes_6cycore_3apu_APUNoise), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_APUNoise) __PYX_ERR(2, 101, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_APUNoise = (struct __pyx_vtabstruct_3nes_6cycore_3apu_APUNoise*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_APUNoise->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_APUNoise)) __PYX_ERR(2, 101, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_3apu_APUDMC = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "APUDMC", sizeof(struct __pyx_obj_3nes_6cycore_3apu_APUDMC), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_APUDMC) __PYX_ERR(2, 117, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_APUDMC = (struct __pyx_vtabstruct_3nes_6cycore_3apu_APUDMC*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_APUDMC->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_APUDMC)) __PYX_ERR(2, 117, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_3apu_NESAPU = __Pyx_ImportType(__pyx_t_1, "nes.cycore.apu", "NESAPU", sizeof(struct __pyx_obj_3nes_6cycore_3apu_NESAPU), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3apu_NESAPU) __PYX_ERR(2, 139, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3apu_NESAPU = (struct __pyx_vtabstruct_3nes_6cycore_3apu_NESAPU*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3apu_NESAPU->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3apu_NESAPU)) __PYX_ERR(2, 139, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("nes.cycore.carts"); if (unlikely(!__pyx_t_1)) __PYX_ERR(3, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3nes_6cycore_5carts_NESCart = __Pyx_ImportType(__pyx_t_1, "nes.cycore.carts", "NESCart", sizeof(struct __pyx_obj_3nes_6cycore_5carts_NESCart), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_5carts_NESCart) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_5carts_NESCart = (struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_5carts_NESCart->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_5carts_NESCart)) __PYX_ERR(3, 17, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_5carts_NESCart0 = __Pyx_ImportType(__pyx_t_1, "nes.cycore.carts", "NESCart0", sizeof(struct __pyx_obj_3nes_6cycore_5carts_NESCart0), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_5carts_NESCart0) __PYX_ERR(3, 42, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_5carts_NESCart0 = (struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart0*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_5carts_NESCart0->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_5carts_NESCart0)) __PYX_ERR(3, 42, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_5carts_NESCart1 = __Pyx_ImportType(__pyx_t_1, "nes.cycore.carts", "NESCart1", sizeof(struct __pyx_obj_3nes_6cycore_5carts_NESCart1), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_5carts_NESCart1) __PYX_ERR(3, 98, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_5carts_NESCart1 = (struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart1*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_5carts_NESCart1->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_5carts_NESCart1)) __PYX_ERR(3, 98, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_5carts_NESCart2 = __Pyx_ImportType(__pyx_t_1, "nes.cycore.carts", "NESCart2", sizeof(struct __pyx_obj_3nes_6cycore_5carts_NESCart2), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_5carts_NESCart2) __PYX_ERR(3, 135, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_5carts_NESCart2 = (struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart2*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_5carts_NESCart2->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_5carts_NESCart2)) __PYX_ERR(3, 135, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_5carts_NESCart4 = __Pyx_ImportType(__pyx_t_1, "nes.cycore.carts", "NESCart4", sizeof(struct __pyx_obj_3nes_6cycore_5carts_NESCart4), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_5carts_NESCart4) __PYX_ERR(3, 177, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_5carts_NESCart4 = (struct __pyx_vtabstruct_3nes_6cycore_5carts_NESCart4*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_5carts_NESCart4->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_5carts_NESCart4)) __PYX_ERR(3, 177, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("nes.cycore.system"); if (unlikely(!__pyx_t_1)) __PYX_ERR(4, 14, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3nes_6cycore_6system_InterruptListener = __Pyx_ImportType(__pyx_t_1, "nes.cycore.system", "InterruptListener", sizeof(struct __pyx_obj_3nes_6cycore_6system_InterruptListener), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_6system_InterruptListener) __PYX_ERR(4, 14, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_6system_InterruptListener = (struct __pyx_vtabstruct_3nes_6cycore_6system_InterruptListener*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_6system_InterruptListener->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_6system_InterruptListener)) __PYX_ERR(4, 14, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_6system_NES = __Pyx_ImportType(__pyx_t_1, "nes.cycore.system", "NES", sizeof(struct __pyx_obj_3nes_6cycore_6system_NES), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_6system_NES) __PYX_ERR(4, 43, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_6system_NES = (struct __pyx_vtabstruct_3nes_6cycore_6system_NES*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_6system_NES->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_6system_NES)) __PYX_ERR(4, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("nes.cycore.ppu"); if (unlikely(!__pyx_t_1)) __PYX_ERR(5, 77, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3nes_6cycore_3ppu_NESPPU = __Pyx_ImportType(__pyx_t_1, "nes.cycore.ppu", "NESPPU", sizeof(struct __pyx_obj_3nes_6cycore_3ppu_NESPPU), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_3ppu_NESPPU) __PYX_ERR(5, 77, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_3ppu_NESPPU = (struct __pyx_vtabstruct_3nes_6cycore_3ppu_NESPPU*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_3ppu_NESPPU->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_3ppu_NESPPU)) __PYX_ERR(5, 77, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyImport_ImportModule("nes.cycore.memory"); if (unlikely(!__pyx_t_1)) __PYX_ERR(6, 13, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_ptype_3nes_6cycore_6memory_MemoryBase = __Pyx_ImportType(__pyx_t_1, "nes.cycore.memory", "MemoryBase", sizeof(struct __pyx_obj_3nes_6cycore_6memory_MemoryBase), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_6memory_MemoryBase) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_6memory_MemoryBase = (struct __pyx_vtabstruct_3nes_6cycore_6memory_MemoryBase*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_6memory_MemoryBase->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_6memory_MemoryBase)) __PYX_ERR(6, 13, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_6memory_NESMappedRAM = __Pyx_ImportType(__pyx_t_1, "nes.cycore.memory", "NESMappedRAM", sizeof(struct __pyx_obj_3nes_6cycore_6memory_NESMappedRAM), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_6memory_NESMappedRAM) __PYX_ERR(6, 34, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_6memory_NESMappedRAM = (struct __pyx_vtabstruct_3nes_6cycore_6memory_NESMappedRAM*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_6memory_NESMappedRAM->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_6memory_NESMappedRAM)) __PYX_ERR(6, 34, __pyx_L1_error)
  __pyx_ptype_3nes_6cycore_6memory_NESVRAM = __Pyx_ImportType(__pyx_t_1, "nes.cycore.memory", "NESVRAM", sizeof(struct __pyx_obj_3nes_6cycore_6memory_NESVRAM), __Pyx_ImportType_CheckSize_Warn);
   if (!__pyx_ptype_3nes_6cycore_6memory_NESVRAM) __PYX_ERR(6, 68, __pyx_L1_error)
  __pyx_vtabptr_3nes_6cycore_6memory_NESVRAM = (struct __pyx_vtabstruct_3nes_6cycore_6memory_NESVRAM*)__Pyx_GetVtable(__pyx_ptype_3nes_6cycore_6memory_NESVRAM->tp_dict); if (unlikely(!__pyx_vtabptr_3nes_6cycore_6memory_NESVRAM)) __PYX_ERR(6, 68, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_modinit_variable_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_variable_import_code", 0);
  /*--- Variable import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}

static int __Pyx_modinit_function_import_code(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_modinit_function_import_code", 0);
  /*--- Function import code ---*/
  __Pyx_RefNannyFinishContext();
  return 0;
}


#ifndef CYTHON_NO_PYINIT_EXPORT
#define __Pyx_PyMODINIT_FUNC PyMODINIT_FUNC
#elif PY_MAJOR_VERSION < 3
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" void
#else
#define __Pyx_PyMODINIT_FUNC void
#endif
#else
#ifdef __cplusplus
#define __Pyx_PyMODINIT_FUNC extern "C" PyObject *
#else
#define __Pyx_PyMODINIT_FUNC PyObject *
#endif
#endif


#if PY_MAJOR_VERSION < 3
__Pyx_PyMODINIT_FUNC initmos6502(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC initmos6502(void)
#else
__Pyx_PyMODINIT_FUNC PyInit_mos6502(void) CYTHON_SMALL_CODE; /*proto*/
__Pyx_PyMODINIT_FUNC PyInit_mos6502(void)
#if CYTHON_PEP489_MULTI_PHASE_INIT
{
  return PyModuleDef_Init(&__pyx_moduledef);
}
static CYTHON_SMALL_CODE int __Pyx_check_single_interpreter(void) {
    #if PY_VERSION_HEX >= 0x030700A1
    static PY_INT64_T main_interpreter_id = -1;
    PY_INT64_T current_id = PyInterpreterState_GetID(PyThreadState_Get()->interp);
    if (main_interpreter_id == -1) {
        main_interpreter_id = current_id;
        return (unlikely(current_id == -1)) ? -1 : 0;
    } else if (unlikely(main_interpreter_id != current_id))
    #else
    static PyInterpreterState *main_interpreter = NULL;
    PyInterpreterState *current_interpreter = PyThreadState_Get()->interp;
    if (!main_interpreter) {
        main_interpreter = current_interpreter;
    } else if (unlikely(main_interpreter != current_interpreter))
    #endif
    {
        PyErr_SetString(
            PyExc_ImportError,
            "Interpreter change detected - this module can only be loaded into one interpreter per process.");
        return -1;
    }
    return 0;
}
static CYTHON_SMALL_CODE int __Pyx_copy_spec_to_module(PyObject *spec, PyObject *moddict, const char* from_name, const char* to_name, int allow_none) {
    PyObject *value = PyObject_GetAttrString(spec, from_name);
    int result = 0;
    if (likely(value)) {
        if (allow_none || value != Py_None) {
            result = PyDict_SetItemString(moddict, to_name, value);
        }
        Py_DECREF(value);
    } else if (PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Clear();
    } else {
        result = -1;
    }
    return result;
}
static CYTHON_SMALL_CODE PyObject* __pyx_pymod_create(PyObject *spec, CYTHON_UNUSED PyModuleDef *def) {
    PyObject *module = NULL, *moddict, *modname;
    if (__Pyx_check_single_interpreter())
        return NULL;
    if (__pyx_m)
        return __Pyx_NewRef(__pyx_m);
    modname = PyObject_GetAttrString(spec, "name");
    if (unlikely(!modname)) goto bad;
    module = PyModule_NewObject(modname);
    Py_DECREF(modname);
    if (unlikely(!module)) goto bad;
    moddict = PyModule_GetDict(module);
    if (unlikely(!moddict)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "loader", "__loader__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "origin", "__file__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "parent", "__package__", 1) < 0)) goto bad;
    if (unlikely(__Pyx_copy_spec_to_module(spec, moddict, "submodule_search_locations", "__path__", 0) < 0)) goto bad;
    return module;
bad:
    Py_XDECREF(module);
    return NULL;
}


static CYTHON_SMALL_CODE int __pyx_pymod_exec_mos6502(PyObject *__pyx_pyinit_module)
#endif
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  static PyThread_type_lock __pyx_t_3[8];
  int __pyx_lineno = 0;
  const char *__pyx_filename = NULL;
  int __pyx_clineno = 0;
  __Pyx_RefNannyDeclarations
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  if (__pyx_m) {
    if (__pyx_m == __pyx_pyinit_module) return 0;
    PyErr_SetString(PyExc_RuntimeError, "Module 'mos6502' has already been imported. Re-initialisation is not supported.");
    return -1;
  }
  #elif PY_MAJOR_VERSION >= 3
  if (__pyx_m) return __Pyx_NewRef(__pyx_m);
  #endif
  #if CYTHON_REFNANNY
__Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
if (!__Pyx_RefNanny) {
  PyErr_Clear();
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
  if (!__Pyx_RefNanny)
      Py_FatalError("failed to import 'refnanny' module");
}
#endif
  __Pyx_RefNannySetupContext("__Pyx_PyMODINIT_FUNC PyInit_mos6502(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pxy_PyFrame_Initialize_Offsets
  __Pxy_PyFrame_Initialize_Offsets();
  #endif
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_AsyncGen_USED
  if (__pyx_AsyncGen_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  __pyx_m = __pyx_pyinit_module;
  Py_INCREF(__pyx_m);
  #else
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("mos6502", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_b);
  __pyx_cython_runtime = PyImport_AddModule((char *) "cython_runtime"); if (unlikely(!__pyx_cython_runtime)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_cython_runtime);
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_nes__cycore__mos6502) {
    if (PyObject_SetAttr(__pyx_m, __pyx_n_s_name_2, __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "nes.cycore.mos6502")) {
      if (unlikely(PyDict_SetItemString(modules, "nes.cycore.mos6502", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global type/function init code ---*/
  (void)__Pyx_modinit_global_init_code();
  (void)__Pyx_modinit_variable_export_code();
  (void)__Pyx_modinit_function_export_code();
  if (unlikely(__Pyx_modinit_type_init_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  if (unlikely(__Pyx_modinit_type_import_code() < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
  (void)__Pyx_modinit_variable_import_code();
  (void)__Pyx_modinit_function_import_code();
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "nes/cycore/mos6502.pyx":4
 * 
 * from .system cimport OAM_DMA, DMC_DMA, DMC_DMA_DURING_OAM_DMA
 * from nes.instructions import INSTRUCTION_SET, NamedInstruction, AddressModes             # <<<<<<<<<<<<<<
 * 
 * cdef class MOS6502:
 */
  __pyx_t_1 = PyList_New(3); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_INSTRUCTION_SET);
  __Pyx_GIVEREF(__pyx_n_s_INSTRUCTION_SET);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_INSTRUCTION_SET);
  __Pyx_INCREF(__pyx_n_s_NamedInstruction);
  __Pyx_GIVEREF(__pyx_n_s_NamedInstruction);
  PyList_SET_ITEM(__pyx_t_1, 1, __pyx_n_s_NamedInstruction);
  __Pyx_INCREF(__pyx_n_s_AddressModes);
  __Pyx_GIVEREF(__pyx_n_s_AddressModes);
  PyList_SET_ITEM(__pyx_t_1, 2, __pyx_n_s_AddressModes);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_nes_instructions, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_INSTRUCTION_SET); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_INSTRUCTION_SET, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_NamedInstruction); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_NamedInstruction, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_AddressModes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_AddressModes, __pyx_t_1) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":1
 * def __pyx_unpickle_MOS6502(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_3nes_6cycore_7mos6502_1__pyx_unpickle_MOS6502, NULL, __pyx_n_s_nes_cycore_mos6502); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_MOS6502, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "nes/cycore/mos6502.pyx":1
 * #import pyximport; pyximport.install()             # <<<<<<<<<<<<<<
 * 
 * from .system cimport OAM_DMA, DMC_DMA, DMC_DMA_DURING_OAM_DMA
 */
  __pyx_t_2 = __Pyx_PyDict_NewPresized(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_2) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "View.MemoryView":209
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_array_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 *     def __dealloc__(array self):
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_array_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_array_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(1, 209, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_array_type);

  /* "View.MemoryView":286
 *         return self.name
 * 
 * cdef generic = Enum("<strided and direct or indirect>")             # <<<<<<<<<<<<<<
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(generic);
  __Pyx_DECREF_SET(generic, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":287
 * 
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default             # <<<<<<<<<<<<<<
 * cdef indirect = Enum("<strided and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 287, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(strided);
  __Pyx_DECREF_SET(strided, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":288
 * cdef generic = Enum("<strided and direct or indirect>")
 * cdef strided = Enum("<strided and direct>") # default
 * cdef indirect = Enum("<strided and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 288, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect);
  __Pyx_DECREF_SET(indirect, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":291
 * 
 * 
 * cdef contiguous = Enum("<contiguous and direct>")             # <<<<<<<<<<<<<<
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(contiguous);
  __Pyx_DECREF_SET(contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":292
 * 
 * cdef contiguous = Enum("<contiguous and direct>")
 * cdef indirect_contiguous = Enum("<contiguous and indirect>")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_MemviewEnum_type), __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 292, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_XGOTREF(indirect_contiguous);
  __Pyx_DECREF_SET(indirect_contiguous, __pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __pyx_t_2 = 0;

  /* "View.MemoryView":316
 * 
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0             # <<<<<<<<<<<<<<
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [
 *     PyThread_allocate_lock(),
 */
  __pyx_memoryview_thread_locks_used = 0;

  /* "View.MemoryView":317
 * DEF THREAD_LOCKS_PREALLOCATED = 8
 * cdef int __pyx_memoryview_thread_locks_used = 0
 * cdef PyThread_type_lock[THREAD_LOCKS_PREALLOCATED] __pyx_memoryview_thread_locks = [             # <<<<<<<<<<<<<<
 *     PyThread_allocate_lock(),
 *     PyThread_allocate_lock(),
 */
  __pyx_t_3[0] = PyThread_allocate_lock();
  __pyx_t_3[1] = PyThread_allocate_lock();
  __pyx_t_3[2] = PyThread_allocate_lock();
  __pyx_t_3[3] = PyThread_allocate_lock();
  __pyx_t_3[4] = PyThread_allocate_lock();
  __pyx_t_3[5] = PyThread_allocate_lock();
  __pyx_t_3[6] = PyThread_allocate_lock();
  __pyx_t_3[7] = PyThread_allocate_lock();
  memcpy(&(__pyx_memoryview_thread_locks[0]), __pyx_t_3, sizeof(__pyx_memoryview_thread_locks[0]) * (8));

  /* "View.MemoryView":549
 *         info.obj = self
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryview_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(1, 549, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryview_type);

  /* "View.MemoryView":995
 *         return self.from_object
 * 
 *     __pyx_getbuffer = capsule(<void *> &__pyx_memoryview_getbuffer, "getbuffer(obj, view, flags)")             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_2 = __pyx_capsule_create(((void *)(&__pyx_memoryview_getbuffer)), ((char *)"getbuffer(obj, view, flags)")); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem((PyObject *)__pyx_memoryviewslice_type->tp_dict, __pyx_n_s_pyx_getbuffer, __pyx_t_2) < 0) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  PyType_Modified(__pyx_memoryviewslice_type);

  /* "(tree fragment)":1
 * def __pyx_unpickle_Enum(__pyx_type, long __pyx_checksum, __pyx_state):             # <<<<<<<<<<<<<<
 *     cdef object __pyx_PickleError
 *     cdef object __pyx_result
 */
  __pyx_t_2 = PyCFunction_NewEx(&__pyx_mdef_15View_dot_MemoryView_1__pyx_unpickle_Enum, NULL, __pyx_n_s_View_MemoryView); if (unlikely(!__pyx_t_2)) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_pyx_unpickle_Enum, __pyx_t_2) < 0) __PYX_ERR(1, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "(tree fragment)":11
 *         __pyx_unpickle_Enum__set_state(<Enum> __pyx_result, __pyx_state)
 *     return __pyx_result
 * cdef __pyx_unpickle_Enum__set_state(Enum __pyx_result, tuple __pyx_state):             # <<<<<<<<<<<<<<
 *     __pyx_result.name = __pyx_state[0]
 *     if len(__pyx_state) > 1 and hasattr(__pyx_result, '__dict__'):
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init nes.cycore.mos6502", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_CLEAR(__pyx_m);
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init nes.cycore.mos6502");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if CYTHON_PEP489_MULTI_PHASE_INIT
  return (__pyx_m != NULL) ? 0 : -1;
  #elif PY_MAJOR_VERSION >= 3
  return __pyx_m;
  #else
  return;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule(modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, "RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* PyObjectGetAttrStr */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (__Pyx_PyUnicode_GET_LENGTH(**name) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (__Pyx_PyUnicode_GET_LENGTH(**argname) != __Pyx_PyUnicode_GET_LENGTH(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* ExtTypeTest */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(__Pyx_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = __Pyx_PyFrame_GetLocalsplus(f);
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, Py_ssize_t nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, (int)nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif
#endif

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || __Pyx_CyFunction_Check(func)))
#else
    if (likely(PyCFunction_Check(func)))
#endif
    {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    int flags = PyCFunction_GET_FLAGS(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (flags & ~(METH_CLASS | METH_STATIC | METH_COEXIST | METH_KEYWORDS | METH_STACKLESS)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    if ((PY_VERSION_HEX < 0x030700A0) || unlikely(flags & METH_KEYWORDS)) {
        return (*((__Pyx_PyCFunctionFastWithKeywords)(void*)meth)) (self, args, nargs, NULL);
    } else {
        return (*((__Pyx_PyCFunctionFast)(void*)meth)) (self, args, nargs);
    }
}
#endif

/* PyObjectCallOneArg */
#if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
    if (likely(PyCFunction_Check(func))) {
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (__Pyx_PyFastCFunction_Check(func)) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(__Pyx_PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyObjectGetMethod */
static int __Pyx_PyObject_GetMethod(PyObject *obj, PyObject *name, PyObject **method) {
    PyObject *attr;
#if CYTHON_UNPACK_METHODS && CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_PYTYPE_LOOKUP
    PyTypeObject *tp = Py_TYPE(obj);
    PyObject *descr;
    descrgetfunc f = NULL;
    PyObject **dictptr, *dict;
    int meth_found = 0;
    assert (*method == NULL);
    if (unlikely(tp->tp_getattro != PyObject_GenericGetAttr)) {
        attr = __Pyx_PyObject_GetAttrStr(obj, name);
        goto try_unpack;
    }
    if (unlikely(tp->tp_dict == NULL) && unlikely(PyType_Ready(tp) < 0)) {
        return 0;
    }
    descr = _PyType_Lookup(tp, name);
    if (likely(descr != NULL)) {
        Py_INCREF(descr);
#if PY_MAJOR_VERSION >= 3
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr) || (Py_TYPE(descr) == &PyMethodDescr_Type)))
        #endif
#else
        #ifdef __Pyx_CyFunction_USED
        if (likely(PyFunction_Check(descr) || __Pyx_CyFunction_Check(descr)))
        #else
        if (likely(PyFunction_Check(descr)))
        #endif
#endif
        {
            meth_found = 1;
        } else {
            f = Py_TYPE(descr)->tp_descr_get;
            if (f != NULL && PyDescr_IsData(descr)) {
                attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
                Py_DECREF(descr);
                goto try_unpack;
            }
        }
    }
    dictptr = _PyObject_GetDictPtr(obj);
    if (dictptr != NULL && (dict = *dictptr) != NULL) {
        Py_INCREF(dict);
        attr = __Pyx_PyDict_GetItemStr(dict, name);
        if (attr != NULL) {
            Py_INCREF(attr);
            Py_DECREF(dict);
            Py_XDECREF(descr);
            goto try_unpack;
        }
        Py_DECREF(dict);
    }
    if (meth_found) {
        *method = descr;
        return 1;
    }
    if (f != NULL) {
        attr = f(descr, obj, (PyObject *)Py_TYPE(obj));
        Py_DECREF(descr);
        goto try_unpack;
    }
    if (descr != NULL) {
        *method = descr;
        return 0;
    }
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(name));
#endif
    return 0;
#else
    attr = __Pyx_PyObject_GetAttrStr(obj, name);
    goto try_unpack;
#endif
try_unpack:
#if CYTHON_UNPACK_METHODS
    if (likely(attr) && PyMethod_Check(attr) && likely(PyMethod_GET_SELF(attr) == obj)) {
        PyObject *function = PyMethod_GET_FUNCTION(attr);
        Py_INCREF(function);
        Py_DECREF(attr);
        *method = function;
        return 1;
    }
#endif
    *method = attr;
    return 0;
}

/* PyObjectCallMethod0 */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method = NULL, *result = NULL;
    int is_method = __Pyx_PyObject_GetMethod(obj, method_name, &method);
    if (likely(is_method)) {
        result = __Pyx_PyObject_CallOneArg(method, obj);
        Py_DECREF(method);
        return result;
    }
    if (unlikely(!method)) goto bad;
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
static CYTHON_INLINE int __Pyx_unpack_tuple2_exact(
        PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2, int decref_tuple) {
    PyObject *value1 = NULL, *value2 = NULL;
#if CYTHON_COMPILING_IN_PYPY
    value1 = PySequence_ITEM(tuple, 0);  if (unlikely(!value1)) goto bad;
    value2 = PySequence_ITEM(tuple, 1);  if (unlikely(!value2)) goto bad;
#else
    value1 = PyTuple_GET_ITEM(tuple, 0);  Py_INCREF(value1);
    value2 = PyTuple_GET_ITEM(tuple, 1);  Py_INCREF(value2);
#endif
    if (decref_tuple) {
        Py_DECREF(tuple);
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
#if CYTHON_COMPILING_IN_PYPY
bad:
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
#endif
}
static int __Pyx_unpack_tuple2_generic(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                       int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    iternextfunc iternext;
    iter = PyObject_GetIter(tuple);
    if (unlikely(!iter)) goto bad;
    if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
    iternext = Py_TYPE(iter)->tp_iternext;
    value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
    value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
    if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
    Py_DECREF(iter);
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
    if (is_dict) {
#if !CYTHON_COMPILING_IN_PYPY
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
#elif PY_MAJOR_VERSION >= 3
        static PyObject *py_items = NULL, *py_keys = NULL, *py_values = NULL;
        PyObject **pp = NULL;
        if (method_name) {
            const char *name = PyUnicode_AsUTF8(method_name);
            if (strcmp(name, "iteritems") == 0) pp = &py_items;
            else if (strcmp(name, "iterkeys") == 0) pp = &py_keys;
            else if (strcmp(name, "itervalues") == 0) pp = &py_values;
            if (pp) {
                if (!*pp) {
                    *pp = PyUnicode_FromString(name + 4);
                    if (!*pp)
                        return NULL;
                }
                method_name = *pp;
            }
        }
#endif
    }
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* PyDictVersioning */
#if CYTHON_USE_DICT_VERSIONS && CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PY_UINT64_T __Pyx_get_tp_dict_version(PyObject *obj) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    return likely(dict) ? __PYX_GET_DICT_VERSION(dict) : 0;
}
static CYTHON_INLINE PY_UINT64_T __Pyx_get_object_dict_version(PyObject *obj) {
    PyObject **dictptr = NULL;
    Py_ssize_t offset = Py_TYPE(obj)->tp_dictoffset;
    if (offset) {
#if CYTHON_COMPILING_IN_CPYTHON
        dictptr = (likely(offset > 0)) ? (PyObject **) ((char *)obj + offset) : _PyObject_GetDictPtr(obj);
#else
        dictptr = _PyObject_GetDictPtr(obj);
#endif
    }
    return (dictptr && *dictptr) ? __PYX_GET_DICT_VERSION(*dictptr) : 0;
}
static CYTHON_INLINE int __Pyx_object_dict_version_matches(PyObject* obj, PY_UINT64_T tp_dict_version, PY_UINT64_T obj_dict_version) {
    PyObject *dict = Py_TYPE(obj)->tp_dict;
    if (unlikely(!dict) || unlikely(tp_dict_version != __PYX_GET_DICT_VERSION(dict)))
        return 0;
    return obj_dict_version == __Pyx_get_object_dict_version(obj);
}
#endif

/* GetModuleGlobalName */
#if CYTHON_USE_DICT_VERSIONS
static PyObject *__Pyx__GetModuleGlobalName(PyObject *name, PY_UINT64_T *dict_version, PyObject **dict_cached_value)
#else
static CYTHON_INLINE PyObject *__Pyx__GetModuleGlobalName(PyObject *name)
#endif
{
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x030500A1
    result = _PyDict_GetItem_KnownHash(__pyx_d, name, ((PyASCIIObject *) name)->hash);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    } else if (unlikely(PyErr_Occurred())) {
        return NULL;
    }
#else
    result = PyDict_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
#endif
#else
    result = PyObject_GetItem(__pyx_d, name);
    __PYX_UPDATE_DICT_CACHE(__pyx_d, result, *dict_cached_value, *dict_version)
    if (likely(result)) {
        return __Pyx_NewRef(result);
    }
    PyErr_Clear();
#endif
    return __Pyx_GetBuiltinName(name);
}

/* PyErrFetchRestore */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* WriteUnraisableException */
static void __Pyx_WriteUnraisable(const char *name, CYTHON_UNUSED int clineno,
                                  CYTHON_UNUSED int lineno, CYTHON_UNUSED const char *filename,
                                  int full_traceback, CYTHON_UNUSED int nogil) {
    PyObject *old_exc, *old_val, *old_tb;
    PyObject *ctx;
    __Pyx_PyThreadState_declare
#ifdef WITH_THREAD
    PyGILState_STATE state;
    if (nogil)
        state = PyGILState_Ensure();
#ifdef _MSC_VER
    else state = (PyGILState_STATE)-1;
#endif
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&old_exc, &old_val, &old_tb);
    if (full_traceback) {
        Py_XINCREF(old_exc);
        Py_XINCREF(old_val);
        Py_XINCREF(old_tb);
        __Pyx_ErrRestore(old_exc, old_val, old_tb);
        PyErr_PrintEx(1);
    }
    #if PY_MAJOR_VERSION < 3
    ctx = PyString_FromString(name);
    #else
    ctx = PyUnicode_FromString(name);
    #endif
    __Pyx_ErrRestore(old_exc, old_val, old_tb);
    if (!ctx) {
        PyErr_WriteUnraisable(Py_None);
    } else {
        PyErr_WriteUnraisable(ctx);
        Py_DECREF(ctx);
    }
#ifdef WITH_THREAD
    if (nogil)
        PyGILState_Release(state);
#endif
}

/* PyObjectCall2Args */
static CYTHON_UNUSED PyObject* __Pyx_PyObject_Call2Args(PyObject* function, PyObject* arg1, PyObject* arg2) {
    PyObject *args, *result = NULL;
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyFunction_FastCall(function, args, 2);
    }
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(function)) {
        PyObject *args[2] = {arg1, arg2};
        return __Pyx_PyCFunction_FastCall(function, args, 2);
    }
    #endif
    args = PyTuple_New(2);
    if (unlikely(!args)) goto done;
    Py_INCREF(arg1);
    PyTuple_SET_ITEM(args, 0, arg1);
    Py_INCREF(arg2);
    PyTuple_SET_ITEM(args, 1, arg2);
    Py_INCREF(function);
    result = __Pyx_PyObject_Call(function, args, NULL);
    Py_DECREF(args);
    Py_DECREF(function);
done:
    return result;
}

/* GetItemInt */
static PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyList_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    Py_ssize_t wrapped_i = i;
    if (wraparound & unlikely(i < 0)) {
        wrapped_i += PyTuple_GET_SIZE(o);
    }
    if ((!boundscheck) || likely(__Pyx_is_valid_index(wrapped_i, PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, wrapped_i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely(__Pyx_is_valid_index(n, PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely(__Pyx_is_valid_index(n, PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AndObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a & b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_and(op1, op2);
            }
        }
                x = a & b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla & llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceAnd : PyNumber_And)(op1, op2);
}
#endif

/* ObjectGetItem */
#if CYTHON_USE_TYPE_SLOTS
static PyObject *__Pyx_PyObject_GetIndex(PyObject *obj, PyObject* index) {
    PyObject *runerr;
    Py_ssize_t key_value;
    PySequenceMethods *m = Py_TYPE(obj)->tp_as_sequence;
    if (unlikely(!(m && m->sq_item))) {
        PyErr_Format(PyExc_TypeError, "'%.200s' object is not subscriptable", Py_TYPE(obj)->tp_name);
        return NULL;
    }
    key_value = __Pyx_PyIndex_AsSsize_t(index);
    if (likely(key_value != -1 || !(runerr = PyErr_Occurred()))) {
        return __Pyx_GetItemInt_Fast(obj, key_value, 0, 1, 1);
    }
    if (PyErr_GivenExceptionMatches(runerr, PyExc_OverflowError)) {
        PyErr_Clear();
        PyErr_Format(PyExc_IndexError, "cannot fit '%.200s' into an index-sized integer", Py_TYPE(index)->tp_name);
    }
    return NULL;
}
static PyObject *__Pyx_PyObject_GetItem(PyObject *obj, PyObject* key) {
    PyMappingMethods *m = Py_TYPE(obj)->tp_as_mapping;
    if (likely(m && m->mp_subscript)) {
        return m->mp_subscript(obj, key);
    }
    return __Pyx_PyObject_GetIndex(obj, key);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_SubtractObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a - b);
            if (likely((x^a) >= 0 || (x^~b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_subtract(op1, op2);
            }
        }
                x = a - b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla - llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("subtract", return NULL)
            result = ((double)a) - (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceSubtract : PyNumber_Subtract)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_RshiftObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a >> b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_rshift(op1, op2);
            }
        }
                x = a >> b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla >> llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceRshift : PyNumber_Rshift)(op1, op2);
}
#endif

/* RaiseException */
#if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
    if (cause) {
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = __Pyx_PyThreadState_Current;
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_OrObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
            return PyInt_FromLong(a | b);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_or(op1, op2);
            }
        }
                x = a | b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla | llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    return (inplace ? PyNumber_InPlaceOr : PyNumber_Or)(op1, op2);
}
#endif

/* PyIntBinop */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_AddObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, int inplace, int zerodivision_check) {
    (void)inplace;
    (void)zerodivision_check;
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long x;
        long a = PyInt_AS_LONG(op1);
            x = (long)((unsigned long)a + b);
            if (likely((x^a) >= 0 || (x^b) >= 0))
                return PyInt_FromLong(x);
            return PyLong_Type.tp_as_number->nb_add(op1, op2);
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a, x;
#ifdef HAVE_LONG_LONG
        const PY_LONG_LONG llb = intval;
        PY_LONG_LONG lla, llx;
#endif
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = -(PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
#ifdef HAVE_LONG_LONG
                    } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                        lla = (PY_LONG_LONG) (((((((((unsigned PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (unsigned PY_LONG_LONG)digits[0]));
                        goto long_long;
#endif
                    }
                    CYTHON_FALLTHROUGH;
                default: return PyLong_Type.tp_as_number->nb_add(op1, op2);
            }
        }
                x = a + b;
            return PyLong_FromLong(x);
#ifdef HAVE_LONG_LONG
        long_long:
                llx = lla + llb;
            return PyLong_FromLongLong(llx);
#endif
        
        
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            double result;
            PyFPE_START_PROTECT("add", return NULL)
            result = ((double)a) + (double)b;
            PyFPE_END_PROTECT(result)
            return PyFloat_FromDouble(result);
    }
    return (inplace ? PyNumber_InPlaceAdd : PyNumber_Add)(op1, op2);
}
#endif

/* PyErrExceptionMatches */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx_PyErr_ExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        if (__Pyx_PyErr_GivenExceptionMatches(exc_type, PyTuple_GET_ITEM(tuple, i))) return 1;
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    if (unlikely(PyTuple_Check(err)))
        return __Pyx_PyErr_ExceptionMatchesTuple(exc_type, err);
    return __Pyx_PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* GetAttr */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_USE_TYPE_SLOTS
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* GetAttr3 */
static PyObject *__Pyx_GetAttr3Default(PyObject *d) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (unlikely(!__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        return NULL;
    __Pyx_PyErr_Clear();
    Py_INCREF(d);
    return d;
}
static CYTHON_INLINE PyObject *__Pyx_GetAttr3(PyObject *o, PyObject *n, PyObject *d) {
    PyObject *r = __Pyx_GetAttr(o, n);
    return (likely(r)) ? r : __Pyx_GetAttr3Default(d);
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_MAJOR_VERSION < 3
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if ((1) && (strchr(__Pyx_MODULE_NAME, '.'))) {
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_MAJOR_VERSION < 3
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, (PyObject *)NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* HasAttr */
static CYTHON_INLINE int __Pyx_HasAttr(PyObject *o, PyObject *n) {
    PyObject *r;
    if (unlikely(!__Pyx_PyBaseString_Check(n))) {
        PyErr_SetString(PyExc_TypeError,
                        "hasattr(): attribute name must be string");
        return -1;
    }
    r = __Pyx_GetAttr(o, n);
    if (unlikely(!r)) {
        PyErr_Clear();
        return 0;
    } else {
        Py_DECREF(r);
        return 1;
    }
}

/* GetTopmostException */
#if CYTHON_USE_EXC_INFO_STACK
static _PyErr_StackItem *
__Pyx_PyErr_GetTopmostException(PyThreadState *tstate)
{
    _PyErr_StackItem *exc_info = tstate->exc_info;
    while ((exc_info->exc_type == NULL || exc_info->exc_type == Py_None) &&
           exc_info->previous_item != NULL)
    {
        exc_info = exc_info->previous_item;
    }
    return exc_info;
}
#endif

/* SaveResetException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = __Pyx_PyErr_GetTopmostException(tstate);
    *type = exc_info->exc_type;
    *value = exc_info->exc_value;
    *tb = exc_info->exc_traceback;
    #else
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    #endif
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = type;
    exc_info->exc_value = value;
    exc_info->exc_traceback = tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* ArgTypeTest */
static int __Pyx__ArgTypeTest(PyObject *obj, PyTypeObject *type, const char *name, int exact)
{
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    else if (exact) {
        #if PY_MAJOR_VERSION == 2
        if ((type == &PyBaseString_Type) && likely(__Pyx_PyBaseString_CheckExact(obj))) return 1;
        #endif
    }
    else {
        if (likely(__Pyx_TypeCheck(obj, type))) return 1;
    }
    PyErr_Format(PyExc_TypeError,
        "Argument '%.200s' has incorrect type (expected %.200s, got %.200s)",
        name, type->tp_name, Py_TYPE(obj)->tp_name);
    return 0;
}

/* BytesEquals */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result;
#if CYTHON_USE_UNICODE_INTERNALS
            Py_hash_t hash1, hash2;
            hash1 = ((PyBytesObject*)s1)->ob_shash;
            hash2 = ((PyBytesObject*)s2)->ob_shash;
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                return (equals == Py_NE);
            }
#endif
            result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
#if CYTHON_USE_UNICODE_INTERNALS
        {
            Py_hash_t hash1, hash2;
        #if CYTHON_PEP393_ENABLED
            hash1 = ((PyASCIIObject*)s1)->hash;
            hash2 = ((PyASCIIObject*)s2)->hash;
        #else
            hash1 = ((PyUnicodeObject*)s1)->hash;
            hash2 = ((PyUnicodeObject*)s2)->hash;
        #endif
            if (hash1 != hash2 && hash1 != -1 && hash2 != -1) {
                goto return_ne;
            }
        }
#endif
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        #if PY_MAJOR_VERSION < 3
        Py_XDECREF(owned_ref);
        #endif
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* decode_c_string */
static CYTHON_INLINE PyObject* __Pyx_decode_c_string(
         const char* cstring, Py_ssize_t start, Py_ssize_t stop,
         const char* encoding, const char* errors,
         PyObject* (*decode_func)(const char *s, Py_ssize_t size, const char *errors)) {
    Py_ssize_t length;
    if (unlikely((start < 0) | (stop < 0))) {
        size_t slen = strlen(cstring);
        if (unlikely(slen > (size_t) PY_SSIZE_T_MAX)) {
            PyErr_SetString(PyExc_OverflowError,
                            "c-string too long to convert to Python");
            return NULL;
        }
        length = (Py_ssize_t) slen;
        if (start < 0) {
            start += length;
            if (start < 0)
                start = 0;
        }
        if (stop < 0)
            stop += length;
    }
    if (unlikely(stop <= start))
        return __Pyx_NewRef(__pyx_empty_unicode);
    length = stop - start;
    cstring += start;
    if (decode_func) {
        return decode_func(cstring, length, errors);
    } else {
        return PyUnicode_Decode(cstring, length, encoding, errors);
    }
}

/* GetException */
#if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb)
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb)
#endif
{
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    #if CYTHON_USE_EXC_INFO_STACK
    {
        _PyErr_StackItem *exc_info = tstate->exc_info;
        tmp_type = exc_info->exc_type;
        tmp_value = exc_info->exc_value;
        tmp_tb = exc_info->exc_traceback;
        exc_info->exc_type = local_type;
        exc_info->exc_value = local_value;
        exc_info->exc_traceback = local_tb;
    }
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    #endif
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
#if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    #if CYTHON_USE_EXC_INFO_STACK
    _PyErr_StackItem *exc_info = tstate->exc_info;
    tmp_type = exc_info->exc_type;
    tmp_value = exc_info->exc_value;
    tmp_tb = exc_info->exc_traceback;
    exc_info->exc_type = *type;
    exc_info->exc_value = *value;
    exc_info->exc_traceback = *tb;
    #else
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    #endif
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* FastTypeChecks */
#if CYTHON_COMPILING_IN_CPYTHON
static int __Pyx_InBases(PyTypeObject *a, PyTypeObject *b) {
    while (a) {
        a = a->tp_base;
        if (a == b)
            return 1;
    }
    return b == &PyBaseObject_Type;
}
static CYTHON_INLINE int __Pyx_IsSubtype(PyTypeObject *a, PyTypeObject *b) {
    PyObject *mro;
    if (a == b) return 1;
    mro = a->tp_mro;
    if (likely(mro)) {
        Py_ssize_t i, n;
        n = PyTuple_GET_SIZE(mro);
        for (i = 0; i < n; i++) {
            if (PyTuple_GET_ITEM(mro, i) == (PyObject *)b)
                return 1;
        }
        return 0;
    }
    return __Pyx_InBases(a, b);
}
#if PY_MAJOR_VERSION == 2
static int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject* exc_type2) {
    PyObject *exception, *value, *tb;
    int res;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&exception, &value, &tb);
    res = exc_type1 ? PyObject_IsSubclass(err, exc_type1) : 0;
    if (unlikely(res == -1)) {
        PyErr_WriteUnraisable(err);
        res = 0;
    }
    if (!res) {
        res = PyObject_IsSubclass(err, exc_type2);
        if (unlikely(res == -1)) {
            PyErr_WriteUnraisable(err);
            res = 0;
        }
    }
    __Pyx_ErrRestore(exception, value, tb);
    return res;
}
#else
static CYTHON_INLINE int __Pyx_inner_PyErr_GivenExceptionMatches2(PyObject *err, PyObject* exc_type1, PyObject *exc_type2) {
    int res = exc_type1 ? __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type1) : 0;
    if (!res) {
        res = __Pyx_IsSubtype((PyTypeObject*)err, (PyTypeObject*)exc_type2);
    }
    return res;
}
#endif
static int __Pyx_PyErr_GivenExceptionMatchesTuple(PyObject *exc_type, PyObject *tuple) {
    Py_ssize_t i, n;
    assert(PyExceptionClass_Check(exc_type));
    n = PyTuple_GET_SIZE(tuple);
#if PY_MAJOR_VERSION >= 3
    for (i=0; i<n; i++) {
        if (exc_type == PyTuple_GET_ITEM(tuple, i)) return 1;
    }
#endif
    for (i=0; i<n; i++) {
        PyObject *t = PyTuple_GET_ITEM(tuple, i);
        #if PY_MAJOR_VERSION < 3
        if (likely(exc_type == t)) return 1;
        #endif
        if (likely(PyExceptionClass_Check(t))) {
            if (__Pyx_inner_PyErr_GivenExceptionMatches2(exc_type, NULL, t)) return 1;
        } else {
        }
    }
    return 0;
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches(PyObject *err, PyObject* exc_type) {
    if (likely(err == exc_type)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        if (likely(PyExceptionClass_Check(exc_type))) {
            return __Pyx_inner_PyErr_GivenExceptionMatches2(err, NULL, exc_type);
        } else if (likely(PyTuple_Check(exc_type))) {
            return __Pyx_PyErr_GivenExceptionMatchesTuple(err, exc_type);
        } else {
        }
    }
    return PyErr_GivenExceptionMatches(err, exc_type);
}
static CYTHON_INLINE int __Pyx_PyErr_GivenExceptionMatches2(PyObject *err, PyObject *exc_type1, PyObject *exc_type2) {
    assert(PyExceptionClass_Check(exc_type1));
    assert(PyExceptionClass_Check(exc_type2));
    if (likely(err == exc_type1 || err == exc_type2)) return 1;
    if (likely(PyExceptionClass_Check(err))) {
        return __Pyx_inner_PyErr_GivenExceptionMatches2(err, exc_type1, exc_type2);
    }
    return (PyErr_GivenExceptionMatches(err, exc_type1) || PyErr_GivenExceptionMatches(err, exc_type2));
}
#endif

/* None */
static CYTHON_INLINE void __Pyx_RaiseUnboundLocalError(const char *varname) {
    PyErr_Format(PyExc_UnboundLocalError, "local variable '%s' referenced before assignment", varname);
}

/* PyObject_GenericGetAttrNoDict */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject *__Pyx_RaiseGenericGetAttributeError(PyTypeObject *tp, PyObject *attr_name) {
    PyErr_Format(PyExc_AttributeError,
#if PY_MAJOR_VERSION >= 3
                 "'%.50s' object has no attribute '%U'",
                 tp->tp_name, attr_name);
#else
                 "'%.50s' object has no attribute '%.400s'",
                 tp->tp_name, PyString_AS_STRING(attr_name));
#endif
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GenericGetAttrNoDict(PyObject* obj, PyObject* attr_name) {
    PyObject *descr;
    PyTypeObject *tp = Py_TYPE(obj);
    if (unlikely(!PyString_Check(attr_name))) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    assert(!tp->tp_dictoffset);
    descr = _PyType_Lookup(tp, attr_name);
    if (unlikely(!descr)) {
        return __Pyx_RaiseGenericGetAttributeError(tp, attr_name);
    }
    Py_INCREF(descr);
    #if PY_MAJOR_VERSION < 3
    if (likely(PyType_HasFeature(Py_TYPE(descr), Py_TPFLAGS_HAVE_CLASS)))
    #endif
    {
        descrgetfunc f = Py_TYPE(descr)->tp_descr_get;
        if (unlikely(f)) {
            PyObject *res = f(descr, obj, (PyObject *)tp);
            Py_DECREF(descr);
            return res;
        }
    }
    return descr;
}
#endif

/* PyObject_GenericGetAttr */
#if CYTHON_USE_TYPE_SLOTS && CYTHON_USE_PYTYPE_LOOKUP && PY_VERSION_HEX < 0x03070000
static PyObject* __Pyx_PyObject_GenericGetAttr(PyObject* obj, PyObject* attr_name) {
    if (unlikely(Py_TYPE(obj)->tp_dictoffset)) {
        return PyObject_GenericGetAttr(obj, attr_name);
    }
    return __Pyx_PyObject_GenericGetAttrNoDict(obj, attr_name);
}
#endif

/* SetVTable */
static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* PyObjectGetAttrStrNoError */
static void __Pyx_PyObject_GetAttrStr_ClearAttributeError(void) {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    if (likely(__Pyx_PyErr_ExceptionMatches(PyExc_AttributeError)))
        __Pyx_PyErr_Clear();
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStrNoError(PyObject* obj, PyObject* attr_name) {
    PyObject *result;
#if CYTHON_COMPILING_IN_CPYTHON && CYTHON_USE_TYPE_SLOTS && PY_VERSION_HEX >= 0x030700B1
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro == PyObject_GenericGetAttr)) {
        return _PyObject_GenericGetAttrWithDict(obj, attr_name, NULL, 1);
    }
#endif
    result = __Pyx_PyObject_GetAttrStr(obj, attr_name);
    if (unlikely(!result)) {
        __Pyx_PyObject_GetAttrStr_ClearAttributeError();
    }
    return result;
}

/* SetupReduce */
static int __Pyx_setup_reduce_is_named(PyObject* meth, PyObject* name) {
  int ret;
  PyObject *name_attr;
  name_attr = __Pyx_PyObject_GetAttrStr(meth, __pyx_n_s_name_2);
  if (likely(name_attr)) {
      ret = PyObject_RichCompareBool(name_attr, name, Py_EQ);
  } else {
      ret = -1;
  }
  if (unlikely(ret < 0)) {
      PyErr_Clear();
      ret = 0;
  }
  Py_XDECREF(name_attr);
  return ret;
}
static int __Pyx_setup_reduce(PyObject* type_obj) {
    int ret = 0;
    PyObject *object_reduce = NULL;
    PyObject *object_reduce_ex = NULL;
    PyObject *reduce = NULL;
    PyObject *reduce_ex = NULL;
    PyObject *reduce_cython = NULL;
    PyObject *setstate = NULL;
    PyObject *setstate_cython = NULL;
#if CYTHON_USE_PYTYPE_LOOKUP
    if (_PyType_Lookup((PyTypeObject*)type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#else
    if (PyObject_HasAttr(type_obj, __pyx_n_s_getstate)) goto __PYX_GOOD;
#endif
#if CYTHON_USE_PYTYPE_LOOKUP
    object_reduce_ex = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#else
    object_reduce_ex = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce_ex); if (!object_reduce_ex) goto __PYX_BAD;
#endif
    reduce_ex = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce_ex); if (unlikely(!reduce_ex)) goto __PYX_BAD;
    if (reduce_ex == object_reduce_ex) {
#if CYTHON_USE_PYTYPE_LOOKUP
        object_reduce = _PyType_Lookup(&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#else
        object_reduce = __Pyx_PyObject_GetAttrStr((PyObject*)&PyBaseObject_Type, __pyx_n_s_reduce); if (!object_reduce) goto __PYX_BAD;
#endif
        reduce = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_reduce); if (unlikely(!reduce)) goto __PYX_BAD;
        if (reduce == object_reduce || __Pyx_setup_reduce_is_named(reduce, __pyx_n_s_reduce_cython)) {
            reduce_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_reduce_cython);
            if (likely(reduce_cython)) {
                ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce, reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_reduce_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
            } else if (reduce == object_reduce || PyErr_Occurred()) {
                goto __PYX_BAD;
            }
            setstate = __Pyx_PyObject_GetAttrStr(type_obj, __pyx_n_s_setstate);
            if (!setstate) PyErr_Clear();
            if (!setstate || __Pyx_setup_reduce_is_named(setstate, __pyx_n_s_setstate_cython)) {
                setstate_cython = __Pyx_PyObject_GetAttrStrNoError(type_obj, __pyx_n_s_setstate_cython);
                if (likely(setstate_cython)) {
                    ret = PyDict_SetItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate, setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                    ret = PyDict_DelItem(((PyTypeObject*)type_obj)->tp_dict, __pyx_n_s_setstate_cython); if (unlikely(ret < 0)) goto __PYX_BAD;
                } else if (!setstate || PyErr_Occurred()) {
                    goto __PYX_BAD;
                }
            }
            PyType_Modified((PyTypeObject*)type_obj);
        }
    }
    goto __PYX_GOOD;
__PYX_BAD:
    if (!PyErr_Occurred())
        PyErr_Format(PyExc_RuntimeError, "Unable to initialize pickling for %s", ((PyTypeObject*)type_obj)->tp_name);
    ret = -1;
__PYX_GOOD:
#if !CYTHON_USE_PYTYPE_LOOKUP
    Py_XDECREF(object_reduce);
    Py_XDECREF(object_reduce_ex);
#endif
    Py_XDECREF(reduce);
    Py_XDECREF(reduce_ex);
    Py_XDECREF(reduce_cython);
    Py_XDECREF(setstate);
    Py_XDECREF(setstate_cython);
    return ret;
}

/* TypeImport */
#ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(PyObject *module, const char *module_name, const char *class_name,
    size_t size, enum __Pyx_ImportType_CheckSize check_size)
{
    PyObject *result = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    result = PyObject_GetAttrString(module, class_name);
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if ((size_t)basicsize < size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    if (check_size == __Pyx_ImportType_CheckSize_Error && (size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        goto bad;
    }
    else if (check_size == __Pyx_ImportType_CheckSize_Warn && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. "
            "Expected %zd from C header, got %zd from PyObject",
            module_name, class_name, size, basicsize);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(result);
    return NULL;
}
#endif

/* GetVTable */
static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* CLineInTraceback */
#ifndef CYTHON_CLINE_IN_TRACEBACK
static int __Pyx_CLineForTraceback(CYTHON_NCP_UNUSED PyThreadState *tstate, int c_line) {
    PyObject *use_cline;
    PyObject *ptype, *pvalue, *ptraceback;
#if CYTHON_COMPILING_IN_CPYTHON
    PyObject **cython_runtime_dict;
#endif
    if (unlikely(!__pyx_cython_runtime)) {
        return c_line;
    }
    __Pyx_ErrFetchInState(tstate, &ptype, &pvalue, &ptraceback);
#if CYTHON_COMPILING_IN_CPYTHON
    cython_runtime_dict = _PyObject_GetDictPtr(__pyx_cython_runtime);
    if (likely(cython_runtime_dict)) {
        __PYX_PY_DICT_LOOKUP_IF_MODIFIED(
            use_cline, *cython_runtime_dict,
            __Pyx_PyDict_GetItemStr(*cython_runtime_dict, __pyx_n_s_cline_in_traceback))
    } else
#endif
    {
      PyObject *use_cline_obj = __Pyx_PyObject_GetAttrStr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback);
      if (use_cline_obj) {
        use_cline = PyObject_Not(use_cline_obj) ? Py_False : Py_True;
        Py_DECREF(use_cline_obj);
      } else {
        PyErr_Clear();
        use_cline = NULL;
      }
    }
    if (!use_cline) {
        c_line = 0;
        PyObject_SetAttr(__pyx_cython_runtime, __pyx_n_s_cline_in_traceback, Py_False);
    }
    else if (use_cline == Py_False || (use_cline != Py_True && PyObject_Not(use_cline) != 0)) {
        c_line = 0;
    }
    __Pyx_ErrRestoreInState(tstate, ptype, pvalue, ptraceback);
    return c_line;
}
#endif

/* CodeObjectCache */
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, ((size_t)new_max) * sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
#include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    PyThreadState *tstate = __Pyx_PyThreadState_Current;
    if (c_line) {
        c_line = __Pyx_CLineForTraceback(tstate, c_line);
    }
    py_code = __pyx_find_code_object(c_line ? -c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? -c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        tstate,            /*PyThreadState *tstate,*/
        py_code,           /*PyCodeObject *code,*/
        __pyx_d,    /*PyObject *globals,*/
        0                  /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
#define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

#if PY_MAJOR_VERSION < 3
static int __Pyx_GetBuffer(PyObject *obj, Py_buffer *view, int flags) {
    if (PyObject_CheckBuffer(obj)) return PyObject_GetBuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_array_type)) return __pyx_array_getbuffer(obj, view, flags);
        if (__Pyx_TypeCheck(obj, __pyx_memoryview_type)) return __pyx_memoryview_getbuffer(obj, view, flags);
    PyErr_Format(PyExc_TypeError, "'%.200s' does not have the buffer interface", Py_TYPE(obj)->tp_name);
    return -1;
}
static void __Pyx_ReleaseBuffer(Py_buffer *view) {
    PyObject *obj = view->obj;
    if (!obj) return;
    if (PyObject_CheckBuffer(obj)) {
        PyBuffer_Release(view);
        return;
    }
    if ((0)) {}
    view->obj = NULL;
    Py_DECREF(obj);
}
#endif


/* MemviewSliceIsContig */
static int
__pyx_memviewslice_is_contig(const __Pyx_memviewslice mvs, char order, int ndim)
{
    int i, index, step, start;
    Py_ssize_t itemsize = mvs.memview->view.itemsize;
    if (order == 'F') {
        step = 1;
        start = 0;
    } else {
        step = -1;
        start = ndim - 1;
    }
    for (i = 0; i < ndim; i++) {
        index = start + step * i;
        if (mvs.suboffsets[index] >= 0 || mvs.strides[index] != itemsize)
            return 0;
        itemsize *= mvs.shape[index];
    }
    return 1;
}

/* OverlappingSlices */
static void
__pyx_get_array_memory_extents(__Pyx_memviewslice *slice,
                               void **out_start, void **out_end,
                               int ndim, size_t itemsize)
{
    char *start, *end;
    int i;
    start = end = slice->data;
    for (i = 0; i < ndim; i++) {
        Py_ssize_t stride = slice->strides[i];
        Py_ssize_t extent = slice->shape[i];
        if (extent == 0) {
            *out_start = *out_end = start;
            return;
        } else {
            if (stride > 0)
                end += stride * (extent - 1);
            else
                start += stride * (extent - 1);
        }
    }
    *out_start = start;
    *out_end = end + itemsize;
}
static int
__pyx_slices_overlap(__Pyx_memviewslice *slice1,
                     __Pyx_memviewslice *slice2,
                     int ndim, size_t itemsize)
{
    void *start1, *end1, *start2, *end2;
    __pyx_get_array_memory_extents(slice1, &start1, &end1, ndim, itemsize);
    __pyx_get_array_memory_extents(slice2, &start2, &end2, ndim, itemsize);
    return (start1 < end2) && (start2 < end1);
}

/* Capsule */
static CYTHON_INLINE PyObject *
__pyx_capsule_create(void *p, CYTHON_UNUSED const char *sig)
{
    PyObject *cobj;
#if PY_VERSION_HEX >= 0x02070000
    cobj = PyCapsule_New(p, sig, NULL);
#else
    cobj = PyCObject_FromVoidPtr(p, NULL);
#endif
    return cobj;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_char(unsigned char value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned char) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned char) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned char) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned char),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE unsigned char __Pyx_PyInt_As_unsigned_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const unsigned char neg_one = (unsigned char) -1, const_zero = (unsigned char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(unsigned char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(unsigned char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (unsigned char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char, digit, digits[0])
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 2 * PyLong_SHIFT) {
                            return (unsigned char) (((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 3 * PyLong_SHIFT) {
                            return (unsigned char) (((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) >= 4 * PyLong_SHIFT) {
                            return (unsigned char) (((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (unsigned char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (unsigned char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(unsigned char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(unsigned char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(unsigned char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(unsigned char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                            return (unsigned char) ((((((unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(unsigned char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(unsigned char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(unsigned char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) (((unsigned char)-1)*(((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(unsigned char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(unsigned char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(unsigned char) - 1 > 4 * PyLong_SHIFT) {
                            return (unsigned char) ((((((((((unsigned char)digits[3]) << PyLong_SHIFT) | (unsigned char)digits[2]) << PyLong_SHIFT) | (unsigned char)digits[1]) << PyLong_SHIFT) | (unsigned char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(unsigned char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(unsigned char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(unsigned char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            unsigned char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (unsigned char) -1;
        }
    } else {
        unsigned char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (unsigned char) -1;
        val = __Pyx_PyInt_As_unsigned_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to unsigned char");
    return (unsigned char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to unsigned char");
    return (unsigned char) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const int neg_one = (int) -1, const_zero = (int) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_PY_LONG_LONG(PY_LONG_LONG value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(PY_LONG_LONG),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
static CYTHON_INLINE PY_LONG_LONG __Pyx_PyInt_As_PY_LONG_LONG(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const PY_LONG_LONG neg_one = (PY_LONG_LONG) -1, const_zero = (PY_LONG_LONG) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(PY_LONG_LONG) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (PY_LONG_LONG) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, digit, digits[0])
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) >= 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (PY_LONG_LONG) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (PY_LONG_LONG) 0;
                case -1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(PY_LONG_LONG,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(PY_LONG_LONG) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(PY_LONG_LONG) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(PY_LONG_LONG) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) (((PY_LONG_LONG)-1)*(((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(PY_LONG_LONG) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(PY_LONG_LONG, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(PY_LONG_LONG) - 1 > 4 * PyLong_SHIFT) {
                            return (PY_LONG_LONG) ((((((((((PY_LONG_LONG)digits[3]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[2]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[1]) << PyLong_SHIFT) | (PY_LONG_LONG)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(PY_LONG_LONG) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(PY_LONG_LONG) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(PY_LONG_LONG, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            PY_LONG_LONG val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (PY_LONG_LONG) -1;
        }
    } else {
        PY_LONG_LONG val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (PY_LONG_LONG) -1;
        val = __Pyx_PyInt_As_PY_LONG_LONG(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to PY_LONG_LONG");
    return (PY_LONG_LONG) -1;
}

/* CIntFromPy */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntToPy */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const long neg_one = (long) -1, const_zero = (long) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* MemviewSliceCopyTemplate */
static __Pyx_memviewslice
__pyx_memoryview_copy_new_contig(const __Pyx_memviewslice *from_mvs,
                                 const char *mode, int ndim,
                                 size_t sizeof_dtype, int contig_flag,
                                 int dtype_is_object)
{
    __Pyx_RefNannyDeclarations
    int i;
    __Pyx_memviewslice new_mvs = { 0, 0, { 0 }, { 0 }, { 0 } };
    struct __pyx_memoryview_obj *from_memview = from_mvs->memview;
    Py_buffer *buf = &from_memview->view;
    PyObject *shape_tuple = NULL;
    PyObject *temp_int = NULL;
    struct __pyx_array_obj *array_obj = NULL;
    struct __pyx_memoryview_obj *memview_obj = NULL;
    __Pyx_RefNannySetupContext("__pyx_memoryview_copy_new_contig", 0);
    for (i = 0; i < ndim; i++) {
        if (unlikely(from_mvs->suboffsets[i] >= 0)) {
            PyErr_Format(PyExc_ValueError, "Cannot copy memoryview slice with "
                                           "indirect dimensions (axis %d)", i);
            goto fail;
        }
    }
    shape_tuple = PyTuple_New(ndim);
    if (unlikely(!shape_tuple)) {
        goto fail;
    }
    __Pyx_GOTREF(shape_tuple);
    for(i = 0; i < ndim; i++) {
        temp_int = PyInt_FromSsize_t(from_mvs->shape[i]);
        if(unlikely(!temp_int)) {
            goto fail;
        } else {
            PyTuple_SET_ITEM(shape_tuple, i, temp_int);
            temp_int = NULL;
        }
    }
    array_obj = __pyx_array_new(shape_tuple, sizeof_dtype, buf->format, (char *) mode, NULL);
    if (unlikely(!array_obj)) {
        goto fail;
    }
    __Pyx_GOTREF(array_obj);
    memview_obj = (struct __pyx_memoryview_obj *) __pyx_memoryview_new(
                                    (PyObject *) array_obj, contig_flag,
                                    dtype_is_object,
                                    from_mvs->memview->typeinfo);
    if (unlikely(!memview_obj))
        goto fail;
    if (unlikely(__Pyx_init_memviewslice(memview_obj, ndim, &new_mvs, 1) < 0))
        goto fail;
    if (unlikely(__pyx_memoryview_copy_contents(*from_mvs, new_mvs, ndim, ndim,
                                                dtype_is_object) < 0))
        goto fail;
    goto no_fail;
fail:
    __Pyx_XDECREF(new_mvs.memview);
    new_mvs.memview = NULL;
    new_mvs.data = NULL;
no_fail:
    __Pyx_XDECREF(shape_tuple);
    __Pyx_XDECREF(temp_int);
    __Pyx_XDECREF(array_obj);
    __Pyx_RefNannyFinishContext();
    return new_mvs;
}

/* MemviewSliceInit */
static int
__Pyx_init_memviewslice(struct __pyx_memoryview_obj *memview,
                        int ndim,
                        __Pyx_memviewslice *memviewslice,
                        int memview_is_new_reference)
{
    __Pyx_RefNannyDeclarations
    int i, retval=-1;
    Py_buffer *buf = &memview->view;
    __Pyx_RefNannySetupContext("init_memviewslice", 0);
    if (unlikely(memviewslice->memview || memviewslice->data)) {
        PyErr_SetString(PyExc_ValueError,
            "memviewslice is already initialized!");
        goto fail;
    }
    if (buf->strides) {
        for (i = 0; i < ndim; i++) {
            memviewslice->strides[i] = buf->strides[i];
        }
    } else {
        Py_ssize_t stride = buf->itemsize;
        for (i = ndim - 1; i >= 0; i--) {
            memviewslice->strides[i] = stride;
            stride *= buf->shape[i];
        }
    }
    for (i = 0; i < ndim; i++) {
        memviewslice->shape[i]   = buf->shape[i];
        if (buf->suboffsets) {
            memviewslice->suboffsets[i] = buf->suboffsets[i];
        } else {
            memviewslice->suboffsets[i] = -1;
        }
    }
    memviewslice->memview = memview;
    memviewslice->data = (char *)buf->buf;
    if (__pyx_add_acquisition_count(memview) == 0 && !memview_is_new_reference) {
        Py_INCREF(memview);
    }
    retval = 0;
    goto no_fail;
fail:
    memviewslice->memview = 0;
    memviewslice->data = 0;
    retval = -1;
no_fail:
    __Pyx_RefNannyFinishContext();
    return retval;
}
#ifndef Py_NO_RETURN
#define Py_NO_RETURN
#endif
static void __pyx_fatalerror(const char *fmt, ...) Py_NO_RETURN {
    va_list vargs;
    char msg[200];
#ifdef HAVE_STDARG_PROTOTYPES
    va_start(vargs, fmt);
#else
    va_start(vargs);
#endif
    vsnprintf(msg, 200, fmt, vargs);
    va_end(vargs);
    Py_FatalError(msg);
}
static CYTHON_INLINE int
__pyx_add_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)++;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE int
__pyx_sub_acquisition_count_locked(__pyx_atomic_int *acquisition_count,
                                   PyThread_type_lock lock)
{
    int result;
    PyThread_acquire_lock(lock, 1);
    result = (*acquisition_count)--;
    PyThread_release_lock(lock);
    return result;
}
static CYTHON_INLINE void
__Pyx_INC_MEMVIEW(__Pyx_memviewslice *memslice, int have_gil, int lineno)
{
    int first_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None))
        return;
    if (unlikely(__pyx_get_slice_count(memview) < 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    first_time = __pyx_add_acquisition_count(memview) == 0;
    if (unlikely(first_time)) {
        if (have_gil) {
            Py_INCREF((PyObject *) memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_INCREF((PyObject *) memview);
            PyGILState_Release(_gilstate);
        }
    }
}
static CYTHON_INLINE void __Pyx_XDEC_MEMVIEW(__Pyx_memviewslice *memslice,
                                             int have_gil, int lineno) {
    int last_time;
    struct __pyx_memoryview_obj *memview = memslice->memview;
    if (unlikely(!memview || (PyObject *) memview == Py_None)) {
        memslice->memview = NULL;
        return;
    }
    if (unlikely(__pyx_get_slice_count(memview) <= 0))
        __pyx_fatalerror("Acquisition count is %d (line %d)",
                         __pyx_get_slice_count(memview), lineno);
    last_time = __pyx_sub_acquisition_count(memview) == 1;
    memslice->data = NULL;
    if (unlikely(last_time)) {
        if (have_gil) {
            Py_CLEAR(memslice->memview);
        } else {
            PyGILState_STATE _gilstate = PyGILState_Ensure();
            Py_CLEAR(memslice->memview);
            PyGILState_Release(_gilstate);
        }
    } else {
        memslice->memview = NULL;
    }
}

/* CIntFromPy */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
static CYTHON_INLINE char __Pyx_PyInt_As_char(PyObject *x) {
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic push
#pragma GCC diagnostic ignored "-Wconversion"
#endif
    const char neg_one = (char) -1, const_zero = (char) 0;
#ifdef __Pyx_HAS_GCC_DIAGNOSTIC
#pragma GCC diagnostic pop
#endif
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(char) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(char, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (char) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case  1: __PYX_VERIFY_RETURN_INT(char, digit, digits[0])
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 2 * PyLong_SHIFT) {
                            return (char) (((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 3 * PyLong_SHIFT) {
                            return (char) (((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) >= 4 * PyLong_SHIFT) {
                            return (char) (((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (char) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(char) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (char) 0;
                case -1: __PYX_VERIFY_RETURN_INT(char, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(char,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(char) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(char) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                            return (char) ((((((char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(char) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(char) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                            return (char) ((((((((char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(char) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) (((char)-1)*(((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(char) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(char, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(char) - 1 > 4 * PyLong_SHIFT) {
                            return (char) ((((((((((char)digits[3]) << PyLong_SHIFT) | (char)digits[2]) << PyLong_SHIFT) | (char)digits[1]) << PyLong_SHIFT) | (char)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(char) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(char) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(char, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            char val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (char) -1;
        }
    } else {
        char val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (char) -1;
        val = __Pyx_PyInt_As_char(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to char");
    return (char) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to char");
    return (char) -1;
}

/* CheckBinaryVersion */
static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        if (PyObject_Hash(*t->p) == -1)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE const char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
#if !CYTHON_PEP393_ENABLED
static const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    char* defenc_c;
    PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
    if (!defenc) return NULL;
    defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    {
        char* end = defenc_c + PyBytes_GET_SIZE(defenc);
        char* c;
        for (c = defenc_c; c < end; c++) {
            if ((unsigned char) (*c) >= 128) {
                PyUnicode_AsASCIIString(o);
                return NULL;
            }
        }
    }
#endif
    *length = PyBytes_GET_SIZE(defenc);
    return defenc_c;
}
#else
static CYTHON_INLINE const char* __Pyx_PyUnicode_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
    if (unlikely(__Pyx_PyUnicode_READY(o) == -1)) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
    if (likely(PyUnicode_IS_ASCII(o))) {
        *length = PyUnicode_GET_LENGTH(o);
        return PyUnicode_AsUTF8(o);
    } else {
        PyUnicode_AsASCIIString(o);
        return NULL;
    }
#else
    return PyUnicode_AsUTF8AndSize(o, length);
#endif
}
#endif
#endif
static CYTHON_INLINE const char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
        return __Pyx_PyUnicode_AsStringAndSize(o, length);
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrueAndDecref(PyObject* x) {
    int retval;
    if (unlikely(!x)) return -1;
    retval = __Pyx_PyObject_IsTrue(x);
    Py_DECREF(x);
    return retval;
}
static PyObject* __Pyx_PyNumber_IntOrLongWrongResultType(PyObject* result, const char* type_name) {
#if PY_MAJOR_VERSION >= 3
    if (PyLong_Check(result)) {
        if (PyErr_WarnFormat(PyExc_DeprecationWarning, 1,
                "__int__ returned non-int (type %.200s).  "
                "The ability to return an instance of a strict subclass of int "
                "is deprecated, and may be removed in a future version of Python.",
                Py_TYPE(result)->tp_name)) {
            Py_DECREF(result);
            return NULL;
        }
        return result;
    }
#endif
    PyErr_Format(PyExc_TypeError,
                 "__%.4s__ returned non-%.4s (type %.200s)",
                 type_name, type_name, Py_TYPE(result)->tp_name);
    Py_DECREF(result);
    return NULL;
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_Check(x) || PyLong_Check(x)))
#else
  if (likely(PyLong_Check(x)))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = m->nb_int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = m->nb_long(x);
  }
  #else
  if (likely(m && m->nb_int)) {
    name = "int";
    res = m->nb_int(x);
  }
  #endif
#else
  if (!PyBytes_CheckExact(x) && !PyUnicode_CheckExact(x)) {
    res = PyNumber_Int(x);
  }
#endif
  if (likely(res)) {
#if PY_MAJOR_VERSION < 3
    if (unlikely(!PyInt_Check(res) && !PyLong_Check(res))) {
#else
    if (unlikely(!PyLong_CheckExact(res))) {
#endif
        return __Pyx_PyNumber_IntOrLongWrongResultType(res, name);
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(b);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyBool_FromLong(long b) {
  return b ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False);
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
